//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzwanyunheng@corp.netease.com
//	Author: wanyunheng
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<CheckboardResolve.fx>
///	@path	~/System/Shader/
///	@date	2019/09/16
///	@brief	CheckboardResolve

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Shader
{
	int4		TextureSize				: layout(c0);
};

///////////////// samplers //////////////////
texture2D tBaseMap : layout(t0);
sampler sBaseSampler : layout(s0)
{
	Texture = tBaseMap;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

Texture2DMS<uint> ObjectIDBuffer : layout(t1);

RWTexture2D<float4> TextureOut	: layout(u0);


#define UP 0
#define DOWN 1
#define LEFT 2
#define RIGHT 3
#define SAMPLE_COUNT 4

uint SampleID(int x, int y, int checkboardId)
{
	return ObjectIDBuffer.Load(int2(x / 2, y), (x + y + checkboardId) & 0x00000001);
}

uint GetObjId(uint id)
{
	return id & uint(0xffff0000);
}

float3 RGB2YCoCg(float3 rgb)
{
	float R  = rgb.r; float G = rgb.g; float B = rgb.b;
	float T = 0.25*R + 0.25*B;
	float Y  = 0.5*G + T;
	float Co = 0.5*(R-B);
	float Cg = 0.5*G - T;
	return float3(Y, Co, Cg);
}

float3 YCoCg2RGB(float3 YCoCg)
{
	float Y  = YCoCg.x; float Co = YCoCg.y; float Cg = YCoCg.z;
	float t = Y - Cg;
	float R = t + Co;
	float G = Y + Cg;
	float B = t - Co;

	return float3(R, G, B);
}

float GetColorDiffWeight(float3 color1, float3 color2)
{
	float3 diff = color1 - color2;
	float len = sqrt(dot(diff, diff));
	return 1.0 / (len + 0.001);
}

[numthreads(8, 8, 1)]
void CheckerboardResolveCS(int3 GlobalID : SV_DispatchThreadID)
{
	// dispatch tBaseMap coordinate (x, y), fill TextureOut (2x, y) and (2x + 1, y)
	int ux = GlobalID.x * 2;
	int uy = GlobalID.y;
	int checkboardId = TextureSize.z & 0x00000001;

	// checker board pattern, "O": sample0, shaded, "*": sample1, unshaded
	// checkboardId = 0
	// O * O *
	// * O * O
	// checkboardId = 1
	// * O * O
	// O * O *

	if (ux < TextureSize.x && uy < TextureSize.y)
	{
		int ox = (uy + checkboardId) & 0x00000001;
		int shaded_ux = ux + ox;
		int unshaded_ux = ux + 1 - ox;

		// prim id
		uint ids[SAMPLE_COUNT] =
		{
			SampleID(unshaded_ux, uy - 1, checkboardId),
			SampleID(unshaded_ux, uy + 1, checkboardId),
			SampleID(unshaded_ux - 1, uy, checkboardId),
			SampleID(unshaded_ux + 1, uy, checkboardId)
		};
		uint shaded_id = SampleID(unshaded_ux, uy, checkboardId);
		// object id
		uint shaded_object_id = GetObjId(shaded_id);
		// color
		float3 colors[SAMPLE_COUNT] =
		{
			RGB2YCoCg(tBaseMap.Load(GlobalID.xy + int2(   0, -1)).rgb),
			RGB2YCoCg(tBaseMap.Load(GlobalID.xy + int2(   0,  1)).rgb),
			RGB2YCoCg(tBaseMap.Load(GlobalID.xy + int2( -ox,  0)).rgb),
			RGB2YCoCg(tBaseMap.Load(GlobalID.xy + int2(1-ox,  0)).rgb)
		};
		int shaded_index = LEFT;
		if (ox & 0x00000001) shaded_index = RIGHT;
		float3 shaded_color = colors[shaded_index];

		float INF = 9999.0;
		float3 prim_color_min = INF;
		float3 prim_color_max = -INF;
		for (int i = 0; i < SAMPLE_COUNT; ++i)
		{
			if (ids[i] == shaded_id)
			{
				prim_color_min = min(prim_color_min, colors[i]);
				prim_color_max = max(prim_color_max, colors[i]);
			}
		}

		float weights[2] = {1.7, 1.7};
		if (GetObjId(ids[UP]) == GetObjId(ids[DOWN]))
		{
			weights[0] = GetColorDiffWeight(colors[UP], colors[DOWN]);
		}
		if (GetObjId(ids[LEFT]) == GetObjId(ids[RIGHT]))
		{
			weights[1] = GetColorDiffWeight(colors[LEFT], colors[RIGHT]);
		}

		float3 result = 0.0;
		// get prim color
		float3 prim_color = 0.0;
		float prim_weight_total = 0.0;
		int prim_count_total = 0;
		if (shaded_id == ids[UP] || (all(colors[UP] >= prim_color_min) && all(colors[UP] <= prim_color_max)))
		{
			prim_color += colors[UP] * weights[0];
			prim_weight_total += weights[0];
			++prim_count_total;
		}
		if (shaded_id == ids[DOWN] || (all(colors[DOWN] >= prim_color_min) && all(colors[DOWN] <= prim_color_max)))
		{
			prim_color += colors[DOWN] * weights[0];
			prim_weight_total += weights[0];
			++prim_count_total;
		}
		if (shaded_id == ids[LEFT] || (all(colors[LEFT] >= prim_color_min) && all(colors[LEFT] <= prim_color_max)))
		{
			prim_color += colors[LEFT] * weights[1];
			prim_weight_total += weights[1];
			++prim_count_total;
		}
		if (shaded_id == ids[RIGHT] || (all(colors[RIGHT] >= prim_color_min) && all(colors[RIGHT] <= prim_color_max)))
		{
			prim_color += colors[RIGHT] * weights[1];
			prim_weight_total += weights[1];
			++prim_count_total;
		}

		if (prim_count_total > 1)
		{
			// use prim color
			result = prim_color / prim_weight_total;
		}
		else
		{
			// get object color
			float3 object_color = 0.0;
			float object_weight_total = 0.0;
			int object_count_total = 0;
			if (shaded_object_id == GetObjId(ids[UP]))
			{
				object_color += colors[UP] * weights[0];
				object_weight_total += weights[0];
				++object_count_total;
			}
			if (shaded_object_id == GetObjId(ids[DOWN]))
			{
				object_color += colors[DOWN] * weights[0];
				object_weight_total += weights[0];
				++object_count_total;
			}
			if (shaded_object_id == GetObjId(ids[LEFT]))
			{
				object_color += colors[LEFT] * weights[1];
				object_weight_total += weights[1];
				++object_count_total;
			}
			if (shaded_object_id == GetObjId(ids[RIGHT]))
			{
				object_color += colors[RIGHT] * weights[1];
				object_weight_total += weights[1];
				++object_count_total;
			}

			if (prim_count_total == 1)
			{
				// blend prim color and object color
				result = prim_color / prim_weight_total;
				if (object_count_total > 0)
				{
					object_color = object_color / object_weight_total;
					float3 diff = result - object_color;
					float scale = min(1.0, dot(diff, diff) / 0.025);
					result = result * scale + object_color * (1.0 - scale);
				}
			}
			else if (object_count_total > 0)
			{
				// use object color
				result = object_color / object_weight_total;
			}
			else
			{
				result = ((colors[UP] + colors[DOWN]) * weights[0] + (colors[LEFT] + colors[RIGHT]) * weights[1]) / (2.0 * (weights[0] + weights[1]));
			}
		}

		// output shaded sample
		TextureOut.Store(int2(shaded_ux, uy), float4(YCoCg2RGB(shaded_color), 1.0));
		// output unshaded sample
		TextureOut.Store(int2(unshaded_ux, uy), float4(YCoCg2RGB(result), 1.0));
	}
}

technique CheckerboardResolve
{
	ComputeShader = CheckerboardResolveCS();
}
