
#pragma shader material
#pragma shader category("ClothEditor")

#include "Constants.fxh"

////////////////////////////////Constant buffer///////////////////////////
cbuffer Batch
{
    float4x3 World                 : World      ;

    float4 BrushShow               : Param      ;
    float4 BrushParam              : Param2     ;
};


// Surface textures
texture2D tDiffuseMap : layout(t0) = "White";
sampler sDiffuseSampler : layout(s0, s1)
{
	Texture = tDiffuseMap;
	Filter = PointLinear;
	Address = Wrap;
	SRGB = false;
};


#include "Common.fxh"

///////////////////////////////Assemble Input/////////////////////////////
struct VertexAssemble
{
    float4      Position        : POSITION;
    float4      Normal          : NORMAL; 
    float2      TexCoord        : TEXCOORD;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

	float3 LocalNormal;
};

struct InterMediateGeometry
{
    float4      Center;

    float4      Color; 

    float       Radius;

    int         Split;
};

//////////////////////////////Vertex Output///////////////////////////////
struct VertexShadingOutput
{
    float4		HPosition		: SV_Position;			// Position in homogeneous space. For output to rasterizer.
	
	float4		ClothColor      : TEXCOORDN;			// Cloth Vertex Color for surface.

    float4      WorldPosition   : TEXCOORDN;            // World Position.
};

//////////////////////////////Geometry Output///////////////////////////////
struct GeometryShadingOutput
{
    float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		ClothColor      : TEXCOORDN;			//	Cloth Vertex Color for surface.

    float4      TexCoord        : TEXCOORDN;            // For wire render

    float4      WorldPosition   : TEXCOORDN;            // World Position.
};

///////////////////////////////////Functions/////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);

	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);

	// Convert normal to color, x is value, y is hight light
	OUT.ClothColor.xyz = 0.5 + 0.5 * INTM.LocalNormal;
}

float4 EncodeVetexColor(float4 ClothColor)
{
    float4 OUT;
    float4 StaticColor = float4(0.917, 0.352, 0.859, 1);
    float4 ShiftColor = float4(1,1,1,1);
    float cons = ClothColor.r;
    OUT = lerp(
           StaticColor,/*Static point*/
           lerp(float4(cons, cons, cons, 1)/*Dynamic point*/, 
                ShiftColor,
                0.2),
          step(0.0000001, abs(cons)));                                 
    return OUT;
}

void GeometryExtendFaceToSphere(in InterMediateGeometry INTM, float4 a, float4 b, 
    float4 c, inout TriangleStream<VertexShadingOutput> OUT)
{
    // Tesslation level
    /*
                    1 --- 3
           (UpDir) / \   / \ .....
                  /   \ /   \ (DownDir)
     (StartPoint)0 --- 2     
    */
    // split < length - 1 
    float4 buffer[10] = {
        float4(0,0,0,0), float4(0,0,0,0), float4(0,0,0,0), float4(0,0,0,0),
        float4(0,0,0,0), float4(0,0,0,0),float4(0,0,0,0), float4(0,0,0,0),
        float4(0,0,0,0), float4(0,0,0,0)
    };
    VertexShadingOutput tmp;
    [@_HLSL]
    tmp = (VertexShadingOutput)0;
    tmp.ClothColor = float4(0,0,1,0);
    int split = INTM.Split;
    const float4 updir = (a - c)/(float)split;
    const float4 downdir = (b - a)/(float)split;
    buffer[0] = a;
    int v, h;
    float4 nextPos = float4(0,0,0,0);
    for (v = 1; v <= split; ++v)
    {
        // To avoid compile error
        if (v != 1)
            buffer[v-1] = nextPos;
        nextPos = buffer[0] - updir;
        for (h = 0; h < v; ++h)
        {
            tmp.HPosition = nextPos;
            OUT.Append(tmp);
            tmp.HPosition = buffer[h];
            OUT.Append(tmp);

            float4 tmpPos = nextPos;
            nextPos = buffer[h] + downdir;
            buffer[h] = tmpPos; }
        tmp.HPosition = nextPos;
        OUT.Append(tmp);
        OUT.RestartStrip();
        // Compile error
        //buffer[v] = nextPos;
    }
}

void GeometryExtendVertexToSphere(in InterMediateGeometry INTM, inout TriangleStream<VertexShadingOutput> OUT)
{
    //                             a (0.0,1.0,0.0);
    //                           / |\ 
    //                          /  | \
    //   (-0.75, -0.5, -0.433) b---|--d (0.75, -0.5, -0.433);
    //                          \  | /
    //                           \ |/
    //                             c (0.0, -0.5, 0.866);
    //const float radius = 0.2;
    float4 a = INTM.Center + INTM.Radius * float4(0.0, 1.0, 0.0, 0.0);
    float4 b = INTM.Center + INTM.Radius * float4(-0.75, -0.5, -0.433, 0.0);
    //float4 c = INTM.Center + INTM.Radius * float4(0.0, -0.5, 0.866, 0.0);
    float4 c = INTM.Center + INTM.Radius * float4(0.0, -0.5, 11.866, 0.0);
    float4 d = INTM.Center + INTM.Radius * float4(0.75, -0.5, -0.433, 0.0);
    GeometryExtendFaceToSphere(INTM, a, c, b, OUT);
    //GeometryExtendFaceToSphere(INTM, float4(0,0.2,0,1), float4(0.2,0,0,1), float4(-0.2,0,0,1), OUT);
    //GeometryExtendFaceToSphere(INTM, a, d, c, OUT);
    //GeometryExtendFaceToSphere(INTM, a, b, d, OUT);
    //GeometryExtendFaceToSphere(INTM, c, d, b, OUT);
}
  
///////////////////////////////////Box function////////////////////////////////////
void GeometryMakeNewQuad(in InterMediateGeometry INTM, in int a, in int b, in int c, in int d, inout TriangleStream<GeometryShadingOutput> OUT)
{
    //              0---------1
    //             /|        /|
    //            / |       / |
    //           3---------2  |
    //           |  4------|--5 
    //           | /       | /
    //           |/        |/
    //           7---------6   
    float r = INTM.Radius;
    float4 Vertices[8] = {float4(-r,r,-r,0), float4(r,r,-r,0), 
                          float4(r,r,r,0), float4(-r,r,r,0),
                          float4(-r,-r,-r,0), float4(r,-r,-r,0), 
                          float4(r,-r,r,0), float4(-r,-r,r,0)};
    [@_HLSL]
    GeometryShadingOutput TmpOut = (GeometryShadingOutput)0;
    TmpOut.ClothColor = INTM.Color;
    TmpOut.TexCoord = float4(1, 1, 1, 0);
    TmpOut.WorldPosition= float4(9999999.0, 0, 0, 0);

    TmpOut.HPosition = Vertices[a] + INTM.Center;
    OUT.Append(TmpOut);
    TmpOut.HPosition = Vertices[b] + INTM.Center;
    OUT.Append(TmpOut);
    TmpOut.HPosition = Vertices[c] + INTM.Center;
    OUT.Append(TmpOut);
    OUT.RestartStrip();
    TmpOut.HPosition = Vertices[a] + INTM.Center;
    OUT.Append(TmpOut);
    TmpOut.HPosition = Vertices[c] + INTM.Center;
    OUT.Append(TmpOut);
    TmpOut.HPosition = Vertices[d] + INTM.Center;
    OUT.Append(TmpOut);
    OUT.RestartStrip();
}
void GeometryExtendVertexToBox(in InterMediateGeometry INTM, inout TriangleStream<GeometryShadingOutput> OUT)
{
    GeometryMakeNewQuad(INTM, 3, 2, 6, 7, OUT);
}
///////////////////////////////////Shading////////////////////////////////////
VertexShadingOutput ClothShowValueVertexShading(in VertexAssemble IN)
{
    VertexShadingOutput OUT;
    InterMediateVertex INTM;
    VertexUnifyInterMediate(IN, INTM);
    VertexUnifyOutput(IN, INTM, OUT);
    return OUT;
}

[maxvertexcount(50)]
void ClothShowValueGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<GeometryShadingOutput> OUT)
{
    // Original Cloth Mesh
    {
        GeometryShadingOutput tmpOut;

        tmpOut.TexCoord = float4(1, 0, 0, 0);
        tmpOut.HPosition = IN[0].HPosition;
        tmpOut.ClothColor = EncodeVetexColor(IN[0].ClothColor);
        tmpOut.WorldPosition = IN[0].WorldPosition;
        OUT.Append(tmpOut);

        tmpOut.TexCoord = float4(0, 1, 0, 0);
        tmpOut.HPosition = IN[1].HPosition;
        tmpOut.ClothColor = EncodeVetexColor(IN[1].ClothColor);
        tmpOut.WorldPosition = IN[1].WorldPosition;
        OUT.Append(tmpOut);

        tmpOut.TexCoord = float4(0, 0, 1, 0);
        tmpOut.HPosition = IN[2].HPosition;
        tmpOut.ClothColor = EncodeVetexColor(IN[2].ClothColor);
        tmpOut.WorldPosition = IN[2].WorldPosition;
        OUT.Append(tmpOut);

        OUT.RestartStrip();
    }
    float rAverage = length(IN[0].HPosition.xy - IN[1].HPosition.xy)
                   + length(IN[1].HPosition.xy - IN[2].HPosition.xy)
                   + length(IN[0].HPosition.xy - IN[2].HPosition.xy);
    rAverage = 0.33333*rAverage;
    // Marker
    {
        InterMediateGeometry INTM; 
        INTM.Split = 5;
        INTM.Radius = rAverage*0.05;
        INTM.Color = float4(0,0,1,0);

        INTM.Center = IN[0].HPosition;
        INTM.Color = float4(0.0, IN[0].ClothColor.y, 1.0 - IN[0].ClothColor.y, 1);
        //GeometryExtendVertexToSphere(INTM, OUT);
        GeometryExtendVertexToBox(INTM, OUT);
        INTM.Center = IN[1].HPosition;
        INTM.Color = float4(0.0, IN[1].ClothColor.y, 1.0 - IN[1].ClothColor.y, 1);
        //GeometryExtendVertexToSphere(INTM, OUT);
        //GeometryExtendVertexToSphere(INTM, OUT);
        GeometryExtendVertexToBox(INTM, OUT);
        INTM.Center = IN[2].HPosition;
        INTM.Color = float4(0.0, IN[2].ClothColor.y, 1.0 - IN[2].ClothColor.y, 1);
        //GeometryExtendVertexToSphere(INTM, OUT);
        GeometryExtendVertexToBox(INTM, OUT);
    }
}

float4 ClothShowValuePixelShading(in GeometryShadingOutput IN) : SV_Target
{
    float4 OUT = float4(0, 0, 0, 1);

    float3 difference = fwidth(IN.TexCoord.xyz);
	float3 distance = IN.TexCoord.xyz / difference;
	float d = min(min(distance.x, distance.y), distance.z);
    d = step(0.5, d);

    OUT.rgb = d * IN.ClothColor;
    float3 bright = float3(1, 0, 0);
    // y channel is the current brush value
    float3 value = float3(BrushShow.y, BrushShow.y, BrushShow.y);
    float split = 10;

    if (length(IN.WorldPosition) < 99999 && BrushParam.w > 0.0000001)
    {
        float3 r = BrushParam.xyz - IN.WorldPosition.xyz;
        float distance = length(r);
        float outter = BrushParam.w;
        float width = 0.1 * outter;
        float inner = outter - width;
        float swapinner = inner * sin(BrushShow*PI);
        float swapoutter = swapinner + width;
        if (distance < outter && distance > inner)
        {
            OUT.rgb = bright; 
        }
        else
        {
            if (distance < swapoutter && distance > swapinner)
            {
                OUT.rgb = lerp(value, OUT.rgb, 0.5);
            }
        }
    }

    return OUT;
}

technique Shading
{
    VertexShader = ClothShowValueVertexShading();
    GeometryShader = ClothShowValueGeometryShading();
    PixelShader = ClothShowValuePixelShading();
}
