
#pragma shader draw

[format("RGBA8")]
texture2D TextureIn				: layout(t0);

RWTexture2D<float4> TextureOut	: layout(u0);

#define GROUP_SIZE    64

#define BLUR_KERNEL   9
#define BLUR_RADIUS   4

cbuffer Shader
{
	int4	TextureSize				;
	float4	Weights[BLUR_KERNEL]	;
};

groupshared float4 Samples[BLUR_RADIUS + GROUP_SIZE + BLUR_RADIUS];

[numthreads(GROUP_SIZE, 1, 1)]
void GaussianBlurH(int3 GroupID : SV_GroupThreadID, int3 GlobalID : SV_DispatchThreadID)
{
	int2 Idx = GlobalID.xy;
	Idx.x = min(GlobalID.x, TextureSize.x - 1);

	// Load samples into group shared cache.
	if (GroupID.x < BLUR_RADIUS)
	{
		int x = max(0, Idx.x - BLUR_RADIUS);
		Samples[GroupID.x] = TextureIn.Load(int2(x, Idx.y));
	}

	Samples[BLUR_RADIUS + GroupID.x] = TextureIn.Load(Idx);

	if (GroupID.x >= (GROUP_SIZE - BLUR_RADIUS))
	{
		int x = min(TextureSize.x - 1, Idx.x + BLUR_RADIUS);
		Samples[BLUR_RADIUS + GroupID.x + BLUR_RADIUS] = TextureIn.Load(int2(x, Idx.y));
	}

	// synchronize all threads
	GroupMemoryBarrierWithGroupSync();

	float4 color = float4(0, 0, 0, 0);
	[unroll]
	for (int i = 0; i < BLUR_KERNEL; ++i)
	{
		color += Weights[i].x * Samples[GroupID.x + i];
	}

	if (GlobalID.x < TextureSize.x)
		TextureOut.Store(Idx, color);
}

[numthreads(1, GROUP_SIZE, 1)]
void GaussianBlurV(int3 GroupID : SV_GroupThreadID, int3 GlobalID : SV_DispatchThreadID)
{
	int2 Idx = GlobalID.xy;
	Idx.y = min(GlobalID.y, TextureSize.y - 1);

	// Load samples into group shared cache.
	if (GroupID.y < BLUR_RADIUS)
	{
		int y = max(0, Idx.y - BLUR_RADIUS);
		Samples[GroupID.y] = TextureIn.Load(int2(Idx.x, y));
	}

	Samples[BLUR_RADIUS + GroupID.y] = TextureIn.Load(Idx);

	if (GroupID.y >= (GROUP_SIZE - BLUR_RADIUS))
	{
		int y = min(TextureSize.y - 1, Idx.y + BLUR_RADIUS);
		Samples[BLUR_RADIUS + GroupID.y + BLUR_RADIUS] = TextureIn.Load(int2(Idx.x, y));
	}

	// synchronize all threads
	GroupMemoryBarrierWithGroupSync();

	float4 color = float4(0, 0, 0, 0);
	[unroll]
	for (int i = 0; i < BLUR_KERNEL; ++i)
	{
		color += Weights[i].x * Samples[GroupID.y + i];
	}

	if (GlobalID.y < TextureSize.y)
		TextureOut.Store(Idx, color);
}

technique HorizontalBlur
{
	ComputeShader = GaussianBlurH();
}

technique VerticalBlur
{
	ComputeShader = GaussianBlurV();
}

