//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Depth.fx>
///	@path	~/System/Shader/
///	@date	2015/04/08
///	@brief	Generate linear depth.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

@_IsOrthographic
{
	Name = IsOrthographic;
	Description = "Is orthographic camera?";
};

///////////////// samplers //////////////////
texture2D DepthBuffer : layout(t0);
sampler DepthBuffer_Sampler : layout(s0)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2DMS MSDepthBuffer : layout(t1);

//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

// {H75} MSAA
struct PixelScreenOutput
{
	float4		Color		: SV_Target;
	float		Depth		: SV_Depth;
};
// {H75} end

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Resolve depth
[target("R32F")]
float4 DepthResolvePS(VertexScreenOutput IN) : SV_Target
{
	float depth = SampleColorMap(DepthBuffer, DepthBuffer_Sampler, IN.TC).r;
	float near = CameraInfo.x;
	float far = CameraInfo.y;
	float linearDepth;
	if (@_IsOrthographic)
	{
		linearDepth = near * depth / far + (1 - depth);
	}
	else
	{
		linearDepth = near / (depth * (far - near) + near);
	}
	return EncodeDepth(linearDepth);
}

technique DepthResolve
{
	VertexShader = ScreenVS();
	PixelShader = DepthResolvePS();
}

// {H75} MSAA
// Resolve MSAA device depth
[target("R32F")]
PixelScreenOutput MSAADeviceDepthResolvePS(VertexScreenOutput IN)
{
	PixelScreenOutput OUT;
	OUT.Color = float4(0, 0, 0, 0);
	OUT.Depth = 0.f;
	if (@_D3D11 || @_PS4)
	{
		int2 tc = int2(IN.TC.x * ScreenInfo.x, IN.TC.y * ScreenInfo.y);
		// For MSAA, use depth of the first sample (default behavior of Metal).
		float depth = MSDepthBuffer.Load(tc, 0).r;
		OUT.Depth = depth;
	}
	return OUT;
}

technique MSAADeviceDepthResolve
{
	VertexShader = ScreenVS();
	PixelShader = MSAADeviceDepthResolvePS();
}
// {H75} end

//////////////////////////////////////////////////////////////////////////
// Resolve Multisample depth
[target("R32F")]
float4 MS2XDepthResolvePS(VertexScreenOutput IN) : SV_Target
{
	if (@_D3D11 || @_PS4)
	{
		int2 tc = int2(IN.TC.x * ScreenInfo.x, IN.TC.y * ScreenInfo.y);
		float depth = MSDepthBuffer.Load(tc, 0).r;
		depth += MSDepthBuffer.Load(tc, 1).r;
		depth = depth * 0.5;
		float near = CameraInfo.x;
		float far = CameraInfo.y;
		float linearDepth;
		if (@_IsOrthographic)
		{
			linearDepth = near * depth / far + (1 - depth);
		}
		else
		{
			linearDepth = near / (depth * (far - near) + near);
		}
		return EncodeDepth(linearDepth);
	}
	else
	{
		return float4(0, 0, 0, 0);
	}
}

technique MS2XDepthResolve
{
	VertexShader = ScreenVS();
	PixelShader = MS2XDepthResolvePS();
}

[target("R32F")]
float4 MS4XDepthResolvePS(VertexScreenOutput IN) : SV_Target
{
	if (@_D3D11 || @_PS4)
	{
		int2 tc = int2(IN.TC.x * ScreenInfo.x, IN.TC.y * ScreenInfo.y);
		float depth = MSDepthBuffer.Load(tc, 0).r;
		depth += MSDepthBuffer.Load(tc, 1).r;
		depth += MSDepthBuffer.Load(tc, 2).r;
		depth += MSDepthBuffer.Load(tc, 3).r;
		depth = depth * 0.25;
		float near = CameraInfo.x;
		float far = CameraInfo.y;

		float linearDepth;
		if (@_IsOrthographic)
		{
			linearDepth = near * depth / far + (1 - depth);
		}
		else
		{
			linearDepth = near / (depth * (far - near) + near);
		}
		return EncodeDepth(linearDepth);
	}
	else
	{
		return float4(0, 0, 0, 0);
	}
}

technique MS4XDepthResolve
{
	VertexShader = ScreenVS();
	PixelShader = MS4XDepthResolvePS();
}

// {H75}
//////////////////////////////////////////////////////////////////////////
// Clear depth
VertexScreenOutput DepthStencilClearVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, 0, 1);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 DepthStencilClearPS(VertexScreenOutput IN) : SV_Target
{
	return float4(0, 0, 0, 1);
}

technique DepthStencilClear
{
	VertexShader = DepthStencilClearVS();
	PixelShader = DepthStencilClearPS();
}
// {H75} end
