//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Basic.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
}

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = true;
}

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Basic
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  InvFadeDistance		 = 1;
	float  BloomScale			 = 0.5;
	float2 _Padding				 = { 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 fadeColor			 = { 0, 0, 0, 0 };
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D ScreenTexture : SceneColor : layout(t2) = "Transparent";
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D MaskMap : layout(t4) = "White";
sampler MaskMap_Sampler : layout(s4)
{
	Texture = MaskMap;
	Filter = Linear;
	Address = FromTexture;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Basic

void PS_GetParameters(in const VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT = float4(1,0,1,1);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;

	float Mask = SampleColorMap(MaskMap, MaskMap_Sampler, IN.TexCoord.xy).r;
	float4 effectColor = GetEffectColor(IN);	
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	
	OUT *= ModulateColor * effectColor * IN.VertexColor * Mask;
	materialParamters.BaseColor = OUT;
}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectUnifyPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Distortion

float4 EffectDistortionShading(in const VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT;
	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;

	float H = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy).r;
	H *= ModulateColor.x * GetEffectColor(IN).x;

	// hacked refraction
	float3 V = IN.ViewDirection;
	float3 N = IN.WorldNormal;
	float3 R = cross(N, V);
	float3 D = cross(R, V);

	float3 VR = cross(float3(0, 1, 0), V);
	float3 VU = normalize(cross(V, VR));
	VR = cross(VU, V);

	float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
	offset *= (500 / IN.TexCoord.z) * H;
	offset = offset * 0.5 + 0.5;

	EncodeFloatRG(offset.x, OUT.rg);
	EncodeFloatRG(offset.y, OUT.ba);

	return OUT;
}

technique Distortion
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectDistortionShading();
}

