//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Basic.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

#pragma shader override("@_SOFT", "false")
#pragma shader override("@_DISTORTION", "false")

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Basic
cbuffer Batch
{
	float4x3 World 				: World  						;
	float4 EffectColor			: Color  						;
	float4 EffectParam			: Param  						;
	float4 SubUV				: Param2 						;
	float4 cTintColor1			: TintColor1 					;
	float4 cPointCloud[6]		: PointCloudData				;
	float4 cPlaceHold											;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  InvFadeDistance		 = 1;
	float  BloomScale			 = 0.5;
	float2 _Padding				 = { 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  Roughness			 < Widget = "FloatSlider"; Min = "0.08"; Max = "0.92"; Step = "0.01"; Precision = "3"; > = 0.9;
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 fadeColor			 = { 0, 0, 0, 0 };
	float  Metallic				  < Widget = "FloatSlider"; Min = "0.001"; Max = "0.999"; Step = "0.001"; Precision = "3"; > = 0.01;
	float  NormalScale			 < Widget = "FloatSlider"; Min = "0.0"; Max = "5.0"; Step = "0.001"; Precision = "3"; > = 1;
	float  AlphaScale			 < Widget = "FloatSlider"; Min = "0.0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 1;
	float  Thickness			 < Widget = "FloatSlider"; Min = "0.2"; Max = "1.0"; Step = "0.01"; Precision = "3"; > = 1.0;
	float  Scattering			 < Widget = "FloatSlider"; Min = "0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 0.0;
	float  ScatteringPhaseG	 	 < Widget = "FloatSlider"; Min = "0.005"; Max = "0.995"; Step = "0.001"; Precision = "3"; > = 0.7;
	float  cFov					 = 1.0;
};

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is masking map enabled?";
	AffectedTex = MaskMap;
	Visible = true;
};

// @_SOFT
// {
// 	Name = Soft;
// 	Description = "Soft particle";
// 	Visible = true;
// }

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D ScreenTexture : SceneColor : layout(t2) = "Transparent";
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D MaskMap : layout(t4) = "White";
sampler MaskMap_Sampler : layout(s4)
{
	Texture = MaskMap;
	Filter = Linear;
	Address = FromTexture;
};

texture2D tShadowMask : ShadowMask : layout(t5) = "White";
sampler sShadowMaskSampler : layout(s5)
{
	Texture = tShadowMask;
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D tNormalMap : layout(t6) = "NormalMap";
sampler sNormalMap_Sampler : layout(s6)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

#include "BRDF.fxh"
//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Basic

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	InitBaseColorCommon(IN,materialParamters);
	materialParamters.BaseColor *= ModulateColor;

	[@_TANGENT]
	InitWorldNormal(IN,materialParamters);

	//Apply Mask
	if (@_MaskEnable)
	{
		float mask = SampleColorMap(MaskMap, MaskMap_Sampler, IN.TexCoord.xy).r;
		materialParamters.BaseColor.w *= mask;
	}

	materialParamters.Metallic = Metallic;
	materialParamters.Roughness = Roughness;

	//IN.VertexColor.a -> Element : Alpha By Life
	materialParamters.BaseColor.w *= GetEffectColor(IN).a;

}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectLightingPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Distortion

float4 EffectDistortionShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT;
	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;

	float H = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy).r;
	H *= ModulateColor.x * GetEffectColor(IN).x;

	// hacked refraction
	float3 V = IN.ViewDirection;
	float3 N = IN.WorldNormal;
	float3 R = cross(N, V);
	float3 D = cross(R, V);

	float3 VR = cross(float3(0, 1, 0), V);
	float3 VU = normalize(cross(V, VR));
	VR = cross(VU, V);

	float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
	offset *= (500 / IN.TexCoord.z) * H;
	offset = offset * 0.5 + 0.5;

	EncodeFloatRG(offset.x, OUT.rg);
	EncodeFloatRG(offset.y, OUT.ba);

	return OUT;
}

technique Distortion
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectDistortionShading();
};
