//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Dissipated.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

#pragma shader override("@_DISTORTION", "false")
#pragma shader override("@_SOFT", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_TwistEnable
{
	Name = TwistEnable;
	Description = "Is twisting enabled?";
	AffectedTex = TwistTexture;
	Visible = false;
};

@_Mask02Enable
{
	Name = Mask02Enable;
	Description = "Is Mask02 enabled?";
	AffectedTex = MaskTexture02;
	Visible = false;
};

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is Mask enabled?";
	AffectedTex = MaskTexture;
	Visible = false;
};

@_UseControlByLife
{
	Name = UseControlByLife;
	Description = "Use *Control By Life* instead of Alpha";
	Visible = true;
};

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Dissipated
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float4 DissipateUVScolling	 = { 1, 1, 0, 0 };
	float4 MaskUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
	float  Softness				 = 0.0;
	float  DissipateEdge		 = 0.01;
	float4 DissipateColor		 = {1, 1, 1, 1};
	// float  BloomScale			 = 0.5;
	float2 _Padding				 = { 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D DissipateTexture : layout(t1) = "White";
sampler DissipateTexture_Sampler : layout(s1)
{
	Texture = DissipateTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture : layout(t2) = "White";
sampler MaskTexture_Sampler : layout(s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture02 : layout(t3) = "White";
sampler MaskTexture02_Sampler : layout(s3)
{
	Texture = MaskTexture02;
	Filter = Linear;
	Address = FromTexture;
};

texture2D TwistTexture : layout(t4) = "White";
sampler TwistTexture_Sampler : layout(s4)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = FromTexture;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Dissipated.

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT = float4(0,0,0,0);
	float4 effectParam = GetEffectParam(IN);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + effectParam.y * DiffuseUVScolling.zw;
	float2 dissipateTC = IN.TexCoord.xy * DissipateUVScolling.xy + effectParam.y * DissipateUVScolling.zw;
	float2 maskTC = IN.TexCoord.xy * MaskUVScolling.xy + effectParam.y * MaskUVScolling.zw;

	if (@_TwistEnable)
	{
		float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + effectParam.y * TwistUVScolling.zw;
		float2 twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;
		diffuseTC += twist * TwistStrength;
		dissipateTC += twist * TwistStrength;
	}

	float mask = 1;
	if(@_MaskEnable)
	{
		mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).r;
	}

	float mask02 = 1;
	if(@_Mask02Enable)
	{
		mask02 = SampleColorMap(MaskTexture02, MaskTexture02_Sampler, IN.TexCoord.xy).r;
	}

	float4 effectColor = GetEffectColor(IN);
	float4 diffuseRGBA = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	diffuseRGBA.xyz *= ModulateColor.rgb * effectColor.rgb * IN.VertexColor.rgb;
	
	float alpha = SampleColorMap(DissipateTexture, DissipateTexture_Sampler, dissipateTC).r;
	
	//issue : VertexColor.w is used for both Addtive and Dissipated Control
	float dissipatedControl =  effectColor.w;
	float alphaRef = 1;
	if(@_UseControlByLife)
	{
		//Using Control By Life For Dissipated Control
		//Alpha will be used as Alpha as usual
		dissipatedControl = effectParam.w;
		alphaRef = effectColor.w;
	}

	//Extend Range of Dissipated Effect
	//Rescale dissipatedControl from range[0,1] to range[-DissipatedEdge-Softness , 1+DissipatedEdge+Softness]
	dissipatedControl = dissipatedControl * 2*(DissipateEdge + Softness + 0.5) - DissipateEdge - Softness;

	/*	rc1->		Aera Alpha=1
		rc2->		Aera Edge + Alpha=1
		rc2-rc1->	Aera Edge
		1-rc1->		Aera Edge + Alpha=0			*/
	float rc1 =  1-smoothstep(dissipatedControl-Softness,dissipatedControl+Softness,alpha);
	float rc2 =  1-smoothstep(dissipatedControl+DissipateEdge-Softness,dissipatedControl+DissipateEdge+Softness,alpha);
	diffuseRGBA.xyz = lerp(diffuseRGBA.xyz, DissipateColor.xyz , 1-rc1 );//Lerp Dissipated Edge Color,AND premultiply Alpha
	OUT = float4(diffuseRGBA.xyz,saturate((rc2-rc1 )* DissipateColor.w + rc1 * ModulateColor.w ) * alphaRef * mask * mask02 * diffuseRGBA.w);//Calc Alpha , ModulateColor.w Used Here

	materialParamters.BaseColor = OUT;
}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectUnifyPixelShading();
}
