//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Flow.fx>
///	@path	~/System/Shader/Effect
/// @LastChange 2020/05/14
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

//////////////////////////////////Properties////////////////////////////////////////

@_MaskEnable
{
	Name = Mask02Enable;
	Description = "Is masking map enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

@_TwistEnable
{
	Name = TwistEnable;
	Description = "Is mask map enabled?";
	AffectedTex = TwistTexture;
	Visible = true;
};

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = true;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
};

// @_FlowScrolling
// {
// 	Name = FlowScrolling;
// 	Description = "Enable Flow Scrolling instead of Flow Behavior";
// 	Visible = true;
// }

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Flowed
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	// float4 MaskUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 	 = { 0.1, 0.1 };
	float4 FlowUVScolling		 = { 1, 1, 0, 0 };
	float2 FlowStrength		 	 = { 0.5, 0.5 };
	float  FlowSpeed				 = 1.0;
	float  BloomScale			 = 0.5;
	float  InvFadeDistance		 = 1;
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  _Padding				 = 0;
	float4 fadeColor			 = { 0, 0, 0, 0 };
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D FlowTexture : layout(t2) = "White";
sampler FlowTexture_Sampler : layout(s2)
{
	Texture = FlowTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D TwistTexture : layout(t3) = "White";
sampler TwistTexture_Sampler : layout(s3)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture : layout(t4) = "White";
sampler MaskTexture_Sampler : layout(s4)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D DepthBuffer : SceneDepth : layout(t5);
sampler DepthBuffer_Sampler : layout(s5)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D ScreenTexture : SceneColor : layout(t6) = "Transparent";
sampler ScreenTexture_Sampler : layout(s6)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Flowed.

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT = float4(0,0,0,0);
	float4 effectParam = GetEffectParam(IN);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + effectParam.y * DiffuseUVScolling.zw;
	float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + effectParam.y * TwistUVScolling.zw;
	float2 FlowTC = IN.TexCoord.xy * FlowUVScolling.xy + GameTime * FlowUVScolling.zw;
	float2 Flow = SampleColorMap(FlowTexture, FlowTexture_Sampler, FlowTC).xy;
	float2 twist = float2(0,0);
	
	if(@_TwistEnable)
		twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;

	diffuseTC += twist * TwistStrength;
	Flow = (Flow * 2 - 1) * FlowStrength;

	float TimeIn = frac(GameTime * FlowSpeed );
	float TimeIn2 = frac(GameTime * FlowSpeed +0.5);

	//Issue : float2 + float2 * float may cause Invalid Subtitle ".xy"
	float4 Diffuse1 = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler,  diffuseTC + (Flow * TimeIn) );
	float4 Diffuse2 = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler,  diffuseTC + (Flow * TimeIn2));

	TimeIn = sin(TimeIn * PI);
	TimeIn2 = sin(TimeIn2 * PI);

	float M = TimeIn + TimeIn2;

	OUT = saturate(Diffuse1 * (TimeIn/M) + Diffuse2 * (TimeIn2/M) );

	OUT *= ModulateColor * GetEffectColor(IN) * IN.VertexColor;

	if (@_MaskEnable)
	{
		float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.xy).r;
		OUT.a *= mask;
	}

	materialParamters.BaseColor = OUT;

}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectUnifyPixelShading();
}
