//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Ice.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2020/05/22
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")
#pragma shader override("@_SOFT", "false")

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
}

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is Mask enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

@_DissipateEnable
{
	Name = DissipateEnable;
	Description = "Is Mask enabled?";
	AffectedTex = DissipateTexture;
	Visible = true;	
}

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Basic
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  InvFadeDistance		 = 1;
	float  BloomScale			 = 0.5;
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float4 fadeColor			 = { 0, 0, 0, 0 };

	float4 DiffuseUVScrolling	 = { 1, 1, 0, 0 };
	float4 MaskUVScrolling		 = { 1, 1, 0, 0 };
	float4 CrystalUVScrolling	 = { 1, 1, 0, 0 };
	float4 DissipateUVScolling	 = { 1, 1, 0, 0 };

	float  DissipateEdge		 = 0.01;
	float  Softness				 = 0.0;
	float4 DissipateColor		 = {1, 1, 1, 1};
	float4 SurfaceColor 	 	 < Widget = "Color4"; > = {0.8, 0.8, 0.8, 1.0};
	float  HDRScale				 = 1.0;
	float3 RimColor		 		 < Widget = "Color3"; > = {0.137, 0.2, 0.263};
	float  RimIntensity	 	 	 = 1.5;
	float3 IceColor		 	 	 < Widget = "Color3"; > = {0.388, 0.67, 0.737};
	float  Transparent           < Widget = "FloatSlider"; Min = "0"; Max = "1"; Step = "0.01"; Precision = "2"; > = 0.3;
	float  DistortionIntensity   < Widget = "FloatSlider"; Min = "0"; Max = "5.0"; Step = "0.01"; Precision = "2"; > = 0.0;
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "Black";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D CrystalTexture : layout(t2) = "Black";
sampler CrystalTexture_Sampler : layout(s2)
{
	Texture = CrystalTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D DepthBuffer : SceneDepth : layout(t4);
sampler DepthBuffer_Sampler : layout(s4)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D ScreenTexture : SceneColor : layout(t5) = "Transparent";
sampler ScreenTexture_Sampler : layout(s5)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D MaskTexture : layout(t6) = "White";
sampler MaskTexture_Sampler : layout(s6)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D DissipateTexture : layout(t7) = "White";
sampler DissipateTexture_Sampler : layout(s7)
{
	Texture = DissipateTexture;
	Filter = Linear;
	Address = Wrap;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Basic

float4 EffectCustomPixelShading(in VertexShadingOutput IN,in float4 screenCoord : SV_Position) : SV_Target
{
	// 1. Init Shading Parameters Struct
    EffectShadingParameters materialParamters;
    InitEffectShadingParameters(materialParamters);

	half4 OUT = half4(0,0,0,1);
	float4 effectColor = GetEffectColor(IN);
	float4 effectParam = GetEffectParam(IN);
	float2 DiffuseUV = IN.TexCoord.xy * DiffuseUVScrolling.xy + DiffuseUVScrolling.zw * effectParam.y;
	half4 diffuse = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, DiffuseUV);

	half3 V = IN.ViewDirection.xyz;
	half3 N = IN.WorldNormal.xyz;

	half NoV = dot(V,N);

	half2 DistortUV = IN.TexCoord.xy*CrystalUVScrolling.xy + CrystalUVScrolling.zw * effectParam.y;

	half3 innerLayerMask0 = SampleColorMap(CrystalTexture,CrystalTexture_Sampler, DistortUV);

	half DistortionMask = 1-innerLayerMask0;

	//Hack Tangent...
	half3 tangentV = 0.0;
	tangentV.x = dot( V,V);
	tangentV.yz = NoV;
	V = tangentV;

	DistortUV = DistortUV-V.xy/V.z *(DistortionMask)*DistortionIntensity;

	half innerLayerMask = SampleColorMap(CrystalTexture,CrystalTexture_Sampler, DistortUV).r;

	float3 targetColor = IceColor ;

	if(@_DISTORTION)
	{
		half2 spos = screenCoord.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;

		// hacked refraction
		half3 V = IN.ViewDirection;
		half3 R = cross(N, V);
		half3 D = cross(R, V);

		half3 VR = cross(float3(0, 1, 0), V);
		half3 VU = normalize(cross(V, VR));
		VR = cross(VU, V);

		half2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
		offset *= (500 / IN.TexCoord.z) * DistortionIntensity * innerLayerMask;

		targetColor = (SampleColorMap(ScreenTexture, ScreenTexture_Sampler, stc + offset ));         

		targetColor = lerp(targetColor ,IceColor ,Transparent);
	}

	OUT.rgb = lerp(targetColor , SurfaceColor.rgb * diffuse.xyz , SurfaceColor.a * diffuse.r);  //blend surface Color
	OUT.rgb = lerp(OUT.rgb ,innerLayerMask*innerLayerMask*RimIntensity* RimColor ,1-NoV);   // blend fake reflection <- Fresnel like

	half mask = 1;
	if(@_MaskEnable)
	{
		float2 maskTC = IN.TexCoord.xy * MaskUVScrolling.xy + effectParam.y * MaskUVScrolling.zw;
		mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).r;
	}

	if(@_DissipateEnable)
	{
		// return float4( effectParam.w,0,0,1);
		//Using Control By Life For Dissipated Control
		//Alpha will be used as Alpha as usual
		float dissipatedControl = effectParam.w;
		float alphaRef = effectColor.w;
		
		float2 dissipateTC = IN.TexCoord.xy * DissipateUVScolling.xy + effectParam.y * DissipateUVScolling.zw;
		float alpha = SampleColorMap(DissipateTexture, DissipateTexture_Sampler, dissipateTC).r;

		dissipatedControl = dissipatedControl * 2*(DissipateEdge + Softness + 0.5) - DissipateEdge - Softness;
		float rc1 =  1-smoothstep(dissipatedControl-Softness,dissipatedControl+Softness,alpha);
		float rc2 =  1-smoothstep(dissipatedControl+DissipateEdge-Softness,dissipatedControl+DissipateEdge+Softness,alpha);
		OUT.xyz = lerp(OUT.xyz, DissipateColor.xyz , 1-rc1 );//Lerp Dissipated Edge Color,AND premultiply Alpha
		OUT = float4(OUT.xyz,saturate((rc2-rc1 )* DissipateColor.w + rc1 * ModulateColor.w ) * alphaRef * mask * OUT.w);//Calc Alpha , ModulateColor.w Used Here
	
		OUT.xyz *= ModulateColor.xyz * effectColor.xyz * IN.VertexColor.xyz;
		// OUT.xyz *= ModulateColor.w * effectColor.w * mask;
		// OUT.w *= ModulateColor.w * effectColor.w * mask;
		OUT.rgb *= OUT.w;
	}
	else
	{
		OUT.xyz *= ModulateColor.xyz * effectColor.xyz * IN.VertexColor.xyz;
		OUT.xyz *= ModulateColor.w * effectColor.w * mask;
		OUT.w = ModulateColor.w * effectColor.w * mask;
	}

	materialParamters.BaseColor = OUT;

	EncodeHDRParticle(materialParamters.BaseColor);

	if((@_TRANSPARENT && !@_DISTORTION) || !@_DEFERRED_SHADING)
	{
		InitFogLightingEffect(IN,materialParamters.BaseColor);
	}

    return materialParamters.BaseColor;
}

#include "Effect_Shading.fxh"

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectCustomPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Distortion

float4 EffectDistortionShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT;
	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;

	float H = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy).r;
	H *= ModulateColor.x * GetEffectColor(IN).x;

	// hacked refraction
	float3 V = IN.ViewDirection;
	float3 N = IN.WorldNormal;
	float3 R = cross(N, V);
	float3 D = cross(R, V);

	float3 VR = cross(float3(0, 1, 0), V);
	float3 VU = normalize(cross(V, VR));
	VR = cross(VU, V);

	float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
	offset *= (500 / IN.TexCoord.z) * H;
	offset = offset * 0.5 + 0.5;

	EncodeFloatRG(offset.x, OUT.rg);
	EncodeFloatRG(offset.y, OUT.ba);

	return OUT;
}

technique Distortion
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectDistortionShading();
}

