//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Twisted.fx>
///	@path	~/System/Shader/Effect
///	@date	2014/11/10
/// @LastChange 2019/11/15
///	@brief	Effect shading model.

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader material
#pragma shader category("Effect")

//////////////////////////////////Properties////////////////////////////////////////
@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is masking map enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

@_Mask02Enable
{
	Name = Mask02Enable;
	Description = "Is masking map enabled?";
	AffectedTex = MaskTexture02;
	Visible = true;
};

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = true;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
};

//////////////////////////////////////////////////////////////////////////
// Effect shading model - Twisted
cbuffer Batch
{
	float4x3 World 				: World  ;
	float4 EffectColor			: Color  ;			// 	Effect color. Controlled by "Alpha By Life / Color By Life / Emission By Life etc."
	float4 EffectParam			: Param  ;			//	Effect param.	x->LifeTime y->Time w->Control By Life
	float4 SubUV				: Param2 ;			
	float4 cTintColor1			: TintColor1 ;		//	ParameterByLife. xyz->ParameterByLife
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float4 MaskUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
	float  BloomScale			 = 0.5;
	float  InvFadeDistance		 = 1;
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  _Padding				 = 0;
	float4 fadeColor			 = { 0, 0, 0, 0 };
	float  InvertCenter			= 0.5;
	float  InvertExtend			= 0.5;
	float4 InverterColor		= { 0.5, 0.5, 0.5, 1 };
	float  InverterScale		= 1.0;
	float  cFov					 = 1.0;
};

// Effect textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D TwistTexture : layout(t2) = "White";
sampler TwistTexture_Sampler : layout(s2)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture : layout(t3) = "White";
sampler MaskTexture_Sampler : layout(s3)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture02 : layout(t4) = "White";
sampler MaskTexture02_Sampler : layout(s4)
{
	Texture = MaskTexture02;
	Filter = Linear;
	Address = FromTexture;
};

texture2D DepthBuffer : SceneDepth : layout(t5);
sampler DepthBuffer_Sampler : layout(s5)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D ScreenTexture : SceneColor : layout(t6) = "Transparent";
sampler ScreenTexture_Sampler : layout(s6)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.
#include "Effect_Common.fxh"
#include "Effect.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect pixel shader - Twisted.

void PS_GetParameters(in VertexShadingOutput IN,inout EffectShadingParameters materialParamters)
{
	float4 OUT;
	float4 effectParam = GetEffectParam(IN);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + effectParam.y * DiffuseUVScolling.zw;
	float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + effectParam.y * TwistUVScolling.zw;
	float2 maskTC = IN.TexCoord.xy * MaskUVScolling.xy + EffectParam.y * MaskUVScolling.zw;
	float2 twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;

	diffuseTC += twist * TwistStrength;
	maskTC += twist * TwistStrength;

	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);

	float lerpValue = 0;
	float4 DColor = float4(1,0,1,0.2);
	OUT.xyz = (OUT.xyz * GetEffectColor(IN).xyz);

	if (@_MaskEnable)
	{
		float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).r;
		OUT.a *= mask;
	}

	if (@_Mask02Enable)
	{
		float mask02 = SampleColorMap(MaskTexture02, MaskTexture02_Sampler, IN.TexCoord.xy).r;
		OUT.a *= mask02;
	}

	if(@_DISTORTION)
	{	
		float2 spos = IN.HPosition.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;
		float4 screen = SampleColorMap(ScreenTexture, ScreenTexture_Sampler, stc);
		screen.rgb = EncodeGamma(screen.rgb);
		DColor.xyz = (screen.x + screen.y + screen.z)*(InverterScale*0.333);
		DColor.xyz += OUT.xyz * OUT.a;
		DColor = clamp(DColor,0,MAX_EMISSIVE_VALUE);
		lerpValue = smoothstep((InvertCenter-InvertExtend)*(InverterScale+1),(InvertCenter+InvertExtend)*(InverterScale+1),max(DColor.x,max(DColor.y,DColor.z)));
	}
	else
	{
		DColor.xyz = OUT.xyz * OUT.a;
		lerpValue = smoothstep(InvertCenter-InvertExtend,InvertCenter+InvertExtend,max(DColor.x,max(DColor.y,DColor.z)));
	}

	float4 MultipleColor = lerp(ModulateColor,InverterColor,(lerpValue));
	OUT *= MultipleColor;

	IN.TexCoord.w = saturate(IN.TexCoord.w); 

	OUT.xyz *= OUT.a;
	OUT.a = OUT.a * GetEffectColor(IN).a;
	materialParamters.BaseColor = OUT;

}

#include "Effect_Shading.fxh"

float4 EffectCustomPixelShading(in VertexShadingOutput IN,in float4 screenCoord : SV_Position) : SV_Target
{
	//Here Distortion is Different
	//Using ScreenColor to Invert Color 
	//Not Offset ScreenUV
	//Particle_Shading.fx
    return UnlitPixelShading(IN,screenCoord);
}

technique Shading
{
	VertexShader = EffectUnifyVertexShading();
	PixelShader = EffectCustomPixelShading();
}
