//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<hdrsky_star.fx>
///	@path	~/System/Shader/
///	@date	2014/10/27
///	@brief	star.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////
//%_HDR, system property

///////////////// const buffer //////////////////
cbuffer Batch
{
	//worldviewproj
	float4x3 World 						: World	;
};

cbuffer Shader
{
	float4 	cCloudLight					;
	float4 	cCloudAmbient				;
	// skyfog param [SkyFogBegin, SkyFogScale]
	float4  cFogParam					;
};

#define SkyFogBegin 	cFogParam.x
#define SkyFogScale 	cFogParam.y
#define CLOUDGSCALE		cCloudLight.w
#define CLOUDBLOOMSCALE cCloudAmbient.w

texture2D tCloudMap : layout(t0) = "Transparent";
sampler sCloudSampler : layout(s0, s1)
{
	Texture = tCloudMap;
	MinFilter = Linear;
	MagFilter = Linear;
	MipFilter = None;
	Address = Clamp;
	SRGB = false;
};

///////////////// io signatures //////////////////
struct a2v
{
	float4 Position 			: POSITION;
	float4 ColorAndIntensity	: COLOR;
	float3 Normal				: NORMAL;		//quad center to vertice, normalize
	float2 UV					: TEXCOORD0;
};

struct v2f
{
	float4 Position 		: SV_Position;
	float3 Color    		: COLOR;
	float2 UV				: TEXCOORDN;
	half3  WorldNormal		: TEXCOORDN;
	half3  WorldTangent		: TEXCOORDN;
	half3  WorldBinormal	: TEXCOORDN;
};

///////////////// general pass //////////////////
v2f CloudVS(a2v IN)
{
	v2f OUT;

	[@_HLSL]
	OUT = (v2f)0;

	//apply size
	float4 pos = IN.Position;

	// Position in screen space.
	float4 worldPos = float4(MulMat3x4(World, pos).xyz, 1);
	OUT.Position = MulMat4x4(ViewProj, worldPos);
	OUT.Position.z = OUT.Position.w;

	//Color uv
	OUT.Color.rgb = IN.ColorAndIntensity.rgb * IN.ColorAndIntensity.a * 10.0;
	OUT.UV = IN.UV;

	//normal
	OUT.WorldNormal = normalize(-pos.xyz);
	OUT.WorldTangent = normalize(cross(float3(0, 1, 0), OUT.WorldNormal));
	OUT.WorldBinormal = cross(OUT.WorldNormal, OUT.WorldTangent);

	return OUT;
}

float4 CloudPS(v2f IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	half4 SampleColor = tCloudMap.SampleRGBA(sCloudSampler, IN.UV).rgba;
	SampleColor.rg = 2.0 * SampleColor.rg - 1;
	half3 WorldNormal = normalize(IN.WorldTangent*SampleColor.r + IN.WorldBinormal*SampleColor.g + IN.WorldNormal);
	OUT.a = SampleColor.a;

	//sun lighting 1
	half G = lerp(0, 0.99, 1-SampleColor.b);
	half3 V = normalize(-IN.WorldNormal);
	float cosine = dot( SunDirection.xyz, V );
	float cosine2 = cosine * cosine;
	float F = 1.5 * (1 - G*G) / (2 + G*G);
	F *= (1 + cosine2) * pow(1 + G*G - 2*G*cosine, -1.5);
	OUT.rgb = cCloudLight.rgb * CLOUDGSCALE * F * IN.Color.rgb;
	float bloomScale = GetLuminance(OUT.rgb) * CLOUDBLOOMSCALE;

	//sun lighting 2
	half saturation = dot(SunDirection.xyz, -IN.WorldNormal)*0.2 + 0.8;
	half3 sunDir = normalize(1.5*SunDirection.xyz + IN.WorldNormal);
	half lighting = pow(0.5*dot(sunDir, WorldNormal)+0.5, 0.25);
	OUT.rgb += cCloudLight.rgb * lighting * SampleColor.b + cCloudAmbient.rgb;

	//with cloud color
	OUT.rgb *= IN.Color.rgb;

	// calc fog
	half skyY = V.y * -1.0 + 1.0;
	half fogRate = saturate((skyY-SkyFogBegin) * SkyFogScale);
	OUT.rgb = lerp(OUT.rgb, FogColor.rgb, fogRate);

	// screen color tint
	OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);

	OUT.rgb *= OUT.a;
	OUT.a = EncodeTransparentBloomFactor(OUT.a, bloomScale);

	return OUT;
}

///////////////// techniques //////////////////
technique CloudTech
{
	VertexShader = CloudVS();
	PixelShader = CloudPS();
}
///////////////// eof //////////////////
