//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<hdrsky_all.fx>
///	@path	~/System/Shader/
///	@date	2014/10/24
///	@brief	HDR SKY.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

///////////////// Properties //////////////////
@_HAS_GLOBALMAP
{
	Name = HasGlobalMap;
	AffectedTex = tSkyLayer;
	Description = "Do we has global skymap";
	Visible = true;
};

@_HAS_LOCALMAP
{
	Name = HasLocalMap;
	AffectedTex = tLocalSkyLayer;
	Description = "Do we has local skymap";
	Visible = true;
};
//%_HDR, system property

///////////////// const buffer //////////////////
#define BeginV cBeginLenUV.y
#define LenV cBeginLenUV.z
#define SkyFogBegin 	cFogParam.x
#define SkyFogScale 	cFogParam.y
#define LocalRate		cFogParam.z
//#define BloomScale 		cFogParam.w

cbuffer Batch
{
	//world
	float4x3 World 									: World	;

	float4   cBeginLenUV							;
	// skyfog param [SkyFogBegin, SkyFogScale]
	float4   cFogParam								;
	float	 cGlobalUBegin							;
	float	 cLocalUBegin							;
	float	 cGlobalURepeat							;
	float	 cLocalURepeat							;

	float 	 cGlobalVBegin							;
	float 	 cGlobalVLen							;
	float 	 cLocalVBegin							;
	float 	 cLocalVLen								;

	float    cGlobalHDR								 = 1.0;
	float    cLocalHDR								 = 1.0;
};

///////////////// samplers //////////////////
texture2D tSkyLayer : layout(t0) = "Black";
sampler sSkyLayer : layout(s0, s1)
{
	Texture = tSkyLayer;
	Filter = Linear;
	Address = Wrap;
};
texture2D tLocalSkyLayer : layout(t2) = "Transparent";
sampler sLocalSkyLayer : layout(s2, s3)
{
	Texture = tLocalSkyLayer;
	Filter = Linear;
	Address = Wrap;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 baseTC   : TEXCOORD0;
};

struct v2f_hdr
{
	float4 Position	: SV_Position;  //in projection space
	float3 mapTC	: TEXCOORD0;
	float fogT 		: TEXCOORD1;
};

///////////////// general pass //////////////////

v2f_hdr SkyHDRVS(a2v IN)
{
	v2f_hdr OUT;
	
	[@_HLSL]
	OUT = (v2f_hdr)0;

	// Position in unit sphere space.
	float4 vPos = IN.Position;
	float beginTheta = BeginV * PI * 0.5;
	float lenTheta = LenV * PI * 0.5;
	float theta = beginTheta + lenTheta * vPos.y;
	float sinTheta, cosTheta;
	sincos(theta, sinTheta, cosTheta);
	vPos.y = cosTheta;
	vPos.xz *= sinTheta;
	
	// Position in screen space.
	float4 worldPos = float4(MulMat3x4(World, vPos).xyz, 1);
	OUT.Position = MulMat4x4(ViewProj, worldPos);
	OUT.Position.z = (EPSILON*5)*OUT.Position.w;//es2.0 device doesn't support ZFunc=Equal/LessEqual.

	// TC
	OUT.mapTC.xy = IN.baseTC.xy;
	OUT.mapTC.z = BeginV + IN.baseTC.y * LenV;

	OUT.fogT = FogTransmittance(vPos.xyz * CameraInfo.y);
	return OUT;
}

float4 SkyHDRPS(v2f_hdr IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	//input
	float2 baseTC1 = IN.mapTC.xy;
	baseTC1.x = baseTC1.x * cGlobalURepeat + cGlobalUBegin;
	baseTC1.y = baseTC1.y * cGlobalVLen + cGlobalVBegin;
	float2 baseTC2 = IN.mapTC.xy;
	baseTC2.x = baseTC2.x * cLocalURepeat + cLocalUBegin;
	baseTC2.y = baseTC2.y * cLocalVLen + cLocalVBegin;
	float skyY = IN.mapTC.z;

	//texture
	if (@_HAS_GLOBALMAP)
	{
		float4 raw = tSkyLayer.SampleRGBA(sSkyLayer, baseTC1).rgba;
		float globalHDR = 1.0;
		[@_HDR]
		globalHDR = cGlobalHDR;
		OUT.rgba += float4(DecodeGamma(raw.rgb), raw.a) * (1.0 - LocalRate) * globalHDR;
	}
	
	if (@_HAS_LOCALMAP)
	{
		float4 raw = tLocalSkyLayer.SampleRGBA(sLocalSkyLayer, baseTC2).rgba;
		float localHDR = 1.0;
		[@_HDR]
		localHDR = cLocalHDR;
		OUT.rgba += float4(DecodeGamma(raw.rgb), raw.a) * LocalRate * localHDR;
	}
	
	// rain rate
	if (@_Weather == WEATHER_RAIN)
		OUT.rgb *= 1 - 0.8*RainRate;

	// calc fog
	if (!@_DEFERRED_SHADING && !@_VolumetricFog)
	{
		OUT.rgb = FogLighting(IN.fogT, OUT.rgb);
	}

	// screen color tint
	OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);

	//output hdr space
	EncodeHDR(OUT.rgb);

	OUT.rgb *= OUT.a;
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);

	return OUT;
}

///////////////// techniques ///////////
//ZEnable          = TRUE;
//ZWriteEnable     = FALSE;
//CullMode 		 = NONE;
//FOGENABLE 		 = FALSE;

//ALPHATESTENABLE  = FALSE;
//ALPHABLENDENABLE = FALSE;
technique SkyLayerHDRTech
{
	VertexShader = SkyHDRVS();
	PixelShader = SkyHDRPS();
}
///////////////// eof //////////////////
