//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GPUParticle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	GPU particle shader.

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

#pragma shader draw

@_SCATTER
{
	Name = Scatter;
	Description = "Scattering mode";
	Visible = true;
}

@_AERODYNAMIC
{
	Name = AeroDynamic;
	Description = "Enable Aero Dynamic Behaviour";
	Visible = false;
}

@_SNAPTOHEIGHTMAP
{
	Name = SnapToHeightMap;
	Description = "init position on heightmap";
	Visible = false;
}

// constants
cbuffer Batch
{
	float4x3 World 				: World;
	float4	SheltermapMinPos	:SheltermapMinPos;
	float4	SheltermapMaxPos	:SheltermapMaxPos;
};

cbuffer Shader
{
	float4x3 PositionTransform	;
	float4x3 VelocityTransform	;

	float4   PositionScaleMin   ;
	float4   PositionScaleMax   ;

	float4   VelocityScaleMin   ;
	float4   VelocityScaleMax   ;

	float4   VelocityScatter    ;

	// min life, max life, velocity scattering scale, 1 / particle buffer size;
	float4   Param0				;
	// min: size.xy, rotation, rotation rate.
	float4   Param1				;
	// max: size.xy, rotation, rotation rate.
	float4   Param2				;
	float4   InitRotParam0      ;
	float4   InitRotParam1   	;
};

#define InitRotAngleMin	(InitRotParam0.x / 180.0 * PI)
#define InitRotAngleMax	(InitRotParam0.y / 180.0 * PI)
#define InitRotAxiesMin	(float3(InitRotParam0.zw,InitRotParam1.x))
#define InitRotAxiesMax	(float3(InitRotParam1.yzw))

// Particle textures.
texture2D PositionTexture : layout(t0);
sampler PositionTexture_Sampler : layout(s0)
{
	Texture = PositionTexture;
	Filter = Point;
	Address = Wrap;
};

texture2D VelocityTexture : layout(t1);
sampler VelocityTexture_Sampler : layout(s1)
{
	Texture = VelocityTexture;
	Filter = Point;
	Address = Wrap;
};

texture2D RandomTexture : layout(t2);
sampler RandomTexture_Sampler : layout(s2)
{
	Texture = RandomTexture;
	Filter = Point;
	Address = Wrap;
};

texture2D tShelterMap: ShelterMap : layout(t3) = "Black";
sampler ShelterMap_Sampler : layout(s3)
{
	Texture = tShelterMap;
	Filter = Linear;
	Address = Clamp;
};


struct VertexAssemble
{
	float4	Position	: POSITION;
	float2	TexCoord	: TEXCOORDN;
};

struct SimpleVSOutput
{
	float4	Position	: SV_Position;
	float2	TexCoord	: TEXCOORDN;
};

struct ParticleOutput
{
	float4	Position	: SV_Target0;  // rgb: position, a: life time phase.
	float4	Velocity	: SV_Target1;  // rgb: velocity, a: life time total.
	float4	Property	: SV_Target2;  // size.xy, rotation, rotation rate.
									   // size.xy  rotation.xy 				  (when behaviour = aredynamic)
	float4	SecProperty	: SV_Target3;  // rotation.z angularVel.xyz			(when behaviour = aredynamic)
};

SimpleVSOutput SimpleVS(VertexAssemble IN)
{
	SimpleVSOutput OUT;
	OUT.Position = IN.Position;
	OUT.TexCoord = IN.TexCoord;
	return OUT;
}

float3 ProjectToHeightMap(float3 position)
{
	float2 uv = 1 - (position.xz - SheltermapMinPos.xz) * SheltermapMaxPos.xz;
	float sceneDepth = tShelterMap.SampleLevel(ShelterMap_Sampler, uv, 0).r;
	float sceneHeight = lerp(SheltermapMaxPos.y, SheltermapMinPos.y, sceneDepth);
	//sceneHeight += SheltermapMinPos.y;
	return float3(position.x,sceneHeight,position.z);
}

[target("RGBA32F")]
ParticleOutput SpawnPS(in float4 screenCoord : SV_Position)
{
	ParticleOutput OUT;

	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 tc = spos * Param0.w;

	// remap the sampling coord
	// assume particle buffer size = 512 and distribution map size = 256
	float idx = tc.y * 512 + tc.x;
	float2 tmp = float2(frac(idx), idx);
	float2 rtc = tmp * float2(326, 1.27344) + float2(-163.316, -0.636009);

	float3 rand0 = RandomTexture.SampleLevel(RandomTexture_Sampler, rtc + 0.7023 * GameTime, 0).xyz;
	float3 rand1 = RandomTexture.SampleLevel(RandomTexture_Sampler, rtc + 0.9077 * GameTime, 0).xyz;

	// position distribution
	float3 pos = PositionTexture.SampleLevel(PositionTexture_Sampler, rtc + 0.3013 * GameTime, 0).xyz;
	float3 posScale = lerp(PositionScaleMin.xyz, PositionScaleMax.xyz, rand0.x);
	pos = (pos * 2.0 - 1.0);
	OUT.Position = float4(MulMat3x4(PositionTransform, float4(pos, 1)).xyz, 0);
	if (@_SNAPTOHEIGHTMAP)
	{
		OUT.Position.xyz = ProjectToHeightMap(OUT.Position.xyz);
	}

	// velocity distribution
	float3 vel = VelocityTexture.SampleLevel(VelocityTexture_Sampler, rtc + 0.5057 * GameTime, 0).xyz;
	float3 velScale = lerp(VelocityScaleMin.xyz, VelocityScaleMax.xyz, rand0.y);
	vel = (vel * 2.0 - 1.0);
	OUT.Velocity.xyz = MulMat3x4(VelocityTransform, float4(vel, 1)).xyz;

	if (@_SCATTER)
	{
		float3 vec = OUT.Position - OUT.Velocity.xyz;
		vec = vec * lerp(1.0, 1.0 / length(vec), VelocityScatter.w);
		OUT.Velocity.xyz = vec * VelocityScatter.xyz;
	}

	// life time total
	float life = lerp(Param0.x, Param0.y, rand0.z);
	OUT.Velocity.w = life;

	// property : size.xy, rotation, rotation rate.
	OUT.Property = lerp(Param1, Param2, rand1.xxyz);
	OUT.Property.w *= life;

	OUT.SecProperty = float4(0,0,0,0);

	if (@_AERODYNAMIC)
	{
		//property.zw:rotation.xy
		//secProperty.x:rotation.z  yzw:angularVel
		float angle = lerp(InitRotAngleMin,InitRotAngleMax,rand0.x);
		float3 axies = lerp(InitRotAxiesMin,InitRotAxiesMax,rand1.xyz);
		if (dot(axies,axies) == 0)
			axies = float3(1,0,0);
		axies = normalize(axies);
		OUT.Property.zw = axies.xy * sin(angle * 0.5);
		OUT.SecProperty.x = axies.z * sin(angle * 0.5);
	}

	return OUT;
}

technique Spawn
{
	VertexShader = SimpleVS();
	PixelShader = SpawnPS();
}


