//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GPUParticle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	GPU particle shader.

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"
#include "Noise.fxh"

#pragma shader draw

@_TURBULENCE
{
	Name = Turbulence;
	Description = "Enable Turbulence";
	Visible = true;
}

@_VELOCITY
{
	Name = Velocity;
	Description = "Facing Vecloity";
	Visible = true;
}

@_COLLISION
{
	Name = Collision;
	Description = "Enable Collision";
	Visible = true;
}

@_DYNAMICS
{
	Name = Dynamics;
	Description = "Enable Dynamics";
	Visible = true;
}

@_AERODYNAMIC
{
	Name = AeroDynamic;
	Description = "Enable Aero Dynamic Behaviour";
	Visible = false;
}

cbuffer Batch
{
	float4	 FoliageCenter      : FoliageCenter		;
	float4	SheltermapMinPos	:SheltermapMinPos;
	float4	SheltermapMaxPos	:SheltermapMaxPos;
};

cbuffer Shader
{
	float4   Param0				;
	float4   Param1				;
	float4   Param2				;
	float4   Param3				;
	float4   Param4				;
	float4   Param5				;
	float4   Param6				;
	float4 	 Param7				;
	float4   vWBasisX			;
	float4   vWBasisY			;
	float4   vWBasisZ			;
	float4   vWOrthoZ			;
};

#define ElapsedTime			(Param0.x)
#define DeltaTime			(Param0.y)
#define UpdateTime			(Param0.z)
#define InvBufferSize		(Param0.w)

#define Turbulence			(Param1.x)
#define SpaceFreqency		(Param1.y)
#define TimeFreqency		(Param1.z)
#define Phase				(Param1.w)

#define Elasticity			(Param2.x)
#define CollideDistance		(Param2.y)
#define KillVelocity		(Param2.z)
#define WorldWindRatio		(Param2.w)

#define Gravity				(Param3.xyz)
#define DragCoefficient		(Param3.w)
#define Mass				(Param3.w)

#define Attractor			(Param4.xyz)
#define Wind				(Param4.xyz)
#define AttractorStrength	(Param4.w)
#define FrictionFactorPre	(Param4.w)
#define AttractorScale		(Param5.xyz)
#define AttractorPower		(Param5.w)
#define FrictionFactorRatio	(Param5.x)
#define AeroSize			(Param5.y)
#define AirDensityRatio		(Param5.z)
#define FrictionFactorAngular (Param5.w)
#define KineticFriction		(Param6.x)
#define GroundYOffset		(Param6.y)
#define PenetrationDepthToKill	(Param6.z)
#define WindNoise			 (Param6.w)
#define CollisionInertia	(Param7.x)
#define ButtomY				(Param7.y)

// Particle textures.
texture2D PositionTexture : layout(t0);
sampler PositionTexture_Sampler : layout(s0)
{
	Texture = PositionTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D VelocityTexture : layout(t1);
sampler VelocityTexture_Sampler : layout(s1)
{
	Texture = VelocityTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D PropertyTexture : layout(t2);
sampler PropertyTexture_Sampler : layout(s2)
{
	Texture = PropertyTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D SecPropertyTexture : layout(t3);
sampler SecPropertyTexture_Sampler : layout(s3)
{
	Texture = SecPropertyTexture;
	Filter = Point;
	Address = Clamp;
};


texture2D tLocalHeightmap: LocalHeightmap : layout(t4);
sampler LocalHightMap_Sampler : layout(s4)
{
	Texture = tLocalHeightmap;
	Filter = Linear;
	Address = Clamp;
};


texture2D tFoliageMap : FoliageMap : layout(t5) = "Black";
sampler sFoliageMapSampler : layout(s5)
{
	Texture = tFoliageMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

texture2D tShelterMap: ShelterMap : layout(t7) = "Black";
sampler ShelterMap_Sampler : layout(s7)
{
	Texture = tShelterMap;
	Filter = Point;
	Address = Clamp;
};


struct VertexAssemble
{
	float4	Position	: POSITION;
	float2	TexCoord	: TEXCOORDN;
};

struct SimpleVSOutput
{
	float4	Position	: SV_Position;
	float2	TexCoord	: TEXCOORDN;
};

struct ParticleOutput
{
	float4	Position	: SV_Target0;  // rgb: position, a: life time phase.
	float4	Velocity	: SV_Target1;  // rgb: velocity, a: life time total.
	float4	Property	: SV_Target2;  // size.xy, rotation, rotation rate.
									   // size.xy, velocity.x, velocity.z.    (when facing = Velocity)
									   // size.xy  rotation.xy 				  (when behaviour = aredynamic)
	float4	SecProperty	: SV_Target3;  // rotation.z angularVel.xyz			(when behaviour = aredynamic)					   
};

// (worldPostion - camPos) -> (screen coord xy, linear depth)
float3 CalcSreenPosAndDepth(float3 pos)
{
	float3 scDepth = float3(dot(pos, vWBasisX.xyz), dot(pos, vWBasisY.xyz), dot(pos, vWOrthoZ.xyz));
	scDepth *= float3(vWBasisX.w, vWBasisY.w, vWOrthoZ.w);
	scDepth.xy = scDepth.xy / scDepth.z + ScreenInfo.xy * 0.5;
	return scDepth;
}

// (screen coord xy, linear depth) -> (worldPostion - camPos)
float3 CalcWorldPos(float3 scDepth)
{
	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * scDepth.x) + (vWBasisY.xyz * scDepth.y);
	pos *= scDepth.z;
	return pos;
}

float3 ProjectToHeightMap(float3 position)
{
	float2 uv = 1 - (position.xz - SheltermapMinPos.xz) * SheltermapMaxPos.xz;
	float2 clamped = saturate(uv);
	if (any(uv != clamped))
		return  float3(position.x,ButtomY,position.z);
	float sceneDepth = tShelterMap.SampleLevel(ShelterMap_Sampler, uv, 0).r;
	float sceneHeight = lerp(SheltermapMaxPos.y, SheltermapMinPos.y, sceneDepth);
	//sceneHeight += SheltermapMinPos.y;
	return float3(position.x,max(sceneHeight + GroundYOffset,ButtomY),position.z);
}

SimpleVSOutput SimpleVS(VertexAssemble IN)
{
	SimpleVSOutput OUT;
	OUT.Position = IN.Position;
	OUT.TexCoord = IN.TexCoord;
	return OUT;
}

void UpdateRegular(float3 position,float3 velocity,float4 property,out float3 newPosition,out float3 newVelocity,out float4 newProperty,inout float newLife)
{
	newVelocity = velocity;
	newProperty = property;
	newPosition = position;

	if (@_TURBULENCE)
	{
		float3 noiseVelocity = CurlNoise(newPosition.xyz * SpaceFreqency, ElapsedTime * TimeFreqency + Phase);
		newVelocity.xyz += normalize(noiseVelocity) * Turbulence;
	}

	// update position
	newPosition = float4(newPosition.xyz + newVelocity.xyz * DeltaTime,0.0f);

	newPosition.xyz += newVelocity.xyz * DeltaTime;

	if (@_VELOCITY)
	{
		float3 velocityDir = normalize(newVelocity.xyz);
		float ysign = velocityDir.y < 0 ? -1 : 1;
		newProperty.zw = (velocityDir.xz + 2) * ysign;
	}

	if (@_DYNAMICS)
	{
		// gravity & drag
		float3 force = Gravity - newVelocity * DragCoefficient;

		// attractor
		float3 attraction = (Attractor - newPosition) * AttractorScale;
		float distSqr = max(dot(attraction, attraction), 0.01);
		float strength = AttractorStrength * pow(distSqr, AttractorPower);
		force += attraction * strength;

		newVelocity += force * DeltaTime;
	}
	
	if (@_COLLISION)
	{
		float3 pixelPosition = ProjectToHeightMap(newPosition);

		float3 positionH = ProjectToHeightMap(newPosition + float3(1.0/256,0,0));
		float3 positionV = ProjectToHeightMap(newPosition + float3(0,0,1.0/256));
		float3 pixelNormal = normalize(cross(positionV - pixelPosition, positionH - pixelPosition));


		float3 collideStep = newVelocity * (DeltaTime * 0.6);
		float threshold = max(dot(collideStep, collideStep), CollideDistance);
		float3 dist = pixelPosition - newPosition;

		if (newPosition.y - pixelPosition.y < threshold )//&& dot(pixelNormal, newVelocity) < 0)
		{
			newVelocity = reflect(newVelocity, pixelNormal) * Elasticity;

			// kill the particle if the bounced velocity is too small
			if (dot(newVelocity, newVelocity) < KillVelocity)
			{
				newLife = 1.0;
			}
		}
	}
}
/*
//https://en.wikipedia.org/wiki/Euler_angles Tait–Bryan angles X1Y2Z3
float3x3 RotationMatrixFromEuler(float3 rotation)
{
	float c1 = cos(rotation.x);
	float c2 = cos(rotation.y);
	float c3 = cos(rotation.z);
	float s1 = sin(rotation.x);
	float s2 = sin(rotation.y);
	float s3 = sin(rotation.z);
	return float3x3(
		c2*c3,			-c2*s3,			s2,
		c1*s3+c3*s1*s2, c1*c3-s1*s2*s3,	-c2*s1,
		s1*s3-c1*c3*s2,	c3*s1+c1*s2*s3,	c1*c2
	);
}*/

float3x3 RotationMatrixFromAxiesAngle(float3 axies,float angle)
{
	float x = axies.x;
	float y = axies.y;
	float z = axies.z;
	float sinTheta = sin(angle);
	float cosTheta = cos(angle);
	return float3x3 (
		cosTheta + x*x*(1- cosTheta),			x * y * (1- cosTheta) - z * sinTheta,	x * z * (1- cosTheta) + y * sinTheta,
		x * y * (1 - cosTheta) + z * sinTheta,	cosTheta + y * y * ( 1 - cosTheta),		y * z * (1 - cosTheta) - x * sinTheta,
		x * z * (1 - cosTheta) - y * sinTheta,	y * z * (1 - cosTheta) + x * sinTheta,	cosTheta * z * z * (1 - cosTheta)
	);
}

//https://www.gregslabaugh.net/publications/euler.pdf
float3 EulerFromRotationMatrix(float3x3 mat)
{
	float3 rotation = float3(0,0,0);
	if (abs(mat[0][2]) != 1)
	{
		float y1 = asin(mat[0][2]);
		float cy1 = cos(y1);
		float y2 = PI - y1;
		float cy2 = cos(y2);
		float z1 = atan2(-mat[0][1]/cy1,mat[0][0]/cy1);
		float z2 = atan2(-mat[0][1]/cy2,mat[0][0]/cy2);
		float x1 = atan2(-mat[1][2]/cy1,mat[2][2]/cy1);
		float x2 = atan2(-mat[1][2]/cy2,mat[2][2]/cy2);
		rotation = float3(x1,y1,z1);
	} else {
		rotation.x = 0.0f;
		if (mat[0][2] > 0)
		{
			rotation.y = PI / 2;
			rotation.z = rotation.x + atan2(mat[2][1],-mat[2][0]);
		} else {
			rotation.y = - PI / 2;
			rotation.z = -rotation.x + atan2(-mat[2][1],mat[2][0]);
		}
	}
	return rotation;
}

#include "quaternion.fxh"

float4 quaternion_from_float3(float3 value)
{
	float w = sqrt(1.0 - saturate(dot(value,value)));
	return float4(value,w);
}

float3 quaternion_to_float3(float4 quat)
{
	return quat.xyz;
}

float4 apply_angularVelocity(float4 orientation,float3 angularVel,float deltaTime)
{
	float magnitude = length(angularVel);
	if (magnitude == 0)
		return orientation;
	float3 axis = normalize(angularVel);
	float angle = magnitude * deltaTime;
	float4 rotation = rotate_angle_axis(angle,axis);
	return qmul(rotation,orientation);
}

float3 get_interactive_wind(float3 position,float groundHeight)
{
	float2 uv = (position.xz-FoliageCenter.xy)/128.0+0.5;
	float2 origin_wind = tFoliageMap.SampleLevel(sFoliageMapSampler, uv, 0).xy;
	float effective_height = 0.8f;
	float weight = 1.0f - saturate((position.y - groundHeight) / effective_height);
	return float3(origin_wind.x,-length(origin_wind),origin_wind.y) * weight;
}

void UpdateAeroDynamic(float deltaTime,float3 position,float3 velocity,float4 property,float4 secProperty,out float3 newPosition,out float3 newVelocity,out float4 newProperty,out float4 newSecProperty,inout float newLife)
{
	//deltaTime = deltaTime * 0.1f;
	//velocity = float3(0,0.1,0);
	newVelocity = velocity;
	newPosition = position;
	newProperty = property;
	newSecProperty = secProperty;
	
	float4 orientation = quaternion_from_float3(float3(property.zw,secProperty.x));
	
	float3 angularVel = secProperty.yzw;
	
	///parameter block
	float3 gravity = Gravity;
	float3 wind = WindParam.xyz * WindParam.w * WorldWindRatio + Wind;
	float mass = Mass;
	float collisionInertia = CollisionInertia;
	float frictionFactorPre = FrictionFactorPre;
	float frictionFactorAngular = FrictionFactorAngular;
	float frictionFactorRatio = FrictionFactorRatio;
	float len = AeroSize;
	float halfLen = len * 0.5;
	float densityRatio = AirDensityRatio;
	/////////////////////////////////////////////////////
	//////////////wind fade on the ground////////////////
	
	//float3 rand = RandomTexture.SampleLevel(RandomTexture_Sampler, position.xy + 0.7023 * GameTime, 0).xyz;
	//wind += rand * 1.5f;
	///////////////////////////////////////
	float worldHeightX = -1000;
	float worldHeightZ = -1000;
	float worldHeightnX = -1000;
	float worldHeightnZ = -1000;
	float centerHeight = -1000;
	
	if (true)
	{

		worldHeightX = ProjectToHeightMap(newPosition + float3(len,0,0)).y;
		worldHeightZ = ProjectToHeightMap(newPosition + float3(0,0,len)).y;
		worldHeightnX = ProjectToHeightMap(newPosition + float3(-len,0,0)).y;
		worldHeightnZ = ProjectToHeightMap(newPosition + float3(0,0,-len)).y;
		centerHeight = ProjectToHeightMap(newPosition).y;
	}

	float particleHeight = max(newPosition.y - centerHeight,0);
	bool mayCollide = particleHeight < 0.5 * len;//may have collision
	
	//float wind_weight = 1.0f;
	if (true)
	{
	float wind_weight = saturate((newPosition.y - (centerHeight)) / 0.5f);
	float curl = CurlNoise(newPosition,GameTime * 0.1);
	//wind += (curl * noise) * ( 1- wind_weight);
	wind += curl * WindNoise;
	wind = wind * wind_weight;
	float3 interactiveWind = get_interactive_wind(newPosition,centerHeight) * 7000.0f;
	float noise = cnoise(float3(newPosition.x,newPosition.y,GameTime * 0.01));
	wind += (wind * noise) * ( 1- wind_weight);
	//angularVel.y += interactiveWind.y;
	wind += interactiveWind;
	}
	
	
	float densityAir = mass / len * densityRatio;
	
	float3 velocityInAir = velocity - wind;
	float velocityInAirLength = length(velocityInAir);
	bool hasVelocity = velocityInAirLength > 0.001f ;
	
	float3x3 rotationMat = quaternion_to_matrix3(orientation);
	float3 localY = mul(rotationMat,float3(0,1,0));
	float3 dirVelocityInAir = float3(0,0,0);
	if (hasVelocity)
	{
		dirVelocityInAir = normalize(velocityInAir);
	}
	
	float3 friction = float3(0,0,0);
	float3 groundFriction = float3(0,0,0);
	float3 lift = float3(0,0,0);
	float3 dirLift = float3(0,0,0);
	float3 dirLiftSide = float3(0,0,0);
	float cosAttackAngle = 0;
	float sinAttackAngle = 0;
	
	////////////////////////////  linear acceleration /////////////////////////////////////////////

	if (hasVelocity)
	{//friction
		float projVelocityPre = dot(localY,velocityInAir);
		float3 velocityPre = projVelocityPre * localY;//velocity alone localY
		float3 velocityPar = velocityInAir - velocityPre;//velocity alone local XZ plane
		float lenVelocityPar = length(velocityPar);
		float frictionPre = frictionFactorPre * abs(projVelocityPre) * mass;
        float frictionPar = frictionFactorPre * frictionFactorRatio * lenVelocityPar * mass;
		if (lenVelocityPar < 0.001f)
			lenVelocityPar = 1.0f;//avoid devide by zero;
		float3 dirVelocityPre = float3(0,0,0);
		float lenVelocityPre = length(velocityPre);
		if (lenVelocityPre < 0.001f)
			lenVelocityPre = 1.0f;//avoid devide by zero;
		friction = -((frictionPre * velocityPre / lenVelocityPre) + (velocityPar * frictionPar / lenVelocityPar));
	}
	bool hasLift = (abs(dot(dirVelocityInAir,localY)) < 0.99f);
	if (hasVelocity && hasLift)
	{//lift
		float3 realUp = localY;
		
		{
			if (dot(dirVelocityInAir,localY) < 0)
				realUp = -localY;
			float3 sideDir = cross(realUp,dirVelocityInAir);
			sideDir = normalize(sideDir);
			dirLiftSide = sideDir;
			dirLift = cross(dirVelocityInAir,sideDir);
		}
		
		cosAttackAngle = dot(dirLift,realUp);
		sinAttackAngle = length(cross(dirLift,realUp));
		float flift = densityAir * len * PI * velocityInAirLength * velocityInAirLength * cosAttackAngle;
		lift = dirLift * flift;
	}
	
	float3 totalLinearAccelaration = (lift + friction) / mass + gravity;
	//if (mayCollide)
	//	totalLinearAccelaration = gravity;
	
	////////////////////////////  angular acceleration /////////////////////////////////////////////
	float3 frictionAngularAcc = float3(0,0,0);
	{//rotational friction
		frictionAngularAcc = -angularVel * frictionFactorAngular;
	}
	float3 aeroAngularAcc = float3(0,0,0);
	//if (!mayCollide)
	{
		aeroAngularAcc = dirLiftSide * 3 * densityRatio * PI * (velocityInAirLength * velocityInAirLength) / len * cosAttackAngle * sinAttackAngle;
	}
	////////////////////////////////////////////////////////////////////////////////////////////////
	
	newVelocity += totalLinearAccelaration * deltaTime;
	float3 newAngularVel = angularVel + (frictionAngularAcc + aeroAngularAcc) * deltaTime;
	//newAngularVel = angularVel;
	float4 newOrientation = orientation;
	//rotationMat = quaternion_to_matrix3(orientation);
	//localY = mul(rotationMat,float3(0,1,0));
	
	//velocity based collision constrains
	{
		float3 groundDirX = float3(2 * len,clamp(worldHeightX - worldHeightnX, -len, len),0);
		groundDirX = normalize(groundDirX);
		float3 groundDirZ = float3(0,clamp(worldHeightZ - worldHeightnZ, -len, len),2 * len);
		groundDirZ = normalize(groundDirZ);
		float3 groundNormal = cross(groundDirZ,groundDirX);
		
		// we assume leaf is a dist with radius = size
		//get lowest point P on leaf to get penetration depth
		
		//we assume contact normal is always equal to worldSpace Y since we can only get pentrationDepth alone world Y
		
		float penetrationDepth = newPosition.y - 0.0 - centerHeight;
		float3 penetrateOffset = float3(0,-0.0,0);
		float3 contactNormal = groundNormal;
		float normalDiff = dot(contactNormal,localY);
		if (normalDiff < 0.99)
		{
			float3 forward = (1,0,0);
			forward= cross(localY,float3(0,1,0));
			
			forward = normalize(forward);
			float3 side = normalize(cross(localY,forward))* halfLen;
			if (side.y > 0)
				side = -side;
			penetrateOffset = side;
			float terrainHeightAtPoint = ProjectToHeightMap(newPosition + side).y;
			penetrationDepth = (newPosition.y + side.y) - terrainHeightAtPoint;
		}
		
		float alpha = max(normalDiff - 0.99,0);
		alpha *= 1.0 / (1.0 - 0.99);
		penetrateOffset = lerp(penetrateOffset,float3(0,0,0),alpha);
		penetrationDepth = lerp(penetrationDepth,newPosition.y - 0.0 - centerHeight,alpha);
		
		float invMass = 1.0/ mass;
		float I = collisionInertia;
		float invI = 1.0/I;
		float3 v1 = newVelocity + cross(newAngularVel,penetrateOffset);
		
		float lambdaN = 0;
		if (penetrationDepth < 0)// do penetrate
		{
			

			v1 = newVelocity + cross(newAngularVel,penetrateOffset);
			
			//penetration-avoid constrain
			float3 crossRN = cross(penetrateOffset,contactNormal);
			float effectiveMass = 1.0 / (invMass + dot(crossRN,crossRN) * invI );
			float constrain = dot (v1,contactNormal) + dot(crossRN,newAngularVel);
			float lambda = effectiveMass * -constrain;
			lambdaN = lambda;
			
			newVelocity = newVelocity + (contactNormal * lambda * invMass);
			newAngularVel = newAngularVel + (crossRN * invI * lambda );
			//if (penetrationDepth < 0)
			float beta = 0.5;
			newVelocity.y += min(0.2,(beta/deltaTime) * max(0,centerHeight - (newPosition.y + newVelocity.y * deltaTime)));//position constrain
			
			
			v1 = newVelocity + cross(newAngularVel,penetrateOffset);
			//friction
			float frictionCoff = KineticFriction;
			float3 velOnSurface = v1 - (dot(newVelocity,contactNormal) * contactNormal);
			if (length(velOnSurface) > 0.001)
			{
			float3 u = normalize(velOnSurface);
			float3 crossRU = cross(penetrateOffset,u);
            float effectMassFric =1.0 / (invMass + dot(crossRU,crossRU) * invI );
            float constraintFric = dot(v1,u);
			float lambdaFric = -effectMassFric * constraintFric;
			lambdaFric = clamp(lambdaFric, - frictionCoff * lambdaN,frictionCoff * lambdaN);
			
			newVelocity = newVelocity + (u * lambdaFric * invMass);
			newAngularVel = newAngularVel + (crossRU * invI * lambdaFric );
			}
			
		}
		
		
		//friction
		if (false && penetrationDepth <= halfLen)
		{
			float frictionCoff = 0.9;
			float maxFriction = 0.9;
			float3 velOnSurface = v1 - (dot(newVelocity,contactNormal) * contactNormal);
			float forceOnSurface = dot(gravity * mass,-contactNormal) * saturate(- penetrationDepth * 5);
			float3 fricForce = -(velOnSurface) * clamp(forceOnSurface * frictionCoff,0,maxFriction);
			float3 angularAcc = cross(penetrateOffset,fricForce);
			newAngularVel = newAngularVel + angularAcc;
				
			float3 fricVel = -velOnSurface * clamp(forceOnSurface * frictionCoff,0,maxFriction);
				
			newVelocity = newVelocity + fricVel * normalDiff * normalDiff;
			////newVelocity = newVelocity - velOnSurface;
		}
		
		if (penetrationDepth < -PenetrationDepthToKill)
			newLife = 1.0;
		
		newVelocity *= 0.99;
		newAngularVel *= 0.98;
	}
	newOrientation = apply_angularVelocity(orientation,newAngularVel,deltaTime);
	newPosition = position + newVelocity * deltaTime;
	
	newProperty.zw = newOrientation.xy;
	newSecProperty.x = newOrientation.z;
	newSecProperty.yzw = newAngularVel.xyz;
}

[target("RGBA32F")]
ParticleOutput UpdatePS(in float4 screenCoord : SV_Position)
{
	ParticleOutput OUT;

	float2 spos = screenCoord.xy;
	[@_D3D9]
	spos += 0.5;
	float2 tc = spos * InvBufferSize;

	float4 position = PositionTexture.SampleLevel(PositionTexture_Sampler, tc, 0);
	//if (position.w > 1.0)
	//{
	//	OUT.Position = position;
	//	OUT.Velocity = 0.0;
	//	OUT.Property = 0.0;
	//	OUT.SecProperty = 0.0;
	//	return OUT;
	//}

	float4 velocity = VelocityTexture.SampleLevel(VelocityTexture_Sampler, tc, 0);
	float4 property = PropertyTexture.SampleLevel(PropertyTexture_Sampler, tc, 0);
	
	float newLife = position.w + DeltaTime / velocity.w;
	float3 newPosition = position.xyz;
	float3 newVelocity = velocity.xyz;
	float4 newProperty = property;
	float4 newSecProperty = float4(0,0,0,0);
	if (@_AERODYNAMIC)
	{
		float4 secProperty = SecPropertyTexture.SampleLevel(SecPropertyTexture_Sampler, tc, 0);
		newSecProperty = secProperty;
		//UpdateAeroDynamic(DeltaTime,position.xyz,velocity.xyz,property,secProperty,newPosition,newVelocity,newProperty,newSecProperty,newLife);
		
		float interval = 1.0/60.0;
		for (int i = 0 ; i < UpdateTime;i++)
		{
			UpdateAeroDynamic(interval,position.xyz,velocity.xyz,property,secProperty,newPosition,newVelocity,newProperty,newSecProperty,newLife);
			position.xyz = newPosition;
			velocity.xyz = newVelocity;
			property = newProperty;
			secProperty = newSecProperty;
		}
	} else {
		UpdateRegular(position.xyz,velocity.xyz,property,newPosition,newVelocity,newProperty,newLife);
	}
	
	OUT.Position = float4(newPosition, newLife);
	OUT.Velocity = float4(newVelocity, velocity.w);
	OUT.Property = newProperty;
	OUT.SecProperty = newSecProperty;

	return OUT;
}

technique Update
{
	VertexShader = SimpleVS();
	PixelShader = UpdatePS();
}

