//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Failure.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#pragma shader category("Editor")

#define USE_LIGHTMAP			0
#define USE_SPECULAR			0

#include "Constants.fxh"

#include "Common.fxh"  
#include "BRDF.fxh"
#include "Lighting.fxh"

cbuffer Batch
{ 
	float4x3 World 					: World 			;
	float4 cParamter				: Param				;
	float4 cParamter2				: Param2			;
};

cbuffer Shader
{
	float3 Color					 < IsColor = "true"; > = {1, 0, 0};
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.
 
	float4 		WorldPosition 	: TEXCOORDN;			//  World Position

	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
	
	float4 		DiffLighting 	: TEXCOORDN;			//  PerVertex Diffuse Lighting.

	[@_TANGENT]
	half3		WorldTangent	: TEXCOORDN;			//	Output tangents in world space for normal mapping.
	[@_TANGENT]
	half3		WorldBinormal	: TEXCOORDN;			//	Output binormal in world space for normal mapping.

	[@_CastSunShadow]
	float4 		LitUVZ			: TEXCOORDN; 
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD5;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

	float3 LocalNormal;

	[@_TANGENT]
	float3 LocalTangent;

	[@_TANGENT]
	float3 LocalBinormal;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);

	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}

void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal.xyz = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent.xyz = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal.xyz = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
		}
		/*
		// Hard skinning.
		if (@_DF_SKIN == 2)
		{
			// Fetch.
			float4 Mx, My, Mz;
			GetSkeletonMatrix4x3(i[0], Mx, My, Mz);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalBinormal);
		}
		// Dual-quaternion skinning.
		if (@_DF_SKIN == 3)
		{
			// Fetch.
			float4 Q0[2], Q1[2], Q2[2], Q3[2];
			GetSkeletonDualQuaternion(i[0], Q0[0], Q0[1]);
			GetSkeletonDualQuaternion(i[1], Q1[0], Q1[1]);
			GetSkeletonDualQuaternion(i[2], Q2[0], Q2[1]);
			GetSkeletonDualQuaternion(i[3], Q3[0], Q3[1]);
			// Amend orientations.
			w *= float4(1.f, sign(dot(Q0[0], Q1[0])), sign(dot(Q0[0], Q2[0])), sign(dot(Q0[0], Q3[0])));

			// Blend.
			float4 Qn = LinearBlend(w, Q0[0], Q1[0], Q2[0], Q3[0]);
			float4 Qd = LinearBlend(w, Q0[1], Q1[1], Q2[1], Q3[1]); 

			// Apply.
			INTM.LocalPosition.xyz = DualQuaternionTransformP(Qn, Qd, INTM.LocalPosition.xyz);
			INTM.LocalNormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = DualQuaternionTransformN(Qn, Qd, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalBinormal);
		}
		*/
	} 
} 

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	//position
	if (@_INSTANCED)
	{
		OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}

	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);

	//TexCoord
	OUT.TexCoord = IN.TexCoord.xyxy;

	//normal
	if (@_INSTANCED)
	{
		OUT.WorldNormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
		if (@_TANGENT)
		{
			OUT.WorldTangent.xyz  = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalTangent));
			OUT.WorldBinormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalBinormal));	
		}
	}
	else
	{
		OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
		if (@_TANGENT)
		{
			OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, INTM.LocalTangent));
			OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalBinormal));	
		}
	}

	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;

	//EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;

	OUT.DiffLighting.xyz = half3(0, 0, 0);

	//PerVertex Diffuse Lighting
	OUT.DiffLighting.rgb = LightingVS(OUT.WorldPosition.xyz, OUT.WorldNormal.xyz);

	//Fog T
	OUT.DiffLighting.w = FogTransmittance(OUT.ViewDirection.xyz);

	[@_CastSunShadow]
	OUT.LitUVZ = MulMat4x4(ShadowViewProjTexs0, float4(OUT.WorldPosition.xyz,1));
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput SurfaceVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float4 SurfacePixelShading(in VertexShadingOutput IN) : SV_Target
{
	return float4(Color, 0);
}

technique Shading
{
	VertexShader = SurfaceVertexShading();
	PixelShader = SurfacePixelShading();
}

/////////////////////////////////////GBuffer/////////////////////////////////////

struct GBufferOUTPUT
{
	float4 BaseColor 	: SV_Target0;//.a is roughness & Skin
	float4 Normal 		: SV_Target1;//.a is metallic
};

GBufferOUTPUT GBufferPixelShading(in VertexShadingOutput IN)
{
	GBufferOUTPUT OUT;
	
	OUT.BaseColor = float4(1,0,0,0);
	OUT.Normal = float4(1,1,1,1);

	return OUT;
}

technique GBuffer
{
	VertexShader = SurfaceVertexShading();
	PixelShader = GBufferPixelShading();
}

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		SlopeDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//position
	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalPosition.xyz), 1);
	}
	else
	{
		WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	}
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);
	
	//slope depth
	float3 WorldNormal;
	if (@_INSTANCED)
	{
		WorldNormal = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, INTM.LocalNormal));
	}
	else
	{
		WorldNormal = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	}
	float4 SlopePos = MulMat4x4(ViewProj, WorldPosition + float4(WorldNormal, 1));
	OUT.SlopeDepth = SlopePos.zw;

	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	float slopeDepth = IN.SlopeDepth.x / IN.SlopeDepth.y;
	slopeDepth = abs(slopeDepth - depth);
	// depth += SHADOW_BIAS + SHADOW_SLOPEBIAS * slopeDepth;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

float4 DebugPixelShading(in VertexShadingOutput IN) : SV_Target
{
	return float4(Color, 0);
}

technique Debug
{
	VertexShader = SurfaceVertexShading();
	PixelShader = DebugPixelShading();
}


//////////////////////////////////////////////////////////////////////////
// Clip
float4 ClipPixelShading(in VertexShadingOutput IN) : SV_Target
{
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = SurfaceVertexShading();
	PixelShader = ClipPixelShading();
}


//////////////////////////////////////////////////////////////////////////
// Wireframe
[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].TexCoord = float4(1, 0, 0, 0);
	OUT.Append(IN[0]);
	IN[1].TexCoord = float4(0, 1, 0, 0);
	OUT.Append(IN[1]);
	IN[2].TexCoord = float4(0, 0, 1, 0);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 WireframePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 difference = fwidth(IN.TexCoord.xyz);
	float3 distance = IN.TexCoord.xyz / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-1.5 * d * d);
	return float4(cParamter2.xyz, a);
}

technique Wireframe
{
	VertexShader = SurfaceVertexShading();
	GeometryShader = WireframeGeometryShading();
	PixelShader = WireframePixelShading();
};

//////////////////////////////////////////////////////////////////////////
// Frozen
float4 FrozenPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	return OUT;
}

technique Frozen
{
	VertexShader = SurfaceVertexShading();
	PixelShader = FrozenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Burnt
float4 BurntPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	return OUT;
}

technique Burnt
{
	VertexShader = SurfaceVertexShading();
	PixelShader = BurntPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Highlight
float4 HighlightPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(1, 0, 0, 0);
	return OUT;
}

technique Highlight
{
	VertexShader = SurfaceVertexShading();
	PixelShader = HighlightPixelShading();
}
