//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GaussianDOF.fx>
///	@path	~/System/Shader/
///	@date	2015/04/08
///	@brief	Gaussian depth of field rendering.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

// constants
cbuffer Batch
{
	float4x3 World 				: World ;
};

cbuffer Shader
{
	float4 DoFParams			;
};

// textures.
texture2D DiffuseTexture : layout(t0);
sampler DiffuseTexture_Sampler : layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthBuffer : SceneDepth : layout(t1);
sampler DepthBuffer_Sampler : layout(s1)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D BlurredTexture : layout(t2);
sampler BlurredTexture_Sampler : layout(s2)
{
	Texture = BlurredTexture;
	Filter = Linear;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Setup pass

#define FocalDistance			(DoFParams.x)
#define FocalRegion				(DoFParams.y)
#define NearTransitionRegion	(DoFParams.z)
#define FarTransitionRegion		(DoFParams.w)

float ComputeDOFNearFocalMask(float SceneDepth)
{
	float NearFocalPlane = FocalDistance;
	return saturate((NearFocalPlane - SceneDepth) / NearTransitionRegion);
}

float ComputeDOFFarFocalMask(float SceneDepth)
{
	// Don't blur skybox
	//if (SceneDepth >= CameraInfo.y)
	//	return 0;

	float FarFocalPlane = FocalDistance + FocalRegion;
	return saturate((SceneDepth - FarFocalPlane) / FarTransitionRegion);
}

float ComputeDOFFarNearFocalMask(float SceneDepth)
{
	float NearFocalPlane = FocalDistance;
	float NearFactor = saturate((NearFocalPlane - SceneDepth) / NearTransitionRegion);
	float factor = saturate(sign(NearFactor));

	float FarFocalPlane = FocalDistance + FocalRegion;
	float FarFactor = saturate((SceneDepth - FarFocalPlane) / FarTransitionRegion);

	return factor * NearFactor + (1-factor)*FarFactor;
}

// Encode color before Gaussian Blur to avoid color banding
#define EncodeBlurred(blurred) sqrt(blurred)
#define DecodeBlurred(blurred) (blurred * blurred)

float4 DoFSetupPS(VertexScreenOutput IN) : SV_Target
{
	float2 stc = IN.TC;
	float2 offset = ScreenInfo.zw;

	float4 depthQuad;
	depthQuad.x = DecodeDepth(DepthBuffer.SampleLevel(DepthBuffer_Sampler, stc + float2(-offset.x,  offset.y), 0));
	depthQuad.y = DecodeDepth(DepthBuffer.SampleLevel(DepthBuffer_Sampler, stc + float2( offset.x,  offset.y), 0));
	depthQuad.z = DecodeDepth(DepthBuffer.SampleLevel(DepthBuffer_Sampler, stc + float2( offset.x, -offset.y), 0));
	depthQuad.w = DecodeDepth(DepthBuffer.SampleLevel(DepthBuffer_Sampler, stc + float2(-offset.x, -offset.y), 0));
	depthQuad *= CameraInfo.y;

	float4 OUT = float4(0, 0, 0, 0);
	float4 tap = float4(0, 0, 0, 1);
	float mask = 0;

	mask = ComputeDOFFarNearFocalMask(depthQuad.x);
	tap.xyz = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, stc + float2(-offset.x,  offset.y)).xyz;
	OUT += tap * mask;

	mask = ComputeDOFFarNearFocalMask(depthQuad.y);
	tap.xyz = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, stc + float2( offset.x,  offset.y)).xyz;
	OUT += tap * mask;

	mask = ComputeDOFFarNearFocalMask(depthQuad.z);
	tap.xyz = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, stc + float2( offset.x, -offset.y)).xyz;
	OUT += tap * mask;

	mask = ComputeDOFFarNearFocalMask(depthQuad.w);
	tap.xyz = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, stc + float2(-offset.x, -offset.y)).xyz;
	OUT += tap * mask;

	// average 4 taps
	OUT = OUT * 0.25;
	return EncodeBlurred(OUT);
}

technique DoFSetup
{
	VertexShader = ScreenVS();
	PixelShader = DoFSetupPS();
}


//////////////////////////////////////////////////////////////////////////
// Recombine pass

float4 RecombinePS(VertexScreenOutput IN) : SV_Target
{
	//float4 Focused = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TC);
	//float4 OUT = Focused;

	float depth = DecodeDepth(DepthBuffer.SampleLevel(DepthBuffer_Sampler, IN.TC, 0));
	depth *= CameraInfo.y;

	float4 Blurred = SampleColorMap(BlurredTexture, BlurredTexture_Sampler, IN.TC);
	Blurred = DecodeBlurred(Blurred);

	float FocusedMask = 1 - ComputeDOFFarNearFocalMask(depth);
	float BlurredMask = Blurred.a;

	float DivBias = 0.001;
	//float3 BlurredColor = (Focused.rgb * DivBias + Blurred.rgb) / (BlurredMask + DivBias);
	float3 BlurredColor = Blurred.rgb / (BlurredMask + DivBias);

	float Smash = 0.25;
	FocusedMask = saturate((FocusedMask - (1.0 - Smash)) * (1.0 / Smash));
	FocusedMask *= FocusedMask;

	//OUT.rgb = lerp(BlurredColor, Focused.rgb, FocusedMask);
	float4 OUT;
	OUT.rgb = BlurredColor;
	OUT.a = 1 - FocusedMask;
	return OUT;
}

technique DoFRecombine
{
	VertexShader = ScreenVS();
	PixelShader = RecombinePS();
}


