//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GlobalFog.fx>
///	@path	~/System/Shader/
///	@date	2016/10/27
///	@brief	Motin Blur.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////
@_VolumetricFogEnable
{
	Name = VolumetricFogEnable;
	Description = "Apply global volumetic fog";
	Fallback = false;
	Default = false;
}

@_EnableVolumetricShadow
{
	Name = EnableVolumetricShadow;
	Description = "Apply volumetic shadow";
	Fallback = false;
	Default = false;
}

@_IsFirstFrame
{
	Name = IsFirstFrame;
	Description = "Apply volumetic shadow";
	Fallback = false;
	Default = false;
}

@_PhysicallyFalloff
{
	Name = PhysicallyFalloff;
	Description = "";
	Default = true;
};

@_CameraInLight
{
	Name = CameraInLight;
	Description = "";
	Default = false;
};

@_UseTAA
{
	Name = UseTAA;
	Description = "";
	Default = false;
}

///////////////// samplers //////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tShadowMap : layout(t1);
sampler sShadowSampler : layout(s1)
{
	Texture = tShadowMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tScatteringMap : layout(t2);
sampler sScatteringSampler : layout(s2)
{
	Texture = tScatteringMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tLastScatteringMap : layout(t3);
sampler sLastScatteringSampler : layout(s3)
{
	Texture = tLastScatteringMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tRandom2D : layout(t4);
sampler sRandom2D : layout(s4)
{
	Texture = tRandom2D;
	Filter = Point;
	Address = Wrap;
};

texture2D tVolumetricDepth : layout(t2);
sampler sVolumetricDepthSampler : layout(s2)
{
	Texture = tVolumetricDepth;
	Filter = Point;
	Address = Clamp;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 vWBasisX						;
	float4 vWBasisY						;
	float4 vWBasisZ						;
	float4 vFogInfo		: FogInfo 		;
	float3 cVolumetricFogAlbedo			;
	float cVolumetricFogPhaseG			;
	float cFogDensity					;
	float cFogBaseHeight				;
	float cVolumetricFogExtinctionScale	;
	float cVolumetricFogDistance		;
	float4x4 cShadowViewProjTex			;
	float3 cVolumetricEmissive			;
	float cJitter						;
	float4x4 lastViewProjTex			;
	float4 cRTSize						;	//xy-curr size, zw-orig size
	float4x3 World 					 	 ;
	float4 cLightData[4]				;
	float cShadowDistance 				;	//xy-curr size, zw-orig size
	float cLocalLightStepLength			;
	float cLocalLightExtinction			;
	float cVolumetricScatteringScale	;
	float cFarPlane						;
	float cLightRadius2					;
	float cDistanceToCam				;
};

// #define FOG_OFFSET	(vFogInfo.x)
// #define FOG_DENSITY 	(vFogInfo.y)
// #define FOG_EXP_K	(vFogInfo.z)//5.0/FOG_HEIGHT
// #define FOG_EXP_BASE	(vFogInfo.w)//d*exp(-k*(CameraPos.y-BaseH))/k
#define MinVolumetricFogDistance 2

///////////////////////////functions///////////////////////
float IsotropicPhase()
{
	return 1 / (4 * PI);
}

float HenyeyGreensteinPhase(float g, float CosTheta)
{
	g = -g;
	return (1 - g * g) / (4 * PI * pow(1 + g * g - 2 * g * CosTheta, 1.5));
}

float SchlickPhase(float k, float CosTheta)
{
	float Inner = (1 + k * CosTheta);
	return (1 - k * k) / (4 * PI * Inner * Inner);
}

float RaleighPhase(float CosTheta)
{
	return 3.0f * (1.0f + CosTheta * CosTheta) / (16.0f * PI);
}

// Positive g = forward scattering
// Zero g = isotropic
// Negative g = backward scattering
float PhaseFunction(float g, float CosTheta)
{
	return HenyeyGreensteinPhase(g, CosTheta);
}

float FogDensity(in float3 worldPos)
{
	return cFogDensity * exp(-vFogInfo.z*(worldPos.y - cFogBaseHeight));
}

float HeightFogTransmittance(in float3 cameraToWorld)
{
	float D = max(0, length(cameraToWorld) - vFogInfo.x);
	//half tLen = FOG_DENSITY * exp(-k*(CameraPos.y-BaseH)) * D * (1-exp(-k*DIR.y)) / (k*DIR.y)
	float tLen = vFogInfo.w;
	if (abs(cameraToWorld.y) > EPSILON)
		tLen *= (1-exp(-vFogInfo.z*cameraToWorld.y)) / (cameraToWorld.y);
	float T = exp(-D*tLen);
	return T;
}

float VolumetricFogTransmittance(in float3 cameraToWorld, in float length, in float extinctionScale)
{
	float D = length;
	//half tLen = FOG_DENSITY * exp(-k*(CameraPos.y-BaseH)) * D * (1-exp(-k*DIR.y)) / (k*DIR.y)
	float tLen = vFogInfo.w * extinctionScale;
	if (abs(cameraToWorld.y) > EPSILON)
		tLen *= (1-exp(-vFogInfo.z*cameraToWorld.y)) / (cameraToWorld.y);
	float T = exp(-D*tLen);
	return T;
}

float VolumetricShadow(in float3 worldPos)
{
	if (@_EnableVolumetricShadow)
	{
		//light space uvz
		float4 P_LS = mul(float4(worldPos,1), cShadowViewProjTex);
		P_LS.xyz /= P_LS.w;

		if (dot(step(P_LS.xyz, float3(1,1,1)), step(float3(0,0,0), P_LS.xyz)) < 2.99)
			return 1;

		//Shadow
		float shadowDepth = DecodeDepth(tShadowMap.Sample(sShadowSampler, P_LS.xy));
		return step(P_LS.z, shadowDepth);
	}
	return 1;
}

#define SHADOW_SAMPLES 28
#define SCATTERING_SAMPLES 32
float3 GetScatteringColor(in float3 startPos, in float3 ViewDirection, in float ViewDistance, in float jitter)
{
	//∫ Tr(startPos->Pos_t) * [sun intensity * Phase(theta) * VolumetricShadow(Pos_t) + ambient] * ScatteringCoefficient ds
	//ExtinctionCoefficient = cVolumetricFogExtinctionScale * FogDensity(pos);
	//ScatteringCoefficient =  cVolumetricFogExtinctionScale * FogDensity(pos) * cVolumetricFogAlbedo;
	float3 SunRadiance = SunColor.rgb * PhaseFunction(cVolumetricFogPhaseG, dot(SunDirection.xyz, ViewDirection));
	float3 ConstantValue =  cVolumetricFogAlbedo * cVolumetricFogExtinctionScale;
	float3 OutScattering1 = half3(0, 0, 0);
	float3 OutScattering2 = half3(0, 0, 0);
	float ShadowDistance = min(cShadowDistance, ViewDistance*SHADOW_SAMPLES/SCATTERING_SAMPLES);
	float stepDistance = ShadowDistance / SHADOW_SAMPLES;
	float3 stepPos = stepDistance * ViewDirection;
	float3 iterPos = startPos + 0.0001 * stepPos;
	float iterLength = 0.0001;
	
	//shadow samples
	for (int s = 0; s < SHADOW_SAMPLES; ++s)
	{
		OutScattering1 += VolumetricFogTransmittance(iterPos-startPos, iterLength, cVolumetricFogExtinctionScale) * 
						(SunRadiance * VolumetricShadow(iterPos+(jitter)*stepPos) + cVolumetricEmissive.rgb) * 
						FogDensity(iterPos);
		iterPos += stepPos;
		iterLength += stepDistance;
	}
	
	//unshadow samples
	stepDistance = (ViewDistance - ShadowDistance) / (SCATTERING_SAMPLES - SHADOW_SAMPLES);
	stepPos = stepDistance * ViewDirection;
	iterPos = startPos + ShadowDistance * ViewDirection + 0.001 * stepPos;
	iterLength = 0.001*stepDistance;
	for (int s2 = SHADOW_SAMPLES; s2 < SCATTERING_SAMPLES; ++s2)
	{
		OutScattering2 += VolumetricFogTransmittance(iterPos-startPos, iterLength, cVolumetricFogExtinctionScale) * 
						(SunRadiance  + cVolumetricEmissive.rgb) * 
						FogDensity(iterPos);
		iterPos += stepPos;
		iterLength += stepDistance;
	}
	return ConstantValue * (OutScattering1 * ShadowDistance / SHADOW_SAMPLES + OutScattering2 * (ViewDistance - ShadowDistance) / (SCATTERING_SAMPLES - SHADOW_SAMPLES));
}

float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}
/////////////////// Calc Curr Volumetric Fog ///////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 CalcVolumetricFogPS(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos / cRTSize.xy;
	float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc));
	float3 cameraToWorld = CalcHomogeneousPos(depth, stc);
	float3 worldPos = CameraPos.xyz + cameraToWorld;
	float2 jitterUV = frac(float2(worldPos.z+worldPos.y, worldPos.x+worldPos.y)) + float2(cJitter, 0);
	float jitter = tRandom2D.Sample(sRandom2D, jitterUV).x - 0.5;
	
	//CurrScattering
	float ViewDistance = length(cameraToWorld);
	float3 ViewDirection = cameraToWorld / ViewDistance;
	//ViewDistance = clamp(ViewDistance-vFogInfo.x, MinVolumetricFogDistance, min((1-0.9*abs(ViewDirection.y))*CameraInfo.y,cVolumetricFogDistance));
	ViewDistance = clamp(ViewDistance-vFogInfo.x, MinVolumetricFogDistance, cVolumetricFogDistance);
	float3 CurrScattering = GetScatteringColor(CameraPos.xyz, ViewDirection, ViewDistance, jitter);

	return float4(CurrScattering, 1.0);
}

technique CalcVolumetricFog
{
	VertexShader = ScreenVS();
	PixelShader = CalcVolumetricFogPS();
}

/////////////////// TAAFilter ////////////////////
float4 TAAFilterPS(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 stc = spos / cRTSize.xy;
	float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc));
	float3 cameraToWorld = CalcHomogeneousPos(depth, stc);
	float3 worldPos = CameraPos.xyz + cameraToWorld;

	//lastUV
	float4 lastPos = MulMat4x4(lastViewProjTex, float4(worldPos,1));
	float2 lastUV = lastPos.xy / lastPos.w;
	float InLastFrameScreen = lastUV.x >= 0 && lastUV.x <= 1 && lastUV.y >= 0 && lastUV.y <= 1;
	if (@_IsFirstFrame)
		InLastFrameScreen = 0;
	float4 _LastScattering = tLastScatteringMap.Sample(sLastScatteringSampler, lastUV).rgba;
	float3 LastScattering = _LastScattering.rgb;
	float lastDepth = _LastScattering.a;
	float currDepth = lastPos.w / CameraInfo.y;
	float lastFrameRate =  max(0, InLastFrameScreen - 0.01 - abs(1-lastDepth/currDepth));

	//clamp
	float3 s0 = tScatteringMap.Sample(sScatteringSampler, stc).rgb;
	float2 uv = stc - 0.5 / cRTSize.xy;
	float3 s1 = tScatteringMap.Sample(sScatteringSampler, stc + float2(-1/cRTSize.x, -1/cRTSize.y)).rgb;
	float3 s2 = tScatteringMap.Sample(sScatteringSampler, stc + float2(-1/cRTSize.x,  1/cRTSize.y)).rgb;
	float3 s3 = tScatteringMap.Sample(sScatteringSampler, stc + float2( 1/cRTSize.x, -1/cRTSize.y)).rgb;
	float3 s4 = tScatteringMap.Sample(sScatteringSampler, stc + float2( 1/cRTSize.x,  1/cRTSize.y)).rgb;
	float3 minScattering = min(min(min(s0, s1), min(s2, s3)), s4);
	float3 maxScattering = max(max(max(s0, s1), max(s2, s3)), s4);
	LastScattering = clamp(LastScattering, minScattering, maxScattering);

	return float4(lerp(s0, LastScattering.rgb, lastFrameRate), depth);
}
technique TAAFilter
{
	VertexShader = ScreenVS();
	PixelShader = TAAFilterPS();
}

/////////////////// MergeFog ////////////////////
float4 MergeFogPS(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;

	//Transmittance for dest color
	float2 stc = spos * ScreenInfo.zw;
	float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc));
	float3 cameraToWorld = CalcHomogeneousPos(depth, stc);

	//scattering
	if (@_VolumetricFogEnable)
	{
		float ViewDistance = length(cameraToWorld);
		float3 ViewDirection = cameraToWorld / ViewDistance;
		ViewDistance = clamp(ViewDistance-vFogInfo.x, MinVolumetricFogDistance, cVolumetricFogDistance);
		float _T = VolumetricFogTransmittance(ViewDirection*ViewDistance, ViewDistance, 1);
		float T = lerp(FOG_MINT, 0.999, _T);
		float Alpha = 1 - T;
		float3 OutScattering = tScatteringMap.Sample(sScatteringSampler, stc).rgb;
		return float4(OutScattering, Alpha);
	}
	else
	{
		float T = HeightFogTransmittance(cameraToWorld);
		T = lerp(FOG_MINT, 0.999, T);
		float Alpha = 1 - T;
		return float4(FogColor.rgb*Alpha, Alpha);
	}
}

technique MergeFog
{
	VertexShader = ScreenVS();
	PixelShader = MergeFogPS();
}

////////////////////////////////////VolumetricFogLightingDepth////////////////////////////////////
/// light volume vertex to vs
struct VertexAssembleLightVolume
{
	float3		Position		: POSITION;
};
// vertex output to pixel shader.
struct VertexOutputVolumericFogLighting
{
	float4		HPosition		: SV_Position;
	float		DepthInView		: TEXCOORD0;
};

float LocalLightShadow(in float3 worldPos)
{
	if (@_EnableVolumetricShadow)
	{
		//light space uvz
		float4 P_LS = mul(float4(worldPos,1), cShadowViewProjTex);
		P_LS.xyz /= P_LS.w;
		P_LS.z = P_LS.w / cFarPlane;

		if (dot(step(P_LS.xyz, float3(1,1,1)), step(float3(0,0,0), P_LS.xyz)) < 2.99)
			return 0;

		//Shadow
		float shadowDepth = DecodeDepth(tShadowMap.SampleLevel(sShadowSampler, P_LS.xy, 0));
		return step(shadowDepth, P_LS.z);
	}
	return 0;
}

VertexOutputVolumericFogLighting VolumetricFogLightingVS( VertexAssembleLightVolume IN )
{
	VertexOutputVolumericFogLighting OUT;

	// Transform position.
	float4 worldPosition = float4(MulMat3x4(World, float4(IN.Position, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, worldPosition);
	OUT.DepthInView = OUT.HPosition.w / CameraInfo.y;

	//disable farplane culling
	OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);

	return OUT;
}

float4 VolumetricFogLightingDepthPS( VertexOutputVolumericFogLighting IN ) : SV_Target
{
	return float4(IN.DepthInView.rrr, 1);
}

technique VolumetricFogLightingDepth
{
	VertexShader = VolumetricFogLightingVS();
	PixelShader = VolumetricFogLightingDepthPS();
}

////////////////////////////////VolumetricFogPointLighting////////////////////////////////////////
float4 VolumetricFogPointLightingPS( VertexOutputVolumericFogLighting IN ) : SV_Target
{
	//calc startPos, endPos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float3 camToFarPlane = CalcHomogeneousPos(1, stc);
	float et = min(IN.DepthInView, DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc)));//[0-1]
	float st = 0;
	if (!@_CameraInLight)
	{
		float3 centerToCam = CameraPos.xyz - cLightData[0].xyz;
		float A = dot(camToFarPlane, camToFarPlane);
		float B = 2 * dot(centerToCam, camToFarPlane);
		float C = dot(centerToCam, centerToCam) - cLightRadius2;
		st = (-B - sqrt(max(0,B*B - 4*A*C))) / (2*A);
	}
	et = max(et, st+EPSILON);
	float3 startPos = CameraPos.xyz + st * camToFarPlane;
	float3 endPos = CameraPos.xyz + et * camToFarPlane;

	//calc samples
	float RaymarchingDistance = length(endPos - startPos);
	float3 ViewDirection = (endPos - startPos) / RaymarchingDistance;
	float NSamples = floor(max(30-st*CameraInfo.y, 0) + 1);

	//ray marching
	float3 OutScattering = float3(0, 0, 0);
	float stepDistance = RaymarchingDistance / NSamples;
	float3 stepPos = (endPos-startPos)/NSamples;
	float3 iterPos = startPos + 0.5*stepPos;
	float iterLength = 0.5*stepDistance;
	//Albedo * Distance * ∫ Tr(startPos->Pos_t) * [sun intensity * Phase(theta) * VolumetricShadow(Pos_t)] * ScatteringCoefficient dt
	//ExtinctionCoefficient = cLocalLightExtinction;
	//ScatteringCoefficient = cLocalLightExtinction * cVolumetricScatteringScale;
	float3 ConstantValue =  cVolumetricFogAlbedo * RaymarchingDistance * cLocalLightExtinction * cVolumetricScatteringScale;// * IsotropicPhase();
	float3 centerPos = (startPos+endPos)/2;
	[loop] for (int i = 0; i < NSamples; ++i)
	{
		//if (VolumetricShadow(iterPos))
		{
			float3 L = cLightData[0].xyz - lerp(centerPos, iterPos, (NSamples-1) / 30);
			half D2 = dot(L, L);
			half D = sqrt(D2);
			L /= D;
			half Atten = CalcDistanceAttenuation(D2, cLightData) * LightRangeMask(D2, cLightData[0].w);
			float3 inputRadiance = cLightData[1].rgb * Atten;
			OutScattering += exp(-cLocalLightExtinction * iterLength) * inputRadiance.rgb;
		}
		
		iterPos += stepPos;
		iterLength += stepDistance;
	}
	OutScattering = OutScattering * ConstantValue / NSamples;
	return float4(OutScattering, 0);
}

technique VolumetricFogPointLighting
{
	VertexShader = VolumetricFogLightingVS();
	PixelShader = VolumetricFogPointLightingPS();
}

float4 VolumetricFogSpotLightingPS( VertexOutputVolumericFogLighting IN ) : SV_Target
{
	//calc startPos, endPos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float3 camToFarPlane = CalcHomogeneousPos(1, stc);
	float et = min(IN.DepthInView, DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc)));//[0-1]
	float st = 0;
	if (!@_CameraInLight)
	{
		float3 centerToCam = CameraPos.xyz - cLightData[0].xyz;
		float CoV = dot(centerToCam, cLightData[2].xyz);
		float DoV = dot(camToFarPlane, cLightData[2].xyz);
		float CoD = dot(centerToCam, camToFarPlane);
		float A = DoV*DoV - dot(camToFarPlane, camToFarPlane) * cLightData[3].w;
		float B = 2*(DoV*CoV-CoD*cLightData[3].w);
		float C = CoV*CoV - dot(centerToCam, centerToCam)*cLightData[3].w;
		if (abs(A) < EPSILON)
		 	st = -C / B;
		else
		{
			float flag = step(A, 0) * 2 - 1;
		 	st = (-B + flag * sqrt(max(EPSILON,B*B - 4*A*C))) / (2*A);
			if (dot((CameraPos.xyz + st*camToFarPlane - cLightData[0].xyz), cLightData[2].xyz) < 0)
				st = (-B - flag * sqrt(max(EPSILON,B*B - 4*A*C))) / (2*A);
		}
		if (st + 0.0001 > et)
			st = 0;
		st = max(st, 0);
		//st = tVolumetricDepth.Sample(sVolumetricDepthSampler, stc).x;
	}
	et = max(et, st+EPSILON);
	float3 startPos = CameraPos.xyz + st * camToFarPlane;
	float3 endPos = CameraPos.xyz + et * camToFarPlane;

	//calc samples
	float RaymarchingDistance = length(endPos - startPos);
	float3 ViewDirection = (endPos - startPos) / RaymarchingDistance;
	float NSamples = floor(max(28-cDistanceToCam, 0) + 4);

	//ray marching
	float jitter = 0;
	if (@_UseTAA)
	{
		float2 jitterUV = frac(float2(endPos.z+endPos.y, endPos.x+endPos.y)) + float2(cJitter, 0);
		jitter = tRandom2D.SampleLevel(sRandom2D, jitterUV, 0).x;
	}
	float3 OutScattering = float3(0, 0, 0);
	float stepDistance = RaymarchingDistance / NSamples;
	float3 stepPos = (endPos-startPos)/NSamples;
	float3 iterPos = startPos + 0.5*stepPos;
	float3 shadowPos = startPos + jitter*stepPos;
	float iterLength = 0.5;
	//Albedo * Distance * ∫ Tr(startPos->Pos_t) * [sun intensity * Phase(theta) * VolumetricShadow(Pos_t)] * ScatteringCoefficient dt
	//ExtinctionCoefficient = cLocalLightExtinction;
	//ScatteringCoefficient = cLocalLightExtinction * cVolumetricScatteringScale;
	float3 ConstantValue =  cVolumetricFogAlbedo * RaymarchingDistance * cLocalLightExtinction * cVolumetricScatteringScale;// * IsotropicPhase();
	float3 centerPos = (startPos+endPos)/2;
	for (int i = 0; i < NSamples; ++i)
	{
		float shadow = 1;
		if (@_UseTAA)
			shadow = 1 - saturate(1-cDistanceToCam/10) * LocalLightShadow(shadowPos);
		// Spot lighting.
		float3 L = cLightData[0].xyz - iterPos;
		half D2 = dot(L, L);
		half D = sqrt(D2);
		L /= D;
		half DoL = dot( cLightData[2].xyz, -L );
		
		half Atten = CalcDistanceAttenuation(D2, cLightData) * LightRangeMask(D2, cLightData[0].w);
		half Spot = Square(saturate(DoL*cLightData[3].y + cLightData[3].z));
		half3 inputRadiance = cLightData[1].rgb * Atten * Spot * shadow;
		OutScattering += exp(-cLocalLightExtinction * iterLength) * inputRadiance.rgb;
		
		iterPos += stepPos;
		shadowPos += stepPos;
		iterLength += stepDistance;
	}
	OutScattering = OutScattering * ConstantValue / NSamples;
	return float4(OutScattering, 0);
}

technique VolumetricFogSpotLighting
{
	VertexShader = VolumetricFogLightingVS();
	PixelShader = VolumetricFogSpotLightingPS();
}


