//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GodRay.fx>
///	@path	~/System/Shader/
///	@date	2018/06/11
///	@brief	Screen space god ray rendering.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

@_ENABLE_AA
{
	Name = EnableAA;
	Description = "Enable AA";
	Default = false;
};

// constants
cbuffer Batch
{
	float4x3 World 				: World ;
};

cbuffer Shader
{
	float4 GodRayParams			;
	float4 SunPos				;
};

// textures.
texture2D DiffuseTexture : layout(t0);
sampler DiffuseTexture_Sampler : layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthBuffer : SceneDepth : layout(t1);
sampler DepthBuffer_Sampler : layout(s1)
{
	Texture = DepthBuffer;
	Filter = Linear;
	Address = Clamp;
};

texture2D PreviousTexture : layout(t2);
sampler PreviousTexture_Sampler : layout(s2)
{
	Texture = PreviousTexture;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Setup pass

float4 GodRaySetupPS(VertexScreenOutput IN) : SV_Target
{
	float2 stc = IN.TC;
	float3 color = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, stc).xyz;

	float edge = 1.0 - stc.x * (1.0 - stc.x) * stc.y * (1.0 - stc.y) * 8.0;
	float edgeMask = 1.0 - edge * edge * edge * edge;

	float depth = DecodeDepth(DepthBuffer.SampleLevel(DepthBuffer_Sampler, stc, 0));
	float depthMask = saturate((depth - GodRayParams.x) * 32.0);
	float occlusion = min(depthMask, edgeMask);

	float2 distance = (stc - SunPos.xy) * float2(SunPos.z, 1.0);
	float distanceMask = saturate(1 - length(distance) * GodRayParams.y);
	distanceMask = distanceMask * distanceMask;

	float lum = GetLuminance(color);
	float emission = saturate(GodRayParams.z - lum * 2.0);

	return float4(occlusion, depthMask, distanceMask, emission);
}

technique GodRaySetup
{
	VertexShader = ScreenVS();
	PixelShader = GodRaySetupPS();
}

//////////////////////////////////////////////////////////////////////////
// Blur pass

float4 GodRayBlurPS(VertexScreenOutput IN) : SV_Target
{
	float2 stc = IN.TC;
	float2 distance = (SunPos.xy - stc);
	float dist = length(distance);
	float len = min(dist, GodRayParams.y * 8);
	float2 vec = distance * (len / dist);

	float4 s0 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, stc);
	float2 s1 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, (stc + vec * (1.0 / 8))).xy * (7.0 / 8);
	float2 s2 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, (stc + vec * (2.0 / 8))).xy * (6.0 / 8);
	float2 s3 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, (stc + vec * (3.0 / 8))).xy * (5.0 / 8);
	float2 s4 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, (stc + vec * (4.0 / 8))).xy * (4.0 / 8);
	float2 s5 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, (stc + vec * (5.0 / 8))).xy * (3.0 / 8);
	float2 s6 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, (stc + vec * (6.0 / 8))).xy * (2.0 / 8);
	float2 s7 = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, (stc + vec * (7.0 / 8))).xy * (1.0 / 8);

	float2 sum0 = s0.xy + s1 + s2 + s3;
	float2 sum1 = s4 + s5 + s6 + s7;
	float4 ret = float4((sum0 + sum1) / 4.5, s0.zw);

	if (@_ENABLE_AA)
	{
		float prev = SampleColorMap(PreviousTexture, PreviousTexture_Sampler, stc).x;
		ret.x = lerp(prev, ret.x, 0.3);
	}

	return ret;
}

technique GodRayBlur
{
	VertexShader = ScreenVS();
	PixelShader = GodRayBlurPS();
}

//////////////////////////////////////////////////////////////////////////
// Recombine pass

float4 RecombinePS(VertexScreenOutput IN) : SV_Target
{
	// value = (occlusion, depthMask, distanceMask, emission)
	float4 value = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TC);
	float alpha = value.x * value.z;
	float3 color = GodRayParams.rgb * alpha;

	float darkControl = (1 - value.y) * value.z * GodRayParams.w;
	if (@_HDR)
	{
		alpha = darkControl;
	}
	else
	{
		color = color * value.w;
		alpha = darkControl;
	}
	return float4(color, alpha);
}

technique GodRayRecombine
{
	VertexShader = ScreenVS();
	PixelShader = RecombinePS();
}

