//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liuzhongyuan@corp.netease.com
//	Author: Jrenliu
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GpuCloth.fx>
///	@path	~/System/Shader/
///	@date	2020/03/22
///	@brief	GPU cloth shader. solve constaint

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader draw

#define INCLUDE_GPU_PHYSICS_VS
#define INCLUDE_GPU_PHYSICS_DATA
#include "GpuClothCommon.fxh"

texture2D tConstraintTexture : layout(t1) = "Black";
sampler sConstraintSampler : layout(s1)
{
	Texture = tConstraintTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tInputPositionTexture : layout(t2) = "Black";
sampler sInputPositionSampler : layout(s2)
{
	Texture = tInputPositionTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tSkinCache : layout(t3) = "Black";
sampler sSkinCacheSampler : layout(s3)
{
	Texture = tSkinCache;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tRuntimeTexture : layout(t5) = "Black";
sampler sRuntimeTextureSampler : layout(s5)
{
	Texture = tRuntimeTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

float GetCollisionSphereCount(uint runtimeStart)
{
    float2 uv = cvtUv(runtimeStart + GPUPhyRuntimeColHeaderStart, GPUPhyRuntimeTexWidth, GPUPhyRuntimeTexHeight);
    return tRuntimeTexture.SampleLevel(sRuntimeTextureSampler, uv, 0).x;
}

float GetCollisionCapsuleCount(uint runtimeStart)
{
    float2 uv = cvtUv(runtimeStart + GPUPhyRuntimeColHeaderStart, GPUPhyRuntimeTexWidth, GPUPhyRuntimeTexHeight);
    return tRuntimeTexture.SampleLevel(sRuntimeTextureSampler, uv, 0).y;
}

float4 GetCollisionSphere(in uint runtimeStart, in float index)
{
    float2 uv = cvtUv(runtimeStart + GPUPhyRuntimeColDataStart + index, GPUPhyRuntimeTexWidth, GPUPhyRuntimeTexHeight);
    return tRuntimeTexture.SampleLevel(sRuntimeTextureSampler, uv, 0);
}

float2 GetCollisionCapsuleIndex(uint runtimeStart, in float index)
{
    float2 uv = cvtUv(runtimeStart + GPUPhyRuntimeColDataStart + GetCollisionSphereCount(runtimeStart) + index, GPUPhyRuntimeTexWidth, GPUPhyRuntimeTexHeight);
    float4 capsule = tRuntimeTexture.SampleLevel(sRuntimeTextureSampler, uv, 0);
    if (frac(index / 2.0) > 0.0001)
    {
        return capsule.zw;
    }
    return capsule.xy;
}

void solveCollisionCapsuleConstraint(in uint runtimeStart, inout float3 source)
{
    uint shapeNum = (uint)GetCollisionCapsuleCount(runtimeStart);
    for (uint i = 0; i < shapeNum; ++i)
    {
        float2 pair = GetCollisionCapsuleIndex(runtimeStart, i);
        float4 sphere0 = GetCollisionSphere(runtimeStart, pair.x);
        float4 sphere1 = GetCollisionSphere(runtimeStart, pair.y);
        float len_sphere1_sphere0 = length(sphere0 - sphere1);
        float3 dir = (sphere1.xyz - sphere0.xyz) / len_sphere1_sphere0;
        float len_proj_sphere0 = dot(source - sphere0.xyz, dir);
        float len_proj_sphere1 = len_sphere1_sphere0 - len_proj_sphere0;
        if (len_proj_sphere0 > 0 && len_proj_sphere1 > 0)
        {
            float r = (len_proj_sphere0 * sphere1.w - len_proj_sphere1 * sphere0.w) /len_sphere1_sphere0;
            float3 proj = sphere0.xyz + len_proj_sphere0 * dir;
            float3 delta = source - proj;
            float dis = length(delta);
            if (dis < r)
            {
                float pulse = r - dis;
                source = proj + ((r + pulse)/dis) * delta;
            }
        }
    }
}

void solveCollisionSphereConstraint(in uint runtimeStart, inout float3 source)
{
    uint shapeNum = (uint)GetCollisionSphereCount(runtimeStart);
    for (uint i = 0; i < shapeNum; ++i)
    {
        float4 sphere = GetCollisionSphere(runtimeStart, i);
        float3 delta = (source - sphere.xyz);
        float dis = length(delta);
        float r = sphere.w;
        if (dis < r)
        {
            float pulse = r - dis;
            source = sphere.xyz + ((r + pulse) / dis) * delta;
        }
    }
}

struct ClothIterationOutput
{
    float4 PositionResult   : SV_Target0;
};

#define epsilon 1.192092896e-07

void solveSkinnigConstraint(inout float3 source, in float3 skinning, in float rate)
{
    source = lerp(source, skinning, rate);
}

void solveEqualDistanceConstraint(inout float3 source, in float3 target, in float staticLen)
{
    float3 delta = (target - source);
    float dynamicLen = length(delta); 
    float percent = 0.5 * (dynamicLen - staticLen + epsilon) / (dynamicLen + epsilon);
    source += percent * delta;
}

void solveMaxDistanceConstraint(inout float3 source, in float3 target, in float maxDistance)
{
    float3 delta = (target - source);
    float dynamicLen = length(delta); 
    if (dynamicLen > maxDistance)
    {
        float percent = (dynamicLen - maxDistance + epsilon) / (dynamicLen + epsilon);
        source += percent * delta;
    }
}

[target("RGBA32F")]
ClothIterationOutput ClothIterationPs(in SimpleVSOutput IN)
{
    //@ Data prepare
    float4 fabric = tFabricTexture.SampleLevel(sFabricTextureSampler, IN.TexCoord, 0);
    uint particleStart = (uint)fabric.x;
    uint runtimeStart = (uint)fabric.y;
    uint consStart = (uint)fabric.w;
    float4 preIter = tInputPositionTexture.SampleLevel(sInputPositionSampler, IN.TexCoord, 0);
    float3 skinPos = tSkinCache.SampleLevel(sSkinCacheSampler, IN.TexCoord, 0).xyz;
    float4 massNodeConfig = tConstraintTexture.SampleLevel(sConstraintSampler, cvtUv(consStart, GPUPhyConstraintTexWidth, GPUPhyConstraintTexHeight), 0);
    float numCons = massNodeConfig.x;
    int inumCons = (int)numCons;
    float numTethers = massNodeConfig.y;
    int inumTethers = (int)numTethers;
    float maxDis = massNodeConfig.z;
    float dynamic = maxDis == 0.0 ? 0 : 1;
    //@ Solve strech, Jacobian
    float3 result = preIter.xyz;
    float3 inode;
#if 1
    inode = float3(0, 0, 0);
    float tetherScale = 1.1;
    for (int tet = 0; tet < ceil(numTethers/2); ++tet)
    {
        float4 cons = tConstraintTexture.SampleLevel(sConstraintSampler, cvtUv(consStart /*objectOffset*/ + 1/*header*/ + tet/*iterIndex*/ , GPUPhyConstraintTexWidth, GPUPhyConstraintTexHeight), 0);
        {
            float3 ePos = tInputPositionTexture.SampleLevel(sInputPositionSampler, cvtUv(cons.x + particleStart, GPUPhyParticleTexWidth, GPUPhyParticleTexHeight), 0).xyz;
            float staticLen = cons.y;
            float3 tmp = result;
            solveMaxDistanceConstraint(tmp, ePos, tetherScale * staticLen);
            inode += tmp;
        }
        if (frac(numTethers/2.0) < 0.01 || (tet != ceil(numTethers/2) - 1))
        {
            float3 ePos = tInputPositionTexture.SampleLevel(sInputPositionSampler, cvtUv(cons.z + particleStart, GPUPhyParticleTexWidth, GPUPhyParticleTexHeight), 0).xyz;
            float staticLen = cons.w;
            float3 tmp = result;
            solveMaxDistanceConstraint(tmp, ePos, tetherScale * staticLen);
            inode += tmp;
        }
    }
    result = inode / numTethers;
#endif
    inode = float3(0, 0, 0);
    for (int e = 0; e < ceil(numCons/2); ++e)
    {
        float4 cons = tConstraintTexture.SampleLevel(sConstraintSampler, cvtUv(consStart /*objectOffset*/ + 1/*header*/ + numTethers/*stepOverTetherRegion*/+ e/*iterIndex*/ , GPUPhyConstraintTexWidth, GPUPhyConstraintTexHeight), 0);
        {
            float3 ePos = tInputPositionTexture.SampleLevel(sInputPositionSampler, cvtUv(cons.x + particleStart, GPUPhyParticleTexWidth, GPUPhyParticleTexHeight), 0).xyz;
            float staticLen = cons.y;
            float3 tmp = result;
            solveEqualDistanceConstraint(tmp, ePos, staticLen);
            inode += tmp;
        }
        if (frac(numCons/2.0) < 0.01 || (e != ceil(numCons/2) - 1))
        {
            float3 ePos = tInputPositionTexture.SampleLevel(sInputPositionSampler, cvtUv(cons.z + particleStart, GPUPhyParticleTexWidth, GPUPhyParticleTexHeight), 0).xyz;
            float staticLen = cons.w;
            float3 tmp = result;
            solveEqualDistanceConstraint(tmp, ePos, staticLen);
            inode += tmp;
        }
    }
    result  = inode / numCons;

    if (dynamic > 0)
    {     
        //@ Solve maxDistance
        solveMaxDistanceConstraint(result, skinPos, maxDis);
        //@ Solve collision
        solveCollisionSphereConstraint(runtimeStart, result);
        solveCollisionCapsuleConstraint(runtimeStart, result);
    }

    ClothIterationOutput OUT;
    OUT.PositionResult = dynamic * float4(result, 0) + (1 - dynamic) * float4(skinPos, 0);
    //@{ Test
    //OUT.PositionResult = float4(skinPos, 1);
    //@}
    return OUT;
}

technique ClothIteration
{
    VertexShader = SimpleVS();
    PixelShader = ClothIterationPs();
}
