//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liuzhongyuan@corp.netease.com
//	Author: Jrenliu
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GpuCloth.fx>
///	@path	~/System/Shader/
///	@date	2020/03/22
///	@brief	GPU cloth shader. skin, gravity, wind

#include "Constants.fxh"
#include "Common.fxh"

#pragma shader draw

#define cDamping 0.06

#define INCLUDE_GPU_PHYSICS_VS
#define INCLUDE_GPU_PHYSICS_DATA
#define INCLUDE_GPU_PHYSICS_SKINNING_TOOL
#include "GpuClothCommon.fxh"

texture2D tRuntimeTexture : layout(t5) = "Black";
sampler sRuntimeTextureSampler : layout(s5)
{
	Texture = tRuntimeTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tPrePositionTexture : layout(t6);
sampler sPrePositionTextureSampler : layout(s6)
{
	Texture = tPrePositionTexture;
	Filter = Point;
	Address = Clamp;
};

texture2D tPrePrePositionTexture : layout(t7);
sampler sPrePrePositionTextureSampler : layout(s7)
{
	Texture = tPrePrePositionTexture;
	Filter = Point;
	Address = Clamp;
};

// Get 3x4 matrix from skeleton data.
void GetPreToCurMatrix4x3Tex(in uint DataStart, out float4 Mx, out float4 My, out float4 Mz )
{
    float2 uv = cvtUv(DataStart + GPUPhyRuntimePreToCurMatStart, GPUPhyRuntimeTexWidth, GPUPhyRuntimeTexHeight);
	Mx = tRuntimeTexture.SampleLevel(sRuntimeTextureSampler, uv, 0);
    
    uv = cvtUv(1 + DataStart + GPUPhyRuntimePreToCurMatStart, GPUPhyRuntimeTexWidth, GPUPhyRuntimeTexHeight);
	My = tRuntimeTexture.SampleLevel(sRuntimeTextureSampler, uv, 0);
    
    uv = cvtUv(2 + DataStart + GPUPhyRuntimePreToCurMatStart, GPUPhyRuntimeTexWidth, GPUPhyRuntimeTexHeight);
	Mz = tRuntimeTexture.SampleLevel(sRuntimeTextureSampler, uv, 0);
}

struct ClothTickOutput
{
    float4 PositionResult        : SV_Target0;
    float4 SkinCache        : SV_Target1;
};

[target("RGBA32F")]
ClothTickOutput ClothTickPs(in SimpleVSOutput IN)
{
    
    float3 skinning;
    float4 fabric = tFabricTexture.SampleLevel(sFabricTextureSampler, IN.TexCoord, 0);
    uint runtimeStart = (uint)fabric.y;
    uint skeletonStart = (uint)fabric.z;
    float4 static_position = tStaticPositionTexture.SampleLevel(sStaticPositionTextureSampler, IN.TexCoord, 0);
    {
        float4 fweight = tWeightTexture.SampleLevel(sWeightTextureSampler, IN.TexCoord, 0);
        float4 findex = tIndexTexture.SampleLevel(sIndexTextureSampler, IN.TexCoord, 0);
        uint4 index = (uint)(255.5 * findex);
        SkinningTex(skeletonStart, fweight, index , static_position.xyz, skinning);
    }
    float4 prePosition = tPrePositionTexture.SampleLevel(sPrePositionTextureSampler, IN.TexCoord, 0);
    float4 prePrePosition = tPrePrePositionTexture.SampleLevel(sPrePrePositionTextureSampler, IN.TexCoord, 0);
    float3 newPre, newPrePre;
    {
        // Fetch.
        float4 PreToCur[3];
        GetPreToCurMatrix4x3Tex(runtimeStart, PreToCur[0], PreToCur[1], PreToCur[2]);
        newPre = Matrix4x3TransformP(PreToCur[0], PreToCur[1], PreToCur[2], prePosition.xyz);
        newPrePre = Matrix4x3TransformP(PreToCur[0], PreToCur[1], PreToCur[2], prePrePosition.xyz);
    }
    ClothTickOutput OUT;
    // TODO Gravity, Wind and External force
    float timeStep = 0.016;
    float3 velMuldeltaT = (newPre - newPrePre).xyz;
    // delta = t*v = t*(v0+acc*t) = v0*t + acc*t^2
    float3 acc = float3(0, -9.8, 0);
    //float3 delta = timeStep * timeStep * acc;
    float3 delta = velMuldeltaT + timeStep * timeStep * acc;
    // damping
    delta = (1.0 - cDamping) * delta;
    // Dynamic mass-node
    float influenceLocal = 0.05;
    OUT.PositionResult = float4((1.0 - influenceLocal) * prePosition.xyz + influenceLocal * newPre + delta, 1);
    // Save the skinning cache for max distance constraint, and w is the distance.
    OUT.SkinCache = float4(skinning, fabric.w);

    //@{ Test
    //OUT.PositionResult = float4(skinning, 0);
    //OUT.PositionResult = prePosition;
    //OUT.PositionResult = float4(dynamic * prePosition.xyz + (1 - dynamic) * skinning, 0);
    //OUT.PositionResult = float4(skinning, 0);
    //OUT.PositionResult = float4(position.xyz, 0);
    //OUT.SkinCache = float4(static_position.xyz, position.w);
    //uint runtimeRow = runtimeStart / GPUPhyRuntimeTexWidth;
    //uint runtimeCol = runtimeStart % GPUPhyRuntimeTexWidth;
    //OUT.SkinCache = float4(skinning + float3(0.0, runtimeRow, 0.0), fabric.w);
    //OUT.SkinCache = float4(skinning + delta, fabric.w);
    //@}

    return OUT;
}

technique ClothTick
{
    VertexShader = SimpleVS();
    PixelShader = ClothTickPs();
}
