//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<SSAO.fx>
///	@path	~/System/Shader/
///	@date	2018/05/28
///	@brief	SSAO.
//

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

#define MaxRadiusPixels 100.0
#define NumDirections 4
#define  NumSamples 3

///////////////// Properties //////////////////


///////////////// const buffer //////////////////
cbuffer Batch
{
    float2 FocalLen             ;//float2(cot(fovX/2),cot(fovY/2))
    float2 UVToViewA            ;//A = (-2*HalfTanFovX, 2*HalfTanFovY); [glsl]A = (-2*HalfTanFovX, -2*HalfTanFovY);
    float2 UVToViewB            ;//B = (HalfTanFovX, -HalfTanFovY); [glsl]B = (HalfTanFovX, HalfTanFovY)
    float2 AORes                ;
    float2 InvAORes             ;
    float2 NoiseScale           ;
    float AOStrength             = 2.0;
    float _R                      = 0.3;
    float _R2                     = 0.09;
    float _NegInvR2               = -11.111111111111112;
    float TanBias                = 0.5773502691896257;//tan(30.0 * PI / 180.0);
};

#define R 0.6
#define R2 (R*R)
#define NegInvR2 (-1.0 / R2)

///////////////// samplers //////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tRandom2D : layout(t1);
sampler sRandom2D : layout(s1)
{
	Texture = tRandom2D;
	Filter = Point;
	Address = Wrap;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

float ViewSpaceZFromDepth(float d)
{
    return -d * CameraInfo.y;
}

//clipPos.xy = uv * (2, -2) + (-1, 1)
//viewPos.xy = clip.xy * (HalfTanFovX,HalfTanFovY) * -viewPos.z
//A = (-2*HalfTanFovX, 2*HalfTanFovY)
//B = (HalfTanFovX, -HalfTanFovY)
float3 UVToViewSpace(float2 uv, float z)
{
	uv = UVToViewA * uv + UVToViewB;
	return float3(uv * z, z);
}

float3 GetViewPos(float2 uv)
{
    float z = ViewSpaceZFromDepth(DecodeDepth(tDepthBuffer.SampleLevel(sDepthBufferSampler, uv, 0)));
	return UVToViewSpace(uv, z);
}

// float3 GetViewPosPoint(float2 uv)
// {
// 	float2 coord = float2(gl_FragCoord.xy) + uv;
// 	float z = texelFetch(texture0, coord, 0).r;
// 	return UVToViewSpace(uv, z);
// }

float TanToSin(float x)
{
	return x * rsqrt(x*x + 1.0);
}

float InvLength(float2 V)
{
	return rsqrt(dot(V,V));
}

float Tangent(float3 V)
{
	return V.z * InvLength(V.xy);
}

float BiasedTangent(float3 V)
{
	return V.z * InvLength(V.xy) + TanBias;
}

float Tangent(float3 P, float3 S)
{
    return -(P.z - S.z) * InvLength(S.xy - P.xy);
}

float Length2(float3 V)
{
	return dot(V,V);
}

float3 MinDiff(float3 P, float3 Pr, float3 Pl)
{
    float3 V1 = Pr - P;
    float3 V2 = P - Pl;
    return (Length2(V1) < Length2(V2)) ? V1 : V2;
}

//round(x)=floor(x+0.5)
float2 SnapUVOffset(float2 uv)
{
    return round(uv * AORes) * InvAORes;
}

float Falloff(float d2)
{
	return d2 * NegInvR2 + 1.0f;
}

float HorizonOcclusion(float2 texCoord, float2 deltaUV, float3 P, float3 dPdu, float3 dPdv, float randstep, float numSamples)
{
	float ao = 0;

	// Offset the first coord with some noise
	float2 uv = texCoord + SnapUVOffset(randstep*deltaUV);
	deltaUV = SnapUVOffset( deltaUV );

	// Calculate the tangent floattor
	float3 T = deltaUV.x * dPdu + deltaUV.y * dPdv;

	// Get the angle of the tangent floattor from the viewspace axis
	float tanH = BiasedTangent(T);
	float sinH = TanToSin(tanH);

	float tanS;
	float d2;
	float3 S;

	// Sample to find the maximum angle
	for(float s = 1; s <= numSamples; ++s)
	{
		uv += deltaUV;
		S = GetViewPos(uv);
		tanS = Tangent(P, S);
		d2 = Length2(S - P);

		// Is the sample within the radius and the angle greater?
		if(d2 < R2 && tanS > tanH)
		{
			float sinS = TanToSin(tanS);
			// Apply falloff based on the distance
			ao += Falloff(d2) * (sinS - sinH);

			tanH = tanS;
			sinH = sinS;
		}
	}
	
	return ao;
}

float2 RotateDirections(float2 Dir, float2 CosSin)
{
    return float2(Dir.x*CosSin.x - Dir.y*CosSin.y,
                  Dir.x*CosSin.y + Dir.y*CosSin.x);
}

void ComputeSteps(inout float2 stepSizeUv, inout float numSteps, float rayRadiusPix, float rand)
{
    // Avoid oversampling if numSteps is greater than the kernel radius in pixels
    numSteps = min(NumSamples, rayRadiusPix);

    // Divide by Ns+1 so that the farthest samples are not fully attenuated
    float stepSizePix = rayRadiusPix / (numSteps + 1);

    // Clamp numSteps if it is greater than the max kernel footprint
    float maxNumSteps = MaxRadiusPixels / stepSizePix;
    if (maxNumSteps < numSteps)
    {
        // Use dithering to avoid AO discontinuities
        numSteps = floor(maxNumSteps + rand);
        numSteps = max(numSteps, 1);
        stepSizePix = MaxRadiusPixels / numSteps;
    }

    // Step size in uv space
    stepSizeUv = stepSizePix * InvAORes;
}

/////////////////// technique CalcAO ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.HPosition.z = OUT.HPosition.w;

	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 CalcAOPS(v2f IN) : SV_Target
{
	float numDirections = NumDirections;

	float3 P, Pr, Pl, Pt, Pb;
	P 	= GetViewPos(IN.TC);

	// Sample neighboring pixels
    Pr 	= GetViewPos(IN.TC + float2( InvAORes.x, 0));
    Pl 	= GetViewPos(IN.TC + float2(-InvAORes.x, 0));
    Pt 	= GetViewPos(IN.TC + float2( 0, InvAORes.y));
    Pb 	= GetViewPos(IN.TC + float2( 0,-InvAORes.y));

    // Calculate tangent basis floattors using the minimu difference
    float3 dPdu = MinDiff(P, Pr, Pl);
    float3 dPdv = MinDiff(P, Pt, Pb) * (AORes.y * InvAORes.x);

    // Get the random samples from the noise texture
	float2 jitterUV = frac(float2(P.z+P.y, P.x+P.y)*10);
	//jitterUV = IN.TC * 16;
	float3 random = tRandom2D.Sample(sRandom2D, jitterUV).rgb;
	random.xy = random.xy * 2 - 1;

	// Calculate the projected size of the hemisphere
    float2 rayRadiusUV = 0.5 * R * FocalLen / -P.z;
    float rayRadiusPix = rayRadiusUV.x * AORes.x;
    float ao = 1.0;

    // Make sure the radius of the evaluated hemisphere is more than a pixel
    if(rayRadiusPix > 1.0)
    {
    	ao = 0.0;
    	float numSteps;
    	float2 stepSizeUV;

    	// Compute the number of steps
    	ComputeSteps(stepSizeUV, numSteps, rayRadiusPix, random.z);

		float alpha = 2.0 * PI / numDirections;

		// Calculate the horizon occlusion of each direction
		for(float d = 0; d < numDirections; ++d)
		{
			float theta = alpha * d;

			// Apply noise to the direction
			float2 dir = RotateDirections(float2(cos(theta), sin(theta)), random.xy);
			float2 deltaUV = dir * stepSizeUV;

			// Sample the pixels along the direction
			ao += HorizonOcclusion(IN.TC, deltaUV, P, dPdu, dPdv, random.z, numSteps);
		}

		// Average the results and produce the final AO
		ao = max(0, 1.0 - ao / numDirections * 4);
	}

	return float4(ao, ao, ao, 1);
    //return float4(0, 0, 0, saturate(1-ao));
    //out_frag0 = vec2(ao, 30.0 * P.z);
	//return float4(ao, -P.z / CameraInfo.y);//z for smart lowpass filter
}

technique CalcAO
{
	VertexShader = ScreenVS();
	PixelShader = CalcAOPS();
}

