//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<HMotionBlur.fx>
///	@path	~/System/Shader/
///	@date	2018/05/28
///	@brief	High Quality Motion Blur.
//

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

///////////////// Properties //////////////////
#define MotionBlurLevel (@_MotionBlurLevel1*2+@_MotionBlurLevel0)

@_MotionBlurLevel0
{
	Name = MotionBlurLevel0;
	Default = true;
	Visible = false;
};

@_MotionBlurLevel1
{
	Name = MotionBlurLevel1;
	Default = true;
	Visible = false;
};

@_VelocityMapEnable
{
	Name = VelocityMapEnable;
	Default = false;
	Visible = false;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4 vWBasisX				;
	float4 vWBasisY				;
	float4 vWBasisZ				;
	float4x4 lastViewProjTex	;
	float4 cPixelSize			;//1/w,1/h,-1/w,-1/h
	float cExposeTime			;//ExposeTime/FrameTime
	float2 cMinMaxV				;
};

///////////////// samplers //////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Linear;
	Address = Clamp;
};

texture2D tVBuffer : VelocityMap : layout(t1);
sampler sVBufferSampler : layout(s1)
{
	Texture = tVBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tMotionVBuffer : layout(t2);
sampler sMotionVBufferSampler : layout(s2)
{
	Texture = tMotionVBuffer;
	Filter = Linear;
	Address = Clamp;
};

texture2D tNeighborMaxBuffer : layout(t3);
sampler sNeighborMaxBufferSampler : layout(s3)
{
	Texture = tNeighborMaxBuffer;
	Filter = Linear;
	Address = Clamp;
};

texture2D tScreenMap : layout(t4);
sampler sScreenSampler : layout(s4)
{
	Texture = tScreenMap;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

///////////////////////////functions///////////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * depth;
}
float2 DecodeVelocity(float4 x)
{
	// if (@_GLSL)
	// 	return float2(R8G8ToFloat(x.rg)*2-1, R8G8ToFloat(x.ba)*2-1);
	// else
	return x.rg;
}

float4 EncodeVelocity(float2 v)
{
	// if (@_GLSL)
	// 	return float4(FloatToR8G8(v.x*0.5+0.5), FloatToR8G8(v.y*0.5+0.5));
	// else
	return float4(v.rg, abs(v.rg*ScreenInfo.xy));
}
/////////////////// technique InitVBuffer ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);

	OUT.HPosition.z = OUT.HPosition.w;
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 InitVBufferPS(v2f IN) : SV_Target
{
	//rebuild world pos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	float2 tc = spos * ScreenInfo.zw;
	float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, tc));
	float3 worldPos = CalcHomogeneousPos(depth, tc);
	worldPos += CameraPos.xyz;

	//motion vector
	float4 lastPos = MulMat4x4(lastViewProjTex, float4(worldPos,1));
	float2 lastUV = lastPos.xy / lastPos.w;
	if (@_VelocityMapEnable)
	{
		float2 motionVector = tVBuffer.SampleLevel(sVBufferSampler, tc, 0).xy;
		if (motionVector.x != 0 || motionVector.y != 0)
		{
			lastUV = tc + motionVector;
		}
	}
	//frameV * ExposeTime / FrameTime
	float2 v = 0.5*(tc-lastUV) * cExposeTime;
	float lenV = length(v*ScreenInfo.xy);
	float newLenV = clamp(lenV, 0.5, cMinMaxV.y);
	v *= newLenV / (lenV + EPSILON);
	return EncodeVelocity(v);
}

technique InitVBuffer
{
	VertexShader = ScreenVS();
	PixelShader = InitVBufferPS();
}

/////////////////// technique MaxVelocity ////////////////////
float4 MaxVelocityPS(v2f IN) : SV_Target
{
	//4 pixel to 1, cPixelSize(0.25/w,0.25/h,-0.25/w,-0.25/h)
	// {HEX} don't add .xyzw here or Mesa Lang Optimize Error
	float3 v1 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.xy)).xyy;
	float3 v2 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.xw)).xyy;
	float3 v3 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.zy)).xyy;
	float3 v4 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.zw)).xyy;
	v1.z = dot(v1.xy, v1.xy);
	v2.z = dot(v2.xy, v2.xy);
	v3.z = dot(v3.xy, v3.xy);
	v4.z = dot(v4.xy, v4.xy);
	float3 maxV12 = lerp(v1, v2, step(v1.z, v2.z));
	float3 maxV34 = lerp(v3, v4, step(v3.z, v4.z));
	float3 maxV = lerp(maxV12, maxV34, step(maxV12.z, maxV34.z));
	return EncodeVelocity(maxV.xy);
}

technique MaxVelocity
{
	VertexShader = ScreenVS();
	PixelShader = MaxVelocityPS();
}

/////////////////// technique NeighborMax ////////////////////
float4 NeighborMaxPS(v2f IN) : SV_Target
{
	//4 pixel to 1, cPixelSize(1/w,1/h,-1/w,-1/h)
	float3 v1 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.xy)).xyy;
	float3 v2 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.xw)).xyy;
	float3 v3 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.zy)).xyy;
	float3 v4 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + cPixelSize.zw)).xyy;
	float3 v5 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC)).xyy;
	float3 v6 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + float2(cPixelSize.z,0))).xyy;
	float3 v7 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + float2(0,cPixelSize.w))).xyy;
	float3 v8 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + float2(cPixelSize.x,0))).xyy;
	float3 v9 = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, IN.TC + float2(0,cPixelSize.y))).xyy;
	v1.z = dot(v1.xy, v1.xy);
	v2.z = dot(v2.xy, v2.xy);
	v3.z = dot(v3.xy, v3.xy);
	v4.z = dot(v4.xy, v4.xy);
	v5.z = dot(v5.xy, v5.xy);
	v6.z = dot(v6.xy, v6.xy);
	v7.z = dot(v7.xy, v7.xy);
	v8.z = dot(v8.xy, v8.xy);
	v9.z = dot(v9.xy, v9.xy);
	float3 maxV12 = lerp(v1, v2, step(v1.z, v2.z));
	float3 maxV34 = lerp(v3, v4, step(v3.z, v4.z));
	float3 maxV56 = lerp(v5, v6, step(v5.z, v6.z));
	float3 maxV78 = lerp(v7, v8, step(v7.z, v8.z));
	float3 maxV14 = lerp(maxV12, maxV34, step(maxV12.z, maxV34.z));
	float3 maxV58 = lerp(maxV56, maxV78, step(maxV56.z, maxV78.z));
	float3 maxV18 = lerp(maxV14, maxV78, step(maxV14.z, maxV78.z));
	float2 maxV = lerp(maxV18.xy, v9.xy, step(maxV18.z, v9.z));
	return EncodeVelocity(maxV);
}

technique NeighborMax
{
	VertexShader = ScreenVS();
	PixelShader = NeighborMaxPS();
}

/////////////////// technique MotionBlur ////////////////////
float cone(float2 X, float2 Y, float2 V)
{
	return clamp(1 - length(X-Y)/length(V), 0, 1);
}

float cylinder(float2 X, float2 Y, float2 V)
{
	return 1.0 - smoothstep(0.95*length(V), 1.05*length(V), length(X-Y));
}

float cone_len(float lenXY, float lenV)
{
	return clamp(1 - lenXY/lenV, 0, 1);
}

float cylinder_len(float lenXY, float lenV)
{
	return 1.0 - smoothstep(0.95*lenV, 1.05*lenV, lenXY);
}

float softDepthCompare(float za, float zb)
{
	return clamp(1 - (za-zb)/0.001, 0, 1);
}

float4 MotionBlurPS(v2f IN) : SV_Target
{
	float SAMPLES = 4.0;
	if (MotionBlurLevel == 1)
	{
		SAMPLES = 8.0;
	}
	else if (MotionBlurLevel == 2)
	{
		SAMPLES = 16.0;
	}
	else if (MotionBlurLevel == 3)
	{
		SAMPLES = 32.0;
	}
	float2 maxV = DecodeVelocity(tNeighborMaxBuffer.Sample(sNeighborMaxBufferSampler, IN.TC));
	float2 X = IN.TC;
	float4 CX = tScreenMap.Sample(sScreenSampler, X);
	float2 VX = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, X));
	float k = length(maxV*ScreenInfo.xy);
	if (k < cMinMaxV.x)
		return CX;

	float weight = 1 / max(0.5,length(VX*ScreenInfo.xy));
	float4 sumColor = weight * CX;
	float ZX = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, X)) * CameraInfo.y;
	float jitter = frac(ZX*65535 + IN.TC.x * 255) * 2 - 1;
	float2 Y = X + maxV * (-1.0 + jitter/SAMPLES);
	float2 stepUV = maxV * 2.0 / SAMPLES;
	float lenVX = length(VX);
	[unroll]for (float i = 0; i < SAMPLES; ++i)
	{
		//float2 Y = X + maxV * (-1.0+(2.0*i+jitter) / SAMPLES);
		float ZY = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, Y)) * CameraInfo.y;
		float2 VY = DecodeVelocity(tMotionVBuffer.Sample(sMotionVBufferSampler, Y));
		float f = softDepthCompare(ZY, ZX);
		float b = softDepthCompare(ZX, ZY);
		float lenYX = length(Y-X);
		float lenVY = length(VY);
		// float w = 	f * cone(Y, X, VY) + //blur y in front of any x
		// 			b * cone(X, Y, VX) + //any y behind blurry X, estimate background
		// 			cylinder(Y, X, VY) * cylinder(X, Y, VX) * 2; //simultaneously blurry x and y
		float w = 	f * cone_len(lenYX, lenVY) + //blur y in front of any x
					b * cone_len(lenYX, lenVX) + //any y behind blurry X, estimate background
					cylinder_len(lenYX, lenVY) * cylinder_len(lenYX, lenVX) * 2; //simultaneously blurry x and y
		sumColor += tScreenMap.Sample(sScreenSampler, Y) * w;
		weight += w;
		Y += stepUV;
	}
	return float4(sumColor.rgb / weight, CX.a);
}

technique MotionBlur
{
	VertexShader = ScreenVS();
	PixelShader = MotionBlurPS();
}


