#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4 cHiZFactor		;
	float4 cHiZSize			;
	float4x4 cViewProjTex	;
};

texture2D tInputRT : layout(t0);
sampler sInputRT_Sampler : layout(s0)
{
	Texture = tInputRT;
	Filter = Point;
	Address = Clamp;
};

texture2D tCenterRT : layout(t1);
sampler sCenterRT_Sampler : layout(s1)
{
	Texture = tCenterRT;
	Filter = Point;
	Address = Clamp;
};

texture2D tExtendRT : layout(t2);
sampler sExtendRT_Sampler : layout(s2)
{
	Texture = tExtendRT;
	Filter = Point;
	Address = Clamp;
};

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 HZBOcclusionPS(VertexScreenOutput IN) : SV_Target
{
	float4 BoundsCenter = tCenterRT.SampleLevel(sCenterRT_Sampler, IN.TC, 0);
	float4 BoundsExtent = tExtendRT.SampleLevel(sExtendRT_Sampler, IN.TC, 0);

	// Could frustum cull here
	float3 BoundsMin = BoundsCenter.xyz - BoundsExtent.xyz;
	float3 BoundsMax = BoundsCenter.xyz + BoundsExtent.xyz;
	float3 Bounds[2] = { BoundsMin, BoundsMax };

	// Screen rect from bounds
	float3 RectMin = float3(1, 1, 1);
	float3 RectMax = float3(-1, -1, -1);
	[unroll]
	for (int i = 0; i < 8; i++)
	{
		float3 PointSrc;
		PointSrc.x = Bounds[(i >> 0) & 1].x;
		PointSrc.y = Bounds[(i >> 1) & 1].y;
		PointSrc.z = Bounds[(i >> 2) & 1].z;

		float4 PointClip = MulMat4x4(cViewProjTex, float4(PointSrc, 1));
		float3 PointScreen = PointClip.xyz / PointClip.w;

		RectMin = min(RectMin, PointScreen);
		RectMax = max(RectMax, PointScreen);
	}
	float4 Rect = float4(RectMin.xy, RectMax.xy) * cHiZFactor.xyxy;
	float4 RectPixels = Rect * cHiZSize.xyxy;
	float2 RectSize = (RectPixels.zw - RectPixels.xy) * 0.5;	// 0.5 for 4x4
	float Level = max(ceil(log2(max(RectSize.x, RectSize.y))), cHiZFactor.z);

	// Check if we can drop one level lower
	float LevelLower = max(Level - 1, 0);
	float4 LowerRect = RectPixels * exp2(-LevelLower);
	float2 LowerRectSize = ceil(LowerRect.zw) - floor(LowerRect.xy);
	if (all(LowerRectSize <= 4))
	{
		Level = LevelLower;
	}

	// 4x4 samples
	float2 Scale = (Rect.zw - Rect.xy) / 3;
	float2 Bias = Rect.xy;

	float4 MinDepth = float4(1, 1, 1, 1);
	[unroll]
	for (int j = 0; j < 4; j++)
	{
		// TODO could vectorize this
		float4 Depth;
		Depth.x = DecodeDepth(tInputRT.SampleLevel(sInputRT_Sampler, float2(j, 0) * Scale + Bias, Level));
		Depth.y = DecodeDepth(tInputRT.SampleLevel(sInputRT_Sampler, float2(j, 1) * Scale + Bias, Level));
		Depth.z = DecodeDepth(tInputRT.SampleLevel(sInputRT_Sampler, float2(j, 2) * Scale + Bias, Level));
		Depth.w = DecodeDepth(tInputRT.SampleLevel(sInputRT_Sampler, float2(j, 3) * Scale + Bias, Level));
		MinDepth = min(MinDepth, Depth);
	}
	MinDepth.x = min(min(MinDepth.x, MinDepth.y), min(MinDepth.z, MinDepth.w));

	float Out = (RectMax.z >= MinDepth.x) ? 1 : 0;
	return Out.xxxx;
}

technique HZBOcclusionTech
{
	VertexShader = ScreenVS();
	PixelShader = HZBOcclusionPS();
}
