//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzzhengyi@corp.netease.com
//	Author: Bleston
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<PhysicsCollision.fx>
///	@path	~/System/Shader/
///	@date	2019/02/27
///	@brief	Render physics collision

#pragma shader material

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World          : World ;
	float4 cColor           : Color ;
};


//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		WorldNormal		: NORMAL;
	float4		TexCoord		: TEXCOORDN;			//	distance to border
	float3		ViewDirection	: TEXCOORDN;
};


struct ColorVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		Color			: COLOR;				//	Color
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Color 			: COLOR;
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;
};

struct ColorVertexAssemble
{
	float4		Position 		: POSITION;
	float4		Color 			: COLOR;
	float2		TexCoord		: TEXCOORDN;
};

//////////////////////////////////Functions////////////////////////////////////////


//////////////////////////////////Shading////////////////////////////////////////

VertexShadingOutput PhysicsCollisionVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	float4 WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1.0)).xyz, 1.0);
	float4 normal = IN.Normal;
	normal = normal * 2 - 1.0;
	OUT.WorldNormal = float4(MulMat3x4(World, float4(normal.xyz, 0.0)).xyz, 1.0);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.TexCoord = float4(IN.Color.rgba);
	OUT.ViewDirection = normalize(CameraPos.xyz - WorldPosition.xyz);

	return OUT;
}

float4 PhysicsCollisionPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 color = saturate((dot(SunDirection.xyz, IN.WorldNormal.xyz) + 1.0) * 0.3 + 0.4) * cColor.rgb;

	float3 difference = fwidth(IN.TexCoord.xyz);
	float3 distance = IN.TexCoord.xyz / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-0.8 * d * d);

	color = (1-a) * color + a * float3(0, 0, 0);

	float4 OUT = float4(color.rgb, EncodeTransparentBloomFactor(cColor.a, 0));
	OUT.rgb = EncodeGamma(OUT.rgb);
	return OUT;
}

technique DrawBasic
{
	VertexShader = PhysicsCollisionVertexShading();
	PixelShader = PhysicsCollisionPixelShading();
}


ColorVertexShadingOutput ColorVertexShading(in ColorVertexAssemble IN)
{
	ColorVertexShadingOutput OUT;
	float4 WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1.0)).xyz, 1.0);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.Color = IN.Color;
	return OUT;
}

float4 ColorPixelShading(in ColorVertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(IN.Color.rgb, EncodeTransparentBloomFactor(IN.Color.a, 0));
	return OUT;
}

technique DrawBasicColor
{
	VertexShader = ColorVertexShading();
	PixelShader = ColorPixelShading();
}
