//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Impostor.fx>
///	@path	~/System/Shader/
///	@date	2018/04/16
///	@brief	Impostor shading model.

#pragma shader draw

#include "Constants.fxh"
#include "Common.fxh"
#include "BRDF.fxh"
#include "DeferredShadingCommon.fxh"

//////////////////////////////////////////////////////////////////////////
// Impostor shading model.

@_ENABLE_FRAME_BLEND
{
	Name = EnableFrameBlend;
	Description = "Enable frame blend to get smooth impostor transition.";
	Visible = true;
};

@_ENABLE_FREE_ROTATION
{
	Name = EnableFreeRotation;
	Description = "Enable free rotation.";
	Visible = true;
};

cbuffer Shader
{
	float4 ImpostorParam		;
};

#define IMPOSTOR_LAYOUT			(ImpostorParam.x)	// e.g. 16 means 16 x 16 tiles.
#define IMPOSTOR_LAYOUT_INV		(ImpostorParam.y)	// 1 / IMPOSTOR_LAYOUT
#define IMPOSTOR_QUANTIZE		(ImpostorParam.z)	// (IMPOSTOR_LAYOUT - 1) / 2
#define IMPOSTOR_QUANTIZE_INV	(ImpostorParam.w)	// 1 / IMPOSTOR_QUANTIZE 

// Impostor textures.
texture2D DiffuseTexture : layout(t0) = "Transparent";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthTexture : layout(t2) = "Black";
sampler DepthTexture_Sampler : layout(s2)
{
	Texture = DepthTexture;
	Filter = Linear;
	Address = Clamp;
};

struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		VertexColor 	: COLOR;
	float2		TexCoord		: TEXCOORDN;
};

//////////////////////////////////////////////////////////////////////////
#include "Lighting.fxh"

/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate.
	float4		EyeVector		: TEXCOORDN;			//	Camera position - World position.
	float4		SunDirection	: TEXCOORDN;			//	Rotated sun direction & linear depth.

	float4		ViewDir			: TEXCOORDN;			//	Octahedron view direction 1.
	[@_ENABLE_FRAME_BLEND]
	float4		ViewDir2		: TEXCOORDN;			//	Octahedron view direction 2.
	[@_ENABLE_FRAME_BLEND]
	float4		ViewDir3		: TEXCOORDN;			//	Octahedron view direction 3.
	[@_ENABLE_FRAME_BLEND || @_ENABLE_FREE_ROTATION]
	float4		LocalVector		: TEXCOORDN;			//	World position - Impostor center.
};

// Project the hemisphere onto the hemi-octahedron.
float2 float3_to_hemioct(float3 v)
{
	float2 p = v.xz * (1.0 / (abs(v.x) + abs(v.y) + abs(v.z)));
	return float2(p.x + p.y, p.x - p.y);
}

// Map hemi-octahedron to hemisphere vector.
float3 hemioct_to_float3(float2 e)
{
	float2 t = float2(e.x + e.y, e.x - e.y) * 0.5;
	float3 v = float3(t.x, 1.0 - abs(t.x) - abs(t.y), t.y);
	return normalize(v);
}

//////////////////////////////////////////////////////////////////////////
// Common Impostor vertex shader.

void GetRotationMatrix(in VertexAssemble IN, out float3 r0, out float3 r1, out float3 r2)
{
	float rotation = IN.TexCoord.y;
	float sinyaw, cosyaw;
	sincos(rotation, sinyaw, cosyaw);

	if (@_ENABLE_FREE_ROTATION)
	{
		float2 pitchRoll = GetVertexColor(IN.VertexColor).zw * (PI * 2);
		float sinpitch, cospitch, sinroll, cosroll;
		sincos(pitchRoll.x, sinpitch, cospitch);
		sincos(pitchRoll.y, sinroll, cosroll);
		r0.x = cosroll * cosyaw + sinpitch * sinroll * sinyaw;
		r0.y = cospitch * sinroll;
		r0.z = cosyaw * sinpitch * sinroll - cosroll * sinyaw;
		r1.x = -cosyaw * sinroll + cosroll * sinpitch * sinyaw;
		r1.y = cospitch * cosroll;
		r1.z = cosroll * cosyaw * sinpitch + sinroll * sinyaw;
		r2.x = cospitch * sinyaw;
		r2.y = -sinpitch;
		r2.z = cospitch * cosyaw;
	}
	else
	{
		r0 = float3(cosyaw, sinyaw, -sinyaw);
		r1 = float3(0, 0, 0);
		r2 = float3(0, 0, 0);
	}
}

float3 RotateVector(float3 v, float3 r0, float3 r1, float3 r2)
{
	if (@_ENABLE_FREE_ROTATION)
	{
		return float3(
			v.x * r0.x + v.y * r1.x + v.z * r2.x,
			v.x * r0.y + v.y * r1.y + v.z * r2.y,
			v.x * r0.z + v.y * r1.z + v.z * r2.z);
	}
	else
	{
		// r0 = (cosyaw, sinyaw, -sinyaw)
		return float3(dot(v.xz, r0.xy), v.y, dot(v.xz, r0.zx));
	}
}

float3 RotateVectorInverse(float3 v, float3 r0, float3 r1, float3 r2)
{
	if (@_ENABLE_FREE_ROTATION)
	{
		return float3(dot(v, r0), dot(v, r1), dot(v, r2));
	}
	else
	{
		// r0 = (cosyaw, sinyaw, -sinyaw)
		return float3(dot(v.xz, r0.xz), v.y, dot(v.xz, r0.yx));
	}
}

VertexShadingOutput ImpostorVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	float3 pos = IN.Position.xyz;
	float2 tc = GetVertexColor(IN.VertexColor).xy;
	float radius = IN.TexCoord.r;

	float3 r0, r1, r2;
	GetRotationMatrix(IN, r0, r1, r2);

	float3 viewdir = CameraPos.xyz - pos;
	float3 viewrot;
	if (@_ENABLE_FREE_ROTATION)
	{
		viewdir = normalize(viewdir);
		viewrot = RotateVectorInverse(viewdir, r0, r1, r2);
		viewrot.y = max(viewrot.y, 0);
		viewrot = normalize(viewrot);
	}
	else
	{
		viewdir.y = max(viewdir.y, 0);
		viewdir = normalize(viewdir);
		viewrot = RotateVectorInverse(viewdir, r0, r1, r2);
	}

	if (@_ENABLE_FRAME_BLEND)
	{
		// Find the nearest 3 octahedron view vectors.
		float2 hemioct = float3_to_hemioct(viewrot);
		hemioct = max((hemioct + 1.0) * IMPOSTOR_QUANTIZE, float2(0, 0));
		float2 quantized = min(floor(hemioct), float2(IMPOSTOR_LAYOUT - 2, IMPOSTOR_LAYOUT - 2));
		float2 weight = hemioct - quantized;
		float2 offset = float2(1.0, 0.0);
		if (weight.x < weight.y)
		{
			weight.xy = weight.yx;
			offset.xy = offset.yx;
		}

		hemioct = quantized / IMPOSTOR_QUANTIZE - 1.0;
		OUT.ViewDir = float4(hemioct_to_float3(hemioct), 1 - weight.x);
		OUT.ViewDir2 = float4(hemioct_to_float3(hemioct + offset * IMPOSTOR_QUANTIZE_INV), weight.x - weight.y);
		OUT.ViewDir3 = float4(hemioct_to_float3(hemioct + 1.0 * IMPOSTOR_QUANTIZE_INV), weight.y);
		OUT.TexCoord = float4((quantized + 0.5) * IMPOSTOR_LAYOUT_INV, offset);

		// Enlarge the quad a little bit to avoid truncation.
		radius = radius * 1.1f;
	}
	else
	{
		// Locked to original octahedron view vectors.
		float2 hemioct = float3_to_hemioct(viewrot);
		float2 quantized = floor((hemioct + 1.0) * IMPOSTOR_QUANTIZE + 0.5);
		hemioct = quantized / IMPOSTOR_QUANTIZE - 1.0;
		viewdir = hemioct_to_float3(hemioct);

		OUT.ViewDir = float4(viewdir, 1.0);
		OUT.TexCoord = float4(quantized + tc, quantized + 0.5) * IMPOSTOR_LAYOUT_INV;
		viewdir = RotateVector(viewdir, r0, r1, r2);
	}

	float3 right = normalize(cross(float3(0, 1, 0), viewdir));
	float3 up = cross(viewdir, right);
	float2 extend = (tc * 2.0 - 1.0) * radius;
	pos = pos + extend.x * right - extend.y * up;
	OUT.HPosition = MulMat4x4(ViewProj, float4(pos, 1));

	float3 eye = RotateVectorInverse(CameraPos.xyz - pos, r0, r1, r2);
	OUT.EyeVector = float4(eye, FogTransmittance(-eye));
	OUT.SunDirection.xyz = RotateVectorInverse(SunDirection.xyz, r0, r1, r2);
	OUT.SunDirection.w = OUT.HPosition.w * CameraInfo.z;

	if (@_ENABLE_FRAME_BLEND || @_ENABLE_FREE_ROTATION)
	{
		// Output local vector for virtual-frame projection.
		float3 local = RotateVectorInverse(pos - IN.Position.xyz, r0, r1, r2);
		OUT.LocalVector = float4(local, 0.5 / IN.TexCoord.r);
	}

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Common Impostor pixel shader.

float2 GetParallaxOffset(float2 uv, float3 eye, float3 outside, float3 right, float3 up, out float3 normal)
{
	float4 depthNormal = SampleColorMap(DepthTexture, DepthTexture_Sampler, uv) * 2.0 - 1.0;
	float2 nxy = depthNormal.rb;
	float nz = depthNormal.a * sqrt(saturate(1 - dot(nxy, nxy)));
	normal = nxy.x * right + nxy.y * up + nz * outside;

	float depth = -depthNormal.g;
	float proj = dot(eye, outside);
	float2 uvProj = float2(dot(eye, right), dot(eye, up));
	float scale = depth / proj * 0.5 * IMPOSTOR_LAYOUT_INV;
	return uvProj * float2(scale, -scale);
}

float2 ProjectToImpostorUV(float3 eye, float3 outside, float3 local, float2 uvbase, float invsize, out float3 normal)
{
	float3 right = normalize(cross(float3(0, 1, 0), outside));
	float3 up = cross(outside, right);
	float2 offset = float2(dot(local, right), -dot(local, up)) * invsize;
	float2 uv = uvbase + (saturate(offset + 0.5) - 0.5) * IMPOSTOR_LAYOUT_INV;
	float2 parallax = GetParallaxOffset(uv, eye, outside, right, up, normal);
	return uv + parallax;
}

struct ImpostorPixelOutput
{
	float4 Color	: SV_Target0;
	[@_DEFERRED_SHADING]
	float4 GBufferB : SV_Target1;	// GBufferB: Metallic, Specular, Roughness, ShadingModelID
	[@_DEFERRED_SHADING]
	float4 GBufferC : SV_Target2;	// GBufferC: BaseColor.rgb, AO
	[@_DEFERRED_SHADING]
	float4 GBufferD : SV_Target3;	// GBufferD: CustomData
	[@_DEFERRED_SHADING]
	float4 GBufferE : SV_Target4;	// GBufferE: N.xyz, InShadow
	[@_DEFERRED_SHADING]
	float4 GBufferF : SV_Target5;	// GBufferF: TempRT for OnChipMem optimization
	[@_DEFERRED_SHADING]
	float4 GBufferG : SV_Target6;	// GBufferG: Depth(r32f)
};
[target("X16F", "X16F", "X16F", "X16F", "X16F", "X16F", "R32F")]
ImpostorPixelOutput ImpostorPixelShading(in VertexShadingOutput IN)
{
	float4 OUT;
	float3 eye = IN.EyeVector.xyz;
	float3 normal;

	if (@_ENABLE_FRAME_BLEND)
	{	
		float3 local = IN.LocalVector.xyz;
		float3 normal1;
		float2 uvbase1 = IN.TexCoord.xy;
		float2 uv1 = ProjectToImpostorUV(eye, IN.ViewDir.xyz, local, uvbase1, IN.LocalVector.w, normal1);
		float4 color1 = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, uv1);

		float3 normal2;
		float2 uvbase2 = IN.TexCoord.xy + IN.TexCoord.zw * IMPOSTOR_LAYOUT_INV;
		float2 uv2 = ProjectToImpostorUV(eye, IN.ViewDir2.xyz, local, uvbase2, IN.LocalVector.w, normal2);
		float4 color2 = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, uv2);

		float3 normal3;
		float2 uvbase3 = IN.TexCoord.xy + IMPOSTOR_LAYOUT_INV;
		float2 uv3 = ProjectToImpostorUV(eye, IN.ViewDir3.xyz, local, uvbase3, IN.LocalVector.w, normal3);
		float4 color3 = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, uv3);

		// Blend 3 frames together.
		OUT = color1 * IN.ViewDir.w + color2 * IN.ViewDir2.w + color3 * IN.ViewDir3.w;

		// Blend normal?
		normal = normal1 * IN.ViewDir.w + normal2 * IN.ViewDir2.w + normal3 * IN.ViewDir3.w;
		//normal = normal1;
	}
	else if (@_ENABLE_FREE_ROTATION)
	{
		float3 local = IN.LocalVector.xyz;
		float2 uvbase = IN.TexCoord.zw;
		float2 uv = ProjectToImpostorUV(eye, IN.ViewDir.xyz, local, uvbase, IN.LocalVector.w, normal);
		OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, uv);
	}
	else
	{
		float3 right = normalize(cross(float3(0, 1, 0), IN.ViewDir.xyz));
		float3 up = cross(IN.ViewDir.xyz, right);

		float2 uv = IN.TexCoord.xy;
		float2 parallax = GetParallaxOffset(uv, eye, IN.ViewDir.xyz, right, up, normal);
		OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, uv + parallax);
	}

	clip(OUT.a - 0.5);
	
	ImpostorPixelOutput output;
	if (@_DEFERRED_SHADING)
	{
		GBufferData gbuffer;
		gbuffer.ShadingModelID = SHADINGMODELID_DEFAULT_LIT;
		gbuffer.BaseColor = DecodeGamma(OUT.rgb);
		gbuffer.Metallic = 0.0;
		gbuffer.Specular = 0.0;
		gbuffer.Roughness = 1.0;
		gbuffer.WorldNormal = normal;
		gbuffer.Emission = Diffuse_Lambert(gbuffer.BaseColor) * AmbientColor.rgb;
		gbuffer.InShadow = 0;
		gbuffer.AO = 1;
		gbuffer.CustomData = half4(0, 0, 0, 0);

		EncodeGBufferData(gbuffer, output.Color, output.GBufferB, output.GBufferC, output.GBufferD, output.GBufferE, OUT.GBufferF);
		output.GBufferF = float4(0, 0, 0, 0);
		output.GBufferG = EncodeDepth(IN.SunDirection.w);
	}
	else
	{
		float3 light = SunColor.rgb * saturate(dot(IN.SunDirection.xyz, normal)) + AmbientColor.rgb;
		float3 diffuse = Diffuse_Lambert(DecodeGamma(OUT.rgb));
		OUT.rgb = light * diffuse;

		// Fog.
		if (@_FOG && !@_VolumetricFog)
		{
			OUT.rgb = FogLighting(IN.EyeVector.w, OUT.rgb);
		}

		EncodeHDR(OUT);
		OUT.a = EncodeSolidBloomFactor(0);
		output.Color = OUT;
	}

	return output;
}

technique Shading
{
	VertexShader = ImpostorVertexShading();
	PixelShader = ImpostorPixelShading();
}
