#ifndef __FILE_DISK_LIGHTING_FXH__
#define __FILE_DISK_LIGHTING_FXH__

#include "DeferredShadingCommon.fxh"

#define LUT_SIZE (64.0)
#define LUT_SCALE ((LUT_SIZE - 1.0)/LUT_SIZE)
#define LUT_BIAS (0.5/LUT_SIZE)

// An extended version of the implementation from
// "How to solve a cubic equation, revisited"
// http://momentsingraphics.de/?p=105
float3 SolveCubic(float4 Coefficient)
{
    // Normalize the polynomial
    Coefficient.xyz /= Coefficient.w;
    // Divide middle coefficients by three
    Coefficient.yz /= 3.0;

    float A = Coefficient.w;
    float B = Coefficient.z;
    float C = Coefficient.y;
    float D = Coefficient.x;

    // Compute the Hessian and the discriminant
    float3 Delta = float3(
        -Coefficient.z*Coefficient.z + Coefficient.y,
        -Coefficient.y*Coefficient.z + Coefficient.x,
        dot(float2(Coefficient.z, -Coefficient.y), Coefficient.xy)
    );

    float Discriminant = dot(float2(4.0*Delta.x, -Delta.y), Delta.zy);

    float3 RootsA, RootsD;

    float2 xlc, xsc;

    // Algorithm A
    {
        float A_a = 1.0;
        float C_a = Delta.x;
        float D_a = -2.0*B*Delta.x + Delta.y;

        // Take the cubic root of a normalized complex number
        float Theta = atan2(sqrt(Discriminant), -D_a)/3.0;

        float x_1a = 2.0*sqrt(-C_a)*cos(Theta);
        float x_3a = 2.0*sqrt(-C_a)*cos(Theta + (2.0/3.0)*PI);

        float xl;
        if ((x_1a + x_3a) > 2.0*B)
            xl = x_1a;
        else
            xl = x_3a;

        xlc = float2(xl - B, A);
    }

    // Algorithm D
    {
        float A_d = D;
        float C_d = Delta.z;
        float D_d = -D*Delta.y + 2.0*C*Delta.z;

        // Take the cubic root of a normalized complex number
        float Theta = atan2(D*sqrt(Discriminant), -D_d)/3.0;

        float x_1d = 2.0*sqrt(-C_d)*cos(Theta);
        float x_3d = 2.0*sqrt(-C_d)*cos(Theta + (2.0/3.0)*PI);

        float xs;
        if (x_1d + x_3d < 2.0*C)
            xs = x_1d;
        else
            xs = x_3d;

        xsc = float2(-D, xs + C);
    }

    float E =  xlc.y*xsc.y;
    float F = -xlc.x*xsc.y - xlc.y*xsc.x;
    float G =  xlc.x*xsc.x;

    float2 xmc = float2(C*F - B*G, -B*F + C*E);

    float3 Root = float3(xsc.x/xsc.y, xmc.x/xmc.y, xlc.x/xlc.y);

    if (Root.x < Root.y && Root.x < Root.z)
        Root.xyz = Root.yxz;
    else if (Root.z < Root.x && Root.z < Root.y)
        Root.xyz = Root.xzy;

    return Root;
}

float SphereIntegral(float LF, float FZ)
{
    //return max((LF*LF + F.z)/(LF+1.0), 0.0);
	return max((LF*LF + FZ)/(LF+1.0), 0.0);
}

half3 DiskSpecularLightingLTC(in GBufferData GBuffer, in LightingData LitData, in BxDFContext Context)
{
    if (LitData.PBRQualityLevel == PBRQualityLevel_None)
        return half3(0, 0, 0);

	float2 uv = float2(GBuffer.Roughness, sqrt(1.0 - Context.NoV));
    uv = uv*LUT_SCALE + LUT_BIAS;

	float4 LTCMat = tLTCMatTexture.SampleLevel(sLTCMatSampler, uv, 0);
	float4 LTCGF = tLTCGFTexture.SampleLevel(sLTCGFSampler, uv, 0);

	float3x3 LTC = float3x3(
		LTCMat.x, 0, LTCMat.z,
		       0, 1,        0,
		LTCMat.y, 0, LTCMat.w
    );

	 // Rotate to tangent space
	float3 T1 = normalize(LitData.V - LitData.N * dot(LitData.N, LitData.V));
	float3 T2 = cross(LitData.N, T1);
	float3x3 TangentBasis = float3x3(
		T1.x,        T1.y,        T1.z,
		T2.x,        T2.y,        T2.z,
		LitData.N.x, LitData.N.y, LitData.N.z
    );
	float3 L_[4];
	L_[0] = mul(TangentBasis, LitData.Origin - LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	L_[1] = mul(TangentBasis, LitData.Origin + LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	L_[2] = mul(TangentBasis, LitData.Origin + LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);
	L_[3] = mul(TangentBasis, LitData.Origin - LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);

	// init ellipse
    float3 C  = 0.5 * (L_[0] + L_[2]);
    float3 V1 = 0.5 * (L_[1] - L_[2]);
    float3 V2 = 0.5 * (L_[1] - L_[0]);

	// LTC
    C  = mul(LTC, C);
    V1 = mul(LTC, V1);
    V2 = mul(LTC, V2);

	// compute eigenvectors of ellipse
    float a, b;
    float d11 = dot(V1, V1);
    float d22 = dot(V2, V2);
    float d12 = dot(V1, V2);
    if (abs(d12)/sqrt(d11*d22) > 0.0001)
    {
        float tr = d11 + d22;
        float det = -d12*d12 + d11*d22;

        // use sqrt matrix to solve for eigenvalues
        det = sqrt(det);
        float u = 0.5*sqrt(tr - 2.0*det);
        float v = 0.5*sqrt(tr + 2.0*det);
        float e_max = Square(u + v);
        float e_min = Square(u - v);

        float3 V1_, V2_;

        if (d11 > d22)
        {
            V1_ = d12*V1 + (e_max - d11)*V2;
            V2_ = d12*V1 + (e_min - d11)*V2;
        }
        else
        {
            V1_ = d12*V2 + (e_max - d22)*V1;
            V2_ = d12*V2 + (e_min - d22)*V1;
        }

        a = 1.0 / e_max;
        b = 1.0 / e_min;
        V1 = normalize(V1_);
        V2 = normalize(V2_);
    }
    else
    {
        a = 1.0 / dot(V1, V1);
        b = 1.0 / dot(V2, V2);
        V1 *= sqrt(a);
        V2 *= sqrt(b);
    }

    float3 V3 = cross(V1, V2);
    if (dot(C, V3) < 0.0)
        V3 *= -1.0;

    float L  = dot(V3, C);
    float x0 = dot(V1, C) / L;
    float y0 = dot(V2, C) / L;

    float E1 = rsqrt(a);
    float E2 = rsqrt(b);

    a *= L*L;
    b *= L*L;

    float c0 = a*b;
    float c1 = a*b*(1.0 + x0*x0 + y0*y0) - a - b;
    float c2 = 1.0 - a*(1.0 + x0*x0) - b*(1.0 + y0*y0);
    float c3 = 1.0;

    float3 roots = SolveCubic(float4(c0, c1, c2, c3));
    float e1 = roots.x;
    float e2 = roots.y;
    float e3 = roots.z;

    float3 avgDir = float3(a*x0/(a - e2), b*y0/(b - e2), 1.0);

    float3x3 rotate = float3x3(
		V1.x, V2.x, V3.x,
		V1.y, V2.y, V3.y,
		V1.z, V2.z, V3.z
	);
	//mat3 rotate = mat3_from_columns(V1, V2, V3);

    avgDir = mul(rotate, avgDir);
    avgDir = normalize(avgDir);

    float L1 = sqrt(-e2/e3);
    float L2 = sqrt(-e2/e1);
    float LF = L1*L2*rsqrt((1.0 + L1*L1)*(1.0 + L2*L2));
	float backFace = step(0, dot(LitData.Axis[2].xyz, L));
	float FZ = avgDir.z;
    float NoL = SphereIntegral(LF, FZ);
	float Irradiance = LF * NoL;
	// Kill negative and NaN
	Irradiance = -min(-Irradiance, 0.0); 

	// BRDF shadowing and Fresnel(e.g.,G*F)
	float3 SpecularColor = LTCGF.y + ( LTCGF.x - LTCGF.y ) * GBuffer.SpecularColor;

	return LitData.FalloffColor * Irradiance * SpecularColor;
}

/*
LightParams0    Origin, InvSquareRange
LightParams1    Color, Extent.x
LightParams2    AxisX, Extent.y
LightParams3    AxisY, 
AxisZ, 
*/
LightingData GetDiskLightingData(in float3 P, in GBufferData GBuffer, in half3 V, 
	in half InShadow, int PBRQualityLevel, bool TwoSided, in float4 LightParams[4], 
    in float3 AxisZ)
{
    LightingData LitData;
    float3 LitDir = LightParams[0].xyz - P;
	float D2 = dot(LitDir, LitDir);
    LitData.FalloffColor = LightParams[1].rgb * LightRangeMask(D2, LightParams[0].w);
    LitData.L = LitDir;
    LitData.V = V;
    LitData.N = GBuffer.WorldNormal;
    LitData.Origin = LitDir; //move (0,0,0) to P;
    LitData.Axis = float3x3(
        LightParams[2].x, LightParams[2].y, LightParams[2].z,
        LightParams[3].x, LightParams[3].y, LightParams[3].z,
        AxisZ.x,          AxisZ.y,          AxisZ.z
    );
    LitData.Extent.x = LightParams[1].w;
    LitData.Extent.y = LightParams[2].w;

    // Rotate to tangent space
	float3 T1 = normalize(LitData.V - LitData.N * dot(LitData.N, LitData.V));
	float3 T2 = cross(LitData.N, T1);
	float3x3 TangentBasis = float3x3(
		T1.x,        T1.y,        T1.z,
		T2.x,        T2.y,        T2.z,
		LitData.N.x, LitData.N.y, LitData.N.z
    );
	float3 L_[4];
	L_[0] = mul(TangentBasis, LitData.Origin - LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	L_[1] = mul(TangentBasis, LitData.Origin + LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	L_[2] = mul(TangentBasis, LitData.Origin + LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);
	L_[3] = mul(TangentBasis, LitData.Origin - LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);

	// init ellipse
    float3 C  = 0.5 * (L_[0] + L_[2]);
    float3 V1 = 0.5 * (L_[1] - L_[2]);
    float3 V2 = 0.5 * (L_[1] - L_[0]);

	// do not need transforming in diffuse lighting
    // C  = Minv * C;
    // V1 = Minv * V1;
    // V2 = Minv * V2;

	// compute eigenvectors of ellipse
    float a, b;
    float d11 = dot(V1, V1);
    float d22 = dot(V2, V2);
    float d12 = dot(V1, V2);
    if (abs(d12)/sqrt(d11*d22) > 0.0001)
    {
        float tr = d11 + d22;
        float det = -d12*d12 + d11*d22;

        // use sqrt matrix to solve for eigenvalues
        det = sqrt(det);
        float u = 0.5*sqrt(tr - 2.0*det);
        float v = 0.5*sqrt(tr + 2.0*det);
        float e_max = Square(u + v);
        float e_min = Square(u - v);

        float3 V1_, V2_;

        if (d11 > d22)
        {
            V1_ = d12*V1 + (e_max - d11)*V2;
            V2_ = d12*V1 + (e_min - d11)*V2;
        }
        else
        {
            V1_ = d12*V2 + (e_max - d22)*V1;
            V2_ = d12*V2 + (e_min - d22)*V1;
        }

        a = 1.0 / e_max;
        b = 1.0 / e_min;
        V1 = normalize(V1_);
        V2 = normalize(V2_);
    }
    else
    {
        a = 1.0 / dot(V1, V1);
        b = 1.0 / dot(V2, V2);
        V1 *= sqrt(a);
        V2 *= sqrt(b);
    }

    float3 V3 = cross(V1, V2);
    if (dot(C, V3) < 0.0)
        V3 *= -1.0;

    float L  = dot(V3, C);
    float x0 = dot(V1, C) / L;
    float y0 = dot(V2, C) / L;

    float E1 = rsqrt(a);
    float E2 = rsqrt(b);

    a *= L*L;
    b *= L*L;

    float c0 = a*b;
    float c1 = a*b*(1.0 + x0*x0 + y0*y0) - a - b;
    float c2 = 1.0 - a*(1.0 + x0*x0) - b*(1.0 + y0*y0);
    float c3 = 1.0;

    float3 roots = SolveCubic(float4(c0, c1, c2, c3));
    float e1 = roots.x;
    float e2 = roots.y;
    float e3 = roots.z;

    float3 avgDir = float3(a*x0/(a - e2), b*y0/(b - e2), 1.0);

    float3x3 rotate = float3x3(
		V1.x, V2.x, V3.x,
		V1.y, V2.y, V3.y,
		V1.z, V2.z, V3.z
	);
	//mat3 rotate = mat3_from_columns(V1, V2, V3);

    avgDir = mul(rotate, avgDir);
    avgDir = normalize(avgDir);

    float L1 = sqrt(-e2/e3);
    float L2 = sqrt(-e2/e1);
    float LF = L1*L2*rsqrt((1.0 + L1*L1)*(1.0 + L2*L2));
	float backFace = step(0, dot(AxisZ.xyz, L));
	float FZ = avgDir.z;
    float NoL = SphereIntegral(LF, FZ);
    
    LitData.Atten = LF;
    LitData.NoL = LitData._NoL = NoL;
    LitData.InShadow = InShadow;
    LitData.RadiusTan = 0;
	LitData.PBRQualityLevel = PBRQualityLevel;
	LitData.TranslucencyLighting = false;
	LitData.IsSphericalLight = false;
	LitData.IsRectLight = false;
	LitData.IsDiskLight = true;
    LitData.TwoSided = TwoSided;
	return LitData;
}
#endif //__FILE_DISK_LIGHTING_FXH__
