//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liuzhongyuan@corp.netease.com
//	Author: Jrenliu
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<GpuClothCommon.fxh>
///	@path	~/System/Shader/
///	@date	2020/04/28
///	@brief	GPU cloth shader.

/* Flags
    INCLUDE_GPU_PHYSICS_DATA
    INCLUDE_GPU_PHYSICS_SKINNING_TOOL
    INCLUDE_GPU_PHYSICS_COLLISION_TOOL
*/
#ifndef __FILE_GPU_CLOTH_COMMON_FXH__
#define __FILE_GPU_CLOTH_COMMON_FXH__
//@{ Should be same with the GpuCloth.h
#define GPUPhyMaxClothNum 256

#define GPUPhyParticleTexWidth 512
#define GPUPhyParticleTexHeight 512

#define GPUPhyConstraintTexWidth 512
#define GPUPhyConstraintTexHeight 512

#define GPUPhyRuntimeTexWidth 512
#define GPUPhyRuntimeTexHeight 512
#define GPUPhyRuntimeCurMatStart 0
#define GPUPhyRuntimePreToCurMatStart 3
#define GPUPhyRuntimeColHeaderStart 6
#define GPUPhyRuntimeColDataStart 7

#define GPUPhySkeletonTexWidth 256
#define GPUPhySkeletonTexHeight GPUPhyMaxClothNum
//@}

float2 cvtUv(float index, float width, float height)
{
    float2 uv;
	float row = floor(index / width);
	uv.x = (index - width * row + 0.5) / width;
    uv.y = (row + 0.5) / height;
    return uv;
}

#ifdef INCLUDE_GPU_PHYSICS_VS
struct VertexAssemble
{
    float4 Position 		: POSITION;
    float2 TexCoord         : TEXCOORDN;
};

struct SimpleVSOutput
{
    float4 Position         : SV_Position;
    float2 TexCoord         : TEXCOORDN;
};

SimpleVSOutput SimpleVS(VertexAssemble IN)
{
	SimpleVSOutput OUT;
	OUT.Position = IN.Position;
	[@_GLSL]
	OUT.Position.y = -IN.Position.y;
	OUT.TexCoord = IN.TexCoord;
	return OUT;
}
#endif

#ifdef INCLUDE_GPU_PHYSICS_DATA
/*
    \brief Pixel => Particle
    \r cloth index
    \b constriant offset
*/
texture2D tFabricTexture : layout(t0) = "Black";
sampler sFabricTextureSampler : layout(s0)
{
	Texture = tFabricTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

#endif

#ifdef INCLUDE_GPU_PHYSICS_SKINNING_TOOL 

texture2D tSkeletonTexture : layout(t1) = "Black";
sampler sSkeletonTextureSampler : layout(s1)
{
	Texture = tSkeletonTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tStaticPositionTexture : layout(t2) = "Black";
sampler sStaticPositionTextureSampler : layout(s2)
{
	Texture = tStaticPositionTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tWeightTexture : layout(t3) = "Black";
sampler sWeightTextureSampler : layout(s3)
{
	Texture = tWeightTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2D tIndexTexture : layout(t4) = "Black";
sampler sIndexTextureSampler : layout(s4)
{
	Texture = tIndexTexture;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

// Get 3x4 matrix from skeleton data.
void GetSkeletonMatrix4x3Tex(uint idx, out float4 Mx, out float4 My, out float4 Mz )
{
    float2 newUv = cvtUv(idx, GPUPhySkeletonTexWidth, GPUPhySkeletonTexHeight);
	Mx = tSkeletonTexture.SampleLevel(sSkeletonTextureSampler, newUv, 0);
	My = tSkeletonTexture.SampleLevel(sSkeletonTextureSampler, newUv + float2(1.0/GPUPhySkeletonTexWidth, 0.0), 0);
	Mz = tSkeletonTexture.SampleLevel(sSkeletonTextureSampler, newUv + float2(2.0/GPUPhySkeletonTexWidth, 0.0), 0);
}

void SkinningTex(in uint tex_start, in float4 w, in uint4 i, in float3 static_position, inout float3 position)
{		
    // Fetch.
    float4 M0[3], M1[3], M2[3], M3[3];
    GetSkeletonMatrix4x3Tex(tex_start + 3 * i[0], M0[0], M0[1], M0[2]);
    GetSkeletonMatrix4x3Tex(tex_start + 3 * i[1], M1[0], M1[1], M1[2]);
    GetSkeletonMatrix4x3Tex(tex_start + 3 * i[2], M2[0], M2[1], M2[2]);
    GetSkeletonMatrix4x3Tex(tex_start + 3 * i[3], M3[0], M3[1], M3[2]);
    // Blend.
    float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
    float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
    float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);
    // Apply.
    position = Matrix4x3TransformP(Mx, My, Mz, static_position);
}

#endif

texture2D tPhysicsResult : PhysicsResult : layout(t9);
sampler sPhysicsResultSampler : layout(s9)
{
	Texture = tPhysicsResult;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

float3 FetchGpuClothResult(float3 key, uint index, float3 defaultValue)
{
	// x is model type
	if (key.x != -101.0f)
		return defaultValue;
	// z if position offset
    uint GpuIndex = (uint)(key.z) + index;
	float2 PositionUv = cvtUv(GpuIndex, GPUPhyParticleTexWidth, GPUPhyParticleTexHeight);
	float4 deformation =  tPhysicsResult.SampleLevel(sPhysicsResultSampler, PositionUv, 0);
    return deformation.xyz;
}

#endif







