#ifndef __FILE_RECT_LIGHTING_FXH__
#define __FILE_RECT_LIGHTING_FXH__

#include "DeferredShadingCommon.fxh"

#define LUT_SIZE (64.0)
#define LUT_SCALE ((LUT_SIZE - 1.0)/LUT_SIZE)
#define LUT_BIAS (0.5/LUT_SIZE)

// Linearly Transformed Cosines
// [ Hill et al. 2016, "Real-Time Area Lighting: a Journey from Research to Production" ]
float3 IntegrateEdgeVec(float3 v1, float3 v2)
{
    float x = dot(v1, v2);
    float y = abs(x);

    float a = 0.8543985 + (0.4965155 + 0.0145206*y)*y;
    float b = 3.4175940 + (4.1616724 + y)*y;
    float v = a / b;

    float theta_sintheta = (x > 0.0) ? v : 0.5*rsqrt(max(1.0 - x*x, 1e-7)) - v;

    return cross(v1, v2)*theta_sintheta;
}

float IntegrateEdge(float3 v1, float3 v2)
{
    float x = dot(v1, v2);
    float y = abs(x);

    float a = 0.8543985 + (0.4965155 + 0.0145206*y)*y;
    float b = 3.4175940 + (4.1616724 + y)*y;
    float v = a / b;

    float theta_sintheta = (x > 0.0) ? v : 0.5*rsqrt(max(1.0 - x*x, 1e-7)) - v;
	return theta_sintheta;
}

// Optimized Lambert poly irradiance
float3 PolygonIrradiance( float3 Poly[4] )
{
	float3 L0 = normalize( Poly[0] );	//  2 mad, 4 mul, 1 rsqrt
	float3 L1 = normalize( Poly[1] );	//  2 mad, 4 mul, 1 rsqrt
	float3 L2 = normalize( Poly[2] );	//  2 mad, 4 mul, 1 rsqrt
	float3 L3 = normalize( Poly[3] );	//  2 mad, 4 mul, 1 rsqrt
	// 24 alu, 4 rsqrt

#if 0
	float3 L;
	L  = acos( dot( L0, L1 ) ) * normalize( cross( L0, L1 ) );
	L += acos( dot( L1, L2 ) ) * normalize( cross( L1, L2 ) );
	L += acos( dot( L2, L3 ) ) * normalize( cross( L2, L3 ) );
	L += acos( dot( L3, L0 ) ) * normalize( cross( L3, L0 ) );
#else
	float w01 = IntegrateEdge( L0, L1 );
	float w12 = IntegrateEdge( L1, L2 );
	float w23 = IntegrateEdge( L2, L3 );
	float w30 = IntegrateEdge( L3, L0 );

#if 0
	float3 L;
	L  = w01 * cross( L0, L1 );	// 6 mul, 3 mad
	L += w12 * cross( L1, L2 );	// 3 mul, 6 mad
	L += w23 * cross( L2, L3 );	// 3 mul, 6 mad
	L += w30 * cross( L3, L0 );	// 3 mul, 6 mad
#else
	float3 L;
	L  = cross( L1, -w01 * L0 +  w12 * L2 );	// 6 mul, 6 mad
	L += cross( L3,  w30 * L0 + -w23 * L2 );	// 3 mul, 9 mad
#endif
#endif

	// Vector irradiance
	return L;
}

float SphereIntegral(float LF, float3 F)
{
    //return max((LF*LF + F.z)/(LF+1.0), 0.0);
	return max((LF*LF + F.z)/(LF+1.0), 0.0);
}

half3 RectSpecularLightingLTC(in GBufferData GBuffer, in LightingData LitData, in BxDFContext Context)
{
    if (LitData.PBRQualityLevel == PBRQualityLevel_None)
        return half3(0, 0, 0);

	float2 uv = float2(GBuffer.Roughness, sqrt(1.0 - Context.NoV));
    uv = uv*LUT_SCALE + LUT_BIAS;

	float4 LTCMat = tLTCMatTexture.SampleLevel(sLTCMatSampler, uv, 0);
	float4 LTCGF = tLTCGFTexture.SampleLevel(sLTCGFSampler, uv, 0);

	float3x3 LTC = float3x3(
		LTCMat.x, 0, LTCMat.z,
		       0, 1,        0,
		LTCMat.y, 0, LTCMat.w
    );

	// Rotate to tangent space
	float3 T1 = normalize(LitData.V - LitData.N * dot(LitData.N, LitData.V));
	float3 T2 = cross(LitData.N, T1);
	float3x3 TangentBasis = float3x3(
		T1.x,        T1.y,        T1.z,
		T2.x,        T2.y,        T2.z,
		LitData.N.x, LitData.N.y, LitData.N.z
    );
	LTC = mul( LTC, TangentBasis );
	
	float3 Poly[4];
	Poly[0] = mul(LTC, LitData.Origin - LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	Poly[1] = mul(LTC, LitData.Origin + LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	Poly[2] = mul(LTC, LitData.Origin + LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);
	Poly[3] = mul(LTC, LitData.Origin - LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);

	//form factor in main direction |F|
	float3 F = PolygonIrradiance(Poly);
	float LF = length(F);
    F /= LF;
    if (dot(LitData.Axis[2], LitData.L) < 0)
        F.z = -F.z;
    float NoL = SphereIntegral(LF, F);
	float Irradiance = LF * NoL;
	// Kill negative and NaN
	Irradiance = -min(-Irradiance, 0.0); 

	// BRDF shadowing and Fresnel(e.g.,G*F)
	float3 SpecularColor = LTCGF.y + ( LTCGF.x - LTCGF.y ) * GBuffer.SpecularColor;

	return LitData.FalloffColor * Irradiance * SpecularColor;
}

/*
LightParams0    Origin, InvSquareRange
LightParams1    Color, Extent.x
LightParams2    AxisX, Extent.y
LightParams3    AxisY, 
AxisZ, 
*/
LightingData GetRectLightingData(in float3 P, in GBufferData GBuffer, in half3 V, 
	in half InShadow, int PBRQualityLevel, bool TwoSided, in float4 LightParams[4], 
    in float3 AxisZ)
{
    LightingData LitData;
    float3 L = LightParams[0].xyz - P;
	float D2 = dot(L, L);
    LitData.FalloffColor = LightParams[1].rgb * LightRangeMask(D2, LightParams[0].w);
    LitData.L = L * rsqrt(D2);
    LitData.V = V;
    LitData.N = GBuffer.WorldNormal;
    LitData.Origin = L; //move (0,0,0) to P;
    LitData.Axis = float3x3(
        LightParams[2].x, LightParams[2].y, LightParams[2].z,
        LightParams[3].x, LightParams[3].y, LightParams[3].z,
        AxisZ.x,          AxisZ.y,          AxisZ.z
    );
    LitData.Extent.x = LightParams[1].w;
    LitData.Extent.y = LightParams[2].w;

    // Rotate to tangent space
	float3 T1 = normalize(LitData.V - LitData.N * dot(LitData.N, LitData.V));
	float3 T2 = cross(LitData.N, T1);
	float3x3 TangentBasis = float3x3(
		T1.x,        T1.y,        T1.z,
		T2.x,        T2.y,        T2.z,
		LitData.N.x, LitData.N.y, LitData.N.z
    );
	float3 Poly[4];
	Poly[0] = mul(TangentBasis, LitData.Origin - LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	Poly[1] = mul(TangentBasis, LitData.Origin + LitData.Axis[0] * LitData.Extent.x - LitData.Axis[1] * LitData.Extent.y);
	Poly[2] = mul(TangentBasis, LitData.Origin + LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);
	Poly[3] = mul(TangentBasis, LitData.Origin - LitData.Axis[0] * LitData.Extent.x + LitData.Axis[1] * LitData.Extent.y);

	//form factor in main direction |F|
	float3 F = PolygonIrradiance(Poly);
	float LF = length(F);
    F /= LF;
	float backFace = step(0, dot(AxisZ.xyz, L));
	F.z *= backFace * 2 - 1;
    float NoL = SphereIntegral(LF, F);
    
    LitData.Atten = LF;
    LitData.NoL = LitData._NoL = NoL;
    LitData.InShadow = InShadow;
    LitData.RadiusTan = 0;
	LitData.PBRQualityLevel = PBRQualityLevel;
	LitData.TranslucencyLighting = false;
	LitData.IsSphericalLight = false;
	LitData.IsRectLight = true;
	LitData.IsDiskLight = false;
    LitData.TwoSided = TwoSided;
	return LitData;
}
#endif //__FILE_RECT_LIGHTING_FXH__
