//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Project Messiah
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<ShadingMode.fxh>
///	@path	Shader/
///	@date	2019/06/27
///	@brief	shading models definition

#ifndef __SHADING_MODE_FXH__
#define __SHADING_MODE_FXH__

#include "BRDF.fxh"
#include "DeferredShadingCommon.fxh"
#include "RectLighting.fxh"
#include "DiskLighting.fxh"

// [ de Carpentier 2017, "Decima Engine: Advances in Lighting and AA" ]
void SphereMaxNoH(inout BxDFContext Context, float radiusTan, bool bNewtonIteration, bool IsSphericalLight)
{
	if( IsSphericalLight )
	{
		float CosAlpha = rsqrt( 1 + Pow2( radiusTan ) );
		float SinAlpha = CosAlpha * radiusTan;
	
		float RoL = 2 * Context.NoL * Context.NoV - Context.VoL;
		if( RoL >= CosAlpha )
		{
			Context.NoH = 1;
			Context.VoH = abs( Context.NoV );
		}
		else
		{
			float rInvLengthT = SinAlpha * rsqrt( 1 - RoL*RoL );
			float NoTr = rInvLengthT * ( Context.NoV - RoL * Context.NoL );
			float VoTr = rInvLengthT * ( 2 * Context.NoV*Context.NoV - 1 - RoL * Context.VoL );

			if( bNewtonIteration )
			{
				// dot( cross(N,L), V )
				float NxLoV = sqrt( saturate( 1 - Pow2(Context.NoL) - Pow2(Context.NoV) - Pow2(Context.VoL) + 2 * Context.NoL * Context.NoV * Context.VoL ) );

				float NoBr = rInvLengthT * NxLoV;
				float VoBr = rInvLengthT * NxLoV * 2 * Context.NoV;
				float NoLVTr = Context.NoL * CosAlpha + Context.NoV + NoTr;
				float VoLVTr = Context.VoL * CosAlpha + 1   + VoTr;

				float p = NoBr   * VoLVTr;
				float q = NoLVTr * VoLVTr;
				float s = VoBr   * NoLVTr;

				float xNum = q * ( -0.5 * p + 0.25 * VoBr * NoLVTr );
				float xDenom = p*p + s * (s - 2*p) + NoLVTr * ( (Context.NoL * CosAlpha + Context.NoV) * Pow2(VoLVTr) + q * (-0.5 * (VoLVTr + Context.VoL * CosAlpha) - 0.5) );
				float TwoX1 = 2 * xNum / ( Pow2(xDenom) + Pow2(xNum) );
				float SinTheta = TwoX1 * xDenom;
				float CosTheta = 1.0 - TwoX1 * xNum;
				NoTr = CosTheta * NoTr + SinTheta * NoBr;
				VoTr = CosTheta * VoTr + SinTheta * VoBr;
			}

			Context.NoL = Context.NoL * CosAlpha + NoTr;
			Context.VoL = Context.VoL * CosAlpha + VoTr;
			float InvLenH = rsqrt( 2 + 2 * Context.VoL );
			Context.NoH = saturate( ( Context.NoL + Context.NoV ) * InvLenH );
			Context.VoH = saturate( InvLenH + InvLenH * Context.VoL );
		}
	}
}

//Preintegrated Skin
//@tSkinProfileMap input texture for skin lighting
float D_GGX_Skin(float Roughness, float NoH)
{
	float NoHSquare = Square(NoH);
	float m = Roughness * Roughness;
	float m2 = m * m;
	float d = ( NoH * m2 - NoH ) * NoH + 1;	// 2 mad
	float lobe1 = m2 / ( PI*d*d );			// 2 mul, 1 rcp
	m2 *= 0.0625;
	//d = ( NoH * m2 - NoH ) * NoH + 1;	// 2 mad
	//float lobe2 = m2 / ( PI*d*d );			// 2 mul, 1 rcp
	float n = 2 / m2 - 2;
	float lobe2 =  (n+2) / (2*PI) * PhongShadingPow( NoH, n );

	return lerp(lobe1, lobe2, 0.17);
}

half3 SpecularGGX(in GBufferData GBuffer, in LightingData LitData, in BxDFContext Context)
{
	if (LitData.PBRQualityLevel >= PBRQualityLevel_Normal)
	{
		half D;
		if (GBuffer.ShadingModelID == SHADINGMODELID_PREINTEGRATED_SKIN ||
			GBuffer.ShadingModelID == SHADINGMODELID_SSS)
			D = D_GGX_Skin(GBuffer.Roughness, Context.NoH);
		else
			D = D_GGX(GBuffer.Roughness, Context.NoH);
		half G = G_Schlick(GBuffer.Roughness, Context.NoV, LitData.NoL); // {HEX} use G_Schlick instead of G_Implicit

		/* {HEX} always use F_Schlick
		half3 F = F_None( GBuffer.SpecularColor );
		if (LitData.PBRQualityLevel == PBRQualityLevel_High)
			F = F_Schlick(GBuffer.SpecularColor, Context.VoH);
		*/
		half3 F = F_Schlick(GBuffer.SpecularColor, Context.VoH);
		// {HEX} end

		return D*G*F;
	}
	else
	{
		return half3(0, 0, 0);
	}
}

//Standard PBR lighting
void DefaultLitBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	BxDFContext Context;
	Init(Context, LitData);
	// SphereMaxNoH(Context, LitData.RadiusTan, true, LitData.IsSphericalLight);
	Context.NoV = saturate(abs(Context.NoV) + 1e-5);

	// {HEX} add micro shadow
	half microShadow = saturate(GBuffer.AO * GBuffer.AO * 2 + abs(LitData.NoL) - 1);
	LitData.InShadow = max(LitData.InShadow, 1 - microShadow);
	// {HEX} end

	half3 Irradiance = LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * LitData.NoL;
	LitResult.DiffuseLighting += Irradiance;//*Diffuse_Lambert( GBuffer.DiffuseColor )
	if(LitData.IsRectLight)
	{
#if IS_DEFERRED_LIGHTTING
		LitResult.SpecularLighting = RectSpecularLightingLTC(GBuffer, LitData, Context);
#endif
	}
	else if(LitData.IsDiskLight)
	{
#if IS_DEFERRED_LIGHTTING
		LitResult.SpecularLighting = DiskSpecularLightingLTC(GBuffer, LitData, Context);
#endif
	}
	if(@_LOD_LEVEL < LOD_Level_2)
		LitResult.SpecularLighting += Irradiance * SpecularGGX(GBuffer, LitData, Context);
}

//Preintegrated Skin
//@tSkinProfileMap input texture for skin lighting
void PreintegratedSkinBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	//diffuse lighting
	half Opacity = max(0, GBuffer.CustomData.a);
	half2 tc = half2(LitData._NoL * 0.5 + 0.5, 1 - Opacity);
	half3 sss = half3(1, 1, 1);
#if MATERIAL_SHADINGMODEL_PREINTEGRATED_SKIN
	sss = tSkinProfileMap.Sample(sSkinProfileMapSampler, tc).rgb;
#elif IS_DEFERRED_LIGHTTING
	sss = LitData.NoL.rrr;
	if (@_SkinProfileMapEnable)
		sss = tSkinProfileMap.SampleLevel(sSkinProfileMapSampler, tc, 0).rgb;
#endif
	// {HEX} fix shadow
	 sss *= sss;
	 half shadow = 1 - LitData.InShadow;
	 half3 shadow_Lut = half3(lerp(sqrt(shadow.r), shadow.r, 0.5), shadow.rr);
	 LitResult.DiffuseLighting += LitData.FalloffColor * LitData.Atten * sss * shadow_Lut ;

	// {HEX} end
	// {HEX} subsurface
	// half translucency = GBuffer.CustomData.a;
	// half3 profile = half3(0.2, 0.1, 0.05);
	// profile *= translucency;
	// LitResult.DiffuseLighting += LitData.FalloffColor * LitData.Atten * shadow_Lut * profile * saturate(0.5 + -LitData._NoL * 0.5);
	
	half translucency = pow(GBuffer.CustomData.a,3);
	float H = normalize(LitData.V + LitData.L);
	half3 profile = half3(0.5855038349464771, 0.15179027573502996, 0.042034140988957754);
	profile *= translucency;
	LitResult.DiffuseLighting += LitData.FalloffColor * LitData.Atten * shadow_Lut * profile * saturate(0.5 + -LitData._NoL * 0.5); 
	// {HEX} end
	//specular lighting
	if (LitData.PBRQualityLevel >= PBRQualityLevel_Normal)
	{
		half3 SpecularN = normalize(DecodeNormal(GBuffer.CustomData.xyz));
		LitData.N = SpecularN;
		LitData._NoL = dot(LitData.L, LitData.N);
		LitData.NoL = saturate(LitData._NoL);
		BxDFContext Context;
		Init(Context, LitData);
		LitResult.SpecularLighting += LitData.FalloffColor * LitData.Atten * shadow_Lut * LitData.NoL * SpecularGGX(GBuffer, LitData, Context);
	}
	else
	{
		LitResult.SpecularLighting += half3(0, 0, 0);
	}

}

//SSS
void SSSBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	DefaultLitBxDF(LitResult, GBuffer, LitData);

	if (LitData.TranslucencyLighting)
	{
		half translucency = GBuffer.CustomData.r;
		half3 profile = half3(0.5855038349464771, 0.15179027573502996, 0.042034140988957754);//d=0.3
		profile *= translucency;
		LitResult.DiffuseLighting += LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * profile * saturate(0.3 + -LitData._NoL);
	}
}

//clear coat PBR lighting
void ClearCoatBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	half3 Irradiance = LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * LitData.NoL;
	half ClearCoat = GBuffer.CustomData.x;
	half ClearCoatRoughness = GBuffer.CustomData.y;

	BxDFContext Context;
	Init(Context, LitData);
	SphereMaxNoH(Context, LitData.RadiusTan, true, LitData.IsSphericalLight);
	Context.NoV = saturate(abs(Context.NoV) + 1e-5);
	//Context.VoH = AreaLight.bIsRect ? Context.NoV : Context.VoH;
	
	//first layer fresnel
	float Fc = exp2( (-5.55473 * Context.VoH - 6.98316) * Context.VoH );;
	{
		float F0 = 0.04;
		float F = Fc + (1 - Fc) * F0;
		F *= ClearCoat;
		LitResult.SpecularLighting += Irradiance * D_GGX(ClearCoatRoughness, Context.NoH) * G_Implicit() * F;
		Irradiance *= 1 - F;
	}

	LitResult.DiffuseLighting += Irradiance;
	if (LitData.PBRQualityLevel >= PBRQualityLevel_Normal)
	{
		half D = D_GGX(GBuffer.Roughness, Context.NoH);
		half G = G_Implicit();
		half3 F = Fc + (1 - Fc) * GBuffer.SpecularColor;
		LitResult.SpecularLighting += Irradiance * D*G*F;
	}
	else
	{
		LitResult.SpecularLighting += half3(0, 0, 0);
	}
}

//foliage PBR lighting
void TwoSidedBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	BxDFContext Context;
	Init(Context, LitData);
	//SphereMaxNoH(Context, LitData.RadiusTan, true, LitData.IsSphericalLight);
	Context.NoV = saturate(abs(Context.NoV) + 1e-5);

	half3 Irradiance = LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * LitData.NoL;
	LitResult.DiffuseLighting += Irradiance;//*Diffuse_Lambert( GBuffer.DiffuseColor )
	if(LitData.IsRectLight)
	{
#if IS_DEFERRED_LIGHTTING
		LitResult.SpecularLighting = RectSpecularLightingLTC(GBuffer, LitData, Context);
#endif
	}
	else if(LitData.IsDiskLight)
	{
#if IS_DEFERRED_LIGHTTING
		LitResult.SpecularLighting = DiskSpecularLightingLTC(GBuffer, LitData, Context);
#endif
	}
	if(@_LOD_LEVEL <= LOD_Level_1)
	{
		// float LODra = distance(CameraPos.xyz,IN.WorldPosition.xyz)*0.007;
		LitResult.SpecularLighting += Irradiance * SpecularGGX(GBuffer, LitData, Context);
	
		
	// {HEX} add subsurface

	half3 SubsurfaceColor = GBuffer.BaseColor * GBuffer.CustomData.a * 10;
	half Wrap = 0.7;
	half WrapNoL = saturate( ( -LitData._NoL + Wrap ) / Square( 1 + Wrap ) );
	// Scatter distribution
	half VoL = dot(LitData.V, LitData.L);
	half Scatter = D_GGX( 0.8, saturate( -VoL ) );
	LitResult.SpecularLighting += LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * WrapNoL * Scatter * SubsurfaceColor;
	//////////LOD RATE/////////////
	LitResult.SpecularLighting*=(1-saturate(GBuffer.CustomData.r*0.015));
	// {HEX} end
	}
	
}

//Hair PBR lighting
void HairBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	half3 H = normalize(LitData.V + LitData.L);
	half VoL = dot(LitData.V, LitData.L);
	half WrapNoL =  LitData._NoL * 0.5 + 0.5;
	half3 Irradiance = LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * WrapNoL;
	half ScatterScale = GBuffer.CustomData.w;
	//SubsurfaceScatter
	half3 D = SubsurfaceScatter(GBuffer.BaseColor, LitData.L, LitData.V, LitData.N, H) * ScatterScale * 10;
	LitResult.SpecularLighting += D * Irradiance;
	if (LitData.PBRQualityLevel >= PBRQualityLevel_Normal)
	{
		half3 S = 0;
		GBuffer.CustomData.xyz 	= DecodeNormal(GBuffer.CustomData).xyz;
		GBuffer.CustomData.xyz 	= normalize(GBuffer.CustomData).xyz;
		half3 Tangent = GBuffer.CustomData.xyz;
		half ClampedRoughness = clamp(GBuffer.Roughness, 1/255.0f, 1.0f);

		// N is the vector parallel to hair pointing toward root
		half SinThetaL = dot(Tangent, LitData.L);
		half SinThetaV = dot(Tangent, LitData.V);
		half CosThetaD = cos( 0.5 * abs( asinFast( SinThetaV ) - asinFast( SinThetaL ) ) );

		half3 Lp = LitData.L - SinThetaL * Tangent;
		half3 Vp = LitData.V - SinThetaV * Tangent;
		half CosPhi = dot(Lp,Vp) * rsqrt( dot(Lp,Lp) * dot(Vp,Vp) + 1e-4 );
		half CosHalfPhi = sqrt( saturate( 0.5 + 0.5 * CosPhi ) );

		half n = 1.55;
		half n_prime = 1.19 / CosThetaD + 0.36 * CosThetaD;

		half Shift = 0.035;
		half AlphaR = -Shift * 2;
		half AlphaTT = Shift;
		half AlphaTRT = Shift * 4;

		half Area = 0;
		half BR = Area +  pow( ClampedRoughness, 2 );
		half BTT = Area + pow( ClampedRoughness, 2 ) / 2;
		half BTRT = Area + pow( ClampedRoughness, 2 ) * 2;

		// R
		half sa = sin( AlphaR );
		half ca = cos( AlphaR );
		Shift = 2 * sa * ( ca * CosHalfPhi * sqrt( 1 - SinThetaV * SinThetaV ) + sa * SinThetaV );

		half Mp = Hair_g( BR * sqrt(2.0) * CosHalfPhi, SinThetaL + SinThetaV - Shift );
		half Np = 0.25 * CosHalfPhi;
		half Fp = Hair_F( sqrt( saturate( 0.5 + 0.5 * VoL ) ) );
		S += Mp * Np * Fp * GBuffer.Specular * 2 * lerp( 1, GBuffer.AO, saturate(-VoL) );

		// TT
		Mp = Hair_g( BTT, SinThetaL + SinThetaV - AlphaTT );
		half a = 1 / n_prime;
		half h = CosHalfPhi * ( 1 + a * ( 0.6 - 0.8 * CosPhi ) );
		half f = Hair_F( CosThetaD * sqrt( saturate( 1 - h*h ) ) );
		Fp = pow(1 - f, 2);
		half3 Tp = pow( GBuffer.BaseColor, 0.5 * sqrt( 1 - pow(h * a, 2) ) / CosThetaD );
		Np = exp( -3.65 * CosPhi - 3.98 );
		S += Mp * Np * Fp * Tp * GBuffer.AO * WrapNoL;

		//TRT
		Mp = Hair_g( BTRT, SinThetaL + SinThetaV - AlphaTRT );
		f = Hair_F( CosThetaD * 0.5 );
		Fp = pow(1 - f, 2) * f;
		Tp = pow( GBuffer.BaseColor, 0.8 / CosThetaD );
		Np = exp( 17 * CosPhi - 16.78 );
		S += Mp * Np * Fp * Tp;

		LitResult.SpecularLighting += -min(-S, 0) * Irradiance;
	}
	else
	{
		LitResult.SpecularLighting += half3(0, 0, 0);
	}
}

//eye PBR lighting
void EyeBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	//diffuse lighting
	half PupilMask = GBuffer.CustomData.r;
	half Opacity = 1;
	half2 tc = half2(LitData._NoL * 0.5 + 0.5, 1 - Opacity);
	half3 sss = half3(1, 1, 1);
#if MATERIAL_SHADINGMODEL_EYE
	sss = tSkinProfileMap.Sample(sSkinProfileMapSampler, tc).rgb;
#elif IS_DEFERRED_LIGHTTING
	sss = LitData.NoL.rrr;
	if (@_SkinProfileMapEnable)
		sss = tSkinProfileMap.SampleLevel(sSkinProfileMapSampler, tc, 0).rgb;
#endif
	half shadow = 1 - LitData.InShadow;
	half3 shadow_Lut = half3(lerp(sqrt(shadow.r), shadow.r, 0.5), shadow.rr);
	half3 InputRadiance = LitData.FalloffColor * LitData.Atten * shadow_Lut;
	half3 EyeNoL = lerp(1 - LitData.NoL.rrr + half3(0.3,0.3,0.3), saturate(sss * sss), PupilMask);
	LitResult.DiffuseLighting += InputRadiance * EyeNoL;

	//specular lighting
	half3 H = normalize(LitData.V + LitData.L);
	half specularIns = lerp(1, 0.1, PupilMask);
	half3 DirectSpecularAdd = pow(max(0.01,dot(GBuffer.WorldNormal, H)), 500) * 2 * specularIns;
	LitResult.SpecularLighting += InputRadiance.r * LitData.NoL * DirectSpecularAdd;
}

//cloth PBR lighting
void ClothBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	half ClothMask = GBuffer.CustomData.r;
	BxDFContext Context;
	Init(Context, LitData);
	//SphereMaxNoH(Context, LitData.RadiusTan, true, LitData.IsSphericalLight);
	Context.NoV = saturate(abs(Context.NoV) + 0.001);

	half microShadow = saturate(GBuffer.AO * GBuffer.AO * 2 + abs(LitData.NoL) - 1);
	LitData.InShadow = max(LitData.InShadow, 1 - microShadow);

	half3 Irradiance = LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * LitData.NoL;
	LitResult.DiffuseLighting += Irradiance;

	if (LitData.PBRQualityLevel >= PBRQualityLevel_Normal)
	{
		half3 ClothSpec = half3(0,0,0);

		half3 DefaultSpec = half3(0,0,0);
		half3 SilkSpec = GBuffer.CustomData.b * GBuffer.SpecularColor * 10;
		if(ClothMask > 0.1)
		{
			// Cloth - Asperity Scattering - Inverse Beckmann Layer
			half3 FuzzColor = GBuffer.BaseColor.rgb * GBuffer.CustomData.g * 10;
			FuzzColor *= (1 - Context.NoV);
			half D2 = D_InvGGX( pow( GBuffer.Roughness, 4 ), Context.NoH );
			half Vis2 = Vis_Cloth( Context.NoV, LitData.NoL );
			half3 F2 = F_Schlick( FuzzColor, Context.VoH );
			ClothSpec = D2 * Vis2 * F2 * ClothMask;
		}
		else
		{
			half D = D_GGX( GBuffer.Roughness, Context.NoH );
			half G = G_Schlick( GBuffer.Roughness, Context.NoV, LitData.NoL );
			half3 F = F_Schlick( GBuffer.SpecularColor, Context.VoH );
			DefaultSpec = D*G*F * (1 - ClothMask);
		}
		LitResult.SpecularLighting += (ClothSpec + SilkSpec + DefaultSpec) * Irradiance;
	}
	else
	{
		LitResult.SpecularLighting += half3(0, 0, 0);
	}
}

//LOD lighting
void LODLitBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	half microShadow = saturate(GBuffer.AO * GBuffer.AO * 2 + abs(LitData.NoL) - 1);
	LitData.InShadow = max(LitData.InShadow, 1 - microShadow);
	half3 Irradiance = LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * LitData.NoL;
	LitResult.DiffuseLighting += Irradiance;
}

void ImpostorBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	half DirLightMap = GBuffer.CustomData.r;
	half3 Irradiance = LitData.FalloffColor * LitData.Atten * (1-LitData.InShadow) * GBuffer.CustomData.r;// * LitData.NoL;
	LitResult.DiffuseLighting += Irradiance;
	//subsurface
	half3 SubsurfaceColor = GBuffer.BaseColor * 3;
	half Wrap = 0.7;
	half WrapNoL = saturate( ( -LitData._NoL + Wrap ) / Square( 1 + Wrap ) );
	// Scatter distribution
	half VoL = dot(LitData.V, LitData.L);
	half Scatter = D_GGX( 0.8, saturate( -VoL ) );
	LitResult.SpecularLighting += Irradiance * WrapNoL * Scatter * SubsurfaceColor;
}


void IntegrateBxDF(in out LightingResult LitResult, in GBufferData GBuffer, in LightingData LitData)
{
	[branch] if(GBuffer.ShadingModelID == SHADINGMODELID_DEFAULT_LIT)
	{
		DefaultLitBxDF(LitResult, GBuffer, LitData);
	}else if(GBuffer.ShadingModelID == SHADINGMODELID_DEFAULT_LIT_NO_GI)
	{
		DefaultLitBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_TWOSIDED_FOLIAGE)
	{
		TwoSidedBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_IMP_LIT)
	{
		ImpostorBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_PREINTEGRATED_SKIN)
	{
		PreintegratedSkinBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_SSS)
	{
		SSSBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_HAIR)
	{
		HairBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_EYE)
	{
		EyeBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_CLEAR_COAT)
	{
		ClearCoatBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_CLOTH)
	{
		ClothBxDF(LitResult, GBuffer, LitData);
	}
	else if (GBuffer.ShadingModelID == SHADINGMODELID_LOD_LIT)
	{
		LODLitBxDF(LitResult, GBuffer, LitData);
	}
	else
	{
		LitResult.DiffuseLighting += half3(0, 0, 0);
		LitResult.SpecularLighting += half3(0, 0, 0);
	}
}

#endif//__SHADING_MODE_FXH__
