//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Manipulator.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#pragma shader category("Manipulator")

#include "Constants.fxh"

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{ 
	float4x3 World 					: World 			;
};
 
cbuffer Shader
{
	float3 cDiffuse					 < IsColor = "true"; > = {1, 1, 1};
};

// Surface textures.
texture2D tDiffuseMap : layout(t0) = "White";
sampler sDiffuseSampler : layout(s0, s1)
{
	Texture = tDiffuseMap;
	Filter = Point;
	Address = Wrap;
	SRGB = false;
};

#include "Common.fxh"

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

	float3 LocalNormal;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.
 
	float4 		WorldPosition 	: TEXCOORDN;			//  World Position

	float4		WorldNormal		: TEXCOORDN;			//	Outputing normlized normal for lighting.

	float4		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
	float 		LinearZ			: TEXCOORDN;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);

	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);
	// No Far-Plane Clip.
	OUT.HPosition.z = max(OUT.HPosition.z, EPSILON*OUT.HPosition.w);

	// TexCoord
	OUT.TexCoord = IN.TexCoord.xyxy;

	OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));

	// ViewDirection
	OUT.ViewDirection.xyz = OUT.WorldPosition.xyz - CameraPos.xyz;

	// EmissionUV
	OUT.WorldNormal.w = 0;
	OUT.ViewDirection.w = 0;
	OUT.LinearZ = OUT.HPosition.w * CameraInfo.z;
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput SurfaceVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	// Intermediate.
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	// Output.
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float4 SurfacePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	OUT.a = 1.0;

	// Runtime per-pixel parameters.
	half3 V = normalize(-IN.ViewDirection.xyz);

	float4 sampleDiffuse = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TexCoord.xy);

	OUT.rgb = sampleDiffuse.rgb * cDiffuse.rgb * (0.5f * (1.0f + dot(V.xyz, IN.WorldNormal.xyz)));
	return OUT;
}

technique Shading
{
	VertexShader = SurfaceVertexShading();
	PixelShader = SurfacePixelShading();
}
//////////////////////////////////GBuffer////////////////////////////////////////
struct GBufferOUTPUT
{
	float4 GBufferA 	: SV_Target0;	
	float4 GBufferB 	: SV_Target1;	
	float4 GBufferC 	: SV_Target2;	
	float4 GBufferD 	: SV_Target3;	
	float4 GBufferE 	: SV_Target4;	
	float4 GBufferF  	: SV_Target5;	
	float4 GBufferG 	: SV_Target6;	
};
GBufferOUTPUT SurfaceGBufferGenPixelShading(in VertexShadingOutput IN)
{
	GBufferOUTPUT OUT;

	// Runtime per-pixel parameters.
	half3 V = normalize(-IN.ViewDirection.xyz);

	float4 sampleDiffuse = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TexCoord.xy);

	OUT.GBufferA.rgb = sampleDiffuse.rgb * cDiffuse.rgb * (0.5f * (1.0f + dot(V.xyz, IN.WorldNormal.xyz)));
	OUT.GBufferB = float4(1,0,0,0); // roughness默认是1，免得ssr计算
	OUT.GBufferC = float4(0,0,0,0);
	OUT.GBufferD = float4(0,0,0,0);
	OUT.GBufferE = float4(0,0,0,0);
	OUT.GBufferF = float4(0,0,0,0);
	OUT.GBufferG = EncodeDepth(IN.LinearZ);
	OUT.GBufferA.w = EncodeDepth(IN.LinearZ);
	return OUT;
}

technique GBufferGen
{
	VertexShader = SurfaceVertexShading();	
	PixelShader = SurfaceGBufferGenPixelShading();
}
