//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	Particle shading model.

#pragma shader material

@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = false;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = false;
}

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = false;
}

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World				;
//	float4x4 WorldViewProj 		: WorldViewProj		;
	float4   InstanceColor		: Color				;
//	float4   cTintColor1		: TintColor1 	;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  InvFadeDistance		 = 1;
	float  BloomScale			 = 0.5;
	float2 UVScale				 = { 1, 1 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  cFov					 = 1.0;
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "Transparent";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D ScreenTexture : SceneColor : layout(t2);
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
#include "Particle_Common.fxh"

//	  Almost Same as Particle.fx::ParticlePixelShading()
void PS_GetParameters(in VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	float4 OUT = float4(1,0,1,1);
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);

	OUT *= ModulateColor * InstanceColor;

	if(@_HDR && @_ENCODEHDR)
		OUT.a *= OUT.a;

	OUT.rgb *= OUT.a;
	OUT *= IN.VertexColor;

	// OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);

	materialParamters.BaseColor = OUT;

}

#include "Particle_Shading.fxh"

technique Shading
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleUnifyPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Distortion.

float4 ParticleDistortionShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	if (@_DISTORTION)
	{
		float2 spos = screenCoord.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;

		float H = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy).r;
		H *= ModulateColor.x * InstanceColor.x * IN.VertexColor.x;

		// calcuate patch normal
		float3 P = IN.WorldPosition;
		float3 N = cross(ddy(P), ddx(P));
		if (@_GLSL || @_SPIRV)
		{
			// ddx/ddy is incorrect on some android devices.
			float len = length(N);
			N = (len == 0) ? float3(0, 1, 0) : N / len;
		}
		else
		{
			N = normalize(N);
		}

		// hacked refraction
		float3 V = IN.ViewDirection;
		float3 R = cross(N, V);
		float3 D = cross(R, V);

		float3 VR = cross(float3(0, 1, 0), V);
		float3 VU = normalize(cross(V, VR));
		VR = cross(VU, V);

		float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
		offset *= (500 / IN.TexCoord.z) * H;
		offset = offset * 0.5 + 0.5;

		EncodeFloatRG(offset.x, OUT.rg);
		EncodeFloatRG(offset.y, OUT.ba);
	}
	return OUT;
}

technique Distortion
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDistortionShading();
};

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDebugShading();
}

