//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<PostProcess.fx>
///	@path	~/System/Shader/
///	@date	2013/04/24
///	@brief	Post process utils.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

@_LinearFilter
{
	Name = useLinearFilter;
	Description = "use linear filter";
	Visible = true;
};

@_CatmullRom4Encode
{
	Name = catmullRom4Encode;
	Description = "catmullRom4Encode";
	Visible = true;
};
//////////////////////////////////////////////////////////////////////////
// Post process utils.

cbuffer Batch
{
	float4 TexToTexParams0	;
	float4 TexToTexParams1	;
	float4 TexToTexParams2	;
	float4 TexToTexParams3	;

	float4 vsOffsets[16]	;
	float4 psWeights[16]	;
};

// Source target.
texture2D SourcePoint0 : layout(t0);
sampler SourcePoint0_Sampler : layout(s0)
{
	Texture = SourcePoint0;
	Filter = Point;
	Address = Clamp;
};
texture2D SourcePoint1 : layout(t1);
sampler SourcePoint1_Sampler : layout(s1)
{
	Texture = SourcePoint1;
	Filter = Point;
	Address = Clamp;
};
texture2D SourceLinear0 : layout(t2);
sampler SourceLinear0_Sampler : layout(s2)
{
	Texture = SourceLinear0;
	Filter = Linear;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
// Assemble input.

/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float2		TC				: TEXCOORD0;
};

struct VertexGaussianOutput
{
	float4		HPosition		: SV_Position;
	float4		TC0				: TEXCOORDN;    
	float4		TC1				: TEXCOORDN;    
	float4		TC2				: TEXCOORDN;    
	float4		TC3				: TEXCOORDN;    
	float4		TC4				: TEXCOORDN; 
};

/// Pixel output.
struct PixelOutput
{
	float4		Color			: SV_Target;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Texture copy
PixelOutput TextureCopyPS( VertexScreenOutput IN )
{
	PixelOutput OUT;
	if(@_LinearFilter)
		OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC);
	else
		OUT.Color = SampleColorMap(SourcePoint0, SourcePoint0_Sampler, IN.TC);
	if (@_CatmullRom4Encode)
	{
		float2 half_f = 2.0 * frac(0.5 * IN.HPosition.xy) - 1.0;
		OUT.Color = half_f.x * half_f.y > 0.0 ? OUT.Color : -OUT.Color;
		OUT.Color = OUT.Color * 0.5 + 0.5;
	}
	return OUT;
}

technique TextureCopy
{
	VertexShader = ScreenVS();
	PixelShader = TextureCopyPS();
}

//////////////////////////////////////////////////////////////////////////
// Texture copy
PixelOutput BlurDownSamplePS( VertexScreenOutput IN )
{
	PixelOutput OUT;

#if 1//high quality
	OUT.Color = float4(0, 0, 0, 0);
	float sumWeight = 0;
	float2 pixelSize = TexToTexParams0.xy;
	for (int x = -2; x < 2; ++x)
	for (int y = -2; y < 2; ++y)
	{
		float weight = 1.0 / (x*x + y*y + 1);
		OUT.Color += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(x,y)*pixelSize) * weight;
		sumWeight += weight;
	}
	OUT.Color /= sumWeight;
#else//middle quality
	OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC) * 4;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(-pixelSize.x, 0)) * 2;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(+pixelSize.x, 0)) * 2;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(0, -pixelSize.y)) * 2;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(0, +pixelSize.y)) * 2;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(-pixelSize.x, -pixelSize.y)) * 1;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(-pixelSize.x, +pixelSize.y)) * 1;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(+pixelSize.x, -pixelSize.y)) * 1;
	OUT.Color.rgb += SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC + float2(+pixelSize.x, +pixelSize.y)) * 1;
	OUT.Color.rgb /= 16;
#endif
	return OUT;
}

technique BlurDownSample
{
	VertexShader = ScreenVS();
	PixelShader = BlurDownSamplePS();
}

//////////////////////////////////////////////////////////////////////////
// TextureAtals VS
VertexScreenOutput TextureAtlasVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.xy = TexToTexParams0.xy + vPos.xy * TexToTexParams0.zw;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}
// TextureAtals
PixelOutput TextureAtlasPS( VertexScreenOutput IN )
{
	PixelOutput OUT;
	OUT.Color = SampleColorMap(SourcePoint0, SourcePoint0_Sampler, IN.TC);
	return OUT;
}

technique TextureAtlas
{
	VertexShader = TextureAtlasVS();
	PixelShader = TextureAtlasPS();
}

//////////////////////////////////////////////////////////////////////////
// Texture to screen
PixelOutput TextureToScreenPS(in VertexScreenOutput IN)
{
	PixelOutput OUT;

	if (@_LinearFilter)
	{
		OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC);
	}
	else
	{
#if 1	 // Modified bilinear interpolation
		float2 p = IN.TC * ScreenInfo.xy + 0.5;
		float2 i = floor(p);
		float2 f = p - i;

		//f = f * f * (3 - 2 * f);						// 3rd order
		f = f * f * f * (f * (f * 6 - 15) + 10);		// 5th order

		p = i + f;
		float2 tc = (p - 0.5) * ScreenInfo.zw;
		OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, tc);

#else	// Bicubic interpolation
		float2 uv = IN.TC * ScreenInfo.xy;
		float2 texelCenter = floor(uv - 0.5) + 0.5;
		float2 fracOffset = uv - texelCenter;
		float2 fracOffset_x2 = fracOffset * fracOffset;
		float2 fracOffset_x3 = fracOffset * fracOffset_x2;
	
		// Calculate the filter weights (B-Spline weighting function)
		float2 weight0 = fracOffset_x2 - 0.5 * (fracOffset_x3 + fracOffset);
		float2 weight1 = 1.5 * fracOffset_x3 - 2.5 * fracOffset_x2 + 1.0;
		float2 weight3 = 0.5 * (fracOffset_x3 - fracOffset_x2);
		float2 weight2 = 1.0 - weight0 - weight1 - weight3;
	
		// Calculate the texture coordinates
		float2 scalingFactor0 = weight0 + weight1;
		float2 scalingFactor1 = weight2 + weight3;
	
		float2 f0 = weight1 / (weight0 + weight1);
		float2 f1 = weight3 / (weight2 + weight3);
	
		float2 texCoord0 = texelCenter - 1.0 + f0;
		float2 texCoord1 = texelCenter + 1.0 + f1;
	
		texCoord0 *= ScreenInfo.zw;
		texCoord1 *= ScreenInfo.zw;
	
		// Sample the texture (4 taps)
		OUT.Color = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord0.x, texCoord0.y)) * (scalingFactor0.x * scalingFactor0.y)
			      + SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord1.x, texCoord0.y)) * (scalingFactor1.x * scalingFactor0.y)
			      + SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord0.x, texCoord1.y)) * (scalingFactor0.x * scalingFactor1.y)
			      + SampleColorMap(SourceLinear0, SourceLinear0_Sampler, float2(texCoord1.x, texCoord1.y)) * (scalingFactor1.x * scalingFactor1.y);

#endif
	}
	//wenyao: important!!! for mx4
	OUT.Color.a = 1.0;
	return OUT;
}

technique TextureToScreen
{
	VertexShader = ScreenVS();
	PixelShader = TextureToScreenPS();
}

//////////////////////////////////////////////////////////////////////////
// Gaussian blur.

VertexGaussianOutput GaussBlurBilinearVS(in VertexAssembleScreen IN)
{
	VertexGaussianOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	
	OUT.TC0.xy = IN.BaseTC.xy + vsOffsets[0].xy;
	OUT.TC0.zw = IN.BaseTC.xy + vsOffsets[1].xy;
	OUT.TC1.xy = IN.BaseTC.xy + vsOffsets[2].xy;
	OUT.TC1.zw = IN.BaseTC.xy + vsOffsets[3].xy;
	OUT.TC2.xy = IN.BaseTC.xy + vsOffsets[4].xy;
	OUT.TC2.zw = IN.BaseTC.xy + vsOffsets[5].xy;
	OUT.TC3.xy = IN.BaseTC.xy + vsOffsets[6].xy;
	OUT.TC3.zw = IN.BaseTC.xy + vsOffsets[7].xy;

	// Original coordinates
	OUT.TC4.xy = IN.BaseTC.xy;
	OUT.TC4.zw = 0;

	return OUT;
}

PixelOutput GaussBlurBilinearPS(in VertexGaussianOutput IN)
{
	PixelOutput OUT;

	half4 sum = 0;
	half4 col = 0;

	// vanila gaussian blur

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC0.xy);
	sum += col * (half) psWeights[0].x;

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC0.zw);
	sum += col * (half) psWeights[1].x;

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC1.xy);
	sum += col * (half) psWeights[2].x;

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC1.zw);
	sum += col * (half) psWeights[3].x;

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC2.xy);
	sum += col * (half) psWeights[4].x;

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC2.zw);
	sum += col * (half) psWeights[5].x;

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC3.xy);
	sum += col * (half) psWeights[6].x;

	col = SampleColorMap(SourceLinear0, SourceLinear0_Sampler, IN.TC3.zw);
	sum += col * (half) psWeights[7].x;

	OUT.Color = sum;

	return OUT;
}

technique GaussBlurBilinear
{
	VertexShader = GaussBlurBilinearVS();
	PixelShader = GaussBlurBilinearPS();
}
