//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<QuadOverdraw.fx>
///	@path	~/System/Shader/
///	@date	2020/02/18
///	@brief	Quad overdraw visualization.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Parameters
cbuffer Shader
{
	float2 MousePosition = { -1, -1 };
};

texture2D InputTexture : layout(t0);
sampler InputTextureSampler : layout(s0)
{
	Texture = InputTexture;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

//////////////////////////////////////////////////////////////////////////
// Vertex input/output

struct VertexInput 
{
	float4 Position		: POSITION;
	float2 TexCoord		: TEXCOORD0;
};

struct VertexOutput
{
	float4 HPosition	: SV_Position;
	float2 TexCoord		: TEXCOORD0;
};

VertexOutput ScreenVS(VertexInput IN)
{
	VertexOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, vPos.z, 1.0);
	OUT.TexCoord = IN.TexCoord;
	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Print value in shader

float DigitBin(int x)
{
	return x==0?480599.0:x==1?139810.0:x==2?476951.0:x==3?476999.0:x==4?350020.0:
	       x==5?464711.0:x==6?464727.0:x==7?476228.0:x==8?481111.0:x==9?481095.0:0.0;
}

float PrintValue(float2 fragCoord, float2 pixelCoord, float2 fontSize,
				 float value, float digits, float decimals)
{
	float2 charCoord = (fragCoord - pixelCoord) / fontSize;
	charCoord.y = 1 - charCoord.y;
	if (charCoord.y < 0.0 || charCoord.y >= 1.0) return 0.0;
	float bits = 0.0;
	float digitIndex1 = digits - floor(charCoord.x) + 1.0;
	if (-digitIndex1 <= decimals)
	{
		float pow1 = pow(10.0, digitIndex1);
		float absValue = abs(value);
		float pivot = max(absValue, 1.5) * 10.0;
		if (pivot < pow1)
		{
			if (value < 0.0 && pivot >= pow1 * 0.1) bits = 1792.0;
		}
		else if (digitIndex1 == 0.0)
		{
			if (decimals > 0.0) bits = 2.0;
		}
		else
		{
			value = digitIndex1 < 0.0 ? frac(absValue) : absValue * 10.0;
			bits = DigitBin(int(fmod(value / pow1, 10.0)));
		}
	}
	return floor(fmod(bits / pow(2.0, floor(frac(charCoord.x) * 4.0) + floor(charCoord.y * 5.0) * 4.0), 2.0));
}

//////////////////////////////////////////////////////////////////////////
// Pixel shader
float3 OverDrawToColor(uint overdraw)
{
	const float3 colors[10] =
	{
		float3(0,     0,   0),
		float3(2,    25, 147),
		float3(0,   149, 255),
		float3(0,   253, 255),
		float3(142, 250,   0),
		float3(255, 251,   0),
		float3(255, 147,   0),
		float3(255,  38,   0),
		float3(148,  17,   0),
		float3(255,   0, 255)
	};
	return colors[min(overdraw, 9)] / 255.0;
}

float4 QuadOverdrawPS(VertexOutput IN) : SV_Target
{
	float4 source = InputTexture.Sample(InputTextureSampler, IN.TexCoord);
	float4 result = float4(0, 0, 0, 1);
	if (@_HLSL)
	{
		uint2 pos = uint2(IN.HPosition.xy) & 1;
		float s0 = source.x;
		int2 dir = pos ? -1 : 1;
		float s1 = s0 + dir.x * ddx_fine(s0);
		float s2 = s0 + dir.y * ddy_fine(s0);
		float s3 = s2 + dir.x * ddx_fine(s2);

		float overdraw = (s0 + s1 + s2 + s3) * 255.0;
		result.xyz = OverDrawToColor((uint)(overdraw + 0.5));
	
		// Show overdraw count at cursor
		uint2 mpos = uint2(MousePosition) & 0xFFFE;
		float2 mcoord = ((float2)mpos + 0.5) * ScreenInfo.zw;
		float ms0 = InputTexture.Sample(InputTextureSampler, mcoord).x;
		float ms1 = InputTexture.Sample(InputTextureSampler, mcoord + float2(1, 0) * ScreenInfo.zw).x;
		float ms2 = InputTexture.Sample(InputTextureSampler, mcoord + float2(0, 1) * ScreenInfo.zw).x;
		float ms3 = InputTexture.Sample(InputTextureSampler, mcoord + float2(1, 1) * ScreenInfo.zw).x;
		float moverdraw = (ms0 + ms1 + ms2 + ms3) * 255.0;

		float v = PrintValue(IN.HPosition.xy, MousePosition + float2(10, 10),
			float2(16, 20), floor(moverdraw + 0.5), 2.0, 0.0);

		float2 p = abs(IN.HPosition.xy - (MousePosition + float2(38, 20)));
		float2 s = abs(p) - float2(20, 10);
		float d = length(max(s, 0.0)) + min(max(s.x, s.y), 0.0);
		result *= 0.5 + 0.5 * saturate(d * 0.1);

		result.xyz = lerp(result.xyz, float3(1, 1, 1), v);
	}
	return result;
}

technique QuadOverdrawTech
{
	VertexShader = ScreenVS();
	PixelShader = QuadOverdrawPS();
}

