//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<SDF.fx>
///	@path	~/System/Shader/
///	@date	2015/03/16
///	@brief	Signed Distance 

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "SDF.fxh"

///////////////// Properties //////////////////
@_UseLastSDF
{
	Name = UseLastSDF;
	Visible = false;
};

@_CaptureSky
{
	Name = CaptureSky;
	Visible = false;
	Default = false;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float3 cMinXYZ				;
	float cLastWeight			;
	float3 cMaxXYZ				;
	float cVolumeSize			;
	float3 cLastSDFOffset		;
	float cTruncatedU			;
	float4x4 cViewProjTex		;
	float4	cWBasisX			;
	float4	cWBasisY			;
	float4	cWBasisZ			;
	float  	cPixelsScale		;//cVolumeSize * Height / (2*tan(fov/2))
};

///////////////// samplers //////////////////
texture2D tDepthMap : SceneDepth : layout(t0);
sampler sDepthSampler : layout(s0)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tScreenMap : layout(t1);
sampler sScreenSampler : layout(s1)
{
	Texture = tScreenMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tSDFPointMap : layout(t2);
sampler sSDFPointSampler : layout(s2)
{
	Texture = tSDFPointMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tSDFLinearMap : layout(t2);
sampler sSDFLinearSampler : layout(s2)
{
	Texture = tSDFLinearMap;
	Filter = Linear;
	Address = Clamp;
};

texture2D tGBufferNormal : GBufferMapE : layout(t3);
sampler sGBufferNormal : layout(s3)
{
	Texture = tGBufferNormal;
	Filter = Point;
	Address = Clamp;
};

float sdBox(float3 p, float3 edge)
{
	float3 d = abs(p) - edge;
	return min(max(d.x,max(d.y,d.z)),0.0) + length(max(d,0.0));
}

float4 map(in float3 p)
{
	//1, if p not in box return last one
	float3 d = max(max(p-cMaxXYZ, cMinXYZ-p), 0);
	float len = length(d);
	if (len > EPSILON)
		return float4(1, 0, 0, cTruncatedU);
	// if (any(p<cMinXYZ) || any(p>cMaxXYZ))
	//  	return float4(1, 0, 0, cTruncatedU);

	//2, project p to vid
	float3 xyzId = WorldPosToXYZId(p, cVolumeSize, cMinXYZ);
	float fc = frac(xyzId.z);
	xyzId.z -= fc;
	float2 uv; XYZIdToUV(xyzId, uv);
	float4 t1 = tSDFLinearMap.Sample(sSDFLinearSampler, uv);
	xyzId.z += 1; XYZIdToUV(xyzId, uv);
	float4 t2 = tSDFLinearMap.Sample(sSDFLinearSampler, uv);
	return lerp(t1, t2, fc);
}

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition	: SV_Position;
	float2 TC			: TEXCOORD0;
};

/////////////////// BrightPassFilter ////////////////////
v2f ScreenVS(a2v IN)
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, EPSILON, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 SDFUpdatePS(v2f IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, cTruncatedU);

	//id
	int vid_x = min((int)IN.HPosition.x, EDGEVOLUMES*FACE_PERLINE-1);
	int vid_y = min((int)IN.HPosition.y, EDGEVOLUMES*FACE_PERLINE-1);
	int3 xyzId; VolumeIdToXYZId(vid_x, vid_y, xyzId);

	float4 F_K_1;
	float W_K_1;
	bool hasLastVolume = false;
	if (@_UseLastSDF)
	{
		int3 lastXYZId = xyzId + (int3)(cLastSDFOffset);
		hasLastVolume = all(lastXYZId >= 0) && all(lastXYZId <= EDGEVOLUMES-1);
		if (hasLastVolume)
		{
			float2 lastUv; XYZIdToUV(lastXYZId, lastUv);
			F_K_1 = tSDFPointMap.Sample(sSDFPointSampler, lastUv).rgba;
			W_K_1 = cLastWeight;
			OUT = F_K_1;
		}
	}

	//volume uvz
	float3 volumePos = XYZIdToWorldPos(xyzId, cVolumeSize, cMinXYZ);
	float4 volumeUVZ = MulMat4x4(cViewProjTex, float4(volumePos,1.0)).xyzw;
	float volumeZ = volumeUVZ.w;
	float2 volumeUv = volumeUVZ.xy / volumeUVZ.w;
	float2 inRange = abs(volumeUv-0.5) < 0.5;
	if (inRange.x * inRange.y > 0 && volumeZ > 0)//only update inner frustum volumes
	{
		//Gbufer data
		float level = log2(cPixelsScale / volumeZ);
		float sceneZ = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, volumeUv.xy, 0)) * CameraInfo.y;
		float4 _GBufferNormal = tGBufferNormal.Sample(sGBufferNormal, volumeUv.xy);
		float3 N = normalize(DecodeNormal(_GBufferNormal.rgb));
		float Metallic;
		float IsPrincipalRole;
		DecodeFloatAndBit(_GBufferNormal.a, Metallic, IsPrincipalRole);
		sceneZ += IsPrincipalRole * CameraInfo.y;//not inject PrincipalRole

		float4 F_K = float4(0, 0, 0, cTruncatedU);
		F_K.a = (sceneZ - volumeZ);
		if (F_K.a >= -cTruncatedU)//only update [diffZ >= -U]
		{
			if (F_K.a < 0)//inject 
				F_K.rgb = tScreenMap.SampleLevel(sScreenSampler, volumeUv.xy, 0).rgb;
			F_K.a = min(F_K.a, cTruncatedU);
			if(hasLastVolume)
			{
				float W_K = 1.0;//fix weight work better
				OUT = float4(W_K_1 * F_K_1 + W_K * F_K) / (W_K_1 + W_K);
			}
			else
			{
				OUT = F_K;
			}
		}
	}
	return OUT;
}

technique SDFUpdate
{
	VertexShader = ScreenVS();
	PixelShader = SDFUpdatePS();
}

/////////////////////////////////////visualize sdf//////////////////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}
float4 SDFVisualizePS(v2f IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	
	[@_D3D9]  
	spos += 0.5;
	
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, stc, 0));
	float3 D = CalcHomogeneousPos(sceneZ, stc);
	// float len = length(D);
	// D /= len;
	//float3 Orig = CameraPos.xyz + len * D;
	float3 Orig = CameraPos.xyz + D;
	float4 sdf = map(Orig);
	return float4(sdf.rgb, 1.0);

	//reflect
	//D = normalize(reflect(D, float3(0, 1, 0)));

	// float t = cVolumeSize;
	// float4 sdf = float4(0, 0, 0, 1);
	// int i;
	// for (i=0; i < RAYMARCHING_STEPS; ++i)
	// {
	// 	sdf = map(Orig + t*D);
	// 	if(sdf.w < 0.1*cVolumeSize)
	// 		break;
	// 	t += sdf.w;
	// }
	// return sdf;
	// return float4(sdf.rg, sdf.w*0.5 + 0.5, 1.0);
}

technique SDFVisualize
{
	VertexShader = ScreenVS();
	PixelShader = SDFVisualizePS();
}

