//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<SDFGI.fx>
///	@path	~/System/Shader/
///	@date	2016/05/09
///	@brief	Signed Distance Field

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "SDF.fxh"

///////////////// Properties //////////////////
@_LastGIAOEnable
{ 
	Name = LastGIAOEnable;
	Visible = true;
	Predict = true;
};

///////////////// const buffer //////////////////
cbuffer Batch
{
	float4x4 cLastViewProj			;
	float4 cWBasisX					;
	float4 cWBasisY					;
	float4 cWBasisZ					;
	
	//sdf0
	float3 cMinXYZ0					;
	float  cVolumeSize0				;
	float3 cMaxXYZ0					;
	float  cTruncatedU0				;
	//sdf1
	float3 cMinXYZ1					;
	float  cVolumeSize1				;
	float3 cMaxXYZ1					;
	float  cTruncatedU1				;

	//sdf parameters
	float  cAoMinDistance0			;
	float  cAoMaxDistance0			;
	float  cTraceBias0				;
	float  cNormalBias0				;
	float  cAoMinDistance1			;
	float  cAoMaxDistance1			;
	float  cTraceBias1				;
	float  cNormalBias1				;

	//others
	float  cGIScale					 = 1.0;
	float3 cSkyLight				;
};

texture2D tDepthMap : SceneDepth : layout(t0);
sampler sDepthSampler : layout(s0)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

texture2D tGBufferNormal : GBufferMapE : layout(t1);
sampler sGBufferNormal : layout(s1)
{
	Texture = tGBufferNormal;
	Filter = Point;
	Address = Clamp;
};

texture2D tSDFLinearMap0 : layout(t2);
sampler sSDFLinearSampler0 : layout(s2)
{
	Texture = tSDFLinearMap0;
	Filter = Linear;
	Address = Clamp;
};

texture2D tSDFLinearMap1 : layout(t3);
sampler sSDFLinearSampler1 : layout(s3)
{
	Texture = tSDFLinearMap1;
	Filter = Linear;
	Address = Clamp;
};

texture2D tRandom2D : layout(t4);
sampler sRandom2D : layout(s4)
{
	Texture = tRandom2D;
	Filter = Point;
	Address = Wrap;
};

texture2D tLastGIAO : layout(t5);
sampler sLastGIAO : layout(s5)
{
	Texture = tLastGIAO;
	Filter = Linear;
	Address = Wrap;
};

texture2D tCurrGIAO : layout(t6);
sampler sCurrGIAO : layout(s6)
{
	Texture = tCurrGIAO;
	Filter = Point;
	Address = Wrap;
};

///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC		 : TEXCOORD0;
};

///////////////// helper functions //////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}
#define DistanceToBox(p, Layer) length(max(max(p-(cMaxXYZ##Layer-cVolumeSize##Layer), (cMinXYZ##Layer+cVolumeSize##Layer)-p), 0))

#define Map(p, Layer, sdf, level)\
{\
	half3 xyzId = WorldPosToXYZId(p, cVolumeSize##Layer, cMinXYZ##Layer);\
	half fc = frac(xyzId.z);\
	xyzId.z -= fc;\
	half2 uv; XYZIdToUV(xyzId, uv);\
	half4 t1 = tSDFLinearMap##Layer.SampleLevel(sSDFLinearSampler##Layer, uv, level);\
	xyzId.z += 1; XYZIdToUV(xyzId, uv);\
	half4 t2 = tSDFLinearMap##Layer.SampleLevel(sSDFLinearSampler##Layer, uv, level);\
	sdf = lerp(t1, t2, fc);\
}

void ComputeTangent(in half3 vNormal, out half3 tan0, out half3 tan1)
{
	tan0 = float3( 1, 0, 0 );
	tan1 = cross( vNormal, tan0 );
	half lg1 = dot( tan1, tan1 );
	if ( lg1 < HALF_EPSILON )
	{
		tan0 = half3( 0, 0, 1 );
		tan1 = cross( vNormal, tan0 );
		tan1 = normalize( tan1 );
	}
	else
		tan1 *= rsqrt(lg1);
	tan0 = normalize( cross(tan1,vNormal) );
}

half4 SDF_GIAO2(in half3 Orig, in half3 N, in half2 sinCosT)
{
	static const float COT30 = 1.7320508075688772;
	static const float TAN30 = 0.5773502691896257;
	static const float4 s_sampleDir[6] = {
		float4(0, 1, 0, 1/4.0),
		float4(0, 0.5, 0.866025, 3/20.0),
		float4(0.823639, 0.5, 0.267617, 3/20.0),
		float4(0.509037, 0.5, -0.700629, 3/20.0),
		float4(-0.509037, 0.5, -0.700629, 3/20.0),
		float4(-0.823639, 0.5, 0.267617, 3/20.0)
	};

	//jitter rotate
	half3 T, B;
	ComputeTangent(N, T, B);
	half sinT = sinCosT.x;
	half cosT = sinCosT.y; 
	half3 _T, _B;
	_T = T * cosT + B * sinT;
	_B = T * -sinT + B * -cosT;

	//dynamic bias
	half3 OrigBias = Orig;
	half3 InitT;
	[branch]if (DistanceToBox(OrigBias, 0) < HALF_EPSILON)
	{
		OrigBias += N * cNormalBias0;
		InitT = cTraceBias0;
	}
	else if (DistanceToBox(OrigBias, 1) < HALF_EPSILON)
	{
		OrigBias += N * cNormalBias1;
		InitT = cTraceBias1;
	}
	else
		return half4(cSkyLight.rgb*PI, 1);

	//cone trace
	half3 GI = half3(0, 0, 0);
	half AO = 0;
	int cone;
	[unroll]for (cone = 0; cone < 6; ++cone)
	{
		half3 D = s_sampleDir[cone].x * _T + s_sampleDir[cone].y * N + s_sampleDir[cone].z * _B;
		half t = InitT;
		half4 sdf = half4(0, 0, 0, 100);
		half VIS = 1.0;
		half3 radiance = half3(0, 0, 0);
		half3 alpha = 1.0;
		int i;
		for (i=0; i < RAYMARCHING_STEPS; ++i)
		{
			half3 marchPos = OrigBias + t*D;
			half aoDistance;
			[branch]if (DistanceToBox(marchPos, 0) < HALF_EPSILON)
			{
				Map(marchPos, 0, sdf, 0);
				aoDistance = max(cAoMinDistance0, sdf.w);
				[unroll]if (t < cAoMaxDistance0)
					VIS = min(COT30 * aoDistance / t, VIS);
				[branch]if (sdf.w < 0)//hit point
				{
					i = RAYMARCHING_STEPS;
					half level = clamp(log2(2*TAN30*t / cVolumeSize0), 0, 5);
					Map(marchPos, 0, sdf, level);
					radiance += sdf.rgb * alpha;
				}
				else
				{
					half level = clamp(log2(2*TAN30*t / cVolumeSize0), 0, 5);
					Map(marchPos, 0, sdf, level);
					half tAlpha = saturate(sdf.w / (TAN30*t));
					tAlpha *= tAlpha;
					radiance += sdf.rgb * (1 - tAlpha) * alpha;
					alpha *= tAlpha;

					if (i == RAYMARCHING_STEPS - 1)
					{
						GI += cSkyLight.rgb * s_sampleDir[cone].w;
					}
				}
			}
			else if (DistanceToBox(marchPos, 1) < HALF_EPSILON)
			{
				Map(marchPos, 1, sdf, 0);
				aoDistance = max(cAoMinDistance1, sdf.w);
				[unroll]if (t < cAoMaxDistance1)
					VIS = min(COT30 * aoDistance / t, VIS);
				[branch]if (sdf.w < 0)//hit point
				{
					i = RAYMARCHING_STEPS;
					half level = clamp(log2(2*TAN30*t / cVolumeSize1), 0, 5);
					Map(marchPos, 1, sdf, level);
					radiance += sdf.rgb * alpha;
				}
				else
				{
					half level = clamp(log2(2*TAN30*t / cVolumeSize1), 0, 5);
					Map(marchPos, 1, sdf, level);
					half tAlpha = saturate(sdf.w / (TAN30*t));
					tAlpha *= tAlpha;
					radiance += sdf.rgb * (1 - tAlpha) * alpha;
					alpha *= tAlpha;

					if (i == RAYMARCHING_STEPS - 1)
					{
						GI += cSkyLight.rgb * s_sampleDir[cone].w;
					}
				}
			}
			else//sky light
				break;
			t += sdf.w;
		}
		AO += VIS * s_sampleDir[cone].w;
		GI += cGIScale * radiance * s_sampleDir[cone].w;
	}
	return half4(GI*PI, saturate(AO));
}

half4 SDF_GIAO(in half3 Orig, in half3 N, in half2 sinCosT)
{
	static const float COT30 = 1.7320508075688772;
	static const float TAN30 = 0.5773502691896257;
	static const float4 s_sampleDir[6] = {
		float4(0, 1, 0, 1/4.0),
		float4(0, 0.5, 0.866025, 3/20.0),
		float4(0.823639, 0.5, 0.267617, 3/20.0),
		float4(0.509037, 0.5, -0.700629, 3/20.0),
		float4(-0.509037, 0.5, -0.700629, 3/20.0),
		float4(-0.823639, 0.5, 0.267617, 3/20.0)
	};

	//jitter rotate
	half3 T, B;
	ComputeTangent(N, T, B);
	half sinT = sinCosT.x;
	half cosT = sinCosT.y; 
	half3 _T, _B;
	_T = T * cosT + B * sinT;
	_B = T * -sinT + B * -cosT;

	//dynamic bias
	half3 OrigBias = Orig;
	half3 InitT;
	[branch]if (DistanceToBox(OrigBias, 0) < HALF_EPSILON)
	{
		OrigBias += N * cNormalBias0;
		InitT = cTraceBias0;
	}
	else if (DistanceToBox(OrigBias, 1) < HALF_EPSILON)
	{
		OrigBias += N * cNormalBias1;
		InitT = cTraceBias1;
	}
	else
		return half4(cSkyLight.rgb*PI, 1);

	//cone trace
	half3 GI = half3(0, 0, 0);
	half AO = 0;
	int cone;
	[unroll]for (cone = 0; cone < 6; ++cone)
	{
		half3 D = s_sampleDir[cone].x * _T + s_sampleDir[cone].y * N + s_sampleDir[cone].z * _B;
		half t = InitT;
		half4 sdf = half4(0, 0, 0, 100);
		half VIS = 1.0;
		int i;
		[branch]for (i=0; i < RAYMARCHING_STEPS; ++i)
		{
			half3 marchPos = OrigBias + t*D;
			half aoDistance;
			[branch]if (DistanceToBox(marchPos, 0) < HALF_EPSILON)
			{
				Map(marchPos, 0, sdf, 0);
				aoDistance = max(cAoMinDistance0, sdf.w);
				[unroll]if (t < cAoMaxDistance0)
					VIS = min(COT30 * aoDistance / t, VIS);
				[branch]if (sdf.w < 0)//hit point
				{
					half level = clamp(log2(2*TAN30*t / cVolumeSize0), 0, 5);
					Map(marchPos, 0, sdf, level);
					i = RAYMARCHING_STEPS;
					GI += sdf.rgb * cGIScale * s_sampleDir[cone].w;
				}
				else if (i == RAYMARCHING_STEPS - 1)
				{
					GI += cSkyLight.rgb * s_sampleDir[cone].w;
				}
			}
			else if (DistanceToBox(marchPos, 1) < HALF_EPSILON)
			{
				Map(marchPos, 1, sdf, 0);
				aoDistance = max(cAoMinDistance1, sdf.w);
				[unroll]if (t < cAoMaxDistance1)
					VIS = min(COT30 * aoDistance / t, VIS);
				[branch]if (sdf.w < 0)//hit point
				{
					half level = clamp(log2(2*TAN30*t / cVolumeSize1), 0, 5);
					Map(marchPos, 1, sdf, level);
					i = RAYMARCHING_STEPS;
					GI += sdf.rgb * cGIScale * s_sampleDir[cone].w;
				}
				else if (i == RAYMARCHING_STEPS - 1)
				{
					GI += cSkyLight.rgb * s_sampleDir[cone].w;
				}
			}
			else//sky light
				break;
			t += sdf.w;
		}
		AO += VIS * s_sampleDir[cone].w;
	}
	return half4(GI*PI, saturate(AO));
}

/////////////////// gi ao ////////////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 GIAOPS(v2f IN) : SV_Target
{
	//calc WorldPos
	float2 HPosition = IN.HPosition.xy * 2;
	float2 spos = HPosition;

	[@_D3D9]
	spos += 0.5;
	
	//worldPos
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, stc, 0));
	half3 D = CalcHomogeneousPos(sceneZ, stc);
	half3 WorldPos = CameraPos.xyz + D;
	
	//worldNormal
	half4 _GBufferNormal = tGBufferNormal.Sample(sGBufferNormal, stc);
	half3 N = half3(0, 1, 0);
	N = normalize(DecodeNormal(_GBufferNormal.rgb));	
	half Metallic;
	half IsPrincipalRole;
	DecodeFloatAndBit(_GBufferNormal.a, Metallic, IsPrincipalRole);

	if (IsPrincipalRole > 0.5)
		return float4(cSkyLight.rgb*PI, 1);

	//curr giao
	half2 sinCosT = tRandom2D.Sample(sRandom2D, float2(WorldPos.z+WorldPos.y, WorldPos.x+WorldPos.y) * GameTime ).xy * 2 - 1;
	half4 GIAO = SDF_GIAO2(WorldPos, N, sinCosT);
	return GIAO;
}

technique GIAO
{
	VertexShader = ScreenVS();
	PixelShader = GIAOPS();
}

float4 GIAOTSAAPS(v2f IN) : SV_Target
{
	//calc WorldPos
	float2 HPosition = IN.HPosition.xy * 2;
	float2 spos = HPosition;

	[@_D3D9]
	spos += 0.5;
	
	//worldPos
	float2 stc = spos * ScreenInfo.zw;
	float sceneZ = DecodeDepth(tDepthMap.SampleLevel(sDepthSampler, stc, 0));
	half3 D = CalcHomogeneousPos(sceneZ, stc);
	half3 WorldPos = CameraPos.xyz + D;

	//currGIAO
	float4 currGIAO = tCurrGIAO.Sample(sCurrGIAO, stc);

	//last giao
	if (@_LastGIAOEnable)
	{
		half4 lastPos = MulMat4x4(cLastViewProj, float4(WorldPos,1));
		lastPos.xy = lastPos.xy / lastPos.w;
		if (abs(lastPos.x) < 1.0 && abs(lastPos.y) < 1.0)//in screen
		{
			half2 lastUV = lastPos.xy * half2(0.5, -0.5) + half2(0.5, 0.5);
			
			//get border Pixels
			float4 c0 = tCurrGIAO.Sample(sCurrGIAO, stc + float2(ScreenInfo.z*2, 0));
			float4 c1 = tCurrGIAO.Sample(sCurrGIAO, stc + float2(-ScreenInfo.z*2, 0));
			float4 c2 = tCurrGIAO.Sample(sCurrGIAO, stc + float2(0, ScreenInfo.w*2));
			float4 c3 = tCurrGIAO.Sample(sCurrGIAO, stc + float2(0, -ScreenInfo.w*2));
			float4 c4 = currGIAO;

			float4 cMin = min(min(c0,c1), min(c2,c3));
			float4 cMax = max(max(c0,c1), max(c2,c3));
			cMin = min(cMin, c4);
			cMax = max(cMax, c4);

			float4 lastGIAO = tLastGIAO.Sample(sLastGIAO, lastUV);
			lastGIAO = clamp(lastGIAO, cMin, cMax);
			currGIAO = lerp(currGIAO, lastGIAO, 0.99);
		}	
	}
	return currGIAO;
}

technique GIAOTSAA
{
	VertexShader = ScreenVS();
	PixelShader = GIAOTSAAPS();
}

