/**
 * Copyright (C) 2012 Jorge Jimenez (jorge@iryoku.com)
 * Copyright (C) 2012 Diego Gutierrez (diegog@unizar.es)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *    1. Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the following disclaimer
 *       in the documentation and/or other materials provided with the 
 *       distribution:
 *
 *       "Uses Separable SSS. Copyright (C) 2012 by Jorge Jimenez and Diego
 *        Gutierrez."
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS 
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS OR CONTRIBUTORS 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software and documentation are 
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of the copyright holders.
 */


/**
 *                  _______      _______      _______      _______
 *                 /       |    /       |    /       |    /       |
 *                |   (----    |   (----    |   (----    |   (----
 *                 \   \        \   \        \   \        \   \
 *              ----)   |    ----)   |    ----)   |    ----)   |
 *             |_______/    |_______/    |_______/    |_______/
 *
 *        S E P A R A B L E   S U B S U R F A C E   S C A T T E R I N G
 *
 *                           http://www.iryoku.com/
 *
 * Hi, thanks for your interest in Separable SSS!
 *
 * It's a simple shader composed of two components:
 *
 * 1) A transmittance function, 'SSSSTransmittance', which allows to calculate
 *    light transmission in thin slabs, useful for ears and nostrils. It should
 *    be applied during the main rendering pass as follows:
 *
 *        float3 t = albedo.rgb * lights[i].color * attenuation * spot;
 *        color.rgb += t * SSSSTransmittance(...)
 *
 *    (See 'Main.fx' for more details).
 *
 * 2) A simple two-pass reflectance post-processing shader, 'SSSSBlur*', which
 *    softens the skin appearance. It should be applied as a regular
 *    post-processing effect like bloom (the usual framebuffer ping-ponging):
 *
 *    a) The first pass (horizontal) must be invoked by taking the final color
 *       framebuffer as input, and storing the results into a temporal
 *       framebuffer.
 *    b) The second pass (vertical) must be invoked by taking the temporal
 *       framebuffer as input, and storing the results into the original final
 *       color framebuffer.
 *
 *    Note that This SSS filter should be applied *before* tonemapping.
 *
 * Before including SeparableSSS.h you'll have to setup the target. The
 * following targets are available:
 *         SMAA_HLSL_3
 *         SMAA_HLSL_4
 *         SMAA_GLSL_3
 *
 * For more information of what's under the hood, you can check the following
 * URLs (but take into account that the shader has evolved a little bit since
 * these publications):
 *
 * 1) Reflectance: http://www.iryoku.com/sssss/
 * 2) Transmittance: http://www.iryoku.com/translucency/
 *
 * If you've got any doubts, just contact us!
 */
 
#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "Noise.fxh"
#include "BRDF.fxh"

///////////////// properties //////////////////
@_IsBlurX
{
	Name = IsBlurX;
	Description = "blur x";
	Default = false;
	Visible = true;
};
@_EnableSeparateLighting
{
	Name = EnableSeparateLighting;
	Description = "Enable Separate Lighting";  
	Default = false;
};
///////////////// io signatures //////////////////
struct a2v 
{
	float4 Position : POSITION; //in object space
	float2 BaseTC   : TEXCOORD0;
};

struct v2f
{
	float4 HPosition : SV_Position;
	float2 TC	: TEXCOORD0;
};

///////////////// samplers //////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Linear;
	Address = Clamp;
};

texture2D tScreen : layout(t1);
sampler sScreen : layout(s1)
{
	Texture = tScreen;
	Filter = Linear;
	Address = Clamp;
};

texture2D tSSSSpecular : layout(t2);
sampler sSSSSpecular : layout(s2)
{
	Texture = tSSSSpecular;
	Filter = Point;
	Address = Clamp;
};

texture2D tBaseMapAndMetallic : layout(t3);
sampler sBaseMapAndMetallic : layout(s3)
{
	Texture = tBaseMapAndMetallic;
	Filter = Point;
	Address = Clamp;
};

///////////////// const buffer //////////////////
#define	SSSS_N_SAMPLES 17
cbuffer Batch
{
	float  sssWidth							;
	float2 blurDir							;	//(1,0) <-> (0,1)
	float  cotHalfFovY						;		//1.0 / tan(0.5 * radians(SSSS_FOVY))
	float4 kernels[SSSS_N_SAMPLES]			;
};

/////////////////// MediaSetupByGlobalFog ///////////////
v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 DiffusionPS(v2f IN) : SV_Target
{
	float4 colorM = tScreen.Sample(sScreen, IN.TC);

    // Initialize the stencil buffer in case it was not already available:
    // if (initStencil) // (Checked in compile time, it's optimized away)
    //     if (SSSS_STREGTH_SOURCE == 0.0) discard;

    // Fetch linear depth of current pixel:
    float depthM = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, IN.TC)) * CameraInfo.y;
    //float depthM = DecodeDepth(colorM.w) * CameraInfo.y;

    // Calculate the sssWidth scale (1.0 for a unit plane sitting on the projection window):
    float distanceToProjectionWindow = cotHalfFovY;//1.0 / tan(0.5 * radians(SSSS_FOVY));
    float scale = distanceToProjectionWindow / depthM;

    // Calculate the final step to fetch the surrounding pixels:
    float2 finalStep = sssWidth * scale * blurDir;
    //finalStep *= SSSS_STREGTH_SOURCE; // Modulate it using the alpha channel.
    finalStep *= 1.0 / 3.0; // Divide by 3 as the kernels range from -3 to 3.
    

    // Accumulate the center sample:
    float4 colorBlurred = colorM;
    colorBlurred.rgb *= kernels[0].rgb;

    // Accumulate the other samples:
    for (int i = 1; i < SSSS_N_SAMPLES; i++) 
	{
        // Fetch color and depth for current sample:
        float2 offset = IN.TC + kernels[i].a * finalStep;
        float4 color = tScreen.Sample(sScreen, offset);

        // If the difference in depth is huge, we lerp color back to "colorM":
        float depth = DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, offset)) * CameraInfo.y;
		//float depth = DecodeDepth(color.w) * CameraInfo.y;
        float s = saturate(300.0f * distanceToProjectionWindow *sssWidth * abs(depthM - depth));
        color.rgb = lerp(color.rgb, colorM.rgb, s);

        // Accumulate:
        colorBlurred.rgb += kernels[i].rgb * color.rgb;
    }

    if (!@_IsBlurX)
    {
        float4 BaseMapAndMetallic = tBaseMapAndMetallic.Sample(sBaseMapAndMetallic, IN.TC);
        float4 specularSSSAndMetallic = tSSSSpecular.Sample(sSSSSpecular, IN.TC);
        float3 DiffuseColor = Diffuse_Lambert(BaseMapAndMetallic.rgb - BaseMapAndMetallic.rgb * specularSSSAndMetallic.a);
        colorBlurred.rgb *= DiffuseColor;
        
        /*// 在merge lighting那边整合
    	float3 specularSSS = specularSSSAndMetallic.rgb * SD_SCENE_MAX_LUMINANCE;
		EncodeHDR(specularSSS);
		colorBlurred.rgb += specularSSS;
        */
    }

    return colorBlurred;
}

technique DiffusionBlur
{
	VertexShader = ScreenVS();
	PixelShader = DiffusionPS();
}



