//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("Effect")

#include "Constants.fxh"
#include "Common.fxh"

/*{Configs}*/
/*{VaryingMask}*/

#include "ShaderGraphDefinitions.fxh"

//////////////////////////////////////////////////////////////////////////
// Effect shading model.
cbuffer Batch
{
	float4x3 World 				: World;
	float4 EffectColor			: Color;
	float4 EffectParam			: Param;
	float4 SubUV				: Param2;
/*{BatchParameters}*/
};

/*{GlobalParameters}*/

//////////////////////////////////Assemble Input////////////////////////////////////////
struct EffectVertexAssemble
{
	float4		Position 		: POSITION;

	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;
	
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;

	[@_INSTANCED]
	float4		Transform0		: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1		: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2		: TEXCOORD5;
	[@_INSTANCED]
	float4		EffectColor		: TEXCOORD6;
	[@_INSTANCED]
	float4		EffectParam		: TEXCOORD7;
};

//////////////////////////////////////////////////////////////////////////
/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate 
	float4		WorldNormal		: TEXCOORDN;			//	World Normal and fog factor.
#if	ENABLE_WORLDPOSITION
	float4		WorldPosition	: TEXCOORDN;			//	World position and linear depth.
#elif ENABLE_LINEARDEPTH
	float		LinearZ			: TEXCOORDN;			//	linear depth
#endif	
#if	ENABLE_VIEWDIRECTION
	float3		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
#endif
	float4		TintColor		: COLOR;				//	Per-vertex color.
	[@_INSTANCED]
	float4		EffectColor		: TEXCOORDN;			// 	Effect color.
	[@_INSTANCED]
	float4		EffectParam		: TEXCOORDN;			//	Effect param.
};

//////////////////////////////////////////////////////////////////////////

#if (VARYING_MASK & VA_TEXCOORD2) && !defined(VATexCoord2)
#define VATexCoord2(Arg)		Arg.TexCoord.zw
#define VSTexCoord2				VATexCoord2(OUT)
#define PSTexCoord2				VATexCoord2(IN)
#endif

#if ENABLE_WORLDPOSITION && ENABLE_LINEARDEPTH
#define VALinearDepth(Arg)		Arg.WorldPosition.w
#define VSLinearDepth			VALinearDepth(OUT)
#define PSLinearDepth			VALinearDepth(IN)
#endif

#include "ShaderGraphVarying.fxh"
#include "Lighting.fxh"

//////////////////////////////////////////////////////////////////////////
// Common Effect vertex shader.

VertexShadingOutput EffectVertexShading(in EffectVertexAssemble IN)
{
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	float4 LocalPosition = IN.Position;
	float3 LocalNormal = (float3)Expand((half3)IN.Normal.xyz);

	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, LocalPosition.xyz);
			LocalNormal = Matrix4x3TransformN(Mx, My, Mz, LocalNormal);
		}
	}

#if	(ENABLE_WORLDPOSITION == 0)
	float4 WorldPosition;
#endif

	if (@_INSTANCED)
	{
		#if	ENABLE_WORLDPOSITION
			OUT.WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, LocalPosition.xyz), 1);	
		#else
			WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, LocalPosition.xyz), 1);	
		#endif
		OUT.WorldNormal.xyz = normalize(Matrix4x3TransformN(IN.Transform0, IN.Transform1, IN.Transform2, LocalNormal));
		OUT.TexCoord.xy = IN.TexCoord;
		OUT.EffectColor = IN.EffectColor;
		OUT.EffectParam = IN.EffectParam;
	}
	else
	{
		#if	ENABLE_WORLDPOSITION
			OUT.WorldPosition = float4(MulMat3x4(World, float4(LocalPosition.xyz, 1)).xyz, 1);
		#else
			WorldPosition = float4(MulMat3x4(World, float4(LocalPosition.xyz, 1)).xyz, 1);
		#endif
		OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, LocalNormal));
		OUT.TexCoord.xy = IN.TexCoord * SubUV.xy + SubUV.zw;
	}

	[@_LIGHT_MAP]
	OUT.TexCoord.zw = IN.TexCoord2;

/*{VertexShader}*/	

	/*{WorldPositionOffset}*/

#if	ENABLE_WORLDPOSITION
	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);
#else
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
#endif
	OUT.HPosition.z -= CameraInfo.x * /*{DepthBias}*/ / (/*{DepthBias}*/ + OUT.HPosition.w);

	float3 viewDirection;
#if	ENABLE_WORLDPOSITION
	OUT.WorldPosition.w = OUT.HPosition.w;
	viewDirection = OUT.WorldPosition.xyz - CameraPos.xyz;
#elif ENABLE_LINEARDEPTH
	OUT.LinearZ = OUT.HPosition.w;
	viewDirection = WorldPosition.xyz - CameraPos.xyz;
#else
	viewDirection = WorldPosition.xyz - CameraPos.xyz;
#endif	

#if EFFECT_ENABLE_FOG
	float fog = FogTransmittance(viewDirection);
	OUT.WorldNormal.w = pow(fog, /*{FogStrength}*/ + 0.001);
#endif	

	//ViewDirection
#if	ENABLE_VIEWDIRECTION
	OUT.ViewDirection = (-viewDirection);
#endif

	if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}

#if EFFECT_ENCODE_HDR	
	[@_HDR]
	OUT.TintColor.rgb *= /*{HDRScale}*/;
#endif

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Utilities
float4 EncodePremultipliedAlphaAndBloom(float4 IN, float invEmission, float fog, float bloom)
{
	float4 OUT = IN;
	OUT.rgb *= OUT.a;
	OUT.a *= invEmission;
#if EFFECT_ENABLE_FOG	
	OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, fog);
#endif	
	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * bloom);
	return OUT;
}

float4 GetEffectColor(in VertexShadingOutput IN)
{
	if (@_INSTANCED)
	{
		return IN.EffectColor * EffectColor * IN.TintColor;
	}
	else
	{
		return EffectColor * IN.TintColor;
	}
}

float4 GetEffectParam(in VertexShadingOutput IN)
{
	if (@_INSTANCED)
	{
		return IN.EffectParam;
	}
	else
	{
		return EffectParam;
	}
}

//////////////////////////////////////////////////////////////////////////
// technique Clip.
float4 EffectClipShading(in VertexShadingOutput IN) : SV_Target
{
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectClipShading();
}

//////////////////////////////////////////////////////////////////////////
// Common effect pixel shader.
float4 EffectPixelShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
#if	ENABLE_VIEWDIRECTION
	IN.ViewDirection = normalize(IN.ViewDirection);
#endif


/*{PixelShader}*/


	//{AlphaReference} clip(/*{EffectColor}*/.a - /*{AlphaReference}*/);

	if (@_TRANSPARENT)
	{
		float invEmission = GetEffectParam(IN).z;
		/*{EffectColor}*/ = EncodePremultipliedAlphaAndBloom(/*{EffectColor}*/, invEmission, IN.WorldNormal.w, /*{BloomScale}*/);
	}
	else
	{
		/*{EffectColor}*/.a = EncodeSolidBloomFactor(GetLuminance(/*{EffectColor}*/.rgb) * /*{BloomScale}*/);
	}

	if(@_HDR)
	{
		EncodeHDRParticle(/*{EffectColor}*/);
	}
	return /*{EffectColor}*/;
}

technique Shading
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 EffectDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.05, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectDebugShading();
}
