//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("Function")

#include "Constants.fxh"
#include "Common.fxh"

#define VARYING_MASK 0xFFFF

#include "ShaderGraphDefinitions.fxh"

//////////////////////////////////////////////////////////////////////////
// Function shading model.
cbuffer Batch
{
	float4x3 World 				: World;
};

/*{GlobalParameters}*/

//////////////////////////////////Assemble Input////////////////////////////////////////
struct FunctionVertexAssemble
{
	float4		Position 		: POSITION;	
	float4		Color 			: COLOR;
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;
};

//////////////////////////////////////////////////////////////////////////
/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4 		TintColor		: COLOR;				//  Vertex Color.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate and depth for surface.
	float4		WorldNormal		: TEXCOORDN;			//	World Normal and fog factor.
	float4		WorldPosition	: TEXCOORDN;			//	World position and linear depth.
	float3		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
	float4		WorldTangent	: TEXCOORDN;			//	Output tangents in world space for normal mapping.
	float4		WorldBinormal	: TEXCOORDN;			//	Output binormal in world space for normal mapping.

};
//////////////////////////////////Common//////////////////////////////////////

#define VALinearDepth(Arg)		Arg.TexCoord.z
#define VSLinearDepth			VALinearDepth(OUT)
#define PSLinearDepth			VALinearDepth(IN)

#define VATexCoord2(Arg)		Arg.TexCoord.zw
#define VSTexCoord2				VATexCoord2(OUT)
#define PSTexCoord2				VATexCoord2(IN)

#include "ShaderGraphVarying.fxh"

//////////////////////////////////////////////////////////////////////////
// Common Function vertex shader.

VertexShadingOutput FunctionVertexShading(in FunctionVertexAssemble IN)
{
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

/*{VertexShader}*/

	float HDRScale = 2.0;
	/*{HDRScale}*/
	float DepthBias = 0.0;
	/*{DepthBias}*/


	float4 LocalPosition = IN.Position;
	float3 LocalNormal = (float3)Expand((half3)IN.Normal.xyz);

	OUT.TintColor = GetVertexColor(IN.Color);
	OUT.WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1)).xyz, 1);
	OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, LocalNormal));
	OUT.TexCoord.xy = IN.TexCoord;
	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);

	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.HPosition.z -= CameraInfo.x * DepthBias / (DepthBias + OUT.HPosition.w);

	if(@_TANGENT)
	{
		OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, normalize(IN.Tangent.xyz)));
		OUT.WorldTangent.w = 0;
		OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, normalize(IN.Binormal.xyz)));	
		OUT.WorldBinormal.w = 0;
	}

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Common function pixel shader.
float4 FunctionPixelShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
/*{PixelShader}*/
	float4 OUT;

	float  BloomScale = 0.5;
	/*{BloomScale}*/
	float4 InputColor = float4(1.0, 1.0, 1.0, 1.0);
	/*{InputColor}*/

	OUT = InputColor;
	
	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);

	if(@_HDR)
	{
		EncodeHDRParticle(OUT);
	}
	return OUT;
}

technique Shading
{
	VertexShader = FunctionVertexShading();
	PixelShader = FunctionPixelShading();
}
