//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("Particle")

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World;
	float4   InstanceColor		: Color;
};

cbuffer Shader
{
	float4 ModulateColor		= { 1, 1, 1, 1 };
	float  BloomScale			= 0.5;
	float  FogStrength			= 0;
	float  DepthBias			= 0;
	float  HDRScale				= 2.0;
};

/*{GlobalParameters}*/

struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		VertexColor 	: COLOR;
	float2		TexCoord		: TEXCOORDN;
};

struct VertexShadingOutput
{
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4 		Color		 	: COLOR;				//  Vertex Color.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.
};


//////////////////////////////////////////////////////////////////////////
// Common particle vertex shader.

VertexShadingOutput ParticleVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	float4 WorldPosition;
	WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1)).xyz, 1);

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.HPosition.z -= CameraInfo.x * DepthBias / (DepthBias + OUT.HPosition.w);

	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.z = OUT.HPosition.w;
	float fog = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);
	OUT.TexCoord.w = pow(fog, FogStrength + 0.001);

	OUT.Color = IN.VertexColor;
	OUT.Color.rgb *= HDRScale;

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Common particle pixel shader.
float3 GetParticleNormal(in float3 worldPosition)
{
	float3 P = worldPosition;
	float3 N = cross(ddy(P), ddx(P));
	if (@_GLSL || @_SPIRV)
	{
		// ddx/ddy is incorrect on some android devices.
		float len = length(N);
		N = (len == 0) ? float3(0, 1, 0) : N / len;
	}
	else
	{
		N = normalize(N);
	}
	return N;
}

half4 GetFinalParticleColor(in VertexShadingOutput IN, half4 color)
{
	half4 OUT = color * InstanceColor;
	OUT.rgb *= OUT.a;
	OUT *= GetVertexColor(IN.Color);
	return OUT;
}

float4 ParticlePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(1, 0, 0, 1);

	// Customize your pixel shader herer.

	OUT = GetFinalParticleColor(IN, OUT);

/*	Modifify here to enbale fog.
	if (@_VolumetricFog)
	{
		float2 spos = IN.HPosition.xy;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.TexCoord.z;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz * OUT.a;
	}
	else
	{
		OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, IN.TexCoord.w);
	}
*/

	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);
	if(@_HDR)
	{
		EncodeHDRParticle(OUT);
	}
	return OUT;
}

technique Shading
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticlePixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticleDebugShading();
}
