//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("Particle")

#include "Constants.fxh"
#include "Common.fxh"

/*{Configs}*/
/*{VaryingMask}*/

#define ENABLE_WORLDPOSITION (ENABLE_WORLDNORMAL | (VARYING_MASK & VA_WORLD_POSITION))

#include "ShaderGraphDefinitions.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World 		;
	float4   InstanceColor		: Color			;
/*{BatchParameters}*/
};

/*{GlobalParameters}*/

struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		VertexColor 	: COLOR;
	float2		TexCoord		: TEXCOORDN;
};

struct VertexShadingOutput
{
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4 		TintColor		: COLOR;				//  Vertex Color.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.
#if	ENABLE_WORLDPOSITION
	float3 		WorldPosition 	: TEXCOORDN;			//  World Position.
#endif
#if	ENABLE_VIEWDIRECTION
	float3		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
#endif
};

//////////////////////////////////////////////////////////////////////////

#if ENABLE_WORLDNORMAL
#define VAWorldNormal(Arg)		PSWorldNormal
#endif

#if ENABLE_LINEARDEPTH
#define VALinearDepth(Arg)		Arg.TexCoord.z
#define VSLinearDepth			VALinearDepth(OUT)
#define PSLinearDepth			VALinearDepth(IN)
#endif

#include "ShaderGraphVarying.fxh"
#include "Lighting.fxh"

//////////////////////////////////////////////////////////////////////////
// Common particle vertex shader.

VertexShadingOutput ParticleVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	float4 WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1)).xyz, 1);
#if	ENABLE_WORLDPOSITION
	OUT.WorldPosition = WorldPosition.xyz;
#endif

	float3 viewDirection = WorldPosition.xyz - CameraPos.xyz;

#if	ENABLE_VIEWDIRECTION
	OUT.ViewDirection = (-viewDirection);
#endif

/*{VertexShader}*/

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.HPosition.z -= CameraInfo.x * /*{DepthBias}*/ / (/*{DepthBias}*/ + OUT.HPosition.w);

	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = 0;
	OUT.TintColor = IN.VertexColor;

#if PARTICLE_ENCODE_HDR
	[@_HDR]
	OUT.TintColor.rgb *= /*{HDRScale}*/;
#endif	

#if PARTICLE_ENABLE_FOG
	float fog = FogTransmittance(viewDirection);
	OUT.TexCoord.w = pow(fog, /*{FogStength}*/ + 0.001);
#endif

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Common particle pixel shader.
float3 GetParticleNormal(in float3 worldPosition)
{
	float3 P = worldPosition;
	float3 N = cross(ddy(P), ddx(P));
	if (@_GLSL || @_SPIRV)
	{
		// ddx/ddy is incorrect on some android devices.
		float len = length(N);
		N = (len == 0) ? float3(0, 1, 0) : N / len;
	}
	else
	{
		N = normalize(N);
	}
	return N;
}

half4 GetFinalParticleColor(in VertexShadingOutput IN, half4 color)
{
	half4 OUT = color * InstanceColor;
	OUT.rgb *= OUT.a;
	OUT *= GetVertexColor(IN.TintColor);
	return OUT;
}

float4 ParticlePixelShading(in VertexShadingOutput IN) : SV_Target
{
#if ENABLE_WORLDNORMAL
	float3 PSWorldNormal = GetParticleNormal(IN.WorldPosition);
#endif

#if	ENABLE_VIEWDIRECTION
	IN.ViewDirection = normalize(IN.ViewDirection);
#endif

/*{PixelShader}*/

	//{AlphaReference} clip(/*{ParticleColor}*/.a - /*{AlphaReference}*/);

#if PARTICLE_ENABLE_FOG
	if (@_VolumetricFog)
	{
		float2 spos = IN.HPosition.xy;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.TexCoord.z;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		/*{ParticleColor}*/.rgb = /*{ParticleColor}*/.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz * /*{ParticleColor}*/.a;
	}
	else
	{
		/*{ParticleColor}*/.rgb = lerp(FogColor.rgb * /*{ParticleColor}*/.a, /*{ParticleColor}*/.rgb, IN.TexCoord.w);
	}
#endif

	/*{ParticleColor}*/.a = EncodeTransparentBloomFactor(/*{ParticleColor}*/.a, GetLuminance(/*{ParticleColor}*/.rgb) * /*{BloomScale}*/);
#if PARTICLE_ENCODE_HDR	
	if(@_HDR)
	{
		EncodeHDRParticle(/*{ParticleColor}*/);
	}
#endif
	return /*{ParticleColor}*/;
}

technique Shading
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticlePixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticleDebugShading();
}
