//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("Unlit")

#define MATERIAL_SHADINGMODEL_UNLIT 1

#include "Constants.fxh"
#include "Common.fxh"  

//////////////////////////////////////////////////////////////////////////
// Constant buffers.
cbuffer Batch
{
	float4x3 World 					: World;
	float4 cParamter				: Param;
	float4 cParamter2				: Param2;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex;
};

cbuffer Shader
{
	float cBloomScale <Widget="FloatSlider"; Min="0.0"; Max="1.0"; Step="0.01"; Precision="2";>		= { 0.5 };
};

//////////////////////////////////////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;
};

struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		TexCoord		: TEXCOORDN;
	float4		WorldPosition	: TEXCOORDN;
	float4 		WorldNormal		: NORMAL;

	[@_TANGENT]
	float4		WorldTangent	: TEXCOORDN;			//	Output tangents in world space for normal mapping.
	[@_TANGENT]
	float4		WorldBinormal	: TEXCOORDN;			//	Output binormal in world space for normal mapping.
};

#include "Lighting.fxh" 

//////////////////////////////////////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;
	float3 LocalNormal;

	[@_TANGENT]
	float3 LocalTangent;
	[@_TANGENT]
	float3 LocalBinormal;
};

//////////////////////////////////////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	[@_HLSL]
	INTM = (InterMediateVertex)0;

	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}

void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
		}
		/*
		// Hard skinning.
		if (@_DF_SKIN == 2)
		{
			// Fetch.
			float4 Mx, My, Mz;
			GetSkeletonMatrix4x3(i[0], Mx, My, Mz);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalBinormal);
		}
		// Dual-quaternion skinning.
		if (@_DF_SKIN == 3)
		{
			// Fetch.
			float4 Q0[2], Q1[2], Q2[2], Q3[2];
			GetSkeletonDualQuaternion(i[0], Q0[0], Q0[1]);
			GetSkeletonDualQuaternion(i[1], Q1[0], Q1[1]);
			GetSkeletonDualQuaternion(i[2], Q2[0], Q2[1]);
			GetSkeletonDualQuaternion(i[3], Q3[0], Q3[1]);
			// Amend orientations.
			w *= float4(1.f, sign(dot(Q0[0], Q1[0])), sign(dot(Q0[0], Q2[0])), sign(dot(Q0[0], Q3[0])));
	
			// Blend.
			float4 Qn = LinearBlend(w, Q0[0], Q1[0], Q2[0], Q3[0]);
			float4 Qd = LinearBlend(w, Q0[1], Q1[1], Q2[1], Q3[1]); 

			// Apply.
			INTM.LocalPosition.xyz = DualQuaternionTransformP(Qn, Qd, INTM.LocalPosition.xyz);
			INTM.LocalNormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = DualQuaternionTransformN(Qn, Qd, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalBinormal);
		}
		*/
	}
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	// TexCoord
	OUT.TexCoord = IN.TexCoord;

	// Normal
	OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	OUT.WorldNormal.w = 0;

	// Position
	OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);

	if (@_TANGENT)
	{
		OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, INTM.LocalTangent));
		OUT.WorldTangent.w = 0;
		OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalBinormal));	
		OUT.WorldBinormal.w = 0;
	}
}

/////////////////////////////////////////////////////////////////////////
// Shading
VertexShadingOutput UnlitVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	// Intermediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	// Output.
	VertexUnifyOutput(IN, INTM, OUT);

	// Customize your vertex shader here.

	return OUT;
}

float4 UnlitPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(1, 0, 0, 0);

	// Customize your pixel shader here.

	if (@_TRANSPARENT)
	{
		OUT.rgb *= OUT.a;
		OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * cBloomScale);
	}
	else
	{
		OUT.a = EncodeSolidBloomFactor(cBloomScale);
	}
	return OUT;
}

technique Shading
{
	VertexShader = UnlitVertexShading();
	PixelShader = UnlitPixelShading();
}

/////////////////////////////////////////////////////////////////////////
// ShadowGen

struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
};

ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	// InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	// Position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

/////////////////////////////////////////////////////////////////////////
// Stencil

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 StencilPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = StencilPixelShading();
}

/////////////////////////////////////////////////////////////////////////
// VBufferGen
#if 0	// begin VBufferGen. Set to 1 to enable VBuffer.

struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	// Position
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	// Motion vector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	// Curr uv
	float2 spos = IN.HPosition.xy;
	float2 currUV = spos * ScreenInfo.zw;
	
	// Last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy * ScreenInfo.xy);
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}

#endif // end VBufferGen

/////////////////////////////////////////////////////////////////////////
// Debug

technique Debug
{
	VertexShader = UnlitVertexShading();
	PixelShader = UnlitPixelShading();
}
