//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.

#pragma shader graph("Unlit")

#define MATERIAL_SHADINGMODEL_UNLIT 1

#include "Constants.fxh"
#include "Common.fxh"  
#include "BRDF.fxh"
#include "Lighting.fxh" 

/*{Configs}*/
/*{VaryingMask}*/

#include "ShaderGraphDefinitions.fxh"

//////////////////////////////////Properties////////////////////////////////////////
/*{GlobalParameters}*/

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World;
	float4 cParamter2				: Param2;
	float4 cParamter				: Param;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex;
/*{BatchParameters}*/
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;
	
	float4		WorldPosition	: TEXCOORDN;

	float4 		WorldNormal		: NORMAL;

#if	ENABLE_VIEWDIRECTION
	float3		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.
#endif

#if UNLIT_ENABLE_FOG
	float 		T 				: TEXCOORDN;			//	Fog Transmitance
#endif

#if ENABLE_VERTEXCOLOR
	float4		TintColor 		: COLOR;				//	TintColor
#endif

	[@_TANGENT]
	float4		WorldTangent	: TEXCOORDN;			//	Output tangents in world space for normal mapping.

	[@_TANGENT]
	float4		WorldBinormal	: TEXCOORDN;			//	Output binormal in world space for normal mapping.
	[@_UV3UV4]
	float4		MultiUVs		: TEXCOORDN;
/*{costumized uvs}*/
};
//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;

	[@_VERTEXCOLOR]
	float4		Color 			: COLOR;

	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_LIGHT_MAP]
	float2		TexCoord2		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_UV3UV4]
	float4		TexCoord3		: TEXCOORD2;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;
};
//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

	float3 LocalNormal;

	[@_TANGENT]
	float3 LocalTangent;

	[@_TANGENT]
	float3 LocalBinormal;
};
//////////////////////////////////Common//////////////////////////////////////

#if ENABLE_LINEARDEPTH
#define VALinearDepth(Arg)		Arg.WorldPosition.w
#define VSLinearDepth			VALinearDepth(OUT)
#define PSLinearDepth			VALinearDepth(IN)
#endif

#define VATexCoord2(Arg)		Arg.TexCoord.zw
#define VSTexCoord2				VATexCoord2(OUT)
#define PSTexCoord2				VATexCoord2(IN)

#include "ShaderGraphVarying.fxh"
//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);

	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}

void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
		}
	}
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	//TexCoord
	OUT.TexCoord = IN.TexCoord.xyxy;
	[@_LIGHT_MAP]
	OUT.TexCoord.zw = IN.TexCoord2.xy;

	[@_UV3UV4]
	OUT.MultiUVs = IN.TexCoord3;

	//normal
	OUT.WorldNormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalNormal));
	OUT.WorldNormal.w = 0;

#if ENABLE_VERTEXCOLOR
	if (@_VERTEXCOLOR)
	{
		OUT.TintColor = GetVertexColor(IN.Color);
	}
	else
	{
		OUT.TintColor = float4(1, 1, 1, 1);
	}
#endif
	//position
	OUT.WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);

	if (@_TANGENT)
	{
		OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, INTM.LocalTangent));
		OUT.WorldTangent.w = 0;
		OUT.WorldBinormal.xyz = normalize(MulMat3x4((float3x3)World, INTM.LocalBinormal));	
		OUT.WorldBinormal.w = 0;
	}

/*{VertexShader}*/

	/*{WorldPositionOffset}*/

	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);

#if ENABLE_LINEARDEPTH
	OUT.WorldPosition.w = OUT.HPosition.w;
#endif

	//{DepthBias} OUT.HPosition.z -= CameraInfo.x * /*{DepthBias}*/ / (/*{DepthBias}*/ + OUT.HPosition.w);

	//ViewDirection
#if	ENABLE_VIEWDIRECTION
	OUT.ViewDirection = OUT.WorldPosition.xyz - CameraPos.xyz;
#endif

#if UNLIT_ENABLE_FOG
	if (@_VolumetricFog)
	{
		OUT.T = OUT.HPosition.w;
	}else
	{
		//Fog T
	#if	ENABLE_VIEWDIRECTION
		OUT.T = FogTransmittance(OUT.ViewDirection.xyz);
	#else
		float3 viewDirection = OUT.WorldPosition.xyz - CameraPos.xyz;
		OUT.T = FogTransmittance(viewDirection);
	#endif
	}
#endif

#if	ENABLE_VIEWDIRECTION
	OUT.ViewDirection = (-OUT.ViewDirection);
#endif
}
//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput UnlitVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float4 UnlitPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);

#if	ENABLE_VIEWDIRECTION
	IN.ViewDirection = normalize(IN.ViewDirection);
#endif
	
/*{PixelShader}*/

	/*{Emission}*/

	if (@_TRANSPARENT)
	{
		/*{Alpha}*/

		OUT.rgb *= OUT.a;
		OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * /*{BloomScale}*/);
	}
	else
	{
		OUT.a = EncodeSolidBloomFactor(/*{BloomScale}*/);
	}

#if UNLIT_ENABLE_FOG
	// fog shading
	if (@_VolumetricFog)
	{
		float2 spos = IN.HPosition.xy;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.T;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
	}
	else
		OUT.rgb = FogLighting(IN.T, OUT.rgb);

	// screen color tint
	OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);
#endif

	return OUT;
}

technique Shading
{
	VertexShader = UnlitVertexShading();
	PixelShader = UnlitPixelShading();
}
/////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);
	
	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = UnlitVertexShading();
	PixelShader = UnlitPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
float GetRandom(float2 tc)
{
	int2 sp = (int2)tc;
	sp -= (sp / int2(4,4)) * int2(4,4); // 3 slots
	float t = (float)sp.x * 4 + (float)sp.y;
	return frac((t * (34.0 / 16.0) + 1.0 / 16.0) * t); // perfect permute.
}

[barycentric("sample")][RE_Z]
float4 ClipPixelShading(in VertexShadingOutput IN) : SV_Target
{
#if	ENABLE_VIEWDIRECTION
	IN.ViewDirection = normalize(IN.ViewDirection);
#endif



/*{ClipPixelShader}*/

	// Clipping
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		clipValue = /*{ClipAlpha}*/ - ALPHA_REF;
	}
	if (@_DISSOLVING)
	{
		clipValue = min(clipValue, cParamter2.x - GetRandom(IN.HPosition.xy) * sign(cParamter2.x));
	}
	if (@_CLIPPING || @_DISSOLVING)
		clip(clipValue);
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = UnlitVertexShading();
	PixelShader = ClipPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}

