//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : zhaoyukun@corp.netease.com
//	Author: chaos
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<System.fx>
///	@path	~/System/Shader/
///	@date	2010/05/04
///	@brief	System shader.

#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x3 World 			: World				;
	float4 cColor 			: Color				;
	float4 Location2D 							;
	float4 BoxLocation							;
	float4 BoxScale								;
	float4 HPositionTransform					 = { 1.0f, 0.0f, 0.0f, 1.0f };

	float4 TextureMask							 = { 1.0f, 1.0f, 1.0f, 1.0f };
};

texture2D Preview_Texture	: layout(t0) = "White";
texture3D Preview_Texture3D	: layout(t0) = "White";
texture2D HUD_Texture		: layout(t0) = "White";

struct VertexAssemble
{
	float4		Position 	: POSITION;
	float4		Color 		: COLOR;
	float2		TexCoord	: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition	: SV_Position;
	float4		Color		: COLOR;
	float2		TexCoord	: TEXCOORDN;
};

////////////////////////////////////////////////////////////////////////
// Common wire display mesh.

VertexOutput WireDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	float4 WorldPosition = float4(MulMat3x4(World, IN.Position).xyz, 1.f);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	OUT.Color = GetVertexColor(IN.Color);
	OUT.TexCoord = IN.TexCoord;
	return OUT;
}

float4 WireDrawPS(in VertexOutput IN) : SV_Target
{
	float4 OUT = IN.Color;
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique WireDraw
{
	VertexShader = WireDrawVS();
	PixelShader = WireDrawPS();
}

////////////////////////////////////////////////////////////////////////
// Common box.

VertexOutput WireBoxVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	float4 Position = float4(IN.Position.xyz * BoxScale.xyz + BoxLocation.xyz, 1);
	float4 WorldPosition = float4(MulMat3x4(World, Position).xyz, 1.f);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	OUT.Color = GetVertexColor(IN.Color);
	OUT.TexCoord = IN.TexCoord;
	return OUT;
}

float4 WireBoxPS(in VertexOutput IN) : SV_Target
{
	float4 OUT = IN.Color;
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique WireBox
{
	VertexShader = WireBoxVS();
	PixelShader = WireBoxPS();
}

////////////////////////////////////////////////////////////////////////
// Anti-aliased wire draw.

struct GeometryOutput
{
	float4					HPosition	: SV_Position;
	float4					Color		: COLOR;
	noperspective float2	TexCoord	: TEXCOORDN;
};

[maxvertexcount(4)]
void AAWireDrawGS(line VertexOutput IN[2], inout TriangleStream<GeometryOutput> OUT)
{
	VertexOutput P0 = IN[0];
	VertexOutput P1 = IN[1];
	if (P0.HPosition.w > P1.HPosition.w)
	{
		VertexOutput temp = P0;
		P0 = P1;
		P1 = temp;
	}
	if (P0.HPosition.w < CameraInfo.x)
	{
		float ratio = (CameraInfo.x - P0.HPosition.w) / (P1.HPosition.w - P0.HPosition.w);
		P0.HPosition = lerp(P0.HPosition, P1.HPosition, ratio);
	}

	float2 a = P0.HPosition.xy / P0.HPosition.w;
	float2 b = P1.HPosition.xy / P1.HPosition.w;
	float2 c = normalize(float2(a.y - b.y, b.x - a.x)) * ScreenInfo.zw * 3;

	GeometryOutput g0;
	g0.HPosition = float4(P0.HPosition.xy + c * P0.HPosition.w, P0.HPosition.zw);
	g0.Color = P0.Color;
	g0.TexCoord = float2( 1.5, 0.0);

	GeometryOutput g1;
	g1.HPosition = float4(P0.HPosition.xy - c * P0.HPosition.w, P0.HPosition.zw);
	g1.Color = P0.Color;
	g1.TexCoord = float2(-1.5, 0.0);

	GeometryOutput g2;
	g2.HPosition = float4(P1.HPosition.xy + c * P1.HPosition.w, P1.HPosition.zw);
	g2.Color = P1.Color;
	g2.TexCoord = float2( 1.5, 0.0);

	GeometryOutput g3;
	g3.HPosition = float4(P1.HPosition.xy - c * P1.HPosition.w, P1.HPosition.zw);
	g3.Color = P1.Color;
	g3.TexCoord = float2(-1.5, 0.0);

	OUT.Append(g0);
	OUT.Append(g1);
	OUT.Append(g2);
	OUT.Append(g3);
	OUT.RestartStrip();
}

float4 AAWireDrawPS(in GeometryOutput IN) : SV_Target
{
	float d = abs(IN.TexCoord.x);
	float a = exp2(-2.7 * d * d) * IN.Color.a;
	return float4(IN.Color.xyz, a);
}

technique AAWireDraw
{
	VertexShader = WireDrawVS();
	GeometryShader = AAWireDrawGS();
	PixelShader = AAWireDrawPS();
}

////////////////////////////////////////////////////////////////////////
// Common texture preview.

sampler_state TexturePreview_Sample : layout(s0, s1)
{
	Texture = Preview_Texture;
	Filter = Linear;
	Address = FromTexture;
	MaxMipLevel = 0;
};

sampler3D TexturePreview3D_Sample : layout(s0, s1)
{
	Texture = Preview_Texture3D;
	Filter = Linear;
	Address = FromTexture;
	MaxMipLevel = 0;
};

VertexOutput TexturePreviewDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4((vPos.xy * 2 - 1) * BoxScale.xy, vPos.z, 1.0);
	OUT.Color = GetVertexColor(IN.Color);
	OUT.TexCoord = IN.TexCoord;

	return OUT;
}

float4 TexturePreviewPS(in VertexOutput IN) : SV_Target
{
	float4 result = Preview_Texture.SampleLevel(TexturePreview_Sample, IN.TexCoord.xy, BoxScale.w);

	if (TextureMask.a < -1e-3f)
	{
		result.rgb = Dyeing(result.rgb, TextureMask.r);
	}
	else
	{
		if (TextureMask.a < 1e-3f)
		{
			if (abs(dot(TextureMask.rgb, TextureMask.rgb) - 1.0f) < 1e-3f)
				result.rgb = dot(result.rgb, TextureMask.rgb);
			else
				result.rgb *= TextureMask.rgb;
			result.a = 1.0f;
		}
		else
		{
			if (abs(dot(TextureMask.rgb, TextureMask.rgb) - 1.0f) < 1e-3f)
				result.rgb = dot(result.rgb, TextureMask.rgb);
			else if (dot(TextureMask.rgb, TextureMask.rgb) < 1e-3f)
				result.rgba = float4(result.a, result.a, result.a, 1);
			else
				result.rgb *= TextureMask.rgb;
		}
	}

	return result;
}

float4 TexturePreviewPS3D(in VertexOutput IN) : SV_Target
{
	// TODO: only supports slice0
	float4 result = Preview_Texture3D.SampleLevel(TexturePreview3D_Sample, float3(IN.TexCoord.xy, 0), BoxScale.w);

	if (TextureMask.a < -1e-3f)
	{
		result.rgb = Dyeing(result.rgb, TextureMask.r);
	}
	else
	{
		if (TextureMask.a < 1e-3f)
		{
			if (abs(dot(TextureMask.rgb, TextureMask.rgb) - 1.0f) < 1e-3f)
				result.rgb = dot(result.rgb, TextureMask.rgb);
			else
				result.rgb *= TextureMask.rgb;
			result.a = 1.0f;
		}
		else
		{
			if (abs(dot(TextureMask.rgb, TextureMask.rgb) - 1.0f) < 1e-3f)
				result.rgb = dot(result.rgb, TextureMask.rgb);
			else if (dot(TextureMask.rgb, TextureMask.rgb) < 1e-3f)
				result.rgba = float4(result.a, result.a, result.a, 1);
			else
				result.rgb *= TextureMask.rgb;
		}
	}

	return result;
}

technique TexturePreview
{
	VertexShader = TexturePreviewDrawVS();
	PixelShader = TexturePreviewPS();
}

technique TexturePreview3D
{
	VertexShader = TexturePreviewDrawVS();
	PixelShader = TexturePreviewPS3D();
}

////////////////////////////////////////////////////////////////////////
// Display HUD rectangle.

struct VertexScreenTextOutput
{
	float4		HPosition	: SV_Position;
	float4 		Color 		: COLOR;
	float2		TexCoord	: TEXCOORDN;
};

sampler_state HUD_TextSample : layout(s0)
{
	Texture = HUD_Texture;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

VertexScreenTextOutput HUD_TextVS(in VertexAssemble IN)
{
	VertexScreenTextOutput OUT;

	[@_HLSL]
	OUT = (VertexScreenTextOutput)0;

	// Scale & transform vertex.
	const float4 translate = float4(Location2D.xy, 0, 0);
	const float4 scale = float4(Location2D.zw, 1, 1);

	OUT.HPosition = IN.Position*scale + translate;
	OUT.Color = GetVertexColor(IN.Color);
	OUT.TexCoord = IN.TexCoord;

	if (@_VULKAN)
	{
		float2 t = OUT.HPosition.xy;
		OUT.HPosition.x = dot(t, HPositionTransform.xy);
		OUT.HPosition.y = dot(t, HPositionTransform.zw);
	}

	return OUT;
}

float4 HUD_TextPS(in VertexScreenTextOutput IN) : SV_Target
{
	float alpha = IN.Color.a * HUD_Texture.Sample(HUD_TextSample, IN.TexCoord).a;
	return float4(IN.Color.rgb * alpha, alpha);
}

technique HUD_Text
{
	VertexShader = HUD_TextVS();
	PixelShader = HUD_TextPS();
}

////////////////////////////////////////////////////////////////////////
// Basic (for drawing helpers).

struct BasicVertexAssemble
{
	float4		Position 	: POSITION;
	float2		TexCoord	: TEXCOORDN;

	[@_INSTANCED]
	float4		Transform0	: TEXCOORD3;
	[@_INSTANCED]
	float4		Transform1	: TEXCOORD4;
	[@_INSTANCED]
	float4		Transform2	: TEXCOORD5;
	[@_INSTANCED]
	float4		Color		: TEXCOORD6;
};

VertexOutput DrawBasicVS(in BasicVertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	float4 WorldPosition;
	if (@_INSTANCED)
	{
		WorldPosition = float4(Matrix4x3TransformP(IN.Transform0, IN.Transform1, IN.Transform2, IN.Position.xyz), 1.f);
		OUT.Color = IN.Color;
	}
	else
	{
			WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz, 1)).xyz, 1.f);
		OUT.Color = cColor;
	}

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.TexCoord = IN.TexCoord;
	return OUT;
}

float4 DrawBasicPS(in VertexOutput IN) : SV_Target
{
	float4 color = IN.Color;
	return float4(color.rgb, EncodeTransparentBloomFactor(color.a, 0));
}

technique DrawBasic
{
	VertexShader = DrawBasicVS();
	PixelShader = DrawBasicPS();
}

technique DrawBasicLine
{
	VertexShader = DrawBasicVS();
	GeometryShader = AAWireDrawGS();
	PixelShader = AAWireDrawPS();
}

////////////////////////////////////////////DrawReflectionProbe/////////////////////////////////////
struct ReflectionProbeVertexOutput
{
	float4		HPosition	: SV_Position;
	float3		MPosition	: TEXCOORDN;
};

ReflectionProbeVertexOutput DrawReflectionProbeVS(in BasicVertexAssemble IN)
{
	ReflectionProbeVertexOutput OUT;

	[@_HLSL]
	OUT = (ReflectionProbeVertexOutput)0;

	float4 WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz,1)).xyz, 1.f);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.MPosition = IN.Position.xyz;

	return OUT;
}

float4 DrawReflectionProbePS(in ReflectionProbeVertexOutput IN) : SV_Target
{
	half3 R = normalize(IN.MPosition);

	half fSign = R.z > 0;
	half fSign2 = fSign*2 - 1;
	R.xy /= R.z * fSign2 + 1;
	R.xy = R.xy * half2(0.25, -0.25) + (0.25+0.5*fSign);
	half4 srcColor = Preview_Texture.SampleLevel( TexturePreview_Sample, R.xy, 0 );
	half3 srcColorDecoded = RGBMDecode( srcColor, MAX_PBR_RGBM_VALUE ) * 4;
	
	return float4(srcColorDecoded, EncodeTransparentBloomFactor(1, 0));
}

technique DrawReflectionProbe
{
	VertexShader = DrawReflectionProbeVS();
	PixelShader = DrawReflectionProbePS();
}

////////////////////////////////////////////////////////////////////////
// Icon.

texture2D tIconMap : layout(t0) = "White";
sampler sIconSampler : layout(s0)
{
	Texture = tIconMap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

struct IconVertexAssemble
{
	float3		Position 	: POSITION;
	float2		TexCoord	: TEXCOORDN;
};

struct IconVertexOutput
{
	float4		HPosition	: SV_Position;
	float3		TexCoord	: TEXCOORDN;
};

IconVertexOutput DrawIconVS(in IconVertexAssemble IN)
{
	IconVertexOutput OUT;

	[@_HLSL]
	OUT = (IconVertexOutput)0;

	OUT.HPosition.xyz = IN.Position.xyz;
	OUT.HPosition.w = 1.0;
	OUT.TexCoord.xy = IN.TexCoord.xy;
	OUT.TexCoord.z = 1;
	if (OUT.HPosition.x > 2) // selected
	{
		OUT.HPosition.x -= 4;
		OUT.TexCoord.z = 2;
	}
	return OUT;
}

float4 DrawIconPS(in IconVertexOutput IN) : SV_Target
{
	float4 OUT = tIconMap.Sample(sIconSampler, IN.TexCoord.xy);
	[@_HDR]
	{
		OUT.rgb = DecodeGamma(OUT.rgb);
		EncodeHDR(OUT.rgb);
	}
	OUT.rgb *= OUT.a * IN.TexCoord.z;
	clip(OUT.a - 0.6);
	return OUT;
}

technique DrawIcon
{
	VertexShader = DrawIconVS();
	PixelShader = DrawIconPS();
}

////////////////////////////////////////////////////////////////////////
// Debug light map.

texture2D tDebugLightmap : layout(t0) = "Red";
sampler sDebugLightmap : layout(s0)
{
	Texture = tDebugLightmap;
	Filter = Linear;
	Address = FromTexture;
	SRGB = false;
};

struct a2v 
{
	float4 Position    : POSITION; //in object space
	float2 BaseTC      : TEXCOORD0;
};

struct v2f
{
	float4 HPosition   : SV_Position;
	float2 TC          : TEXCOORD0;
};

v2f ScreenVS( a2v IN )
{
	v2f OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1-vPos.y;
	OUT.HPosition = float4(vPos.xy*2-1, vPos.z, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float4 DebugLightmapPS(in v2f IN) : SV_Target
{
	float4 OUT = tDebugLightmap.Sample(sDebugLightmap, IN.TC.xy);
	return float4(OUT.rgb, 0);
}

technique DebugLightmap
{
	VertexShader = ScreenVS();
	PixelShader = DebugLightmapPS();
}

