//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<BoardElement.fx>
///	@path	~/Shaders/UI/
///	@date	2016/02/29
///	@brief	Board element.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj				;
	float4   UVTransform				 = { 1.0f, 0.0f, 0.0f, 1.0f };
};

@_DepthMapEnable
{
	Name = DepthMapEnable;
	Description = "Do we have DepthMap?";
	Visible = false;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D TextTexture					: layout(t2) = "Transparent";
sampler_state TextTexture_Sampler		: layout(s2)
{
	Texture = TextTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tDepthMap : SceneDepth : layout(t3);
sampler sDepthSampler : layout(s3)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color1					: COLOR;
	float4		TexCoord				: TEXCOORDN;
	float2		TexCoord2				: TEXCOORDN;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color1					: COLOR;
	float4		Color2					: TEXCOORDN;
	float2		TexCoord				: TEXCOORDN;
	[@_DepthMapEnable]
	float4		ScreenCoord				: TEXCOORDN;
};

////////////////////////////////////////////////////////////////////////
// Board element draw.

VertexOutput BoardElementDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	OUT.Color1 = GetVertexColor(IN.Color1);
	OUT.Color2 = IN.TexCoord;
	OUT.TexCoord.xy = IN.TexCoord2.xy;

	if (@_DepthMapEnable)
	{
		float2 screenUV = OUT.HPosition.xy / (2.0 * OUT.HPosition.w) + 0.5;
		[!@_GLSL]
		screenUV.y = 1.0 - screenUV.y;

		if (@_VULKAN)
		{
			float2 t = (screenUV - 0.5) * 2.0;
			screenUV.x = dot(t, UVTransform.xy);
			screenUV.y = dot(t, UVTransform.zw);
			screenUV = screenUV * 0.5 + 0.5;
		}

		float near = CameraInfo.x;
		float far = CameraInfo.y;
		float tmp = IN.Position.z + near + far;
		float tmp2 = near * far;

		OUT.ScreenCoord.xy = screenUV;
		OUT.ScreenCoord.z = (tmp2 / tmp - BOARD_DEPTH_BIAS) / far;
		OUT.ScreenCoord.w = near / tmp;
	}

	return OUT;
}

float4 BoardElementDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse;
	float4 c1 = IN.Color1;
	float4 c2 = IN.Color2;
	if (c2.a < 2)
	{
		float2 tex = frac(IN.TexCoord.xy);
		tex = float2(c2.r + tex.x * c2.g, c2.b + tex.y * c2.a);
		if (c2.a < 0)
			tex.xy = tex.yx;

		diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, tex);

		diffuse.rgb = saturate(diffuse.rgb + c1.rgb);
		diffuse.a *= c1.a;
	}
	else
	{
		diffuse = SampleColorMap(TextTexture, TextTexture_Sampler, IN.TexCoord.xy);
		if (@_D3D11 || @_D3D12 || @_VULKAN || @_METAL)
			diffuse.rgba = diffuse.rrrg;
		else
			diffuse.rgba = diffuse.rrra;
		if (length(c1.rgb - c2.rgb) > 0.01)
		{
			float blend = diffuse.a / max(diffuse.r, 0.004);
			float alpha = diffuse.r * c1.a * (1 * blend + (c2.a - 2.0f) * ((1.0f - blend) * (1.f - diffuse.a)));
			diffuse.rgb = c1.rgb * blend + c2.rgb * ((1.0f - blend) * (1.f - diffuse.a));
			diffuse.a = alpha;
		}
		else
		{
			diffuse.a = diffuse.r * c1.a;
			diffuse.rgb = c1.rgb;
		}
	}

	// Depth test
	if (@_DepthMapEnable)
	{
		float sDepth = DecodeDepth(tDepthMap.Sample(sDepthSampler, IN.ScreenCoord.xy));
		diffuse.a *= smoothstep(IN.ScreenCoord.z, IN.ScreenCoord.w, sDepth);
	}

	return diffuse;
}

technique Shading
{
	VertexShader = BoardElementDrawVS();
	PixelShader = BoardElementDrawPS();
}
