//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIBaseImage.fx>
///	@path	~/Shader/UI/
///	@date	2015/01/16
///	@brief	Translucent base UI.

#pragma shader draw
#pragma shader category("UI")

#define CUSTOM_TRAJECTORY			0

#include "Constants.fxh"
#include "Common.fxh"

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4   UVTransform				 = { 1.0f, 0.0f, 0.0f, 1.0f };
	float EffectParam					 = 0.0;
};

@_DiffuseTextureEnable
{
	Name = DiffuseTextureEnable;
	Description = "Is diffuse map enabled?";
	Default = false;
};

@_MaskTextureEnable
{
	Name = MaskTextureEnable;
	Description = "Is masking map enabled?";
	Default = false;
};

@_TrajectoryEnable
{
	Name = TrajectoryEnable;
	Description = "Is trajectory map enabled?";
	Default = false;
};

@_DyeingEnable
{
	Name = DyeingEnable;
	Description = "Is the DiffuseTexture dyeing?";
	Default = false;
};

@_GrayingEnable
{
	Name = GrayingEnable;
	Description = "Is the DiffuseTexture graying?";
	Default = false;
};

@_FullSize
{
	Name = FullSize;
	Description = "Is the DiffuseTexture fullsize?";
	Default = false;
};

@_DepthMapEnable
{
	Name = DepthMapEnable;
	Description = "Do we have DepthMap?";
	Visible = false;
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D DiffuseTexturePoint				: layout(t2) = "Transparent";
sampler_state DiffuseTexturePoint_Sampler	: layout(s2, s3)
{
	Texture = DiffuseTexturePoint;
	Filter = Point;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskTexture					: layout(t4) = "White";
sampler_state MaskTexture_Sampler		: layout(s4, s5)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D tDepthMap 	: SceneDepth 	: layout(t6);
sampler sDepthSampler 					: layout(s6)
{
	Texture = tDepthMap;
	Filter = Point;
	Address = Clamp;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORD0;
	[@_DepthMapEnable]
	float4		ScreenCoord				: TEXCOORD1;
};

////////////////////////////////////////////////////////////////////////
// Translucent base UI draw.

VertexOutput UIDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	
	[@_DiffuseTextureEnable]
		OUT.TexCoord.xy = IN.TexCoord;

	if (@_MaskTextureEnable)
	{
		OUT.TexCoord.zw = IN.TexCoord2;
		if (@_TrajectoryEnable && @_GLSL)
			OUT.TexCoord.w = 1 - OUT.TexCoord.w;
	}

	if (@_DepthMapEnable)
	{
		float2 screenUV = OUT.HPosition.xy / (2.0 * OUT.HPosition.w) + 0.5;
		[!@_GLSL]
		screenUV.y = 1.0 - screenUV.y;

		if (@_VULKAN)
		{
			float2 t = (screenUV - 0.5) * 2.0;
			screenUV.x = dot(t, UVTransform.xy);
			screenUV.y = dot(t, UVTransform.zw);
			screenUV = screenUV * 0.5 + 0.5;
		}

		float near = CameraInfo.x;
		float far = CameraInfo.y;
		float tmp = IN.Position.z + near + far;
		float tmp2 = near * far;

		OUT.ScreenCoord.xy = screenUV;
		OUT.ScreenCoord.z = (tmp2 / tmp - BOARD_DEPTH_BIAS) / far;
		OUT.ScreenCoord.w = near / tmp;
	}

	OUT.Color = GetVertexColor(IN.Color);

	return OUT;
}

float4 UIDrawPS(in VertexOutput IN) : SV_Target
{
	if (@_DiffuseTextureEnable)
	{
		float4 diffuse;
		if (@_FullSize)
		{
			diffuse = SampleColorMap(DiffuseTexturePoint, DiffuseTexturePoint_Sampler, IN.TexCoord.xy);
		}
		else
		{
			diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
		}

		if (@_DyeingEnable)
		{
			diffuse.rgb = Dyeing(diffuse.rgb, EffectParam);
		}
		else if (@_GrayingEnable)
		{
			diffuse.rgb = Graying(diffuse.rgb, EffectParam);
		}
		else
		{
			diffuse.rgb *= IN.Color.rgb;
		}

		if (@_MaskTextureEnable)
		{
			if (@_TrajectoryEnable)
			{
#if CUSTOM_TRAJECTORY
				diffuse = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw);
#else
				float4 mk = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw);
				diffuse.rgb = mk.rgb * mk.a + ((1 - mk.a) * diffuse.a) * diffuse.rgb;
				diffuse.a = max(mk.a, diffuse.a);
#endif
			}
			else
			{
				diffuse.a *= SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw).a;
			}
		}

		diffuse.a *= IN.Color.a;

		if (@_DepthMapEnable)
		{
			float sDepth = DecodeDepth(tDepthMap.Sample(sDepthSampler, IN.ScreenCoord.xy));
			diffuse *= smoothstep(IN.ScreenCoord.z, IN.ScreenCoord.w, sDepth);
		}

		return diffuse;
	}
	else
	{
		float4 diffuse = IN.Color;
		if (@_MaskTextureEnable)
		{
			if (@_TrajectoryEnable)
			{
#if CUSTOM_TRAJECTORY
				diffuse = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw);
#else
				float4 mk = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw);
				diffuse.rgb = mk.rgb * mk.a + ((1 - mk.a) * diffuse.a) * diffuse.rgb;
				diffuse.a = max(mk.a, diffuse.a);
#endif
			}
			else
			{
				diffuse.a *= SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw).a;
			}
		}

		if (@_DepthMapEnable)
		{
			float sDepth = DecodeDepth(tDepthMap.Sample(sDepthSampler, IN.ScreenCoord.xy));
			diffuse *= smoothstep(IN.ScreenCoord.z, IN.ScreenCoord.w, sDepth);
		}

		return diffuse;
	}
}

technique Shading
{
	VertexShader = UIDrawVS();
	PixelShader = UIDrawPS();
}
