//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : gzchenjiating@corp.netease.com
//	Author: Getting
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<UIRenderTexture.fx>
///	@path	~/Shader/UI/
///	@date	2018/06/22
///	@brief	Translucent base UI.

#pragma shader draw
#pragma shader category("UI")

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

@_MaskTextureEnable
{
	Name = MaskTextureEnable;
	Description = "Is masking map enabled?";
	Default = false;
};

@_GrayingEnable
{
	Name = GrayingEnable;
	Description = "Is the result graying?";
	Default = false;
};

@_UI3D
{
	Name = UI3D;
	Description = "Is the result 3D?";
	Default = false;
};

@_DepthMapEnable
{
	Name = DepthMapEnable;
	Description = "Is the result depth enable?";
	Default = false;
}

@_Blur
{
	Name = Blur;
	Description = "Is the result blur?";
	Default = false;
}

@_FogEnable
{
	Name = FogEnable;
	Description = "Is the result fog enable?";
	Default = false;
}

@_BGBlur
{
	Name = BGBlur;
	Description = "Is the background blur?";
	Default = false;
}

cbuffer Batch
{
	float4x4 WorldViewProj		 		;
	float4 BlurParam					 = { 1.0f, 1.0f, 1.0f, 1.0f };
};

texture2D DiffuseTexture				: layout(t0) = "Transparent";
sampler_state DiffuseTexture_Sampler	: layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D MaskTexture					: layout(t1) = "White";
sampler_state MaskTexture_Sampler		: layout(s1, s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

texture2D BackgroundTexture				: layout(t3);
sampler_state BackgroundTexture_Sampler	: layout(s3)
{
	Texture = BackgroundTexture;
	Filter = Linear;
	Address = Clamp;
	MaxMipLevel = 0;
};

struct VertexAssemble
{
	float4		Position 				: POSITION;
	float4		Color 					: COLOR;
	float2		TexCoord				: TEXCOORD0;
	float2		TexCoord2				: TEXCOORD1;
};

struct VertexOutput
{
	float4		HPosition				: SV_Position;
	float4		Color					: COLOR;
	float4		TexCoord				: TEXCOORDN;
	float4		FogParam				: TEXCOORDN;
};

////////////////////////////////////////////////////////////////////////
// Translucent base UI draw.

VertexOutput UIRTDrawVS(in VertexAssemble IN)
{
	VertexOutput OUT;

	[@_HLSL]
	OUT = (VertexOutput)0;

	if (@_UI3D && @_DepthMapEnable)
	{
		OUT.HPosition = MulMat4x4(ViewProj, float4(IN.Position.xyz, 1.0));
	}
	else
	{
		OUT.HPosition = MulMat4x4(WorldViewProj, IN.Position);
	}

	OUT.TexCoord.xy = IN.TexCoord;

	[@_GLSL]
	OUT.TexCoord.y = 1 - OUT.TexCoord.y;

	[@_MaskTextureEnable]
	OUT.TexCoord.zw = IN.TexCoord2;

	OUT.Color = GetVertexColor(IN.Color);

	if (@_UI3D && @_FogEnable)
	{
		OUT.FogParam.xyz = IN.Position.xyz - CameraPos.xyz;
		OUT.FogParam.w = FogTransmittance(normalize(OUT.FogParam.xyz) * BlurParam.z);
	}

	return OUT;
}

float4 UIRTDrawPS(in VertexOutput IN) : SV_Target
{
	float4 diffuse;
	if (@_UI3D)
	{
		diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
	}
	else
	{
		if (@_Blur)
		{
			diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy) * 0.1;

			diffuse += (SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(1.407333, 0) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(-1.407333, 0) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(0, 1.407333) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(0, -1.407333) * BlurParam.xy)) * 0.08;

			diffuse += (SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(1.732051, 1.732051) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(-1.732051, 1.732051) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(1.732051, -1.732051) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(-1.732051, -1.732051) * BlurParam.xy)) * 0.065;

			diffuse += (SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(3.294215, 0) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(-3.294215, 0) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(0, 3.294215) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(0, -3.294215) * BlurParam.xy)) * 0.05;

			diffuse += (SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(3.605551, 3.605551) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(-3.605551, 3.605551) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(3.605551, -3.605551) * BlurParam.xy) +
						SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy + float2(-3.605551, -3.605551) * BlurParam.xy)) * 0.03;
		}
		else
		{
			diffuse = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
		}
	}
	
	float alpha = IN.Color.a;

	if (@_MaskTextureEnable)
	{
		alpha *= SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.zw).a;
	}

	diffuse *= alpha;

	if (@_GrayingEnable)
	{
		diffuse.rgb = Graying(diffuse.rgb, IN.Color.r);
	}
	else
	{
		diffuse.rgb *= IN.Color.rgb;
	}

	if (@_UI3D && @_FogEnable)
	{
		diffuse.rgb = FogLighting(IN.FogParam.w, diffuse.rgb);
	}

	if (@_BGBlur && @_UI3D)
	{
		float2 spos = IN.HPosition.xy;

		float2 stc = spos * ScreenInfo.zw;

		float3 background = SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc).rgb * 0.1;

		background += (SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(5.629332, 0) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(-5.629332, 0) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(0, 5.629332) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(0, -5.629332) * ScreenInfo.zw).rgb) * 0.08;

		background += (SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(6.928204, 6.928204) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(-6.928204, 6.928204) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(6.928204, -6.928204) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(-6.928204, -6.928204) * ScreenInfo.zw).rgb) * 0.065;

		background += (SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(13.17686, 0) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(-13.17686, 0) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(0, 13.17686) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(0, -13.17686) * ScreenInfo.zw).rgb) * 0.05;

		background += (SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(14.422204, 14.422204) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(-14.422204, 14.422204) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(14.422204, -14.422204) * ScreenInfo.zw).rgb +
					SampleColorMap(BackgroundTexture, BackgroundTexture_Sampler, stc + float2(-14.422204, -14.422204) * ScreenInfo.zw).rgb) * 0.03;

		diffuse.rgb = diffuse.rgb + background * (1 - diffuse.a);
		diffuse.a = 1;
	}

	return diffuse;
}

technique Shading
{
	VertexShader = UIRTDrawVS();
	PixelShader = UIRTDrawPS();
}
