//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is CellAlpha part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : lnn4047@corp.netease.com
//	Author: liuning
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<VXGIRender.fx>
///	@path	~/System/Shader/
///	@date	2020/09/18
///	@brief	VXGI render pass.

// {HEX} VXGI.
#pragma shader draw

#include "Config.fxh"

#if SD_ENABLE_VXGI

#include "Constants.fxh"
#include "Common.fxh"
#include "MonteCarlo.fxh"
#include "DeferredShadingCommon.fxh"
#include "VXGI/VXGIConst.fxh"
#include "VXGI/VXGICommon.fxh"

@_UseScreenSpace
{ 
	Name = UseScreenSpace;
	Description = "Ray marching in screen space?";
	Default = true;
};

@_UseVoxelSpace
{ 
	Name = UseVoxelSpace;
	Description = "Ray marching in voxel space?";
	Default = true;
};

texture2D tDepthBufferLinear : HierarchicalZ : layout(t0);
sampler sDepthBufferSamplerLinear : layout(s0)
{
	Texture = tDepthBufferLinear;
	Filter = Point;
	Address = Clamp;
};
texture2D tGBufferE : layout(t1);
sampler sGBufferE : layout(s1)
{
	Texture = tGBufferE;
	Filter = Point;
	Address = Clamp;
};
texture2D tSceneColorMap : layout(t2);
sampler sSceneColorSampler : layout(s2)
{
	Texture = tSceneColorMap;
	Filter = Point;
	Address = Clamp;
	Precision = Medium;
};
texture2D tSceneDepthBuffer : SceneDepth : layout(t3);
sampler sSceneDepthBufferSampler : layout(s3)
{
	Texture = tSceneDepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture3D<uint1> tVoxelColor : VoxelColor;

cbuffer Batch
{
	float4 vWBasisX				: WBasisX;
	float4 vWBasisY				: WBasisY;
	float4 vWBasisZ				: WBasisZ;
	float4 cRTSize				;
	float4 cRTAndHZUVScale		= {1, 1, 1, 1};
	float2 cClipToView			= {0, 0}; //tan(0.5*fovy)*w/h, tan(0.5*fovy)
	float2 cViewToClip 			= {0, 0}; //cot(0.5*fovy)*h/w, cot(0.5*fovy)
	float4x4 cProjTex			;
	float4x3 cWorldToView		: View;
	float4x4 cInvView			: InvView;
};

struct VSInput 
{
	float4 Position		: POSITION;
	float2 BaseTC		: TEXCOORD0;
};

struct VSOutput
{
	float4 HPosition	: SV_Position;
	float2 TC			: TEXCOORD0;
};

VSOutput ScreenVS(VSInput IN)
{
	VSOutput OUT;
	float3 vPos = IN.Position.xyz;
	vPos.y = 1 - vPos.y;
	OUT.HPosition = float4(vPos.xy * 2 - 1, EPSILON, 1.0);
	OUT.TC = IN.BaseTC;
	return OUT;
}

float3 CalcHomogeneousPos(float Depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = vWBasisZ.xyz + (vWBasisX.xyz * screenTC.x) + (vWBasisY.xyz * screenTC.y);
	return pos * Depth;
}

bool GetInsideLayer(float3 WorldPosition, float3 BoundsMin, float3 BoundsMax)
{
	return all(WorldPosition >= BoundsMin) && all(WorldPosition <= BoundsMax);
}

bool IsCellHit(uint Value)
{
	return Value > 0;
}

float4 ClipmapRayMarching(float3 RayDir, float3 WorldPosition, float3 WorldNormal, float DirOffset)
{
	float4 Result = float4(0, 0, 0, 0);

	// Parepare clipmap info.
	float3 BoundsMins[VXGIClipmapCascades];
	float3 BoundsMaxs[VXGIClipmapCascades];
	for (int i = 0; i < VXGIClipmapCascades; ++i)
	{
		VXGIGetGridBounds(VXGIColorCellSizes[i], VXGIColorDimension, VXGICurCenter, BoundsMins[i], BoundsMaxs[i]);
	}
	uint MarchValue = 0;
	bool HitFound = false;

	float3 RayStartPos = WorldPosition;
	float MarchDistance = DirOffset;

	// Slope bias.
	float MaxProj = max(abs(WorldNormal.x), max(abs(WorldNormal.y), abs(WorldNormal.z)));
	const float BaseProj = 1 / sqrt(3);
	float NormalOffset = saturate((MaxProj - BaseProj) / (1 - BaseProj));
	float3 MarchPos = RayStartPos + WorldNormal * NormalOffset * VXGIColorCellSizes[1] + RayDir * DirOffset;

	for (int Layer = 0; !HitFound && Layer < VXGIClipmapCascades; ++Layer)
	{
		float CellSize = VXGIColorCellSizes[Layer];

		for(int StepIndex = 0; StepIndex < VXGIRayMarchSteps; ++StepIndex)
		{
			int3 CellIndex = VXGIGetLocalCellIndex(MarchPos, BoundsMins[Layer], CellSize);
			int3 StartCellIndex = VXGIGetLocalCellIndex(WorldPosition, BoundsMins[Layer], CellSize);
			if (!GetInsideLayer(MarchPos, BoundsMins[Layer], BoundsMaxs[Layer]))
			{
				break;
			}
			// Skip self occlusion.
			if (all(CellIndex == StartCellIndex))
			{
				MarchDistance += CellSize;
				MarchPos = RayStartPos + RayDir * MarchDistance;
				continue;
			}

			// Negate ray direction to find hit position information.
			int Slot = SelectSlot(-RayDir);
			int3 MarchDataIndex = VXGIGetClipmapCoord(MarchPos, CellSize, VXGIColorDimension, Layer, Slot);
			MarchValue = uint(tVoxelColor.Load(MarchDataIndex).r);
			if (IsCellHit(MarchValue))
			{
				HitFound = true;
				Result.a = 1.f - saturate(MarchDistance / VXGIAORange);
				break;
			}
			MarchDistance += CellSize;
			MarchPos = RayStartPos + RayDir * MarchDistance;
		}
	}

	// if (HitFound)
	// {
	// 	float3 D = -RayDir;
	// 	int3 Slots = D > 0;
	// 	float3 Weights = D * D;

	// 	int3 ColorIndex = VXGIGetClipmapCoord(MarchPos, VXGIColorCellSizes[Layer], VXGIColorDimension, Layer, 0);
	// 	float4 Rx = UintToFloat4(uint(tVoxelColor.Load(ColorIndex + int3(0 + Slots.x, 0, 0)).r));
	// 	float4 Ry = UintToFloat4(uint(tVoxelColor.Load(ColorIndex + int3(2 + Slots.y, 0, 0)).r));
	// 	float4 Rz = UintToFloat4(uint(tVoxelColor.Load(ColorIndex + int3(4 + Slots.z, 0, 0)).r));
	// 	float3 WeightedColor = (Rx.rgb * Weights.x + Ry.rgb * Weights.y + Rz.rgb * Weights.z);
	// 	Result.rgb = WeightedColor * VXGIIntensity;
	// }
	// else
	// {
	// 	// Sky color.
	// 	if (RayDir.y > 0)
	// 	{
	// 		float3 SkyColor = lerp(VXGISkyHorizonColor, VXGISkyZenithColor, saturate(dot(float3(0, 1, 0), RayDir)));
	// 		Result.rgb = SkyColor * VXGISkyColorIntensity;
	// 	}
	// }
	return Result;
}

half4 HiZRayMarching(float3 rayStartScreen, float3 rayStepScreen, float compareTolerance, float StepOffset)
{
	float3 rayStartUVz = float3(rayStartScreen.xy * cRTAndHZUVScale.xy, rayStartScreen.z);
	float3 rayStepUVz = float3(rayStepScreen.xy * cRTAndHZUVScale.xy, rayStepScreen.z);
	static const half stepNum = 8;  // 默认八步吧
	static const half batchSize = 4;
	const float stepRatio = 1.0 / stepNum;
	float stepCompareTolerance = compareTolerance * stepRatio;
	float level = 1;
	rayStepUVz *= stepRatio;
	float3 rayUVz = rayStartUVz + rayStepUVz * StepOffset;
	float4 multipleSampleDepthDiff;
	bool4 bMutipleSampleHit;
	bool findHit = false;
	int i;
	[unroll]
	for(i = 0; i < stepNum; i = i + 4)
	{
		float2 samplesUV[4];
		float4 samplesZ;
		float4 samplesMip;
		[unroll]
		for(int j = 0; j < batchSize; ++j)
		{
			samplesUV[j] = rayUVz.xy + (i + j + 1) * rayStepUVz.xy;
			samplesZ[j] = rayUVz.z + (i + j + 1) * rayStepUVz.z;
		}
		samplesMip.xy = level;
		level += 1;
		samplesMip.zw = level;
		level += 1;

		float4 sampleDepth;
		[unroll]
		for(int k = 0; k < batchSize; ++k)
		{
			sampleDepth[k] = tDepthBufferLinear.SampleLevel(sDepthBufferSamplerLinear, samplesUV[k], samplesMip[k]).r;
		}
		multipleSampleDepthDiff = samplesZ - sampleDepth;
		bMutipleSampleHit = abs(multipleSampleDepthDiff + stepCompareTolerance) < stepCompareTolerance;
		findHit = any(bMutipleSampleHit);
		if(findHit)
		{
			break;
		}
	}
	if (findHit)
	{
		float4 hitTime;
		if(@_D3D11)
		{
			hitTime = bMutipleSampleHit.xyzw ? float4(1, 2, 3, 4) : 5; // metal may error...
		}else
		{
			hitTime.x = bMutipleSampleHit.x ? 1 : 5;
			hitTime.y = bMutipleSampleHit.y ? 2 : 5;
			hitTime.z = bMutipleSampleHit.z ? 3 : 5;
			hitTime.w = bMutipleSampleHit.w ? 4 : 5;
		}
		float time1 = i + min(min(min(hitTime.x, hitTime.y), hitTime.z), hitTime.w);
		float3 hitUVz = rayUVz + rayStepUVz * time1;
		hitUVz.xy *= cRTAndHZUVScale.zw;
		half3 sampleColor = tSceneColorMap.Sample(sSceneColorSampler, hitUVz.xy, 0).rgb;
		return half4(sampleColor, saturate(1 - time1 / stepNum));
	}
	return half4(0, 0, 0, 0);
}

float GetStepScreenFactorToClipAtScreenEdge(float2 startUV, float2 stepUV)
{
	float2 rayStartScreen = 2 * startUV - 1;
	float2 rayStepScreen = stepUV * 2;
	const float rayStepScreenInvFactor = 0.5 * length(rayStepScreen);
	const float2 S = 1 - max(abs(rayStepScreen + rayStartScreen * rayStepScreenInvFactor) - rayStepScreenInvFactor, 0.0f) / abs(rayStepScreen);
	const float RayStepFactor = min(S.x, S.y) / rayStepScreenInvFactor;

	return RayStepFactor;
}

float4 VXGICalc(VSOutput IN) : SV_Target
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;
	uint FrameIndex = VXGIFrameIndex;
	
	half2 tc = spos * cRTSize.zw;
	float depth = DecodeDepth(tSceneDepthBuffer.SampleLevel(sSceneDepthBufferSampler, tc, 0));
	if(depth > 1 - EPSILON)
	{
		return float4(0, 0, 0, 0.9960784); //254/255
	}
	float3 WorldPosition = CameraPos.xyz + CalcHomogeneousPos(depth, tc);
	float4 normalAndShadow = tGBufferE.Sample(sGBufferE, tc);
	int filterFlag;
	DecodeFilter(normalAndShadow.a, filterFlag);

	float3 WorldNormal = DecodeNormal(normalAndShadow.xyz);
	half3 vNormal = mul(WorldNormal, (float3x3)cWorldToView);

	float VXAODirOffset = InterleavedGradientNoise(spos, FrameIndex);
	float StepOffset = VXAODirOffset - 0.9;

	// 初始化射线数据
	float startDeviceZ = GetDeviceZ(depth);
	float3 rayStartScreen = float3(tc, startDeviceZ); // 注意uv已经是变换到0-1

	half4 ColorSum = half4(0, 0, 0, 0);
	int ValidRayCount = 0;
	for(int i = 0; i < VXGIGatherRayCount; ++i)
	{
		uint2 randomSeed = ComputeRandomSeed(uint2(spos), FrameIndex);
		float2 hammersleyE = Hammersley16(i, VXGIGatherRayCount, randomSeed);
		half3 ray = ComputeL(vNormal, hammersleyE); // view的ray
		float3 rayWS = normalize(mul(ray, (float3x3)cInvView));
		// Slope bias.
		float MaxProj = max(abs(WorldNormal.x), max(abs(WorldNormal.y), abs(WorldNormal.z)));
		float MaxAngle = asin(MaxProj);
		if (dot(rayWS, WorldNormal) < cos(MaxAngle))
		{
			continue;
		}
		++ValidRayCount;

		float4 rayEndClip = MulMat4x4(cProjTex, float4(ray, 0))+ float4(rayStartScreen, 1);
		float3 rayEndScreen = rayEndClip.xyz / rayEndClip.w;
		float3 rayDepthScreen = 0.5 * (rayStartScreen.xyz + mul(float4(0, 0, 1, 0), cProjTex).xyz); // notice
		float3 rayStepScreen = rayEndScreen - rayStartScreen;

		rayStepScreen *= GetStepScreenFactorToClipAtScreenEdge(rayStartScreen.xy, rayStepScreen.xy);
		
		float compareTolerance = max(abs(rayStepScreen.z), (rayStartScreen.z - rayDepthScreen.z)*2);

		float4 SSAO = float4(0, 0, 0, 0);
		if (@_UseScreenSpace)
		{
			SSAO = HiZRayMarching(rayStartScreen, rayStepScreen, compareTolerance, StepOffset);
		}
		float4 VXAO = float4(0, 0, 0, 0);
		if (@_UseVoxelSpace)
		{
			VXAO = ClipmapRayMarching(rayWS, WorldPosition, WorldNormal, VXAODirOffset);
		}
		float4 CombinedResult = max(SSAO, VXAO);
		ColorSum += CombinedResult;
		++FrameIndex;
	}
	ValidRayCount = max(1, ValidRayCount);
	ColorSum  = ColorSum / ValidRayCount;
	float AO = saturate(1.f - ColorSum.a * VXGIAOIntensity);
	ColorSum.a = EncodeShadowAndFilter(AO, filterFlag);
	return ColorSum;
}

technique VXGIRender
{
	VertexShader = ScreenVS();
	PixelShader = VXGICalc();
}

#else

float4 UnusedVertexShading(float4 position: POSITION) : SV_Position
{
	return float4(0, 0, 0, 1);
}

float4 UnusedPixelShading(float4 position: SV_Position) : SV_Target
{
	return float4(0, 0, 0, 1);
}

technique Unused
{
	VertexShader = UnusedVertexShading();
	PixelShader = UnusedPixelShading();
}

#endif
// {HEX} end
