//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is CellAlpha part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : lnn4047@corp.netease.com
//	Author: liuning
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<VXGIVisualize.fx>
///	@path	~/System/Shader/
///	@date	2020/09/18
///	@brief	VXGI visual debug.

// {HEX} VXGI.
#pragma shader draw

#include "Config.fxh"

#if SD_ENABLE_VXGI

#include "Common.fxh"
#include "Constants.fxh"
#include "VXGI/VXGIConst.fxh"
#include "VXGI/VXGICommon.fxh"

texture3D<uint1> tVoxelColor : VoxelColor;

struct VisualDebugVSOutput
{
	float4	HPosition		: SV_Position;
	float4	WorldPosition	: POSITION;
	int3	CellIndex		: TEXCOORDN;
	float3	CellCenterWS	: TEXCOORDN;
};

cbuffer Shader
{
	int VisualizeLayer;
};

float GetCellSize(int Layer)
{
	return VXGIColorCellSizes[Layer];
}

float GetEdgeThickness(int Layer)
{
	float3 kThickness = float3(0.003, 0.006, 0.01);
	return kThickness[Layer];
}

int3 GetDimension()
{
	return VXGIColorDimension;
}

VisualDebugVSOutput VisualDebugVS(uint VertexIndex : SV_VertexID)
{
	VisualDebugVSOutput OUT;
	[@_HLSL]
	OUT = (VisualDebugVSOutput)0;

	// Config.
	float CellSize = GetCellSize(VisualizeLayer);
	int3 Dimension = GetDimension();

	uint InstanceID = VertexIndex / 36;
	uint VertexID = VertexIndex % 36;
	uint FaceID = VertexID / 6;
	VertexID = VertexID % 6;

	int3 CellIndex = int3(
		InstanceID % int(Dimension.x), 
		(InstanceID % int(Dimension.x * Dimension.y)) / Dimension.x, 
		InstanceID / int(Dimension.x * Dimension.y)
	);
	OUT.CellIndex = CellIndex;

	int3 GridMin, GridMax;
	VXGIGetGridRange(CellSize, Dimension, VXGICurCenter, GridMin, GridMax);
	int3 GlobalIndex = GridMin + CellIndex;
	OUT.CellCenterWS = (GlobalIndex + float3(0.5, 0.5, 0.5)) * CellSize;

	const static float3 kVertices[4] = {
		float3(-1, -1,  1),
		float3( 1, -1,  1),
		float3( 1,  1,  1),
		float3(-1,  1,  1)
	};

	const static float3x3 kRotations[6] = {
		float3x3(float3( 1,  0,  0), float3( 0,  1,  0), float3( 0,  0,  1)),	// front
		float3x3(float3( 1,  0,  0), float3( 0,  0,  1), float3( 0, -1,  0)),	// top
		float3x3(float3( 1,  0,  0), float3( 0, -1,  0), float3( 0,  0, -1)),	// back
		float3x3(float3( 1,  0,  0), float3( 0,  0, -1), float3( 0,  1,  0)),	// bottom
		float3x3(float3( 0,  0, -1), float3( 0,  1,  0), float3( 1,  0,  0)),	// left
		float3x3(float3( 0,  0,  1), float3( 0,  1,  0), float3(-1,  0,  0))	// right
	};

	const static int kIndices[6] = { 0, 2, 1, 0, 3, 2 };
	int Index = kIndices[VertexID];

	float3 VertexPosition = mul(kVertices[Index], kRotations[FaceID]) * CellSize * 0.5;
	OUT.WorldPosition = float4(VertexPosition + OUT.CellCenterWS, 1);
	OUT.HPosition = MulMat4x4(ViewProj, OUT.WorldPosition);

	return OUT;
}

float4 VisualDebugPS(in VisualDebugVSOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 1);

	// Config.
	float CellSize = GetCellSize(VisualizeLayer);
	int3 Dimension = GetDimension();

	float3 CellCenterWS = IN.CellCenterWS;
	int3 LayerCoord = VXGIGetClipmapCoord(CellCenterWS, CellSize, Dimension, VisualizeLayer, 0);

	float CellAlpha = 0;
	CellAlpha = UintToFloat4(tVoxelColor.Load(LayerCoord)).a;
	CellAlpha = max(CellAlpha, UintToFloat4(tVoxelColor.Load(LayerCoord + int3(1, 0, 0))).a);
	CellAlpha = max(CellAlpha, UintToFloat4(tVoxelColor.Load(LayerCoord + int3(2, 0, 0))).a);
	CellAlpha = max(CellAlpha, UintToFloat4(tVoxelColor.Load(LayerCoord + int3(3, 0, 0))).a);
	CellAlpha = max(CellAlpha, UintToFloat4(tVoxelColor.Load(LayerCoord + int3(4, 0, 0))).a);
	CellAlpha = max(CellAlpha, UintToFloat4(tVoxelColor.Load(LayerCoord + int3(5, 0, 0))).a);
	clip(CellAlpha - 0.01);

	// Draw cube edge lines.
	{
		float3 EdgeDistance = CellSize * 0.5 - abs(IN.WorldPosition.xyz - CellCenterWS);
		float Threshold = GetEdgeThickness(VisualizeLayer);
		int3 EdgeFlags = EdgeDistance < Threshold;
		// Edges are where two or more planes intersect.
		if (dot(EdgeFlags, int3(1, 1, 1)) >= 2)
		{
			return float4(0, 0, 0.5, 1.0);
		}
	}

	// Draw cell opacity or color.
	{
		float SphereRadius = CellSize * 0.4;
		float3 ViewDirectionNorm = normalize(IN.WorldPosition.xyz - CameraPos.xyz);
		float3 EyeToCell = CellCenterWS.xyz - CameraPos.xyz;
		float ProjectionLength = dot(ViewDirectionNorm, EyeToCell);
		float3 NearestPoint = CameraPos.xyz + ProjectionLength * ViewDirectionNorm;
		float NearestDistance = length(NearestPoint - CellCenterWS);
		if (NearestDistance <= SphereRadius)
		{
			float3 EyeBallPos = NearestPoint - ViewDirectionNorm * sqrt(SphereRadius * SphereRadius - NearestDistance * NearestDistance);
			int Slot = SelectSlot(normalize(EyeBallPos - CellCenterWS));
			float4 CellColor = UintToFloat4(tVoxelColor.Load(LayerCoord + int3(Slot, 0, 0)));
			CellColor = EncodeGamma(CellColor);
			OUT.rgb = CellColor.rgb;
		}
		else
		{
			clip(-1);
		}
	}
	return OUT;
}

technique VisualDebug
{
	VertexShader = VisualDebugVS();
	PixelShader = VisualDebugPS();
}

#else

float4 UnusedVertexShading(float4 position: POSITION) : SV_Position
{
	return float4(0, 0, 0, 1);
}

float4 UnusedPixelShading(float4 position: SV_Position) : SV_Target
{
	return float4(0, 0, 0, 1);
}

technique Unused
{
	VertexShader = UnusedVertexShading();
	PixelShader = UnusedPixelShading();
}

#endif
// {HEX}
