#pragma shader draw
#pragma shader category("Screen")

#include "Constants.fxh"
#include "Common.fxh"
#include "GPUCompression.fxh"

@_EnableCliff
{
	Name = EnableCliff;
	Description = "Enable cliff?";
	Default = false;
	Visible = false;
};
@_BlockMapEnable
{
	Name = BlockMapEnable;
	Description = "Do we have block map?";
	Visible = false;
};
@_NormalMapEnable
{
	Name = NormalMapEnable;
	Description = "Do we have normal map?";
	Visible = false;
};
@_MixtureMapEnable
{
	Name = MixtureMapEnable;
	Description = "Do we have mixture map?";
	Visible = false;
};
@_SurfaceTextureArrayEnable
{
	Name = SurfaceTextureArrayEnable;
	Description = "Enable texture array for terrain?";
	Visible = false;
};
@_CompressUsingBC
{
	Name = CompressUsingBC;
	Description = "Compress Using BC?";
	Predict = false;
	Default = false;
	Visible = false;
};

// Source target.
texture2D tIndexMap : layout(t0);
sampler sIndexMapSampler : layout(s0)
{
	Texture = tIndexMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};
texture2D tBlockMap : layout(t1);
sampler sBlockMapSampler : layout(s1)
{
	Texture = tBlockMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};
texture2D tMixtureMap : layout(t2) = "Reflection";
sampler sMixtureMapSampler : layout(s2)
{
	Texture = tMixtureMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};
texture2D tNormalMap : layout(t3) = "NormalMap";
sampler sNormalMapSampler : layout(s3)
{
	Texture = tNormalMap;
	Filter = Point;
	Address = Clamp;
	SRGB = false;
};

texture2DArray tSurfaceBlockMap : layout(t4);
sampler2DArray sSurfaceBlockSampler : layout(s4)
{
	Texture = tSurfaceBlockMap;
	Filter = Point;
	Address = Wrap;
};

texture2DArray tSurfaceNormalMap : layout(t5);
sampler2DArray sSurfaceNormalSampler : layout(s5)
{
	Texture = tSurfaceNormalMap;
	Filter = Point;
	Address = Wrap;
};

texture2DArray tSurfaceMixtureMap : layout(t6);
sampler2DArray sSurfaceMixtureSampler : layout(s6)
{
	Texture = tSurfaceMixtureMap;
	Filter = Point;
	Address = Wrap;
};

texture2DArray tDecalMap : layout(t7);
sampler2DArray sDecalSampler : layout(s7)
{
	Texture = tDecalMap;
	Filter = Linear;
	Address = Wrap;
};

cbuffer Shader
{
	float4 cChunkBound			 = {0,0,1,1}; // startx,y width height
	float4 cBlendMapSize		 = { 1, 1, 0, 0 };
	float4 cBlockParams			 = { 0, 0, 0, 0 };
	float4 cBlockShrinkage		 = { 0, 0, 1, 1 };
	float4 cOffsetInBlend		 = { 0, 0, 1, 1 }; // offset and scale
	float4 cBlockScale			 = { 1, 0, 0, 0 };
	float4 cBlockScaleList0		 = {1, 1, 1, 1};
	float4 cBlockScaleList1		 = {1, 1, 1, 1};
	float4 cBlockScaleList2		 = {1, 1, 1, 1};
	float4 cBlockScaleList3		 = {1, 1, 1, 1};
	float4 cBlendSharpness		 = {0.3, 0, 0, 0};

	float4 cDecalParams			 = {0, 0, 0, 0};
	float4 cDecalUVTransfer[16]	;
	float4 cDecalLimit[16]		;
	float4 cDecalIndex[16]		;
	float4 cOffsetInTerrain		;
};

#define DECAL_COUNT cDecalParams.x

// the ratio of block one side margin size to block content size
#define BLOCK_MARGIN_RATIO cBlockParams.x
#define BLOCK_CONTENT_HALF_PIXEL_IN_UV cBlockParams.y
#define BLOCK_PIXEL_IN_UV cBlockParams.z
#define BLOCK_LEVEL cBlockParams.w

//////////////////////////////////////////////////////////////////////////
/// Screen only texcoord.
struct VertexAssembleScreen
{
	float4		Position		: POSITION;
	float2		BaseTC			: TEXCOORD0;
};

/// Screen output.
struct VertexScreenOutput
{
	float4		HPosition		: SV_Position;
	float4		TC				: TEXCOORD0;
};

VertexScreenOutput ScreenVS(in VertexAssembleScreen IN)
{
	VertexScreenOutput OUT;
	float3 vPos = IN.Position;
	vPos.y = 1 - vPos.y;
	
	OUT.HPosition = float4(vPos.xy * 2 - 1, 1.0, 1.0);

	OUT.TC.xy = -BLOCK_MARGIN_RATIO + (1 + 2 * BLOCK_MARGIN_RATIO) * IN.BaseTC;

	return OUT;
}

#define TERRAIN_BAKE 1
#include "Terrain_Common.fxh"

struct BakeOutput
{
	uint4 BaseNormalX: SV_Target0;
	uint4 RoughnessAOHeightNormalY: SV_Target1;
};

BakeOutput TerrainBakePS(in VertexScreenOutput IN)
{
	BakeOutput OUT;

	half3 Base[16];
	half NormalX[16];
	half3 RoughnessAOHeight[16];
	half NormalY[16];

	for(int y = 0; y < 4; ++y)
	{
		for(int x = 0; x < 4; ++x)
		{
			int pixel_index = 4 * y + x;

			float2 uv_offset = float2(x - 1.5, y - 1.5) * BLOCK_PIXEL_IN_UV;

			float2 uv = clamp(IN.TC.xy + uv_offset, BLOCK_CONTENT_HALF_PIXEL_IN_UV, 1 - BLOCK_CONTENT_HALF_PIXEL_IN_UV);

			float2 uv_in_blend = cOffsetInBlend.xy + uv * cOffsetInBlend.zw;
			float2 world_pos = float2(cChunkBound.x + uv_in_blend.x * cChunkBound.z, cChunkBound.y + uv_in_blend.y * cChunkBound.w);
			// BlendMap is Y-flipped
			uv_in_blend.y = 1 - uv_in_blend.y;

			float2 uv_in_terrain = cOffsetInTerrain.xy + uv * cOffsetInTerrain.zw;

			half3 diff = half3(1, 1, 1);
			half3 norm = half3(0.5, 0.5, 1);
			half height  = 0.5;
			half3 mixx = half3(0, 0, 1);

			PS_IndexBlend(float3(0, 1, 0), float3(world_pos.x, 0, world_pos.y), uv_in_blend, 1, diff, norm, height, mixx, 0.25, ComputeFarFactor(BLOCK_LEVEL));

			for(int i = 0; i < DECAL_COUNT; ++i)
			{
				float u_min = cDecalLimit[i].x, u_max = cDecalLimit[i].y, v_min = cDecalLimit[i].z, v_max = cDecalLimit[i].w;
				if(uv_in_terrain.x > u_min && uv_in_terrain.x < u_max && uv_in_terrain.y > v_min && uv_in_terrain.y < v_max)
				{
					float2 uv_in_map = cDecalUVTransfer[i].xy * uv_in_terrain + cDecalUVTransfer[i].zw;
					half4 decal = tDecalMap.Sample(sDecalSampler, float3(uv_in_map, cDecalIndex[i].x));
					diff = lerp(diff, decal.rgb, decal.a);
				}
			}

			half3 normal = normalize(norm * 2 - 1);
			normal.xy = (normal.xy + 1) * 0.5;

			Base[pixel_index] = diff;
			NormalX[pixel_index] = half(normal.x);
			RoughnessAOHeight[pixel_index] = half3(mixx.x, mixx.z, height);
			NormalY[pixel_index] = half(normal.y);
		}
	}

	if(@_CompressUsingBC)
	{
		OUT.BaseNormalX = CompressBC3Block(Base, NormalX);
		OUT.RoughnessAOHeightNormalY = CompressBC3Block(RoughnessAOHeight, NormalY);
	}
	else
	{
		CompressETC2RGBABlock(Base, NormalX, OUT.BaseNormalX);
		CompressETC2RGBABlock(RoughnessAOHeight, NormalY, OUT.RoughnessAOHeightNormalY);
	}

	return OUT;
}

technique GenTexture
{
	VertexShader = ScreenVS();
	PixelShader = TerrainBakePS();
}
