//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<Volumetric_Homogeneous.fx>
///	@path	~/System/Shader/
///	@date	2018/08/24
///	@brief	create a homogeneous fog volume

#pragma shader material

#include "Constants.fxh"
#include "Common.fxh"
#define VOLUMETRIC_SHADING

#pragma shader predict("@_INSTANCED", "false")

//////////////////////////////////Properties////////////////////////////////////////

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cAlbedo	 				 < Widget = "LinearColor3"; > = {0.1, 0.2, 1};
	float cExtinction				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "3"; > = 1;
	float cScatteringScale			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "3"; > = 1;
	float cTileSize					 = 100.0;
	float2 cMoveSpeed				 = {1.0, 1.0};
	float cCutOff					 < Widget = "FloatSlider"; Min = "0"; Max = "0.99"; Step = "0.01"; Precision = "3"; > = 0.7;
	float cStartDistance			 < Widget = "FloatSlider"; Min = "0"; Max = "100"; Step = "0.01"; Precision = "3"; > = 5;
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tNoiseMap : layout(t1);
sampler sNoiseSampler: layout(s1)
{
	Texture = tNoiseMap;
	Filter = Linear;
	Address = Mirror;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput 
{ 
	float4		HPosition		: SV_Position;
    float2      TexCoord        : TEXCOORDN;
	float 		DepthInView		: TEXCOORDN;
};
#include "Volumetric_VS.fxh"
#include "Lighting.fxh" 
//////////////////////////////////PS////////////////////////////////////////
void GetVolumetricCoefficient(in float3 pos, out float3 Scattering, out float Extinction)
{
	float2 uv1 = (pos.xz-cMinBB.xz+cMoveSpeed*GameTime) / cTileSize;
	float2 uv2 = (pos.xz-cMinBB.xz+float2(-cMoveSpeed.y, cMoveSpeed.x)*GameTime) / cTileSize;
	float noise1 = tNoiseMap.Sample(sNoiseSampler, uv1).r;
	float noise2 = tNoiseMap.Sample(sNoiseSampler, uv2).r;
	float noise = max(noise1-cCutOff,0) + max(noise2-cCutOff, 0);
	Extinction = cExtinction * noise;
	Scattering = cAlbedo * Extinction * cScatteringScale;
}

float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}


void PS_GetParameters(in VertexShadingOutput IN, out InterMediatePixel IMP)
{
	InitInterMediatePixel(IMP);	
	
	//calc startPos, endPos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float3 camToFarPlane = CalcHomogeneousPos(1, stc);
	float farPlaneLen = length(camToFarPlane);
	float et = min(IN.DepthInView, DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc)));//[0-1]
	float st = 0;
	if (cMinBB.w < 0.5)//out of box
		st = IntersctionAABB(CameraPos.xyz, camToFarPlane, cMinBB.xyz, cMaxBB.xyz).x;
	st += cStartDistance / farPlaneLen;
	st = max(st, 0);
	et = max(et, st+EPSILON);
	IMP.StartPos = CameraPos.xyz + st * camToFarPlane;
	IMP.EndPos = CameraPos.xyz + et * camToFarPlane;

	// other params
	IMP.RaymarchingDistance = (et - st) * farPlaneLen;
	IMP.NSamples = 16;
	half inShadow;
	DynamicGILight(half3(0,1,0), IMP.IndirectLighing, inShadow);
}
#include "Volumetric_PS.fxh"

