//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration) 
//	--------------------------------------------------------------------
///	@file	<Volumetric_Homogeneous.fx>
///	@path	~/System/Shader/
///	@date	2018/08/24
///	@brief	create a homogeneous fog volume

#pragma shader material

#include "Constants.fxh"
#include "Common.fxh"
#define VOLUMETRIC_SHADING

#pragma shader predict("@_INSTANCED", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_NoiseMapEnable 
{ 
	Name = NoiseMapEnable;
	Description = "Do we have NoiseMap?";  
	AffectedTex = tNoiseMap;
	Fallback = true;
	Visible = true;
};

@_TwoLayerNoise
{ 
	Name = TwoLayerNoise;
	Description = "tow layer noise?";  
	Fallback = true;
	Visible = true;
};
//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Shader
{
	float3 cAlbedo	 				 < Widget = "LinearColor3"; > = {0.1, 0.2, 1};
	float cExtinction				 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "3"; > = 1;
	float cScatteringScale			 < Widget = "FloatSlider"; Min = "0"; Max = "10"; Step = "0.01"; Precision = "3"; > = 1;
	float cTileSizeXZ				 = 100.0;
	float cYRepeats					 = 10;
	float cCutOff					 < Widget = "FloatSlider"; Min = "0"; Max = "0.99"; Step = "0.01"; Precision = "3"; > = 0.7;
	float cStartDistance			 < Widget = "FloatSlider"; Min = "0"; Max = "100"; Step = "0.01"; Precision = "3"; > = 20;
	float3 cScroll1					  = {10, -10, 1};
	float3 cScroll2					  = {-10, 10, -1};
};

//////////////////////////////////Input Texture////////////////////////////////////////
texture2D tDepthBuffer : SceneDepth : layout(t0);
sampler sDepthBufferSampler : layout(s0)
{
	Texture = tDepthBuffer;
	Filter = Point;
	Address = Clamp;
};

texture2D tNoiseMap : layout(t1);
sampler sNoiseSampler: layout(s1)
{
	Texture = tNoiseMap;
	Filter = Linear;
	Address = Mirror;
};

texture2D tVolumetricDepth : FogVolumeDepth : layout(t2);
sampler sVolumetricDepthSampler : layout(s2)
{
	Texture = tVolumetricDepth;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput 
{ 
	float4		HPosition		: SV_Position;
    float2      TexCoord        : TEXCOORDN;
	float 		DepthInView		: TEXCOORDN;
};
#include "Volumetric_VS.fxh"
#include "Lighting.fxh" 
//////////////////////////////////PS////////////////////////////////////////
float3 CalcHomogeneousPos(float depth, float2 screenTC)
{
	[@_GLSL]
	screenTC.y = 1.0 - screenTC.y;

	float3 pos = cWBasisZ.xyz + (cWBasisX.xyz * screenTC.x) + (cWBasisY.xyz * screenTC.y);
	return pos * depth;
}

float GetNoise(in float3 pos, out float3 color)
{
	float2 uv = (pos.xz-cMinBB.xz) / cTileSizeXZ;
	float4 noise4 = tNoiseMap.SampleLevel(sNoiseSampler, uv, 0).rgba;
	float y = frac((pos.y-cMinBB.y) / (cMaxBB.y - cMinBB.y) * cYRepeats);
	y *= 4;
	float4 pos4 = float4(max(0, 1-y)+max(0,1-abs(y-4)), 1-abs(y-1), 1-abs(y-2), 1-abs(y-3));
	pos4.yzw = max(0, pos4.yzw);
	color.xyz = noise4.xyz;
	return dot(noise4, pos4);
}

void GetVolumetricCoefficient(in float3 pos, out float3 Scattering, out float Extinction)
{
	float noise = 1.0;
	float3 color = float3(1,1,1);
	if (@_NoiseMapEnable)
	{
		noise = GetNoise(pos + cScroll1 * GameTime, color);
		if (@_TwoLayerNoise)
		{
			float3 color2;
			noise += GetNoise(pos + cScroll2 * GameTime, color2);
			color *= color2;
		}
	}
	
	Extinction = max(EPSILON, cExtinction * noise - cCutOff);
	Scattering = cAlbedo * Extinction * cScatteringScale * color;
}

void PS_GetParameters(in VertexShadingOutput IN, out InterMediatePixel IMP)
{
	InitInterMediatePixel(IMP);	
	
	//calc startPos, endPos
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 stc = spos * ScreenInfo.zw;
	float3 camToFarPlane = CalcHomogeneousPos(1, stc);
	float farPlaneLen = length(camToFarPlane);
	float et = min(IN.DepthInView, DecodeDepth(tDepthBuffer.Sample(sDepthBufferSampler, stc)));//[0-1]
	float st = 0;
	if (cMinBB.w < 0.5)//out of box
		st = IntersctionAABB(CameraPos.xyz, camToFarPlane, cMinBB.xyz, cMaxBB.xyz).x;
	float preDepth = tVolumetricDepth.Sample(sVolumetricDepthSampler, stc).x;
	if (preDepth > st && preDepth < IN.DepthInView)
		st = preDepth;
	st = max(st, cStartDistance / farPlaneLen);
	et = max(et, st+EPSILON);
	IMP.StartPos = CameraPos.xyz + st * camToFarPlane;
	IMP.EndPos = CameraPos.xyz + et * camToFarPlane;

	// other params
	IMP.RaymarchingDistance = (et - st) * farPlaneLen;
	IMP.NSamples = max(1,IMP.RaymarchingDistance) / max(st*farPlaneLen*0.015, 0.01);
	IMP.NSamples = clamp(IMP.NSamples, 2, 32) * 2;
	half inShadow;
	DynamicGILight(half3(0,1,0), IMP.IndirectLighing, inShadow);
}
#include "Volumetric_PS.fxh"

