
#ifndef __FILE_GPU_COMPRESSION_FXH__
#define __FILE_GPU_COMPRESSION_FXH__

#include "Common.fxh"

// Calculate the final packed indices for a color block
uint GetPackedColorBlockIndices(in half3 Block[16], in half3 MinColor, in half3 MaxColor)
{
	uint PackedIndices = 0;

	// Project onto max->min color vector and segment into range [0,3]
	half3 Range = MinColor - MaxColor;
	half Scale = 3.f / dot(Range, Range);
	half3 ScaledRange = Range * Scale;
	half Bias = (dot(MaxColor, MaxColor) - dot(MaxColor, MinColor)) * Scale;

	for (int i = 15; i >= 0; --i)
	{
		// Compute the distance index for this element
		uint Index = (uint)round(dot(Block[i], ScaledRange) + Bias);
		// Convert distance index into the BC index
		Index += (uint)(Index > 0) - (3 * (uint)(Index == 3));
		// OR into the final PackedIndices
		PackedIndices = (PackedIndices << 2) | Index;
	}

	return PackedIndices;
}

// Calculate the final packed indices for an alpha block
// The results are in their final location of the uint2 indices and will need ORing with the min and max alpha
uint2 GetPackedAlphaBlockIndices( in half Block[16], in half MinAlpha, in half MaxAlpha )
{
	uint2 PackedIndices = 0;

	// Segment alpha max->min into range [0,7]
	half Range = MinAlpha - MaxAlpha;
	half Scale = 7.f / Range;
	half Bias = -Scale * MaxAlpha;

	uint i = 0;
	// The first 5 elements of the block will go into the top 16 bits of the x component
	for (i=0; i<5; ++i)
	{
		// Compute the distance index for this element
		uint Index = (uint)round(Block[i] * Scale + Bias);
		// Convert distance index into the BC index
		Index += (uint)(Index > 0) - (7 * (uint)(Index == 7));
		// OR into the final PackedIndices
		PackedIndices.x |= (Index << (i*3 + 16));
	}

	// The 6th element is split across the x and y components
	{
		i = 5;
		uint Index = (uint)round(Block[i] * Scale + Bias);
		Index += (uint)(Index > 0) - (7 * (uint)(Index == 7));
		PackedIndices.x |= (Index << 31);
		PackedIndices.y |= (Index >> 1);
	}

	// The rest of the elements fill the y component
	for (i=6; i<16; ++i)
	{
		uint Index = (uint)round(Block[i] * Scale + Bias);
		Index += (uint)(Index > 0) - (7 * (uint)(Index == 7));
		PackedIndices.y |= (Index << (i*3 - 16));
	}

	return PackedIndices;
}

// Get min and max values in a single channel block
void GetMinMax( in half Block[16], out half OutMin, out half OutMax )
{
	OutMin = Block[0];
	OutMax = Block[0];

	for (int i=1; i<16; ++i)
	{
		OutMin = min(OutMin, Block[i]);
		OutMax = max(OutMax, Block[i]);
	}
}

// Get min and max values in a single three channel block
void GetMinMax(in half3 Block[16], out half3 OutMin, out half3 OutMax)
{
	OutMin = Block[0];
	OutMax = Block[0];

	for (int i = 1; i < 16; ++i)
	{
		OutMin = min(OutMin, Block[i]);
		OutMax = max(OutMax, Block[i]);
	}
}

// Convert half3 color to 565 uint using 'ceil' arithmetic
// Color parameter is inout and is modified to match the converted value
uint Half3ToUint565_Ceil( inout half3 Color )
{
	half3 Scale         = half3(31.f, 63.f, 31.f);
	half3 ColorScaled   = ceil(saturate(Color) * Scale);
	uint ColorPacked     = ((uint)ColorScaled.r << 11) | ((uint)ColorScaled.g << 5) | (uint)ColorScaled.b;
	Color				 = ColorScaled / Scale;

	return ColorPacked;
}

// Convert half3 color to 565 uint using 'floor' arithmetic
// Color parameter is inout and is modified to match the converted value
uint Half3ToUint565_Floor( inout half3 Color )
{
	half3 Scale         = half3(31.f, 63.f, 31.f);
	half3 ColorScaled   = floor(saturate(Color) * Scale);
	uint ColorPacked     = ((uint)ColorScaled.r << 11) | ((uint)ColorScaled.g << 5) | (uint)ColorScaled.b;
	Color				 = ColorScaled / Scale;

	return ColorPacked;
}

// Compress a BC3 block
uint4 CompressBC3Block( in half3 BlockRGB[16], in half BlockA[16] )
{
	half3 MinColor, MaxColor;
	GetMinMax( BlockRGB, MinColor, MaxColor );

	uint MinColor565 = Half3ToUint565_Floor( MinColor );
	uint MaxColor565 = Half3ToUint565_Ceil( MaxColor );

	half MinAlpha, MaxAlpha;
	GetMinMax( BlockA, MinAlpha, MaxAlpha );

	uint MinAlphaUint = (uint)round(MinAlpha * 255.f);
	uint MaxAlphaUint = (uint)round(MaxAlpha * 255.f);

	uint ColorIndices = 0;
	if (MinColor565 < MaxColor565)
	{
		ColorIndices = GetPackedColorBlockIndices( BlockRGB, MinColor, MaxColor );
	}

	uint2 AlphaIndices = 0;
	if (MinAlphaUint < MaxAlphaUint)
	{
		AlphaIndices = GetPackedAlphaBlockIndices(BlockA, MinAlpha, MaxAlpha);
	}

	return uint4((MinAlphaUint << 8) | MaxAlphaUint | AlphaIndices.x, AlphaIndices.y, (MinColor565 << 16) | MaxColor565, ColorIndices);
}

// Use macro here because messiah doesn't support array parameter in metal and es3
#define GetBlockInfo(BlockRGB, AvgRGB, LumMin, LumMax) \
{ \
	AvgRGB[0] = 0.25 * (BlockRGB[0] + BlockRGB[1] + BlockRGB[4] + BlockRGB[5]); \
	AvgRGB[1] = 0.25 * (BlockRGB[2] + BlockRGB[3] + BlockRGB[6] + BlockRGB[7]); \
	AvgRGB[2] = 0.25 * (BlockRGB[8] + BlockRGB[9] + BlockRGB[12] + BlockRGB[13]); \
	AvgRGB[3] = 0.25 * (BlockRGB[10] + BlockRGB[11] + BlockRGB[14] + BlockRGB[15]); \
	LumMin[0] = min(min(BlockRGB[0], BlockRGB[1]), min(BlockRGB[4], BlockRGB[5])); \
	LumMin[1] = min(min(BlockRGB[2], BlockRGB[3]), min(BlockRGB[6], BlockRGB[7])); \
	LumMin[2] = min(min(BlockRGB[8], BlockRGB[9]), min(BlockRGB[12], BlockRGB[13])); \
	LumMin[3] = min(min(BlockRGB[10], BlockRGB[11]), min(BlockRGB[14], BlockRGB[15])); \
	LumMax[0] = max(max(BlockRGB[0], BlockRGB[1]), max(BlockRGB[4], BlockRGB[5])); \
	LumMax[1] = max(max(BlockRGB[2], BlockRGB[3]), max(BlockRGB[6], BlockRGB[7])); \
	LumMax[2] = max(max(BlockRGB[8], BlockRGB[9]), max(BlockRGB[12], BlockRGB[13])); \
	LumMax[3] = max(max(BlockRGB[10], BlockRGB[11]), max(BlockRGB[14], BlockRGB[15])); \
}

half SqrLength(half3 x)
{
	return dot(x, x);
}

half Max3(half3 RGB)
{
	return max(RGB.x, max(RGB.y, RGB.z));
}

half Min3(half3 RGB)
{
	return min(RGB.x, min(RGB.y, RGB.z));
}

half Sum4(half4 data)
{
	return data.x + data.y + data.z + data.w;
}

half Sum4x4(half4x4 data)
{
	return Sum4(data[0]) + Sum4(data[1]) + Sum4(data[2]) + Sum4(data[3]);
}

uint GetTableIndex(float2 TableEntry, half3 Pixel, half3 RGBSub)
{
	half Error = SqrLength(saturate(RGBSub - half(TableEntry.y)) - Pixel);
	uint Result = 0x10001;
	half NewError = SqrLength(saturate(RGBSub - half(TableEntry.x)) - Pixel);
	if(NewError < Error)
	{
		Error = NewError;
		Result = 0x1;
	}
	NewError = SqrLength(saturate(RGBSub + half(TableEntry.x)) - Pixel);
	if(NewError < Error)
	{
		Error = NewError;
		Result = 0x0;
	}
	NewError = SqrLength(saturate(RGBSub + half(TableEntry.y)) - Pixel);
	if(NewError < Error)
	{
		Error = NewError;
		Result = 0x10000;
	}
	return Result;
}

// Thoese functions are used to fix glslopt bugs
half4x4 Less4x4(half4x4 lhs, half rhs)
{
	// Cannot use '<' to compare half4x4 with half, glslopt's bug
	half4x4 Result;
	Result[0] = half4(lhs[0] < rhs.xxxx);
	Result[1] = half4(lhs[1] < rhs.xxxx);
	Result[2] = half4(lhs[2] < rhs.xxxx);
	Result[3] = half4(lhs[3] < rhs.xxxx);
	return Result;
}
half4x4 Less4x4(half4x4 lhs, half4x4 rhs)
{
	// Cannot use '<' to compare half4x4 with half, glslopt's bug
	half4x4 Result;
	Result[0] = half4(lhs[0] < rhs[0]);
	Result[1] = half4(lhs[1] < rhs[1]);
	Result[2] = half4(lhs[2] < rhs[2]);
	Result[3] = half4(lhs[3] < rhs[3]);
	return Result;
}
half4x4 Abs4x4(half4x4 data)
{
	half4x4 Result;
	Result[0] = abs(data[0]);
	Result[1] = abs(data[1]);
	Result[2] = abs(data[2]);
	Result[3] = abs(data[3]);
	return Result;
}
half4x4 Pow4x4(half4x4 data, half e)
{
	half4x4 Result;
	Result[0] = pow(data[0], e);
	Result[1] = pow(data[1], e);
	Result[2] = pow(data[2], e);
	Result[3] = pow(data[3], e);
	return Result;
}
half4x4 Min4x4(half4x4 lhs, half4x4 rhs)
{
	half4x4 Result;
	Result[0] = min(lhs[0], rhs[0]);
	Result[1] = min(lhs[1], rhs[1]);
	Result[2] = min(lhs[2], rhs[2]);
	Result[3] = min(lhs[3], rhs[3]);
	return Result;
}
int4x4 MatElemMul(int4x4 a, int4x4 b)
{
	int4x4 Result;
	Result[0] = a[0] * b[0];
	Result[1] = a[1] * b[1];
	Result[2] = a[2] * b[2];
	Result[3] = a[3] * b[3];
	return Result;
}

// Use macro here because messiah doesn't support array parameter in metal and es3
#define CompressETC2RGBABlock(BlockRGB, BlockA, Result) \
{ \
	half3 AvgRGB[4]; \
	half3 LumMin[4]; \
	half3 LumMax[4]; \
	GetBlockInfo(BlockRGB, AvgRGB, LumMin, LumMax); \
	half3 LeftRGB = 0.5 * (AvgRGB[0] + AvgRGB[2]); \
	half3 RightRGB = 0.5 * (AvgRGB[1] + AvgRGB[3]); \
	half3 TopRGB = 0.5 * (AvgRGB[0] + AvgRGB[1]); \
	half3 BottomRGB = 0.5 * (AvgRGB[2] + AvgRGB[3]); \
	half3 FinalRGB0, FinalRGB1; \
	uint FlipBit; \
	half3 LeftRGBDiff0 = LeftRGB - AvgRGB[0]; \
	half3 LeftRGBDiff1 = LeftRGB - AvgRGB[2]; \
	half3 RightRGBDiff0 = RightRGB - AvgRGB[1]; \
	half3 RightRGBDiff1 = RightRGB - AvgRGB[3]; \
	half3 TopRGBDiff0 = TopRGB - AvgRGB[0]; \
	half3 TopRGBDiff1 = TopRGB - AvgRGB[1]; \
	half3 BottomRGBDiff0 = BottomRGB - AvgRGB[2]; \
	half3 BottomRGBDiff1 = BottomRGB - AvgRGB[3]; \
	half3 FinalLumMin0, FinalLumMin1, FinalLumMax0, FinalLumMax1; \
	if(SqrLength(LeftRGBDiff0) + SqrLength(LeftRGBDiff1) + SqrLength(RightRGBDiff0) + SqrLength(LeftRGBDiff1) \
		< SqrLength(TopRGBDiff0) + SqrLength(TopRGBDiff1) + SqrLength(BottomRGBDiff0) + SqrLength(BottomRGBDiff1)) \
	{ \
		FinalRGB0 = LeftRGB; \
		FinalRGB1 = RightRGB; \
		FinalLumMin0 = min(LumMin[0], LumMin[2]); \
		FinalLumMax0 = max(LumMax[0], LumMax[2]); \
		FinalLumMin1 = min(LumMin[1], LumMin[3]); \
		FinalLumMax1 = max(LumMax[1], LumMax[3]); \
		FlipBit = 0;\
	} \
	else \
	{ \
		FinalRGB0 = TopRGB; \
		FinalRGB1 = BottomRGB; \
		FinalLumMin0 = min(LumMin[0], LumMin[1]); \
		FinalLumMax0 = max(LumMax[0], LumMax[1]); \
		FinalLumMin1 = min(LumMin[2], LumMin[3]); \
		FinalLumMax1 = max(LumMax[2], LumMax[3]); \
		FlipBit = 1;\
	} \
	uint3 FinalRGBBit0 = (uint3)round(FinalRGB0 * 15); \
	uint3 FinalRGBBit1 = (uint3)round(FinalRGB1 * 15); \
	uint3 FinalRGBBaseBit = (uint3)round(FinalRGB0 * 31); \
	uint3 FinalRGBTargetOtherBit = (uint3)round(FinalRGB1 * 31); \
	int3 FinalRGBDiffBit = clamp((int3)FinalRGBTargetOtherBit - (int3)FinalRGBBaseBit, -4, 3); \
	uint3 FinalRGBOtherBit = FinalRGBBaseBit + FinalRGBDiffBit; \
	half3 IndividualRGBSub0 = ((FinalRGBBit0 << 4) | FinalRGBBit0) / 255.0; \
	half3 IndividualRGBSub1 = ((FinalRGBBit1 << 4) | FinalRGBBit1) / 255.0; \
	half3 DifferentialRGBSub0 = ((FinalRGBBaseBit << 3) | (FinalRGBBaseBit >> 2)) / 255.0; \
	half3 DifferentialRGBSub1 = ((FinalRGBOtherBit << 3) | (FinalRGBOtherBit >> 2)) / 255.0; \
	uint DBit; \
	half3 RGBSub0, RGBSub1; \
	uint SubColorData; \
	if(SqrLength(IndividualRGBSub0 - FinalRGB0) + SqrLength(IndividualRGBSub1 - FinalRGB1) < SqrLength(DifferentialRGBSub0 - FinalRGB0) + SqrLength(DifferentialRGBSub1 - FinalRGB1)) \
	{ \
		DBit = 0; \
		RGBSub0 = IndividualRGBSub0; \
		RGBSub1 = IndividualRGBSub1; \
		SubColorData = (FinalRGBBit0.r << 4) | FinalRGBBit1.r | (FinalRGBBit0.g << 12) | (FinalRGBBit1.g << 8) | (FinalRGBBit0.b << 20) | (FinalRGBBit1.b << 16); \
	} \
	else \
	{ \
		DBit = 1; \
		RGBSub0 = DifferentialRGBSub0; \
		RGBSub1 = DifferentialRGBSub1; \
		uint3 FinalRGBDiffBitUnsigned = ((uint3)FinalRGBDiffBit) & uint3(0x7, 0x7, 0x7); \
		SubColorData = (FinalRGBBaseBit.r << 3) | FinalRGBDiffBitUnsigned.r | (FinalRGBBaseBit.g << 11) | (FinalRGBDiffBitUnsigned.g << 8) | (FinalRGBBaseBit.b << 19) | (FinalRGBDiffBitUnsigned.b << 16); \
	} \
	uint int64bitColorHigh = SubColorData | (DBit << 25) | (FlipBit << 24); \
	half MaxDiffSub0 = max(Max3(FinalRGB0 - FinalLumMin0), Max3(FinalLumMax0 - FinalRGB0)); \
	uint TableCode0; \
	/* glslopt's bug, half is enough, but glslopt only support float constant table, */ \
	/* half table will be convert to float table and cause type error! */ \
	float2 Table[8] = {float2(2, 8) / 255, float2(5, 17) / 255, float2(9, 29) / 255, float2(13, 42) / 255, float2(18, 60) / 255, float2(24, 80) / 255, float2(33, 106) / 255, float2(47, 183) / 255}; \
	for(TableCode0 = 0; TableCode0 < 7; ++TableCode0) \
	{ \
		if(half(Table[TableCode0].y) > MaxDiffSub0) \
			break; \
	} \
	half MaxDiffSub1 = max(Max3(FinalRGB1 - FinalLumMin1), Max3(FinalLumMax1 - FinalRGB1)); \
	uint TableCode1; \
	for(TableCode1 = 0; TableCode1 < 7; ++TableCode1) \
	{ \
		if(half(Table[TableCode1].y) > MaxDiffSub1) \
			break; \
	} \
	int64bitColorHigh |= (TableCode0 << 29) | (TableCode1 << 26); \
	float2 TableEntry0 = Table[TableCode0], TableEntry1 = Table[TableCode1]; \
	uint int64bitColorLow = 0; \
	if(FlipBit == 0) \
	{ \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[0], RGBSub0) << 8; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[4], RGBSub0) << 9; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[8], RGBSub0) << 10; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[12], RGBSub0) << 11; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[1], RGBSub0) << 12; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[5], RGBSub0) << 13; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[9], RGBSub0) << 14; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[13], RGBSub0) << 15; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[2], RGBSub1); \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[6], RGBSub1) << 1; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[10], RGBSub1) << 2; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[14], RGBSub1) << 3; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[3], RGBSub1) << 4; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[7], RGBSub1) << 5; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[11], RGBSub1) << 6; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[15], RGBSub1) << 7; \
	} \
	else \
	{ \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[0], RGBSub0) << 8; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[4], RGBSub0) << 9; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[8], RGBSub1) << 10; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[12], RGBSub1) << 11; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[1], RGBSub0) << 12; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[5], RGBSub0) << 13; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[9], RGBSub1) << 14; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[13], RGBSub1) << 15; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[2], RGBSub0); \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[6], RGBSub0) << 1; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[10], RGBSub1) << 2; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[14], RGBSub1) << 3; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[3], RGBSub0) << 4; \
		int64bitColorLow |= GetTableIndex(TableEntry0, BlockRGB[7], RGBSub0) << 5; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[11], RGBSub1) << 6; \
		int64bitColorLow |= GetTableIndex(TableEntry1, BlockRGB[15], RGBSub1) << 7; \
	} \
	half AlphaMin = 1.0, AlphaMax = 0.0; \
	half AvgAlpha = 0; \
	[unroll] \
	for (int i = 0; i < 16; ++i) \
	{ \
		AvgAlpha += BlockA[i]; \
		AlphaMin = min(AlphaMin, BlockA[i]); \
		AlphaMax = max(AlphaMax, BlockA[i]); \
	} \
	AvgAlpha = round(AvgAlpha * (255.0 / 16.0)) / 255; \
	half4x4 BlockAlphaDiff = half4x4(BlockA[0], BlockA[1], BlockA[2], BlockA[3], \
								BlockA[4], BlockA[5], BlockA[6], BlockA[7], \
								BlockA[8], BlockA[9], BlockA[10], BlockA[11], \
								BlockA[12], BlockA[13], BlockA[14], BlockA[15]); \
	BlockAlphaDiff = BlockAlphaDiff - AvgAlpha; \
	half4x4 AlphaNegative = Less4x4(BlockAlphaDiff, -0.5 / 255); \
	BlockAlphaDiff = Abs4x4(BlockAlphaDiff); \
	half AlphaMinDiff = AlphaMin - AvgAlpha; \
	half AlphaMaxDiff = AlphaMax - AvgAlpha; \
	int FinalAlphaTableCode; \
	half FinalAlphaTableError = 17; \
	int FinalAlphaMultiplier; \
	half4x4 FinalAlphaPixelCode; \
	/* glslopt's bug, half is enough, but glslopt only support float constant table, */ \
	/* half table will be convert to float table and cause type error! */ \
	float4 AlphaTable[16] = {float4(2, 5, 8, 14) / 255, \
							float4(2, 6, 9, 12) / 255, \
							float4(1, 4, 7, 12) / 255, \
							float4(1, 3, 5, 12) / 255, \
							float4(2, 5, 7, 11) / 255, \
							float4(2, 6, 8, 10) / 255, \
							float4(3, 6, 7, 10) / 255, \
							float4(2, 4, 7, 10) / 255, \
							float4(1, 5, 7, 9) / 255, \
							float4(1, 4, 7, 9) / 255, \
							float4(1, 3, 7, 9) / 255, \
							float4(1, 4, 6, 9) / 255, \
							float4(2, 3, 6, 9) / 255, \
							float4(0, 1, 2, 9) / 255, \
							float4(3, 5, 7, 8) / 255, \
							float4(2, 4, 6, 8) / 255}; \
	for (int AlphaTableCode = 0; AlphaTableCode < 16; ++AlphaTableCode) \
	{ \
		float4 TableEntry = AlphaTable[AlphaTableCode]; \
		int Multiplier = min(15, max(1, int(max(floor(AlphaMinDiff / (-TableEntry.w - 1.0 / 255)), floor(AlphaMaxDiff / TableEntry.w))))); \
		TableEntry *= Multiplier; \
		half4x4 NegativeCorrection = AlphaNegative * (Multiplier / 255.0); \
		half4x4 TableError = Pow4x4(BlockAlphaDiff - half(TableEntry[0]) - NegativeCorrection, 2); \
		half4x4 AlphaPixelCode = half4x4(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0); \
		[unroll] \
		for(int i = 1; i < 4; ++i) \
		{ \
			half4x4 NewTableError = Pow4x4(BlockAlphaDiff - half(TableEntry[i]) - NegativeCorrection, 2); \
			half4x4 RefreshPixelCode = Less4x4(NewTableError, TableError); \
			TableError = Min4x4(TableError, NewTableError); \
			AlphaPixelCode = MatElemMul(AlphaPixelCode, (1 - RefreshPixelCode)) + i * RefreshPixelCode; \
		} \
		half NewFinalAlphaTableError = Sum4x4(TableError); \
		if (NewFinalAlphaTableError < FinalAlphaTableError) \
		{ \
			FinalAlphaTableError = NewFinalAlphaTableError; \
			FinalAlphaTableCode = AlphaTableCode; \
			FinalAlphaMultiplier = Multiplier; \
			FinalAlphaPixelCode = AlphaPixelCode + 4 - 4 * AlphaNegative; \
		} \
	} \
	uint int64bitAlphaHigh = (uint)(round(AvgAlpha * 255)) | (uint)(FinalAlphaMultiplier << 12) | (uint)(FinalAlphaTableCode << 8); \
	int64bitAlphaHigh |= (uint)(FinalAlphaPixelCode[0][0]) << 21; /* glslopt's bug, must cast to uint */ \
	int64bitAlphaHigh |= (uint)(FinalAlphaPixelCode[1][0]) << 18; \
	int64bitAlphaHigh |= (uint)((uint)FinalAlphaPixelCode[2][0] >> 1) << 16; \
	int64bitAlphaHigh |= (uint)((uint)FinalAlphaPixelCode[2][0] & (uint)0x1) << 31; \
	int64bitAlphaHigh |= (uint)(FinalAlphaPixelCode[3][0]) << 28; \
	int64bitAlphaHigh |= (uint)(FinalAlphaPixelCode[0][1]) << 25; \
	int64bitAlphaHigh |= (uint)((uint)FinalAlphaPixelCode[1][1] >> 2) << 24; \
	uint int64bitAlphaLow = (uint)((uint)FinalAlphaPixelCode[1][1] & (uint)0x3) << 6; \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[2][1]) << 3; \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[3][1]); \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[0][2]) << 13; \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[1][2]) << 10; \
	int64bitAlphaLow |= (uint)((uint)FinalAlphaPixelCode[2][2] >> 1) << 8; \
	int64bitAlphaLow |= (uint)((uint)FinalAlphaPixelCode[2][2] & (uint)0x1) << 23; \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[3][2]) << 20; \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[0][3]) << 17; \
	int64bitAlphaLow |= (uint)((uint)FinalAlphaPixelCode[1][3] >> 2) << 16; \
	int64bitAlphaLow |= (uint)((uint)FinalAlphaPixelCode[1][3] & (uint)0x3) << 30; \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[2][3]) << 27; \
	int64bitAlphaLow |= (uint)(FinalAlphaPixelCode[3][3]) << 24; \
	Result = uint4(int64bitAlphaHigh, int64bitAlphaLow, int64bitColorHigh, int64bitColorLow); \
}

#endif
