//	--------------------------------------------------------------------
//	Copyright(C) 2006~2019 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : hzzhoukeli@corp.netease.com
//	Author: keli.zhou
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<HexVoxelGI.fx>
///	@path	~/System/Shader/Include/
///	@date	2020/01/08
///	@brief	Monte Carlo for Tracing.

/////////////////////////////////////////////// From UE4 /////////////////////////////////////////////////////
// 3D random number generator inspired by PCGs (permuted congruential generator)
// Using a **simple** Feistel cipher in place of the usual xor shift permutation step
// @param v = 3D integer coordinate
// @return three elements w/ 16 random bits each (0-0xffff).
// ~8 ALU operations for result.x    (7 mad, 1 >>)
// ~10 ALU operations for result.xy  (8 mad, 2 >>)
// ~12 ALU operations for result.xyz (9 mad, 3 >>)
uint3 Rand3DPCG16(int3 p)
{
	// taking a signed int then reinterpreting as unsigned gives good behavior for negatives
	uint3 v = uint3(p);

	// Linear congruential step. These LCG constants are from Numerical Recipies
	// For additional #'s, PCG would do multiple LCG steps and scramble each on output
	// So v here is the RNG state
	v = v * 1664525 + 1013904223;

	// PCG uses xorshift for the final shuffle, but it is expensive (and cheap
	// versions of xorshift have visible artifacts). Instead, use simple MAD Feistel steps
	//
	// Feistel ciphers divide the state into separate parts (usually by bits)
	// then apply a series of permutation steps one part at a time. The permutations
	// use a reversible operation (usually ^) to part being updated with the result of
	// a permutation function on the other parts and the key.
	//
	// In this case, I'm using v.x, v.y and v.z as the parts, using + instead of ^ for
	// the combination function, and just multiplying the other two parts (no key) for
	// the permutation function.
	//
	// That gives a simple mad per round.
	v.x += v.y*v.z;
	v.y += v.z*v.x;
	v.z += v.x*v.y;
	v.x += v.y*v.z;
	v.y += v.z*v.x;
	v.z += v.x*v.y;

	// only top 16 bits are well shuffled
	return v >> 16;
}

uint2 ComputeRandomSeed(uint2 PixelPosition, uint Frame)
{
	return Rand3DPCG16(int3(PixelPosition, Frame)).xy;
}

/** Reverses all the 32 bits. */
uint ReverseBits32( uint bits )
{
#if SM5_PROFILE || COMPILER_METAL
	return reversebits( bits );
#else
	bits = ( bits << uint(16) ) | ( bits >> uint(16) );
	bits = ( (bits & uint(16711935))   << uint(8) ) | ( (bits & (uint)(4278255360)) >> uint(8) );
	bits = ( (bits & uint(252645135))  << uint(4) ) | ( (bits & (uint)(4042322160)) >> uint(4) );
	bits = ( (bits & uint(858993459))  << uint(2) ) | ( (bits & (uint)(3435973836)) >> uint(2) );
	bits = ( (bits & uint(1431655765)) << uint(1) ) | ( (bits & (uint)(2863311530)) >> uint(1) );
	return bits;
	/*
	bits = ( bits << 16) | ( bits >> 16);
	bits = ( (bits & 0x00ff00ff) << 8 ) | ( (bits & 0xff00ff00) >> 8 );
	bits = ( (bits & 0x0f0f0f0f) << 4 ) | ( (bits & 0xf0f0f0f0) >> 4 );
	bits = ( (bits & 0x33333333) << 2 ) | ( (bits & 0xcccccccc) >> 2 );
	bits = ( (bits & 0x55555555) << 1 ) | ( (bits & 0xaaaaaaaa) >> 1 );
	return bits;*/
#endif
}

float2 Hammersley16( uint Index, uint NumSamples, uint2 Random )
{
	float E1 = frac( (float)Index / NumSamples + float( Random.x ) * (1.0 / 65536.0) );
	float E2 = float( ( ReverseBits32(Index) >> uint(16) ) ^ Random.y ) * (1.0 / 65536.0);
	return float2( E1, E2 );
}

// [ Duff et al. 2017, "Building an Orthonormal Basis, Revisited" ]
float3x3 GetTangentBasis( float3 TangentZ )
{
	const float Sign = TangentZ.z >= 0 ? 1 : -1;
	const float a = -rcp( Sign + TangentZ.z );
	const float b = TangentZ.x * TangentZ.y * a;

	float3 TangentX = { 1 + Sign * a * Pow2( TangentZ.x ), Sign * b, -Sign * TangentZ.x };
	float3 TangentY = { b,  Sign + a * Pow2( TangentZ.y ), -TangentZ.y };

	return float3x3( TangentX, TangentY, TangentZ );
}

float2 UniformSampleDiskConcentric( float2 E )
{
	float2 p = 2 * E - 1;
	float Radius;
	float Phi;
	if( abs( p.x ) > abs( p.y ) )
	{
		Radius = p.x;
		Phi = (PI/4) * (p.y / p.x);
	}
	else
	{
		Radius = p.y;
		Phi = (PI/2) - (PI/4) * (p.x / p.y);
	}
	return float2( Radius * cos( Phi ), Radius * sin( Phi ) );
}

float3 ComputeL(float3 N, float2 E)
{
	float3x3 TangentBasis = GetTangentBasis(N);
	
	float3 L;
	L.xy = UniformSampleDiskConcentric( E );
	L.z = sqrt(saturate(1 - dot( L.xy, L.xy ) ));
	L = mul( L, TangentBasis );

	return L;
}

/////////////////////////////////////////////// UE4 End /////////////////////////////////////////////////////

float random( float2 p )
{
    float2 K1 = float2(
        23.14069263277926, // e^pi (Gelfond's constant)
         2.665144142690225 // 2^sqrt(2) (Gelfondâ€“Schneider constant)
    );
    return abs(frac( cos( dot(p,K1) ) * 12345.6789 ));
}

float3 GetRandomDirection(float3 N, float rand1, float rand2)
{
	float3 XVector = abs(N.y) > 0.5 ? float3(1, 0, 0) : float3(0, 1, 0);
	float3 TangentY = N;
	float3 TangentZ = normalize(cross(XVector, TangentY));
	float3 TangentX = normalize(cross(TangentY, TangentZ));

	//float rand1 = random(abs(float2(sin(Seed1 * WPosition.x), cos(Seed2 * WPosition.y))));
	//float rand2 = random(abs(float2(sin(Seed2 * 59 * WPosition.z), cos(Seed1 * 31 * WPosition.x))));

	float phi = 2 * PI * rand1;
	float cosTheta2 = rand2;
	float sinTheta2 = 1 - rand2;

	float y = sqrt(cosTheta2);
	float x = sqrt(sinTheta2) * cos(phi);
	float z = sqrt(sinTheta2) * sin(phi);

	float3 RayDir = normalize(x * TangentX + y * TangentY + z * TangentZ);

	return RayDir;
}

float3 GetRandomDirection2(float3 N, float2 E)
{
	float3 XVector = abs(N.y) > 0.5 ? float3(1, 0, 0) : float3(0, 1, 0);
	float3 TangentY = N;
	float3 TangentZ = normalize(cross(XVector, TangentY));
	float3 TangentX = normalize(cross(TangentY, TangentZ));

	//float rand1 = random(abs(float2(sin(Seed1 * WPosition.x), cos(Seed2 * WPosition.y))));
	//float rand2 = random(abs(float2(sin(Seed2 * 59 * WPosition.z), cos(Seed1 * 31 * WPosition.x))));

	float phi = 2 * PI * E.x;
	float cosTheta2 = E.y;
	float sinTheta2 = 1 - E.y;

	float y = sqrt(cosTheta2);
	float x = sqrt(sinTheta2) * cos(phi);
	float z = sqrt(sinTheta2) * sin(phi);

	float3 RayDir = normalize(x * TangentX + y * TangentY + z * TangentZ);

	return RayDir;
}
