///	@file	<Particle_Common.fx>
///	@path	~/System/Shader/
///	@date	2019/01/--
///	@brief	Particle Common Functions / Vars.
/// @note 	Include Almost All Particle VertexShading 

#ifndef __FILE_PARTICLE_COMMON_FXH__
#define __FILE_PARTICLE_COMMON_FXH__

#include "Common.fxh"
#include "Lighting.fxh"

// Struct And Init Function
struct VertexAssemble
{
	//EVF_P3F_C4B_N4B_T2F
	float4		Position 		: POSITION;
    
	half4		VertexColor 	: COLOR;
	[@_NORMAL]
	half4		Normal 			: NORMAL;

	float2		TexCoord		: TEXCOORDN;
	
	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;


};

struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate, depth and fog for surface.
	[@_NORMAL]
	half4		Normal 			: NORMAL;
	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;
	half4		VertexColor		: COLOR;				//	Per-vertex color.
	half3		ViewDirection	: TEXCOORDN;
	float3		WorldPosition	: TEXCOORDN;
};


struct ParticleShadingParameters
{
    half4 BaseColor;
    half Metallic;
    half Roughness;
    half Specular;
    half3 WorldNormal;
	half Depth;
	half3 Emission;
};

void InitParticleShadingParameters(in out ParticleShadingParameters MaterialParamters)
{
    MaterialParamters.BaseColor = half4(1,1,1,1);
    MaterialParamters.Metallic = 0;
    MaterialParamters.Roughness = 1;
    MaterialParamters.Specular = 0.5;
    MaterialParamters.WorldNormal = half3(1,0,0);
	MaterialParamters.Depth = 0;
	MaterialParamters.Emission = half3(0,0,0);
}

void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{
	OUT.HPosition = float4(0,0,0,0);
	OUT.TexCoord = float4(0,0,0,0);
	OUT.VertexColor = half4(0,0,0,0);
	OUT.ViewDirection = half3(0,0,0);
	OUT.WorldPosition = half3(0,0,0);

	[@_NORMAL]
	OUT.Normal = half4(0,0,0,0);

	[@_TANGENT]
	OUT.Tangent = float4(0,0,0,0);
	[@_TANGENT]
	OUT.Binormal = float4(0,0,0,0);
	
}


VertexShadingOutput ParticleUnifyVertexShading(in const VertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);

	float4 InPosition = float4(IN.Position.xyz, 1);
	float4 WorldPosition = float4(0,0,0,0);

	if (@_LOCAL_SPACE)  // input pos in local space
	{
		WorldPosition = float4(MulMat3x4(World, InPosition).xyz, 1);
		OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	}
	else  // input pos in world space
	{
		WorldPosition = InPosition;
		OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	}
	OUT.HPosition.z -= CameraInfo.x * DepthBias / (DepthBias + OUT.HPosition.w);
	OUT.HPosition.w *= cFov;
	OUT.TexCoord.xy = IN.TexCoord * UVScale;
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = 1;

	OUT.VertexColor = GetVertexColor(IN.VertexColor);

	//WorldPosition.xyz = WorldPosition.xyz - CameraPos.xyz;
	OUT.WorldPosition = WorldPosition.xyz ;

	OUT.ViewDirection = normalize(CameraPos.xyz - WorldPosition.xyz);
	
	[@_HDR]
	OUT.VertexColor.rgb *= HDRScale;

	OUT.TexCoord.w  = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);

	if (@_VolumetricFog)
	{
		OUT.TexCoord.w = OUT.HPosition.w;
	}	
	if (@_LOCAL_SPACE)  // input pos in local space
	{
		if (@_NORMAL)
		{
			//N4B -> Normal in VertexAssemble is Unsigned Byte
			//issue World Matrix using as Normal Transfer May be Wrong
			OUT.Normal = half4((2 * IN.Normal.xyz - 1).xyz,0);
			OUT.Normal =  half4( normalize( MulMat3x4(World, OUT.Normal ).xyz) ,0);
		}
		if (@_TANGENT)
		{
			OUT.Tangent = float4( normalize( MulMat3x4(World, IN.Tangent ).xyz) ,0);
			OUT.Binormal = float4( normalize( MulMat3x4(World, IN.Binormal ).xyz) ,0);
		}
	}
	else
	{
		if (@_NORMAL)
		{
			OUT.Normal = 2 * IN.Normal - 1;
		}
		if (@_TANGENT)
		{
			OUT.Tangent = IN.Tangent;
			OUT.Binormal = IN.Binormal;
		}
	}

    
	return OUT;
}

float3 SafeNormalizeF3(const float3 vin,const float3 safeValue)
{
	float len = length(vin);
	return (len < 0.0001) ? safeValue : vin / len;
}

float2 SafeNormalizeF2(const float2 vin,const float2 safeValue)
{
	float len = length(vin);
	return (len < 0.0001) ? safeValue : vin / len;
}

void GetNormalD(in const VertexShadingOutput IN,out float3 Normal,out float3 BiNormal,out float3 Tangent)
{
	/*
		Get Normal using ddx() and ddy()
		this is a gpu based (PS) Normal / BiNormal / Tangent
	*/
	float3 P = IN.WorldPosition;
	float3 dx = SafeNormalizeF3(ddx(P),float3(1,1,0));
	float3 dy = SafeNormalizeF3(ddy(P),float3(0,-1,-1));
	float2 Edge1uv = SafeNormalizeF2(ddx(IN.TexCoord.xy),float2(1,0));
	float2 Edge2uv = SafeNormalizeF2(ddy(IN.TexCoord.xy),float2(0,1));
	float3 N = float3(0,0,0);
	if (false) // src : if(@_NORMAL)
	{
		N = IN.Normal.xyz;
	} else {
		N = SafeNormalizeF3(cross(dy, dx),float3(0,-1,0));
	}

	float cp = Edge1uv.y * Edge2uv.x - Edge1uv.x * Edge2uv.y;
    float mul =  cp == 0 ? 1 : 1.0f / cp;

    Tangent = SafeNormalizeF3((dx * -Edge2uv.y + dy * Edge1uv.y) * mul,float3(1,0,0));
	BiNormal = SafeNormalizeF3((dx * -Edge2uv.x + dy * Edge1uv.x) * mul,float3(0,0,1));

	Normal = N;	
}

void GetNormal(in const VertexShadingOutput IN, out float3 Normal, out float3 BiNormal,out float3 Tangent)
{
	[@_NORMAL]
	Normal = IN.Normal.xyz;
	[@_TANGENT]
	Tangent= IN.Tangent.xyz;
	[@_TANGENT]
	BiNormal = IN.Binormal.xyz;
}

void InitWorldNormal(in const VertexShadingOutput IN,in out ParticleShadingParameters materialParamters)
{
	float3 Normal = float3(0,1,0);
	float3 BiNormal = float3(1,0,0);
	float3 Tangent = float3(0,0,1);

	//using Input When Normal and Tangent are provided in VertexAssemble
	if(@_TANGENT && @_NORMAL)
		GetNormal(IN,Normal,BiNormal,Tangent);
	else
		GetNormalD(IN,Normal,BiNormal,Tangent);

	float4 rst = tNormalMap.SampleBias(sNormalMap_Sampler, IN.TexCoord.xy, 0);
	float3 tangentNormal = rst.xyz;
	[@_DIRECTX]//always use bc5 for pc
	tangentNormal.z = 1;
	tangentNormal = (2 * tangentNormal) - 1;
	tangentNormal.xy *= NormalScale;

	Normal = normalize(Tangent * tangentNormal.x + BiNormal * tangentNormal.y + Normal * tangentNormal.z);

	materialParamters.WorldNormal = Normal;	
	
}

void InitBaseColorCommon(in const VertexShadingOutput IN,in out ParticleShadingParameters materialParamters)
{

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;
	half4 baseColor = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	baseColor.rgb = DecodeGamma(baseColor.rgb);
	
	materialParamters.BaseColor = baseColor;	
}

void InitFogLightingParticle(/*in float2 stc,*/in const VertexShadingOutput IN, in out float4 OUT)
{
	//issue : IN.DiffLighting.w is different to VolumetricFog-off
// 	if (@_FOG && @_VolumetricFog)
// 	{
// 		// float2 spos = IN.HPosition.xy;
// 		// [@_D3D9]
// 		// spos += 0.5;
// 		// float2 stc = spos * ScreenInfo.zw;
// 		float depth = IN.DiffLighting.w;
// 		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
// 		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
// 	}
	if (@_FOG && !@_VolumetricFog)
	{
		float4 fogFinalColor = float4(0,0,0,0);
		FogLightingAndScatter(IN.TexCoord.w, fogFinalColor.xyz, IN.ViewDirection, SunDirection.xyz);
		OUT.rgb = lerp(fogFinalColor.xyz * OUT.a, OUT.rgb, IN.TexCoord.w);
	}
}

void SoftParticle(in const VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	float2 spos = IN.HPosition.xy;
	[@_D3D9]
	spos += 0.5;

	float2 stc = spos * ScreenInfo.zw;
	float depth = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc));
	depth *= CameraInfo.y;
	float fade = saturate((depth - IN.TexCoord.z) * InvFadeDistance);
	materialParamters.BaseColor *= fade;	
}
#endif //__FILE_PARTICLE_COMMON_FXH__
