///	@file	<Particle_BasicLighing.fx>
///	@path	~/System/Shader/
///	@date	20191205
///	@brief	Particle shading model.

#pragma shader material

// #pragma shader override("@_SOFT", "false")
#pragma shader override("@_DISTORTION", "false")

//////////////////////////////////Properties////////////////////////////////////////
@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = true;
};

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is mask map enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};
@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = true;
}

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World 			;
	float4   InstanceColor		: Color				;
	float4 	cPointCloud[6]		: PointCloudData	;
	float4 	cPlaceHold								;
//	float4x4 WorldViewProj 		: WorldViewProj		;
//	float4   cTintColor1		: TintColor1 		;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float4 MaskUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
	float  NormalScale			 < Widget = "FloatSlider"; Min = "0"; Max = "5.0"; Step = "0.01"; Precision = "3"; > = 1;
	float  InvFadeDistance		 = 1;
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 1;
	float  Roughness			 < Widget = "FloatSlider"; Min = "0.08"; Max = "0.92"; Step = "0.01"; Precision = "3"; > = 0.9;
	float2 UVScale				 = { 1, 1 };
	float  Metallic				  < Widget = "FloatSlider"; Min = "0.001"; Max = "0.999"; Step = "0.01"; Precision = "3"; > = 0.01;
	float  AlphaScale			  < Widget = "FloatSlider"; Min = "0.0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 1;
	float  Thickness			 < Widget = "FloatSlider"; Min = "0.2"; Max = "1.0"; Step = "0.01"; Precision = "3"; > = 1.0;
	float  NearFadeDistance		 = 0;
	float  Scattering			 < Widget = "FloatSlider"; Min = "0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 0.0;
	float  ScatteringPhaseG	 	 < Widget = "FloatSlider"; Min = "0.005"; Max = "0.995"; Step = "0.001"; Precision = "3"; > = 0.7;	
	float  cFov				     = 1;
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture : layout(t3) = "White";
sampler MaskTexture_Sampler : layout(s3)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D tNormalMap : layout(t5) = "NormalMap";
sampler sNormalMap_Sampler : layout(s5)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = FromTexture;
};


texture2D tShadowMask : ShadowMask : layout(t4) = "White";
sampler sShadowMaskSampler : layout(s4)
{
	Texture = tShadowMask;
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

texture2D DepthBuffer : SceneDepth : layout(t6);
sampler DepthBuffer_Sampler : layout(s6)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};
//////////////////////////////////////////////////////////////////////////
#include "Lighting.fxh"
#include "BRDF.fxh"
#include "Particle_Common.fxh"

////////////////////////////////////////////////////////////////////////
// // Common particle vertex shader.

void PS_GetParameters(in VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	InitBaseColorCommon(IN,materialParamters);
	// enable ColorByLife
	materialParamters.BaseColor *= ModulateColor * IN.VertexColor;
	
	InitWorldNormal(IN,materialParamters);

	//Apply Mask Texture
	if (@_MaskEnable)
	{
		float2 maskTC = IN.TexCoord.xy * MaskUVScolling.xy + GameTime * MaskUVScolling.zw;
		float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).r;
		materialParamters.BaseColor.w *= mask;
	}

	materialParamters.Metallic = Metallic;
	materialParamters.Roughness = Roughness;

	//Fade When Near Camera
	float distance = length(CameraPos.xyz - IN.WorldPosition);
	float fadeAlpha = clamp(distance/(NearFadeDistance+0.001),0,1);
	materialParamters.BaseColor.w *= pow(fadeAlpha,3);

	//IN.VertexColor.a -> Element : Alpha By Life
	//issue:Also Affect By Element：Emission By Life / Init Emission 
	materialParamters.BaseColor.w *= IN.VertexColor.a;
}

#include "Particle_Shading.fxh"

technique Shading
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleLightingPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDebugShading();
}

