//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle_Bullet.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	Bullet shading model.

//Warning :
//This Shader is Not used in Graphic.xml
//This Shader is Not Merged Between Versions

#pragma shader material

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////Properties////////////////////////////////////////
@_NormalMapEnable
{
	Name = _NormalMapEnable;
	Description = "Is masking map enabled?";
	AffectedTex = NormalTexture;
	Visible = false;
};

@_DepthMapEnable
{
	Name = _DepthMapEnable;
	Description = "Is masking map enabled?";
	AffectedTex = DepthTexture;
	Visible = false;
};

//////////////////////////////////////////////////////////////////////////
// Bullet shading model.
cbuffer Batch
{
	float4x3 World 				: World 			;
	float4   InstanceColor		: Color				;
#if SD_SH_IBL
	float4 cSHCoefficients[7]	: SHCoefficentsData	;
#else	
	float4 cPointCloud[6]		: PointCloudData	;
#endif
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  BloomScale			 = 0.5;
	float  HoleDepth			 = 0.1;
	float  BumpStrength			 = 0.5;
	float  Roughness			 = 1.0; 
	float  FogStrength			 = 0;
	float  DepthBias			 = -0.001;
	float  cFov					 = 1.0;
};

// Bullet textures.
texture2D DiffuseTexture : layout(t0) = "Transparent";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D NormalTexture : layout(t2);
sampler NormalTexture_Sampler : layout(s2)
{
	Texture = NormalTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthTexture : layout(t3);
sampler DepthTexture_Sampler : layout(s3)
{
	Texture = DepthTexture;
	Filter = Linear;
	Address = Clamp;
};

struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		VertexColor 	: COLOR;
	float2		TexCoord		: TEXCOORDN;
};

//////////////////////////////////////////////////////////////////////////
#define USE_SPECULAR			1
#include "BRDF.fxh"
#include "Lighting.fxh"

/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate, depth and fog for surface.
	float4		WorldPosition	: TEXCOORDN;			//	Output world position.
	float4		VertexColor		: COLOR;				//	Per-vertex color.
};

void InitVertexShadingOutput(inout VertexShadingOutput OUT)
{
	OUT.HPosition = float4(0,0,0,0);
	OUT.TexCoord = float4(0,0,0,0);
	OUT.WorldPosition = float4(0,0,0,0);
	OUT.VertexColor = float4(0,0,0,0);
}
//////////////////////////////////////////////////////////////////////////
// Common Bullet vertex shader.

VertexShadingOutput BulletVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutput(OUT);

	float4 WorldPosition = float4(IN.Position.xyz, 1);
	OUT.WorldPosition = WorldPosition;
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	OUT.HPosition.z -= CameraInfo.x * DepthBias / (DepthBias + OUT.HPosition.w);

	OUT.TexCoord.xy = IN.TexCoord;
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = 0;
	OUT.VertexColor = IN.VertexColor;

	float fog = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);
	OUT.TexCoord.w = pow(fog, FogStrength + 0.001);

	return OUT;
}

//////////////////////////////////////////////////////////////////////////
// Common Bullet pixel shader.

float4 BulletPixelShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT;
	float2 diffuseTC = IN.TexCoord.xy;

	float3 P = ddx(IN.WorldPosition.xyz);
	float3 Q = ddy(IN.WorldPosition.xyz);
	float3 Normal = normalize(cross(Q, P));
	float3 View = normalize(IN.WorldPosition.xyz - CameraPos.xyz);

	float3 Tangent, Bitangent;
	if (@_NormalMapEnable || @_DepthMapEnable)
	{
		// reconstruct tangent space.
		float2 C = ddx(diffuseTC);
		float2 D = ddy(diffuseTC);
		float R = rcp(C.y * D.x - D.y * C.x);
		Tangent = -(P * D.y - Q * C.y) * R;
		Bitangent = (P * D.x - Q * C.x) * R;	
	}

	if (@_DepthMapEnable)
	{
		// Parallax mapping
		float2 offset = float2(dot(Tangent, View), dot(Bitangent, View)) * HoleDepth;
		float depth0 = SampleColorMap(DepthTexture, DepthTexture_Sampler, diffuseTC).x;

#if 0  // simple and fast
		diffuseTC += offset * depth0;
#else  // less effcient and less artifact
		float depth1 = SampleColorMap(DepthTexture, DepthTexture_Sampler, diffuseTC + offset).x;
		diffuseTC += offset * ((depth0 + depth1) * 0.5);
#endif
	}

	if (@_NormalMapEnable)
	{
		float2 bump = SampleColorMap(NormalTexture, NormalTexture_Sampler, diffuseTC).xy;
		bump = (2.0 * bump - 1.0) * BumpStrength;
		Normal = Normal + bump.x * normalize(Tangent) + bump.y * normalize(Bitangent);
		Normal = normalize(Normal);
	}

	// Diffuse lighting
	half3 Lighting = half3(0, 0, 0);
	half InShadow = 0;
	DynamicGILight(Normal, Lighting, InShadow);
	half4 Diffuse = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC) * ModulateColor;
	Diffuse.rgb = DecodeGamma(Diffuse.rgb) * Lighting;

	// Specular lighting
	float3 Reflect = reflect(View, Normal);
	half NoV = saturate(dot(Normal, -View));
	float3 L = SunDirection.xyz;
	half NoL = saturate(dot(Normal, L));
	half3 Specular = SimpleBRDF_SPEC(L, Normal, -View, NoV, NoL, half3(0.6, 0.6, 0.6), Roughness);
	Diffuse.rgb += Specular * Lighting;

	OUT = Diffuse * InstanceColor;
	OUT.rgb = EncodeGamma(OUT.rgb) * OUT.a;
	OUT *= GetVertexColor(IN.VertexColor);

	OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, IN.TexCoord.w);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);

	if(@_HDR)
	{
		EncodeHDRParticle(OUT);
	}
	return OUT;
}

technique Shading
{
	VertexShader = BulletVertexShading();
	PixelShader = BulletPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 BulletDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = BulletVertexShading();
	PixelShader = BulletDebugShading();
}

