//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle_Dissipated.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	Particle shading model.

#pragma shader material

#pragma shader override("@_SOFT", "false")
#pragma shader override("@_DISTORTION", "false")

@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = true;
};

@_TwistEnable
{
	Name = TwistEnable;
	Description = "Is twisting enabled?";
	AffectedTex = TwistTexture;
	Visible = true;
};

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is Mask enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

// @_SoftEdge
// {
// 	Name = SoftEdge;
// 	Description = "Enable Soft Edge";
// 	Visible = true;
// };

// @_FullDissipatedRange
// {
// 	Name = FullDissipatedRange;
// 	Description = "Full";
// 	Visible = true;
// }

// @_AdditiveEdge
// {
// 	Name = AdditiveEdge;
// 	Description = "Edge Blend Mode,true is Additive, false is Blend(lerp)";
// 	Visible = true;
// 	Default = false;
// };

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World				;
//	float4x4 WorldViewProj 		: WorldViewProj		;
	float4   InstanceColor		: Color				;
//	float4   cTintColor1		: TintColor1 	;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  InvFadeDistance		 = 1;
	// float  BloomScale			 = 0.5;
	float2 TwistStrength		 = { 0.1, 0.1 };
	float2 UVScale				 = { 1, 1 };
	float  Softness				 = 0.0;
	float  DissipateEdge		 = 0.01;
	float  HDRScale				 = 1.0;
	float4 DissipateColor		 = { 0, 0, 0, 0 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float4 DissipateUVScolling	 = { 1, 1, 0, 0 };
	float4 MaskUVScolling		 = { 1, 1, 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  cFov			         = 1.0;
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D DissipateTexture : layout(t1) = "White";
sampler DissipateTexture_Sampler : layout(s1)
{
	Texture = DissipateTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture : layout(t2) = "White";
sampler MaskTexture_Sampler : layout(s2)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D TwistTexture : layout(t3) = "White";
sampler TwistTexture_Sampler : layout(s3)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D DepthBuffer : SceneDepth : layout(t4);
sampler DepthBuffer_Sampler : layout(s4)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
#include "Particle_Common.fxh"
/*
Known issue:
1.VertexColor.w
2.ModulateColor.w
3.OUT = float4(rgb,rc)
4.DissipateAlphaBias
5.DissipateTwsit / TwistStrength / DiffuseTwist
*/
float rescale(float vMin,float vMax,float v)
{
	return ((v-vMin) )/ (saturate(vMax - vMin) );
}
float Remap(float original_value, float original_min, float original_max, float new_min, float new_max)
{
	return new_min + (saturate((original_value - original_min) / (original_max - original_min)) * (new_max - new_min));
}


void PS_GetParameters(in VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	float4 OUT = float4(0,0,0,0);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;
	float2 dissipateTC = IN.TexCoord.xy * DissipateUVScolling.xy + GameTime * DissipateUVScolling.zw;

	if(@_TwistEnable)
	{
		float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + GameTime * TwistUVScolling.zw;
		float2 twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;
		diffuseTC += twist * TwistStrength;
		dissipateTC += twist * TwistStrength;
	}

	float mask = 1;    
	if(@_MaskEnable)
	{
		float2 maskTC = IN.TexCoord.xy * MaskUVScolling.xy + GameTime * MaskUVScolling.zw;
		mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).x;	
	}

	float4 diffuseRGBA = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, diffuseTC);
	diffuseRGBA.xyz = diffuseRGBA.rgb * ModulateColor.rgb * InstanceColor.rgb * IN.VertexColor.rgb;
	
	float alpha = SampleColorMap(DissipateTexture, DissipateTexture_Sampler, dissipateTC).r;

	//issue : VertexColor.w is used for both Addtive and Dissipated Control
	float alphaRef =  IN.VertexColor.w;

	//Extend Range of Dissipated Effect
	//Rescale alphaRef from range[0,1] to range[-DissipatedEdge-Softness , 1+DissipatedEdge+Softness]
	alphaRef = alphaRef * 2*(DissipateEdge + Softness + 0.5) - DissipateEdge - Softness;

	/*	rc1->		Aera Alpha=1
		rc2->		Aera Edge + Alpha=1
		rc2-rc1->	Aera Edge
		1-rc1->		Aera Edge + Alpha=0			*/
	float rc1 = 1-smoothstep(alphaRef-Softness,alphaRef+Softness,alpha);
	float rc2 = 1-smoothstep(alphaRef+DissipateEdge-Softness,alphaRef+DissipateEdge+Softness,alpha);
	diffuseRGBA.xyz = lerp(diffuseRGBA.xyz, DissipateColor.xyz ,  1-rc1);
	float tAlpha = saturate((rc2-rc1 )* DissipateColor.w + rc1 * ModulateColor.w ) * mask * diffuseRGBA.w;
	OUT = float4(diffuseRGBA.xyz * tAlpha, tAlpha); 
	// OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, IN.TexCoord.w);

	materialParamters.BaseColor = OUT;
	// materialParamters.BaseColor = float4(IN.VertexColor.xw,0,1);

}

#include "Particle_Shading.fxh"

technique Shading
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleUnifyPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDebugShading();
}

