//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle_Flow.fx>
///	@path	~/System/Shader/
/// @LastChange 2020/05/19
///	@brief	Particle shading model.

#pragma shader material
//#pragma shader category("Effect")

//////////////////////////////////Properties////////////////////////////////////////
@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = true;
};

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is mask map enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

@_TwistEnable
{
	Name = TwistEnable;
	Description = "Is mask map enabled?";
	AffectedTex = TwistTexture;
	Visible = true;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = true;
};

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = false;
}

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World				;
//	float4x4 WorldViewProj 		: WorldViewProj		;
	float4   InstanceColor		: Color				;
//	float4   cTintColor1		: TintColor1 	;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
	float4 FlowUVScolling		 = { 1, 1, 0, 0 };
	float2 FlowStrength		 	 = { 0.5, 0.5 };
	float FlowSpeed				 = 1.0;
	float  BloomScale			 = 0.5;
	float  InvFadeDistance		 = 1;
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 1;
	float2 UVScale				 = {1,1};
	float  cFov					 = 1.0;
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "White";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D TwistTexture : layout(t2) = "White";
sampler TwistTexture_Sampler : layout(s2)
{
	Texture = TwistTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D FlowTexture : layout(t3) = "White";
sampler FlowTexture_Sampler : layout(s3)
{
	Texture = FlowTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D MaskTexture : layout(t4) = "White";
sampler MaskTexture_Sampler : layout(s4)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D ScreenTexture : SceneColor : layout(t5) = "Transparent";
sampler ScreenTexture_Sampler : layout(s5)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t6);
sampler DepthBuffer_Sampler : layout(s6)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
#include "Particle_Common.fxh"

void PS_GetParameters(in VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	float4 OUT = float4(0,0,0,0);

	float2 diffuseTC = IN.TexCoord.xy * DiffuseUVScolling.xy + GameTime * DiffuseUVScolling.zw;
	float2 twistTC = IN.TexCoord.xy * TwistUVScolling.xy + GameTime * TwistUVScolling.zw;
	float2 FlowTC = IN.TexCoord.xy * FlowUVScolling.xy + GameTime * FlowUVScolling.zw;
	float2 Flow = (SampleColorMap(FlowTexture, FlowTexture_Sampler, FlowTC)).xy;
	float2 twist = float2(0,0);
	
	if(@_TwistEnable)
		twist = SampleColorMap(TwistTexture, TwistTexture_Sampler, twistTC).xy;

	diffuseTC += twist * TwistStrength;
	Flow = (Flow * 2 - 1) * FlowStrength;

	float TimeIn = frac(GameTime * FlowSpeed );
	float TimeIn2 = frac(GameTime * FlowSpeed +0.5);

	//Issue : float2 + float2 * float may cause Invalid Subtitle ".xy"
	float4 OUTD1 = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler,  diffuseTC + (Flow * TimeIn));
	float4 OUTD2 = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler,  diffuseTC + (Flow * TimeIn2));

	TimeIn = sin(TimeIn * PI);
	TimeIn2 = sin(TimeIn2 * PI);

	float M = TimeIn + TimeIn2;

	OUT = saturate(OUTD1 * (TimeIn/M) + OUTD2 * (TimeIn2/M) );
	
	OUT.rgb *= OUT.a;
	OUT *= ModulateColor * InstanceColor * IN.VertexColor;

//	LDR Bloom Encode
//	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);

	if (@_MaskEnable)
	{
		float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, IN.TexCoord.xy).r;
		OUT *= mask;
	}

	// OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, IN.TexCoord.w);

	materialParamters.BaseColor = OUT;

}

#include "Particle_Shading.fxh"

technique Shading
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleUnifyPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDebugShading();
}

