//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	Particle shading model.

#pragma shader material

@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = false;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = false;
}

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = false;
}

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World				;
	float4x4 WorldViewProj 		: WorldViewProj		;
	float4   InstanceColor		: Color				;
	float4	 FoliageCenter      : FoliageCenter		;
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  InvFadeDistance		 = 1;
	float  BloomScale			 = 0.5;
	float2 UVScale				 = { 1, 1 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  cFov					 = 1.0;
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "Transparent";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D ScreenTexture : SceneColor : layout(t2);
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Clamp;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};


texture2D tFoliageMap : FoliageMap : layout(t4) = "Black";
sampler sFoliageMapSampler : layout(s4)
{
	Texture = tFoliageMap;
	Filter = Linear;
	Address = Clamp;
	SRGB = false;
};

struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		VertexColor 	: COLOR;
	float2		TexCoord		: TEXCOORDN;
};

//////////////////////////////////////////////////////////////////////////
#include "Lighting.fxh"

/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate, depth and fog for surface.
	float4		VertexColor		: COLOR;				//	Per-vertex color.
	[@_DISTORTION]
	float3		WorldPos		: TEXCOORDN;			//	WorldPos
	[@_DISTORTION]
	float3		ViewDirection	: TEXCOORDN;			//	View dir
};

//////////////////////////////////////////////////////////////////////////
// Common particle vertex shader.

VertexShadingOutput ParticleVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;

	float4 InPosition = float4(IN.Position.xyz, 1);
	float4 WorldPosition;
	
	float3 modelPos = MulMat3x4(World, float4(0,0,0,1)).xyz;
	float2 uv = (modelPos.xz-FoliageCenter.zw)/128.0+0.5;
	float2 burnParams = tFoliageMap.SampleLevel(sFoliageMapSampler, uv, 0).xy;
	uv = step(abs(uv - 0.5), float2(0.5,0.5));
	burnParams = burnParams * uv.x * uv.y;
	float burnParam = saturate(burnParams.x);
	
	InPosition *= step(0.1, burnParam);
	WorldPosition *= step(0.1, burnParam);

	if (@_LOCAL_SPACE)  // input pos in local space
	{
		WorldPosition = float4(MulMat3x4(World, InPosition).xyz, 1);
		OUT.HPosition = MulMat4x4(WorldViewProj, InPosition);
	}
	else  // input pos in world space
	{
		WorldPosition = InPosition;
		OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	}

	OUT.HPosition.z -= DepthBias / (DepthBias + OUT.HPosition.w);

	OUT.TexCoord.xy = IN.TexCoord * UVScale;
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = 0;
	OUT.VertexColor = IN.VertexColor;

	if (@_DISTORTION)
	{
		OUT.WorldPos = WorldPosition.xyz - CameraPos.xyz;
		OUT.ViewDirection = normalize(CameraPos.xyz - WorldPosition.xyz);
	}
	else
	{
		[@_HDR]
		OUT.VertexColor.rgb *= HDRScale;

		float fog = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);
		OUT.TexCoord.w = pow(fog, FogStrength + 0.001);

		if (@_VolumetricFog)
		{
			OUT.TexCoord.w = OUT.HPosition.w;
		}
	}

	return OUT;
}


//////////////////////////////////////////////////////////////////////////
// Common particle pixel shader.

float4 ParticlePixelShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT;
	if (@_DISTORTION)
	{
		float2 spos = screenCoord.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;

		float H = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy).r;
		H *= ModulateColor.x * InstanceColor.x * GetVertexColor(IN.VertexColor).x;

		// calcuate patch normal
		float3 P = IN.WorldPos;
		float3 N = cross(ddy(P), ddx(P));
		if (@_GLSL || @_SPIRV)
		{
			// ddx/ddy is incorrect on some android devices.
			float len = length(N);
			N = (len == 0) ? float3(0, 1, 0) : N / len;
		}
		else
		{
			N = normalize(N);
		}

		// hacked refraction
		float3 V = IN.ViewDirection;
		float3 R = cross(N, V);
		float3 D = cross(R, V);

		float3 VR = cross(float3(0, 1, 0), V);
		float3 VU = normalize(cross(V, VR));
		VR = cross(VU, V);

		float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
		offset *= (500 / IN.TexCoord.z) * H;

		OUT = SampleColorMap(ScreenTexture, ScreenTexture_Sampler, stc + offset);
		OUT.a = EncodeTransparentBloomFactor(1, 0);
	}
	else if (@_TRANSPARENT)
	{
		OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
		OUT *= ModulateColor * InstanceColor;

		if(@_HDR && @_ENCODEHDR)
			OUT.a *= OUT.a;

		OUT.rgb *= OUT.a;
		OUT *= GetVertexColor(IN.VertexColor);

		if (@_SOFT)
		{
			float2 spos = screenCoord.xy;
			[@_D3D9]
			spos += 0.5;

			float2 stc = spos * ScreenInfo.zw;
			float depth = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, stc));
			depth *= CameraInfo.y;
			float fade = saturate((depth - IN.TexCoord.z) * InvFadeDistance);
			OUT *= fade;
		}
		
		// fog shading
		if (@_VolumetricFog)
		{
			float2 spos = IN.HPosition.xy;
			[@_D3D9]
			spos += 0.5;
			float2 stc = spos * ScreenInfo.zw;
			float depth = IN.TexCoord.w;
			float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
			OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz * OUT.a;
		}
		else
		{
			OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, IN.TexCoord.w);
		}
		OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);
	}
	else
	{
		OUT = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
		OUT *= ModulateColor * InstanceColor;
		OUT.xyz *= GetVertexColor(IN.VertexColor).xyz;
		if (@_VolumetricFog)
		{
			float2 spos = IN.HPosition.xy;
			[@_D3D9]
			spos += 0.5;
			float2 stc = spos * ScreenInfo.zw;
			float depth = IN.TexCoord.w;
			float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
			OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
		}
		else
		{
			OUT.rgb = lerp(FogColor.rgb, OUT.rgb, IN.TexCoord.w);
		}
		float lum = GetLuminance(OUT.rgb);
		OUT.a = EncodeSolidBloomFactor(lum * BloomScale);

		//ApplyScreenColorLum(OUT, lum);
	}

	if(!@_DISTORTION)
	{
		EncodeHDRParticle(OUT);
	}
	return OUT;
}

technique Shading
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticlePixelShading();
}


//////////////////////////////////////////////////////////////////////////
// technique Distortion.

float4 ParticleDistortionShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	if (@_DISTORTION)
	{
		float2 spos = screenCoord.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;

		float H = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy).r;
		H *= ModulateColor.x * InstanceColor.x * GetVertexColor(IN.VertexColor).x;

		// calcuate patch normal
		float3 P = IN.WorldPos;
		float3 N = cross(ddy(P), ddx(P));
		if (@_GLSL || @_SPIRV)
		{
			// ddx/ddy is incorrect on some android devices.
			float len = length(N);
			N = (len == 0) ? float3(0, 1, 0) : N / len;
		}
		else
		{
			N = normalize(N);
		}

		// hacked refraction
		float3 V = IN.ViewDirection;
		float3 R = cross(N, V);
		float3 D = cross(R, V);

		float3 VR = cross(float3(0, 1, 0), V);
		float3 VU = normalize(cross(V, VR));
		VR = cross(VU, V);

		float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
		offset *= (500 / IN.TexCoord.z) * H;
		offset = offset * 0.5 + 0.5;

		EncodeFloatRG(offset.x, OUT.rg);
		EncodeFloatRG(offset.y, OUT.ba);
	}
	return OUT;
}

technique Distortion
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticleDistortionShading();
};

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticleDebugShading();
}

