//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	Particle shading model.

#pragma shader material

@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = false;
};

@_DISTORTION
{
	Name = Distortion;
	Description = "Distortion effect";
	Visible = false;
}

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = false;
}

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	// float4x3 Local 				: InvWorld			;
	float4x3 World 				: World				;
//	float4x4 WorldViewProj 		: WorldViewProj		;
	float4   InstanceColor		: Color				;
//	float4   cTintColor1		: TintColor1 	;
};

cbuffer Shader
{
	float4 LaserPointColor	< Visible = "0" ; >	 = { 1, 1, 1, 1 };
	float  InvFadeDistance  < Visible = "0"; > = 1;
	float  BloomScale		< Visible = "0"; > = 0.5;
	float2 UVScale			< Visible = "0"; > = { 1, 1 };
	float  FogStrength		< Visible = "0"; > = 0;
	float  DepthBias		< Visible = "0"; > = 0;
	float  HDRScale			< Visible = "0"; > = 2.0;
	float  cFov				< Visible = "0"; > = 1.0;
	float  cLaserPointScale < Widget = "FloatSlider" ; Max = "1" ; Min = "0" ; Step="0.01" ; Visible = "0" ; >      = 1.0;
	float  cLaserIntensityScale < Widget = "FloatSlider" ; Max = "100" ; Min = "0" ; Step="0.1" ; >      = 100.0;
	float4  cModulateColor  = { 1, 1, 1, 1 };
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "Transparent";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
};

texture2D ScreenTexture : SceneColor : layout(t2);
sampler ScreenTexture_Sampler : layout(s2)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

//////////////////////////////////////////////////////////////////////////
#include "Particle_Common.fxh"

void InitVertexShadingOutputForLaser(inout VertexShadingOutput OUT)
{
	OUT.HPosition = float4(0,0,0,0);
	OUT.TexCoord = float4(0,0,0,0);
	OUT.VertexColor = half4(0,0,0,0);
	OUT.ViewDirection = half3(0,0,0);
	OUT.WorldPosition = half3(0,0,0);

	[@_NORMAL]
	OUT.Normal = half4(0,0,0,0);

	[@_TANGENT]
	OUT.Tangent = float4(0,0,0,0);
	[@_TANGENT]
	OUT.Binormal = float4(0,0,0,0);
}


VertexShadingOutput ParticleUnifyVertexShadingForLaser(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	InitVertexShadingOutputForLaser(OUT);

	float4 InPosition = float4(IN.Position.xyz, 1);
	float4 WorldPosition = float4(0,0,0,0);

	if (@_LOCAL_SPACE)  // input pos in local space
	{
		WorldPosition = float4(MulMat3x4(World, InPosition).xyz, 1);
	}
	else  // input pos in world space
	{
		WorldPosition = InPosition;
		// LocalPosition = float4(MulMat3x4(Local, float4(InPosition.xyz,1)),1);
	}


	/////////////////////// start scale ///////////////////////
	/*
		__
		^ |
		| |                      
		  |					       |
		h1|					     h0|
		  |                        |
		| |                        |                \
		v |________________________|_____________a_(_\ camera
		                           | <-     d0     ->|
		--|<-                    d                 ->|

           h1          h0 
        ________  = _______
           d           d0
                      
                        d
         h1 = h0 * (  ____  )
                       d0

         h1 -> geometry radius size
         
         d  -> distance from object positon to camera position 
         d0 -> a scaled value to adjust angle a
	*/

	// scale particle to fixed size on screen
	float4 ObjectWorldPos = float4(MulMat3x4(World, float4(0,0,0,1)).xyz, 1);
	float objectDistanceToViewPoint = length(CameraPos.xyz-ObjectWorldPos.xyz);

	IN.VertexColor *= pow(0.8,objectDistanceToViewPoint*((100.-cLaserIntensityScale)/100.));


	float dynamicScale = cLaserPointScale/10 /*scale value*/ *(objectDistanceToViewPoint/*LaserPointBaseForwardDistance*/);

	// get local position , as Local in Particle will cause crash ! so can`t use it. 
	float3 LocalPosition = (WorldPosition - ObjectWorldPos).xyz;
	LocalPosition.xyz *= dynamicScale;

	// reback to world position 
	WorldPosition = float4(MulMat3x4(World, float4(LocalPosition,1)).xyz, 1);

	/////////////////////// end scale ///////////////////////

	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	OUT.HPosition.z -= CameraInfo.x * DepthBias / (DepthBias + OUT.HPosition.w);
	OUT.HPosition.w *= cFov;
	OUT.TexCoord.xy = IN.TexCoord * UVScale;
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = 1;

	OUT.VertexColor = GetVertexColor(IN.VertexColor);

	//WorldPosition.xyz = WorldPosition.xyz - CameraPos.xyz;
	OUT.WorldPosition = WorldPosition.xyz ;


	OUT.ViewDirection = normalize(CameraPos.xyz - WorldPosition.xyz);
	
	[@_HDR]
	OUT.VertexColor.rgb *= HDRScale;

	OUT.TexCoord.w  = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);

	if (@_VolumetricFog)
	{
		OUT.TexCoord.w = OUT.HPosition.w;
	}	
	if (@_LOCAL_SPACE)  // input pos in local space
	{
		if (@_NORMAL)
		{
			//N4B -> Normal in VertexAssemble is Unsigned Byte
			//issue World Matrix using as Normal Transfer May be Wrong
			//https://zhuanlan.zhihu.com/p/72734738
			OUT.Normal = half4((2 * IN.Normal.xyz - 1).xyz,0);
			OUT.Normal =  half4( normalize( MulMat3x4(World, OUT.Normal ).xyz) ,0);
		}
		if (@_TANGENT)
		{
			OUT.Tangent = float4( normalize( MulMat3x4(World, IN.Tangent ).xyz) ,0);
			OUT.Binormal = float4( normalize( MulMat3x4(World, IN.Binormal ).xyz) ,0);
		}
	}
	else
	{
		if (@_NORMAL)
		{
			OUT.Normal = 2 * IN.Normal - 1;
		}
		if (@_TANGENT)
		{
			OUT.Tangent = IN.Tangent;
			OUT.Binormal = IN.Binormal;
		}
	}
    
	return OUT;
}


//	  Almost Same as Particle.fx::ParticlePixelShading()
void PS_GetParameters(in VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	float4 OUT = float4(1,0,1,1);
	OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);

	OUT *= LaserPointColor * cModulateColor ;// * IN.LaserItensity;

	if(@_HDR && @_ENCODEHDR)
		OUT.a *= OUT.a;

	OUT.rgb *= OUT.a;
	OUT *= IN.VertexColor;

	OUT.a = EncodeTransparentBloomFactor(OUT.a, GetLuminance(OUT.rgb) * BloomScale);

	materialParamters.BaseColor = OUT;

}

#include "Particle_Shading.fxh"

technique Shading
{
	VertexShader = ParticleUnifyVertexShadingForLaser();
	PixelShader = ParticleUnifyPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Distortion.

float4 ParticleDistortionShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	if (@_DISTORTION)
	{
		float2 spos = screenCoord.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;

		float H = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy).r;
		H *= LaserPointColor.x * InstanceColor.x * IN.VertexColor.x;

		// calcuate patch normal
		float3 P = IN.WorldPosition;
		float3 N = cross(ddy(P), ddx(P));
		if (@_GLSL || @_SPIRV)
		{
			// ddx/ddy is incorrect on some android devices.
			float len = length(N);
			N = (len == 0) ? float3(0, 1, 0) : N / len;
		}
		else
		{
			N = normalize(N);
		}

		// hacked refraction
		float3 V = IN.ViewDirection;
		float3 R = cross(N, V);
		float3 D = cross(R, V);

		float3 VR = cross(float3(0, 1, 0), V);
		float3 VU = normalize(cross(V, VR));
		VR = cross(VU, V);

		float2 offset = float2(dot(D, VR), -dot(D, VU)) * ScreenInfo.zw;
		offset *= (500 / IN.TexCoord.z) * H;
		offset = offset * 0.5 + 0.5;

		EncodeFloatRG(offset.x, OUT.rg);
		EncodeFloatRG(offset.y, OUT.ba);
	}
	return OUT;
}

technique Distortion
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDistortionShading();
};

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDebugShading();
}

