//	--------------------------------------------------------------------
//	Copyright(C) 2006~2016 NetEase. All rights reserved.
//	This file is a part of the project Messiah.
//	Project Messiah
//	Contact : wuyuwen@corp.netease.com
//	Author: wuyuwen
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Particle.fx>
///	@path	~/System/Shader/
///	@date	2014/11/10
///	@brief	Particle shading model.
#pragma shader material

@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = false;
};
@_EnableEditor
{
	Name = EnableEditor;
	Description = "Show in editor";
	Default = false;
	Visible = true;
};

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World 		: layout(vs);
	float4   InstanceColor		: Color			: layout(ps);
};

cbuffer Shader
{

	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float  BloomScale			 = 0.5;
	float2 UVScale				 = { 1, 1 };
	float3 lensOffset			 < Widget = "Float3Slider";Step = "0.0001";Precision = "4";> = { 0, 0, 0 };
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 2.0;
	float  cCameraYaw			 = 0;
	float  cMaxDistance			 = 200;
	float  cFadepow				 = 1;
	float  cLensDepthBias        = 1;

	float3 cCameraPos  			< Visible = "0";>;
	float3 cCameraDir 			< Visible = "0";>;
	float3 cParticlePos			< Visible = "0";>;
	float3 cParticleDir			< Visible = "0";>;
};

// Particle textures.
texture2D DiffuseTexture : layout(t0) = "Transparent";
sampler DiffuseTexture_Sampler : layout(s0, s1)
{
	Texture = DiffuseTexture;
	Filter = Linear;
	Address = Wrap;
	RelaxedPrecision = true;
};

texture2D ScreenTexture : SceneColor : layout(t5) = "Transparent";
sampler ScreenTexture_Sampler : layout(s5)
{
	Texture = ScreenTexture;
	Filter = Linear;
	Address = Mirror;
};

texture2D DepthBuffer : SceneDepth : layout(t6);
sampler DepthBuffer_Sampler : layout(s6)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};

struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		VertexColor 	: COLOR;
	float2		TexCoord		: TEXCOORDN;
};

/// Result after shading vertex process.
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate, depth and fog for surface.
	float4		VertexColor		: COLOR;				//	Per-vertex color.
	float4 		PivotCoord		: TEXCOORDN;
};

/////////////////////////////////Function////////////////////////////////////////
float3 Gamma2Linear(float3 sRGB)
{
	return  (sRGB * (sRGB * (sRGB * 0.305306011 + 0.682171111) + 0.012522878));
}


float PixelShadingCalCameraFade(float3 v,float3 p,float3 dir,float3 d,in VertexShadingOutput IN)
{
	if(@_EnableEditor)
	{
		
		return 1;
	}
	else
	{
		float3 dv = normalize(p - d);
		float3 cv = -v;

		float d1 = max(dot(dv,cv),0);
		d1 = pow(d1,0.1);

		float d2 = max(dot(dv,dir),0);
		d2 = pow(d2,2);

		float d3 = distance(p,d);
		d3 = 1 - clamp(d3/200,0,1);

		return (d1*d2*d3);
	}
	
}


//////////////////////////////////////////////////////////////////////////
// Common particle vertex shader.

VertexShadingOutput ParticleVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;
	[@_HLSL]
	OUT = (VertexShadingOutput)0;
	
	float4 WorldPosition;
	float3 vDir = normalize(cCameraPos.xyz - cParticlePos.xyz - lensOffset - cParticleDir*0.07);

	if (@_LOCAL_SPACE)  // input pos in local space
	{
		WorldPosition = float4(MulMat3x4(World, float4(IN.Position.xyz,1)).xyz, 1);
	}
	else  // input pos in world space
	{
		WorldPosition = float4(IN.Position.xyz, 1);
	}

	float4 PivotPosition = float4(cParticlePos + lensOffset + cParticleDir*0.07 + vDir*cLensDepthBias,1);

	OUT.PivotCoord = MulMat4x4(ViewProj, PivotPosition);

	float2 screenUV = OUT.PivotCoord.xy / (2 * OUT.PivotCoord.w) + 0.5;
	[!@_GLSL]
	screenUV.y = 1 - screenUV.y;

	OUT.PivotCoord.xy = screenUV.xy;


	//Depth offset
	float cDepthbias = distance(cCameraPos.xyz,PivotPosition.xyz);
	
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);
	//OUT.HPosition.z -= -cDepthbias / (cDepthbias + OUT.HPosition.w);
	OUT.HPosition.z = cDepthbias;
	//OUT.HPosition.w = max(OUT.HPosition.w, 1);
	//OUT.HPosition.z = min(OUT.HPosition.z, (1-OUT.HPosition.w*CameraInfo.z)*OUT.HPosition.w);

	OUT.TexCoord.xy = IN.TexCoord * UVScale;
	OUT.TexCoord.z = OUT.HPosition.w;
	OUT.TexCoord.w = 0;

	OUT.VertexColor = GetVertexColor(IN.VertexColor);

	float fog = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);
	OUT.TexCoord.w = pow(fog, FogStrength + 0.001);

	OUT.PivotCoord.w = distance(cCameraPos.xyz,PivotPosition.xyz);
	 
	return OUT;
}


//////////////////////////////////////////////////////////////////////////
// Common particle pixel shader.

float4 ParticlePixelShading(in VertexShadingOutput IN, in float4 screenCoord : SV_Position) : SV_Target
{
	half4 OUT;
	if (@_TRANSPARENT)
	{
		OUT = SampleColorAlphaMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
		OUT.rgb = Gamma2Linear(OUT.rgb);
		OUT *= ModulateColor * InstanceColor;
		OUT.rgb *= OUT.a;
		OUT *= IN.VertexColor;
		OUT.a = saturate(OUT.a);
		OUT.rgb = lerp(FogColor.rgb * OUT.a, OUT.rgb, IN.TexCoord.w);
		EncodeHDR(OUT);
	}
	else
	{
		OUT = SampleColorMap(DiffuseTexture, DiffuseTexture_Sampler, IN.TexCoord.xy);
		OUT.rgb = Gamma2Linear(OUT.rgb);
		OUT *= ModulateColor * InstanceColor;
		OUT.xyz *= GetVertexColor(IN.VertexColor).xyz;
		OUT.rgb = lerp(FogColor.rgb, OUT.rgb, IN.TexCoord.w);
		EncodeHDR(OUT);
		OUT.a = 1;
	}

	//camera lens flare
	float3 vDir = normalize(cCameraPos.xyz - cParticlePos.xyz);
	half Fade = PixelShadingCalCameraFade(cCameraDir,cCameraPos,cParticleDir,cParticlePos,IN);
	//OUT.a = 1;
	half Fade3 = Fade * Fade * Fade;
	//Fade3 = Fade3 > 0.5 ? Fade3 : 0;
	OUT.rgba*= pow(Fade3,cFadepow);
	OUT.rgb *= ReflectionScale;

	float2 spos = IN.PivotCoord.xy;
	//float2 stc = spos * ScreenInfo.zw;
	float depth = DecodeDepth(SampleColorMap(DepthBuffer, DepthBuffer_Sampler, spos));
	depth *= CameraInfo.y;
	depth =(IN.PivotCoord.w > depth ? 0 : 1) ; 
	// OUT.rgb = depth;
	// OUT.a = 1;
	OUT.rgba *= depth;

	return OUT;
}

technique Shading
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticlePixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	//OUT = EncodeTransparentHDR(OUT);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleVertexShading();
	PixelShader = ParticleDebugShading();
}

