///	@file	<Particle_Smoke.fx>
///	@path	~/System/Shader/
///	@date	20200709
///	@brief	Particle shading model.


#pragma shader material

//SOFT Particle Should Be Allowed in Smoke Rendering
#pragma shader override("@_DISTORTION", "false")
//////////////////////////////////Properties////////////////////////////////////////
@_LOCAL_SPACE
{
	Name = Local;
	Description = "Input particle position is in local space";
	Visible = true;
};

@_SOFT
{
	Name = Soft;
	Description = "Soft particle";
	Visible = false;
};

@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is mask map enabled?";
	AffectedTex = MaskTexture;
	Visible = true;
};

@_EnableEmissiveCurve
{
	Name = EnableEmissive;
	Description = "Check EnableEmissive, using PackedSmokeTexture's B Channel for Emsissive";
	Visible = true;
	Default = false;
};

#include "Constants.fxh"
#include "Common.fxh"

//////////////////////////////////////////////////////////////////////////
// Particle shading model.
cbuffer Batch
{
	float4x3 World 				: World 			;
	float4   InstanceColor		: Color				;
	float4 	cPointCloud[6]		: PointCloudData	;
	float4 	cPlaceHold								;
//	float4x4 WorldViewProj 		: WorldViewProj		;
//	float4   cTintColor1		: TintColor1 		; //Removed in New Engine : common.fxh/ApplyScreenColor()
};

cbuffer Shader
{
	float4 ModulateColor		 = { 1, 1, 1, 1 };
	float4 DiffuseUVScolling	 = { 1, 1, 0, 0 };
	float4 TwistUVScolling		 = { 1, 1, 0, 0 };
	float4 MaskUVScolling		 = { 1, 1, 0, 0 };
	float2 TwistStrength		 = { 0.1, 0.1 };
	float  NormalScale			 < Widget = "FloatSlider"; Min = "0"; Max = "5.0"; Step = "0.01"; Precision = "3"; > = 1;
	float  InvFadeDistance		 = 1;
	float  FogStrength			 = 0;
	float  DepthBias			 = 0;
	float  HDRScale				 = 1;
	float  Roughness			 < Widget = "FloatSlider"; Min = "0.08"; Max = "0.92"; Step = "0.01"; Precision = "3"; > = 0.9;
	float2 UVScale				 = { 1, 1 };
	float  Metallic				  < Widget = "FloatSlider"; Min = "0.001"; Max = "0.999"; Step = "0.01"; Precision = "3"; > = 0.001;
	float  AlphaScale			  < Widget = "FloatSlider"; Min = "0.0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 1;

	//For Emissive Color : using Color By Life
	float4 SmokeColor			 < Widget = "Color4"; > = {0.8, 0.8, 0.8, 1};
	float  EmissiveScale		 = 1.0;
	float  Thickness			 < Widget = "FloatSlider"; Min = "0.2"; Max = "1.0"; Step = "0.01"; Precision = "3"; > = 0.5;
	float  NearFadeDistance		 = 0;
	float  Scattering			 < Widget = "FloatSlider"; Min = "0"; Max = "2.0"; Step = "0.001"; Precision = "3"; > = 0.0;
	float  ScatteringPhaseG	 	 < Widget = "FloatSlider"; Min = "0.005"; Max = "0.995"; Step = "0.001"; Precision = "3"; > = 0.7;	
	float  cFov					 = 1.0;
};

// Particle textures.
texture2D MaskTexture : layout(t0) = "White";
sampler MaskTexture_Sampler : layout(s0)
{
	Texture = MaskTexture;
	Filter = Linear;
	Address = FromTexture;
};

texture2D tPackedSmokeMap : layout(t1) = "White";
sampler sPackedSmokeMap_Sampler : layout(s1)
{
	Texture = tPackedSmokeMap;
	Filter = Linear;
	Address = FromTexture;
};
/*
	[xy__] Normalized Normal Info
	[__z_] Alpha Info
	[___w] Emissive Info

	Alpha Channel is more dependency than rgb channel
	so change Alpha info to Z and Emissive Info to W 
*/

texture2D tShadowMask : ShadowMask : layout(t2) = "White";
sampler sShadowMaskSampler : layout(s2)
{
	Texture = tShadowMask;
	Filter = Point; 
	Address = Clamp;
	SRGB = false;
	MaxMipLevel = 0;
};

//for Soft Particle
texture2D DepthBuffer : SceneDepth : layout(t3);
sampler DepthBuffer_Sampler : layout(s3)
{
	Texture = DepthBuffer;
	Filter = Point;
	Address = Clamp;
};
//////////////////////////////////////////////////////////////////////////
#include "Lighting.fxh"
#include "BRDF.fxh"
#include "Particle_Common.fxh"

////////////////////////////////////////////////////////////////////////
// // Common particle vertex shader.

void InitWorldCustomNormal(in VertexShadingOutput IN,in out ParticleShadingParameters materialParamters)
{
	//Get Normal using ddx() / ddy()
	float3 Normal = float3(0,1,0);
	float3 BiNormal = float3(1,0,0);
	float3 Tangent = float3(0,0,1);
	if(@_TANGENT && @_NORMAL)
		GetNormal(IN,Normal,BiNormal,Tangent);
	else
		GetNormalD(IN,Normal,BiNormal,Tangent);

	float3 tangentNormal = materialParamters.WorldNormal.xyz;
	[@_DIRECTX]//always use bc5 for pc
	tangentNormal.z = 0.5;
	tangentNormal.xy = (tangentNormal.xy-0.5) *2.0;
	tangentNormal.z = sqrt(1 - saturate(tangentNormal.x * tangentNormal.x + tangentNormal.y * tangentNormal.y));
	tangentNormal.xy *= NormalScale;
	// tangentNormal = normalize(tangentNormal);
	
	// tangentNormal = normalize(MulMat3x4((float3x3)World, tangentNormal));

	Normal = normalize(Tangent * tangentNormal.x + BiNormal * tangentNormal.y + Normal * tangentNormal.z);

	materialParamters.WorldNormal = Normal;	

	// materialParamters.WorldNormal = tangentNormal;
}

half Remap(half original_value, half original_min, half original_max, half new_min, half new_max)
{
	return new_min + (saturate((original_value - original_min) / (original_max - original_min)) * (new_max - new_min));
}

void PS_GetParameters(in VertexShadingOutput IN,inout ParticleShadingParameters materialParamters)
{
	float4 packedInfo = SampleColorMap(tPackedSmokeMap, sPackedSmokeMap_Sampler, IN.TexCoord.xy);
	// packedInfo.xy += step(packedInfo.w,0) * 0.5;

	if(@_EnableEmissiveCurve)
	{
		materialParamters.BaseColor.xyz = SmokeColor.xyz;
		materialParamters.Emission = (IN.VertexColor.xyz * packedInfo.w * EmissiveScale);
	}
	else
	{
		//Temp
		// packedInfo.z = saturate(packedInfo.z - 0.05);
		packedInfo.w = Remap(packedInfo.w,0.1,1.0,0.0,1.0);
		float invEmissiveIntensity = pow(1 - packedInfo.w,0.333);
		float3 tst = float3(1,1.0/4.0,1.0/16.0) / exp(4.0*(invEmissiveIntensity) - 1.0);
		materialParamters.Emission = tst.xyz * EmissiveScale * packedInfo.w;
		// materialParamters.Emission = tst.xyz * EmissiveScale * packedInfo.z;

		materialParamters.BaseColor.xyz = IN.VertexColor.xyz;
	}
	materialParamters.BaseColor.w = packedInfo.z;

	//Fade When Near Camera
	float distance = length(CameraPos.xyz - IN.WorldPosition);
	float fadeAlpha = clamp(distance/(NearFadeDistance+0.001),0,1);
	materialParamters.BaseColor.w *= fadeAlpha;

	materialParamters.WorldNormal.xy = packedInfo.xy;
	InitWorldCustomNormal(IN,materialParamters);

	//Apply Mask Texture
	if (@_MaskEnable)
	{
		float2 maskTC = IN.TexCoord.xy * MaskUVScolling.xy + GameTime * MaskUVScolling.zw;
		float mask = SampleColorMap(MaskTexture, MaskTexture_Sampler, maskTC).r;
		materialParamters.BaseColor.w *= mask;
	}

	materialParamters.Metallic = Metallic;
	materialParamters.Roughness = Roughness;

	//IN.VertexColor.a -> Element : Alpha By Life
	materialParamters.BaseColor.w *= IN.VertexColor.a;
}


#include "Particle_Shading.fxh"

technique Shading
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleLightingPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// technique Debug.

float4 ParticleDebugShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0.1, 0.1, 0.05, 0.1);
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);
	return OUT;
}

technique Debug
{
	VertexShader = ParticleUnifyVertexShading();
	PixelShader = ParticleDebugShading();
}

