//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : liwenyao@corp.netease.com
//	Author: liwenyao
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Lava.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_UNLIT 1

#include "Constants.fxh"
#include "Common.fxh"
#include "Lighting.fxh"

@_UseDiyFog
{
	Name = UseDiyFog;
	Description = "use diy fog?";
	Visible = true;
};

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 				;
	float4 cParamter2				: Param2				;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex	;
};

cbuffer Shader
{
	float cFogDistance				 < Widget = "FloatSlider"; Min = "0"; Max = "1000"; Step = "0.1"; Precision = "2"; > = 400;
};

// Surface textures.
texture2D tDiffuseMap : layout(t0);
sampler sDiffuseSampler : layout(s0, s1)
{
	Texture = tDiffuseMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float2		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.

	float3		ViewDirection	: TEXCOORDN;			//	View direction eye->worldPos.

	float		T				: TEXCOORDN;			//	FogTransmitance
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Normal			: NORMAL;
	float2		TexCoord		: TEXCOORDN;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//TexCoord
	OUT.TexCoord.xy = IN.TexCoord.xy;

	//ViewDirection
	OUT.ViewDirection = WorldPosition.xyz - CameraPos.xyz;
	
	//Fog T
	if(@_UseDiyFog)
		OUT.T = FogTransmittance(normalize(OUT.ViewDirection.xyz)*cFogDistance);
	else
		OUT.T = FogTransmittance(OUT.ViewDirection.xyz);

	if (@_VolumetricFog)
	{
		OUT.T = OUT.HPosition.w;
	}

	if (@_NoFarPlaneClip)
	{
		OUT.HPosition.z = max(EPSILON*OUT.HPosition.w, OUT.HPosition.z);
	}
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput BillboardVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float4 BillboardPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	
	OUT = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TexCoord.xy);

	// Clipping
	if (@_CLIPPING && @_ES20)
		clip(OUT.a - 0.5);

	// thunder
	OUT.rgb *= ThunderRate;	

	// fog shading
	if (@_VolumetricFog)
	{
		float2 spos = IN.HPosition.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.T;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
	}
	else
		OUT.rgb = FogLighting(IN.T, OUT.rgb);

	// screen color tint
	OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);

	OUT.rgb = saturate(OUT.rgb) * OUT.a;
	OUT.a = EncodeTransparentBloomFactor(OUT.a, 0);

	return OUT;
}

technique Shading
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Clip
float4 ClipPixelShading(in VertexShadingOutput IN) : SV_Target
{
	if (@_CLIPPING)
	{
		half a = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TexCoord.xy).a;
		clip(a - 0.5);
	}
	return float4(0, 0, 0, 0);
}

technique Clip
{
	VertexShader = BillboardVertexShading();
	PixelShader = ClipPixelShading();
}


//////////////////////////////////////////////////////////////////////////
// Wireframe
[maxvertexcount(3)]
void WireframeGeometryShading(triangle VertexShadingOutput IN[3], inout TriangleStream<VertexShadingOutput> OUT)
{
	IN[0].ViewDirection = float3(1, 0, 0);
	OUT.Append(IN[0]);
	IN[1].ViewDirection = float3(0, 1, 0);
	OUT.Append(IN[1]);
	IN[2].ViewDirection = float3(0, 0, 1);
	OUT.Append(IN[2]);

	OUT.RestartStrip();
}

float4 WireframePixelShading(in VertexShadingOutput IN) : SV_Target
{
	float3 difference = fwidth(IN.ViewDirection);
	float3 distance = IN.ViewDirection / difference;
	float d = min(min(distance.x, distance.y), distance.z);
	float a = exp2(-1.5 * d * d);
	return float4(cParamter2.xyz, a);
}

technique Wireframe
{
	VertexShader = BillboardVertexShading();
	GeometryShader = WireframeGeometryShading();
	PixelShader = WireframePixelShading();
};

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);
	
	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = BillboardVertexShading();
	PixelShader = BillboardPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	float clipValue = 1.0;
	if (@_CLIPPING)
	{
		half a = tDiffuseMap.SampleRGBA(sDiffuseSampler, IN.TC.xy).a;
		clip(a - 0.5);
	}

	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}

