//	--------------------------------------------------------------------
//	Copyright(C) 2006~2020 NetEase Games. All rights reserved.
//	This file is a part of the project Messiah.
//	Messiah Engine, Technology Department of NetEase Games.
//	Contact : jiawei@corp.netease.com
//	Author: jiawei
//	$(reservedDeclaration)
//	--------------------------------------------------------------------
///	@file	<Effect.fx>
///	@path	~/System/Shader/
///	@date	2015/01/12
///	@brief	Physically based shading model.

#pragma shader material
#define MATERIAL_SHADINGMODEL_UNLIT 1

//////////////////////////////////Properties////////////////////////////////////////
@_NormalMapEnable  
{
	Name = NormalMapEnable;
	Description = "Do we have NormalMap?"; 
	AffectedTex = tNormalMap;
	Fallback = "@_TANGENT";
	Visible = true;
}; 
@_LerpEnable
{
	Name = LerpEnable;
	Description = "Is Lerp texture enabled?";
	Visible = true;
};
@_MaskEnable
{
	Name = MaskEnable;
	Description = "Is mask texture enabled?";
	Visible = true;
};
@_DistortEnable
{
	Name = DistortEnable;
	Description = "Is Distort texture enabled?";
	Visible = true;
};

#include "Constants.fxh"
#include "Common.fxh"  
#include "BRDF.fxh"
#include "Lighting.fxh" 

//////////////////////////////////Constant buffer////////////////////////////////////////
cbuffer Batch
{
	float4x3 World 					: World 				;
	float4 cParamter2				: Param2				;
	float4 EffectParam				: Param 				;
	float4x4 LastWorldViewProjTex	: LastWorldViewProjTex	;
};


cbuffer Shader
{
	float4 cUvAScolling				 = { 1, 1, 1, 1 };
	float4 cUvBScolling				 = { 1, 1, 1, 1 };
	float4 cMaskUV					 = { 1, 1, 0, 0 };
	float  cBloomScale				 = 1.0;
	float4 cEffectAColor			 = { 1, 1, 0, 0 };
	float4 cEffectBColor			 = { 1, 1, 0, 0 };
	float4 cDistortUV				 = { 1, 1, 0, 0 };
};
// Surface textures.
texture2D tEffectAMap : layout(t0) = "White";
sampler sEffectASampler : layout(s0)
{
	Texture = tEffectAMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
texture2D tEffectBMap : layout(t1) = "White";
sampler sEffectBSampler : layout(s1)
{
	Texture = tEffectBMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};
texture2D tMaskMap : layout(t2) = "White";
sampler sMaskSampler : layout(s2)
{
	Texture = tMaskMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

texture2D tNormalMap : layout(t3) = "NormalMap";
sampler sNormalSampler : layout(s3)
{
	Texture = tNormalMap;
	Filter = Linear;
	Address = Wrap;
	SRGB = false;
};

//////////////////////////////////Vertex Output////////////////////////////////////////
struct VertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	
	float4		TexCoord		: TEXCOORDN;			//	Texture coordinate for surface.
	float4		TexCoordO		: TEXCOORDN;

	float 		T 				: TEXCOORDN;			//	Fog Transmitance

	[@_TANGENT]
	half3		WorldTangent	: TEXCOORDN;			//	Output tangents in world space for normal mapping.

	[@_TANGENT]
	half3		WorldBinormal	: TEXCOORDN;			//	Output binormal in world space for normal mapping.
};

//////////////////////////////////Assemble Input////////////////////////////////////////
struct VertexAssemble
{
	float4		Position 		: POSITION;
	float4		Normal 			: NORMAL;
	float2		TexCoord		: TEXCOORDN;

	[@_TANGENT]
	float4		Tangent			: TANGENT;
	[@_TANGENT]
	float4		Binormal		: BINORMAL;

	[@_DF_SKIN]
	float4		BlendWeights	: BLENDWEIGHT;
	[@_DF_SKIN]
	uint4		BlendIndices	: BLENDINDICES;
};

//////////////////////////////////InterMediateVertex////////////////////////////////////////
struct InterMediateVertex
{
	float4 LocalPosition;

	float3 LocalNormal;

	[@_TANGENT]
	float3 LocalTangent;

	[@_TANGENT]
	float3 LocalBinormal;
};

//////////////////////////////////Functions////////////////////////////////////////
void VertexUnifyInterMediate(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	INTM.LocalPosition = float4(IN.Position.xyz, 1.0);

	INTM.LocalNormal.xyz = (float3)Expand((half3)IN.Normal.xyz);

	if(@_TANGENT)
	{
		INTM.LocalTangent = normalize(IN.Tangent.xyz);
		INTM.LocalBinormal = normalize(IN.Binormal.xyz);
	}
}

void VertexUnifySkinning(in VertexAssemble IN, inout InterMediateVertex INTM)
{
	if (@_DF_SKIN)
	{
		half4 w = IN.BlendWeights;
		half4 i = IN.BlendIndices;
		// Linear-blend skinning.
		if (@_DF_SKIN == 1)
		{
			// Fetch.
			float4 M0[3], M1[3], M2[3], M3[3];
			GetSkeletonMatrix4x3(i[0], M0[0], M0[1], M0[2]);
			GetSkeletonMatrix4x3(i[1], M1[0], M1[1], M1[2]);
			GetSkeletonMatrix4x3(i[2], M2[0], M2[1], M2[2]);
			GetSkeletonMatrix4x3(i[3], M3[0], M3[1], M3[2]);

			// Blend.
			float4 Mx = LinearBlend(w, M0[0], M1[0], M2[0], M3[0]);
			float4 My = LinearBlend(w, M0[1], M1[1], M2[1], M3[1]);
			float4 Mz = LinearBlend(w, M0[2], M1[2], M2[2], M3[2]);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = Matrix4x3TransformN(Mx, My, Mz, INTM.LocalBinormal);
		}
		// Hard skinning.
		if (@_DF_SKIN == 2)
		{
			// Fetch.
			float4 Mx, My, Mz;
			GetSkeletonMatrix4x3(i[0], Mx, My, Mz);

			// Apply.
			INTM.LocalPosition.xyz = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalPosition.xyz);
			INTM.LocalNormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = Matrix4x3TransformP(Mx, My, Mz, INTM.LocalBinormal);
		}
		// Dual-quaternion skinning.
		if (@_DF_SKIN == 3)
		{
			// Fetch.
			float4 Q0[2], Q1[2], Q2[2], Q3[2];
			GetSkeletonDualQuaternion(i[0], Q0[0], Q0[1]);
			GetSkeletonDualQuaternion(i[1], Q1[0], Q1[1]);
			GetSkeletonDualQuaternion(i[2], Q2[0], Q2[1]);
			GetSkeletonDualQuaternion(i[3], Q3[0], Q3[1]);
			// Amend orientations.
			w *= float4(1.f, sign(dot(Q0[0], Q1[0])), sign(dot(Q0[0], Q2[0])), sign(dot(Q0[0], Q3[0])));
	
			// Blend.
			float4 Qn = LinearBlend(w, Q0[0], Q1[0], Q2[0], Q3[0]);
			float4 Qd = LinearBlend(w, Q0[1], Q1[1], Q2[1], Q3[1]); 

			// Apply.
			INTM.LocalPosition.xyz = DualQuaternionTransformP(Qn, Qd, INTM.LocalPosition.xyz);
			INTM.LocalNormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalNormal);

			[@_TANGENT]
			INTM.LocalTangent = DualQuaternionTransformN(Qn, Qd, INTM.LocalTangent);
			[@_TANGENT]
			INTM.LocalBinormal = DualQuaternionTransformN(Qn, Qd, INTM.LocalBinormal);
		}
	}
}

void VertexUnifyOutput(in VertexAssemble IN, in InterMediateVertex INTM, out VertexShadingOutput OUT)
{
	[@_HLSL]
	OUT = (VertexShadingOutput)0; 

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//TexCoord
	OUT.TexCoord.xy = IN.TexCoord.xy * cUvAScolling.xy + cUvAScolling.zw * GameTime;
	OUT.TexCoord.zw = IN.TexCoord.xy * cUvBScolling.xy + cUvBScolling.zw * GameTime;
	OUT.TexCoordO.xy = IN.TexCoord.xy;

	//Fog T
	OUT.T = FogTransmittance(WorldPosition.xyz - CameraPos.xyz);

	if (@_VolumetricFog)
	{
		OUT.T = OUT.HPosition.w;
	}
	
	if (@_TANGENT)
	{
		OUT.WorldTangent.xyz  = normalize(MulMat3x4((float3x3)World, INTM.LocalTangent));
		OUT.WorldBinormal = normalize(MulMat3x4((float3x3)World, INTM.LocalBinormal));	
	}
}

//////////////////////////////////Shading////////////////////////////////////////
VertexShadingOutput EffectVertexShading(in VertexAssemble IN)
{
	VertexShadingOutput OUT;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//OUTPUT
	VertexUnifyOutput(IN, INTM, OUT);

	return OUT;
}

float4 EffectPixelShading(in VertexShadingOutput IN) : SV_Target
{
	float4 OUT = float4(0, 0, 0, 0);
	float2 TC0 = IN.TexCoord.xy;
	float2 TC1 = IN.TexCoord.zw;	

	float2 twistTC = IN.TexCoord.xy * cDistortUV.xy + cDistortUV.zw;
	float2 twist = SampleColorMap(tEffectBMap, sEffectBSampler, twistTC).xy;

	if (@_DistortEnable)
	{
		TC0 += twist;
		TC1 += twist;
	}

	float4 tEffectA = SampleColorAlphaMap(tEffectAMap, sEffectASampler, TC0) * cEffectAColor;
	float4 tEffectB = SampleColorAlphaMap(tEffectBMap, sEffectBSampler, TC1) * cEffectBColor;

	if (@_LerpEnable)
	{
		OUT = tEffectA * tEffectB;
	}
	else
	{
		OUT = lerp(tEffectA, tEffectB, saturate(tEffectB.a));
	}
	if (@_MaskEnable)
	{
		float2 maskUV = IN.TexCoordO.xy * cMaskUV.xy + cMaskUV.zw;
		float mask = SampleColorMap(tMaskMap, sMaskSampler, maskUV).r;
		OUT.a *= mask;
	}
	
	// fog shading
	if (@_VolumetricFog)
	{
		float2 spos = IN.HPosition.xy;
		[@_D3D9]
		spos += 0.5;
		float2 stc = spos * ScreenInfo.zw;
		float depth = IN.T;
		float4 ScatteringAndTransmittance = VolumetricFogLighting(stc, depth);
		OUT.rgb = OUT.rgb * ScatteringAndTransmittance.w + ScatteringAndTransmittance.xyz;
	}
	else
		OUT.rgb = FogLighting(IN.T, OUT.rgb);

	// screen color tint
	OUT.rgb = lerp(OUT.rgb, ScreenColor.rgb, ScreenColor.a);
	// bloom scale
	half bloomScale = cBloomScale * 3 * GetLuminance(OUT.rgb);

	//_TRANSPARENT
	OUT.rgb = OUT.rgb * OUT.a;
	OUT.a = EncodeTransparentBloomFactor(OUT.a, bloomScale);

	return OUT;
}

technique Shading
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectPixelShading();
}

//////////////////////////////////////////////////////////////////////////
struct ShadowVertexShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float2		SceneDepth		: TEXCOORDN;			//	Scene Depth
	float2		TC         		: TEXCOORDN;
};

//viewProjVS is light view now
ShadowVertexShadingOutput ShadowMapGenVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT;
	[@_HLSL]
	OUT = (ShadowVertexShadingOutput)0;

	//InterMediate
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	VertexUnifySkinning(IN, INTM);

	//position
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//depth
	OUT.SceneDepth = float2(OUT.HPosition.z, OUT.HPosition.w);
	
	//TC
	OUT.TC = IN.TexCoord;

	return OUT;
}

float4 ShadowMapGenPixelShading(in ShadowVertexShadingOutput IN) : SV_Target
{
	float4 rt;
	float depth = IN.SceneDepth.x / IN.SceneDepth.y;
	return EncodeDepth(depth);
}

technique ShadowGen
{
	VertexShader = ShadowMapGenVertexShading();
	PixelShader = ShadowMapGenPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Common surface stencil shader.

ShadowVertexShadingOutput StencilVertexShading(in VertexAssemble IN)
{
	ShadowVertexShadingOutput OUT = ShadowMapGenVertexShading(IN);
	OUT.HPosition.z -= 0.001;
	return OUT;
}

float4 SurfacePixelStencil(in ShadowVertexShadingOutput IN) : SV_Target
{
	return cParamter2.rgba;
}

technique Stencil
{
	VertexShader = StencilVertexShading();
	PixelShader = SurfacePixelStencil();
}

technique Debug
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectPixelShading();
}

//////////////////////////////////////////////////////////////////////////
// Special effects

technique Frozen
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectPixelShading();
}

technique Burnt
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectPixelShading();
}

technique Highlight
{
	VertexShader = EffectVertexShading();
	PixelShader = EffectPixelShading();
}

/////////////////////////////////////VBufferGen/////////////////////////////////////
struct VBufferGenShadingOutput
{
	// Most common geometric info.
	float4		HPosition		: SV_Position;			//	Position in homogeneous space. For output to rasterizer.
	float4		LastTex			: TEXCOORDN;			//	Last Tex
	[@_CLIPPING]
	float2 		TC 				: TEXCOORDN;			//  TextureCoord
};

//viewProjVS is light view now
VBufferGenShadingOutput VBufferGenVertexShading(in VertexAssemble IN)
{
	VBufferGenShadingOutput OUT;
	[@_HLSL]
	OUT = (VBufferGenShadingOutput)0;

	//HPosition
	InterMediateVertex INTM;
	VertexUnifyInterMediate(IN, INTM);
	float4 WorldPosition = float4(MulMat3x4(World, float4(INTM.LocalPosition.xyz, 1)).xyz, 1);
	OUT.HPosition = MulMat4x4(ViewProj, WorldPosition);

	//MotionVector
	InterMediateVertex INTM2;
	VertexUnifyInterMediate(IN, INTM2);
	OUT.LastTex = MulMat4x4(LastWorldViewProjTex, INTM2.LocalPosition);

	[@_CLIPPING]
	OUT.TC = IN.TexCoord.xy;

	return OUT;
}

float4 VBufferGenPixelShading(in VBufferGenShadingOutput IN) : SV_Target
{
	//curr uv
	float2 spos = IN.HPosition.xy;
	[@_D3D9]  
	spos += 0.5;
	float2 currUV = spos * ScreenInfo.zw;
	
	//last uv 
	float2 lastUV = IN.LastTex.xy / IN.LastTex.w;
	lastUV += cParamter2.xy * ScreenInfo.zw;
	float2 motionVector = lastUV - currUV;
	return float4(motionVector.xy, motionVector.xy*ScreenInfo.xy);
	//return float4(FloatToR8G8(motionVector.x), FloatToR8G8(motionVector.y));
}

technique VBufferGen
{
	VertexShader = VBufferGenVertexShading();
	PixelShader = VBufferGenPixelShading();
}

